/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef __FileHandling_hh__
#define __FileHandling_hh__

/** File handling tools
@todo: correct handling of Win32 SAMBA paths including \\machine\....
@author Jan Woetzel 02/2005
*/

#include <bias_config.h>

#include <string>
#include <sstream>
#include <vector>
#include <iomanip>

#define DEFAULT_LEADING_ZEROS 5


namespace BIAS {

  class BIASCommon_EXPORT FileHandling
     {
     public:

  static int RemoveDir(const std::string &subdir);
  static int CreateDir(const std::string &subdir);
  static bool FileExists(const std::string &file);

  /**
     Read all files in @param Dir, match against @param suffix and
@param prefix and put into @param content.  If you want the full path
with file name set @param GetFullPath=true, otherwise only the file name is used.

   */
  static int GetDirectoryContent(const std::string &Dir,
                                            const std::string &prefix,
                                            const std::string &suffix,
                                            std::vector<std::string> &content,
                                            const bool GetFullPath=false);



  /* @return true if the file exists can be pendedis readable on file system */

  static bool FileReadable( const std::string & s);

  template <class T>
  inline static std::string toString(const T thenumber,
                              int numzeros=DEFAULT_LEADING_ZEROS) {
    std::stringstream ss;
    ss << std::setfill('0')<<std::setw(numzeros)<<thenumber;
    return ss.str();
  }

  /**  split full path into:
  - directory
  - basename
  - .suffix (inlcluding dot)
  @todo: Extract machine name or IP for SMB network shares
  for WIN32, e.g. \\machine\share and fix backslashes on  Win32.
  @author Jan Woetzel 03/2005 */
  // TODO: \todo WIN: fix to make \, \\ and / work on directories (jw)
  static void
    SplitName(const std::string & fullname
    ,std::string & dir
    ,std::string & base
    ,std::string & suffix );

  /** get Directory from fullname path JW */
  static std::string
    Directory(const std::string & fullname);

  /**  get Basename of file, 
       no suffix, no directory. */
  static std::string
    Basename(const std::string & fullname);

  /**  get .suffix of file including the ".dot",
       no basename, no directory. */
  static std::string
    Suffix(const std::string & fullname);

  /**  get Extension of file without the "dot",
       no basename, no directory. */
  static std::string
    Extension(const std::string & fullname);

  /**  get filename including suffix, 
       no directory. */
  static std::string
    Filename(const std::string & fullname);

  /**  get directory and basename of file,
       no suffix. */
  static std::string
    DirectoryAndBasename(const std::string & fullname);


  /// type correct helper wrapper for tolower required for stl transformations JW
  static char ToLower_c(char c);
  /// type correct helper wrapper for toupper required for stl transformations JW
  static char ToUpper_c(char c);


  /// @return lower case string of input JW
  static std::string  LowerCase(const std::string & input);

  /// @return upper case string of input JW
  static std::string  UpperCase(const std::string & input);

  /** write compilation time/date and all command line arguments
  to a log file/stream JW */
  static void
    LogCommandLine( int argc,char** argv, std::ostream & os);

  static void
    LogCommandLine( int argc,char** argv, const std::string & filename);

  /** Creates a string with leading zeroes from number 
   * @deprecated use FileHandling::toString(n,digits) */
  static std::string
    LeadingZeroString( const int & n,
    const unsigned int & digits=DEFAULT_LEADING_ZEROS ) ;

  /** @brief create a string containing date and time like 20060329_173054
      @author koeser */
  static std::string   CurrentDateTimeString();

  /** @brief extract numeric value fromstring beginning from right,
  e.g. "blah042foo.txt" --> 42 and the pos. of the 0 and 2.
  Useful helper for FilenameSequence JW */
  static int  ExtractNumberFromRight(const std::string & inStr,
    int & number,
    int & posL,
    int & posR );


  /** @brief construct i,g .seq filename from a given filename
  by manipulating the number within base part of RefFilename.
  of an image sequence given a ref. filename and an offset.
  The number of leading zeroes is kept but digits may be added if required.
  @param offsetFromRef is the number of frames forward or backward from ref. image.
  @param stepsize is the number of increment/decrement per offset.

  Example:
  C:/foo/image00100bar.jpg
  ref=100, offset=2, stepsize=7 --> 114
  resultFilename --> C:/foo/image00114bar.jpg
  or:
  foo9   -> foo10
  foo009 -> foo010

  @return 0 for success, other for failure.
  @author Jan Woetzel */
  static int  FilenameSequence(const std::string & refFileName,
    std::string & resultFilename,
    const int & offsetFromRef,
    const unsigned int & stepsize=1 );

  /** @brief exands filename, e.g. tilde to home
  expands:
  tilde to HOME if set and USERPROFILE otherwise
  @author Jan Woetzel */
  static int  FilenameExpansion(const std::string & i, std::string & o );

  /** @return current working directory as C++ string JW */
  static std::string  GetCwd();


  /**  Change current working directory  */
  static int  ChDir(const std::string &dir);

  /** @brief a trailing slash is "removed" from the path if there is one */
  inline static std::string StripTerminatingSlash(const std::string& thepath) {
    if (thepath.empty()) return thepath;
#ifdef WIN32
    static std::string::size_type poslastslash = thepath.rfind('\\'); // separator for directory
#else
    static std::string::size_type poslastslash = thepath.rfind('/'); // separator for directory
#endif
    //std::string::size_type poslastslash = thepath.rfind('/');
    if (poslastslash == thepath.size()-1) {
      // copy everything without the trailing slash
      return thepath.substr(0, poslastslash);
    }
    // already fine, no slash at the end of path
    return thepath;
  }

  /// @return full qualified path of current thread/application/module/exe/dll
  /// @author Jan Woetzel
  static std::string GetApplicationPath();

     }; // class
} // namespace BIAS
#endif // do not add code after this line
