/* This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel

   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.

   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/

/**
   @file TestCompareFloatingPoint.cpp
   @relates CompareFloatingPoint
   @brief Test comparison of floating point numbers
   @ingroup g_tests
   @author MIP
*/

#include <Base/Common/W32Compat.hh>
#include <Base/Debug/Error.hh>
#include <Base/Common/CompareFloatingPoint.hh>

#include <vector>
#include <limits>
#include <map>
#include <iostream>
#include <iomanip>

using namespace std;
using namespace BIAS;

enum ERelation { LESS, LESS_EQUAL, EQUAL, GREATER_EQUAL, GREATER }; 

ostream& operator<<(ostream& os, const enum ERelation& r)
{
  switch (r){
  case LESS: os << "<"; break;
  case LESS_EQUAL: os << "<="; break;
  case EQUAL: os << "=="; break;
  case GREATER: os << ">"; break;
  case GREATER_EQUAL: os << ">="; break;
  }
  return os;
}

/** @relates CompareFloatingPoint */
template <class T>
class TestValuePair
{
public:
  TestValuePair(const T l, const ERelation rel, const T r, const T eps)
    : left(l), right(r), epsilon(eps), relation(rel) 
  { if (rel==LESS_EQUAL || rel==GREATER_EQUAL) { BIASABORT; } }

  /// the two values
  T left, right;
  /// epsilon for comparison
  T epsilon;
  /// the true relation, i.e. GREATER, LESS or EQUAL
  /// the true relation cannot be GREATER_EQUAL or LESS_EQUAL
  ERelation relation;
};


/** @relates CompareFloatingPoint */
template <class T>
void GenTestValues(vector<TestValuePair<T> >& tv)
{
#if defined(max) || defined(min)
# undef max
# undef min
  const T mini = numeric_limits<T>::min();
  const T maxi = numeric_limits<T>::max();
  
# define max(a,b) ((a) >= (b) ? (a) : (b))
# define min(a,b) ((a) <= (b) ? (a) : (b))
#else
  const T mini = numeric_limits<T>::min();
  const T maxi = numeric_limits<T>::max();
#endif
  const T eps  = numeric_limits<T>::epsilon();
  // the true 
  tv.clear();
  tv.push_back(TestValuePair<T>(0.0, EQUAL, 0.0, 0.0));
  tv.push_back(TestValuePair<T>(1.0, EQUAL, ((T)1.0)+eps, eps*((T)2.0)));
  tv.push_back(TestValuePair<T>(mini, LESS, mini*((T)2.0), eps));
  tv.push_back(TestValuePair<T>(mini, EQUAL, mini, eps));
  tv.push_back(TestValuePair<T>(0.0, EQUAL, mini, mini*((T)2.0)));
  tv.push_back(TestValuePair<T>(mini, EQUAL, 0.0, mini*((T)2)));
  tv.push_back(TestValuePair<T>(((T)1e-10), EQUAL,
                                ((T)1e-10)+((T)1e-18), ((T)1e-7)));
  tv.push_back(TestValuePair<T>(0.0, EQUAL, 0.0, mini));
  tv.push_back(TestValuePair<T>(5.0, LESS, ((T)5.0)+eps*((T)10.0), eps));
  tv.push_back(TestValuePair<T>(1.0, EQUAL, 1.0, eps));
  tv.push_back(TestValuePair<T>(0.0, LESS, 1.0, eps));
  tv.push_back(TestValuePair<T>(0.0, LESS, maxi, eps));
  tv.push_back(TestValuePair<T>(maxi, GREATER, 0.0, eps));
  tv.push_back(TestValuePair<T>(maxi, GREATER, 100.0, eps));
  tv.push_back(TestValuePair<T>(maxi, GREATER, -maxi, eps));
  tv.push_back(TestValuePair<T>(-maxi, LESS, maxi, eps));
}

/// builds the map of expected results for the different comparison functions
void BuildMap(map<ERelation, map<ERelation, bool> >& m)
{
  // m[true relation][tested relation]

  m[EQUAL][EQUAL] = true;
  m[EQUAL][LESS] = false;
  m[EQUAL][GREATER] = false;
  m[EQUAL][LESS_EQUAL] = true;
  m[EQUAL][GREATER_EQUAL] = true;

  m[LESS][EQUAL] = false;
  m[LESS][LESS] = true;
  m[LESS][GREATER] = false;
  m[LESS][LESS_EQUAL] = true;
  m[LESS][GREATER_EQUAL] = false;

  m[GREATER][EQUAL] = false;
  m[GREATER][LESS] = false;
  m[GREATER][GREATER] = true;
  m[GREATER][LESS_EQUAL] = false;
  m[GREATER][GREATER_EQUAL] = true;
}

/** @relates CompareFloatingPoint */
template <class T>
void Test(const TestValuePair<T>& t)
{
  if (t.relation==LESS_EQUAL || t.relation == GREATER_EQUAL){
    BIASERR("true relation cannot be greater equal or less equal!");
    BIASABORT;
  }
  
  // m[true relation][tested relation]
  static map<ERelation, map<ERelation, bool> > exp_res;
  if (exp_res.empty()) BuildMap(exp_res);
  
  const bool debug = false;
  if (debug) cout << "\nepsilon: "<<setprecision(30) << t.epsilon 
                  << "   true relation: "<< setprecision(30) <<t.left
                  <<" "<<t.relation<<" "<<setprecision(30) <<t.right<<endl;
  bool res;
  res = Equal(t.left, t.right, t.epsilon);
  if (debug) cout << " "<<setprecision(30) << t.left <<" == "<<setprecision(30) 
                  << t.right << " = "<<boolalpha<<res<<endl;
  if (res != exp_res[t.relation][EQUAL]) { 
    BIASERR("left: "<<t.left<<"  right: "<<t.right<<" true relation: "
            <<t.relation<<"  expected result: "<<boolalpha
            <<exp_res[t.relation][EQUAL]
            <<"  result: "<<boolalpha<<res<<endl);
    cout << "fabs(left-right): "<<fabs(t.left-t.right)<<endl
         << "eps * fabs(right): "<<t.epsilon * fabs(t.right)<<endl
         << "eps * fabs(left): "<<t.epsilon * fabs(t.left)<<endl
         << "right!=0.: "<<boolalpha<<(t.right!=0.) <<endl
         << "left!=0.: "<<boolalpha<<(t.left!=0.) <<endl
         << "right==0.: "<<boolalpha<<(t.right==0.) <<endl
         << "left==0.: "<<boolalpha<<(t.left==0.) <<endl;
    if (!(t.right==0.))
      cout << "fabs(left-right) / fabs(right): "
           <<fabs(t.left-t.right) / fabs(t.right)<<endl;
    if (!(t.left==0.))
      cout << "fabs(left-right) / fabs(left): "
           <<fabs(t.left-t.right) / fabs(t.left)<<endl;
    BIASABORT; 
  }

  res = Less(t.left, t.right, t.epsilon);
  if (debug) cout << " "<<setprecision(30) << t.left <<" < "<<setprecision(30) 
                  << t.right << " = "<<boolalpha<<res<<endl;
  if (res != exp_res[t.relation][LESS]) { 
    BIASERR("left: "<<t.left<<"  right: "<<t.right<<" true relation: "
            <<t.relation<<"  expected result: "<<boolalpha
            <<exp_res[t.relation][LESS]
            <<"  result: "<<boolalpha<<res<<endl);
    BIASABORT; 
  } 
  

  res = LessEqual(t.left, t.right, t.epsilon);
  if (debug) cout << " "<<setprecision(30) << t.left <<" <= "<<setprecision(30) 
                  << t.right << " = "<<boolalpha<<res<<endl;
  if (res != exp_res[t.relation][LESS_EQUAL]) { 
    BIASERR("left: "<<t.left<<"  right: "<<t.right<<" true relation: "
            <<t.relation<<"  expected result: "<<boolalpha
            <<exp_res[t.relation][LESS_EQUAL]
            <<"  result: "<<boolalpha<<res<<endl);
    BIASABORT; 
  } 
  
  res = Greater(t.left, t.right, t.epsilon);
  if (debug) cout << " "<<setprecision(30) << t.left <<" > "<<setprecision(30) 
                  << t.right << " = "<<boolalpha<<res<<endl;
  if (res != exp_res[t.relation][GREATER]) { 
    BIASERR("left: "<<t.left<<"  right: "<<t.right<<" true relation: "
            <<t.relation<<"  expected result: "<<boolalpha
            <<exp_res[t.relation][GREATER]
            <<"  result: "<<boolalpha<<res<<endl);
    BIASABORT; 
  }

  res = GreaterEqual(t.left, t.right, t.epsilon);
  if (debug) cout << " "<<setprecision(30) << t.left <<" >= "<<setprecision(30) 
                  << t.right << " = "<<boolalpha<<res<<endl;
  if (res != exp_res[t.relation][GREATER_EQUAL]) { 
    BIASERR("left: "<<t.left<<"  right: "<<t.right<<" true relation: "
            <<t.relation<<"  expected result: "<<boolalpha
            <<exp_res[t.relation][GREATER_EQUAL]
            <<"  result: "<<boolalpha<<res<<endl);
    BIASABORT; 
  }

  /// there are 3 true relations and hence the size of exp_res must be 3
  BIASASSERT(exp_res.size()==3u);
}

/** @relates CompareFloatingPoint
    @author woelk 02/2008 (c) www.vision-n.de */
int main(int argc, char *argv[])
{
  // test float
  vector<TestValuePair<float> > tvf;
  GenTestValues(tvf);

  vector<TestValuePair<float> >::const_iterator itf;
  for (itf=tvf.begin(); itf!=tvf.end(); itf++){
    Test(*itf);
  }

  // test double
  vector<TestValuePair<double> > tvd;
  GenTestValues(tvd);

  vector<TestValuePair<double> >::const_iterator itd;
  for (itd=tvd.begin(); itd!=tvd.end(); itd++){
    Test(*itd);
  }

  
  return 0;
}
