/* This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel

   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.
   
   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
#ifndef __RMatrixBase_hh__
#define __RMatrixBase_hh__
#include "bias_config.h" 

#include <Base/Math/Vector3.hh>
#include <Base/Math/Matrix3x3.hh>
#include <Base/Debug/Debug.hh>

namespace BIAS {

#define ROTATION_MATRIX_TYPE double
 
#define ROTATION_MATRIX_EPSILON DBL_EPSILON

  template<class T> class BIASGeometryBase_EXPORT Quaternion;
  /** @class RMatrixBase 
      @ingroup g_geometry
      @brief 3D rotation matrix 
    
      @author frahm, woelk **/
  class BIASGeometryBase_EXPORT RMatrixBase 
    : public Matrix3x3<ROTATION_MATRIX_TYPE> 
  {
  public:
    // sets identity
    RMatrixBase();

    /**	deprectade, only identity allowed
        constructor setting identity or zero
        @brief author koeser **/
    explicit RMatrixBase(const MatrixInitType& i);

    /// copy constructor
    RMatrixBase(const RMatrixBase& r);

    RMatrixBase(const Matrix3x3<ROTATION_MATRIX_TYPE>& mat);

    /// initialization from rotation axis w and angle phi (Rodriguez)
    RMatrixBase(const Vector3<ROTATION_MATRIX_TYPE> w, 
                const ROTATION_MATRIX_TYPE phi);

    explicit RMatrixBase(const TNT::Matrix<ROTATION_MATRIX_TYPE>& r);

    virtual ~RMatrixBase();

    /** angles in rad, 
        either: order 1. z, 2. y, 3. x  and  fixed  axes 
        or: order 1. x, 2. y, 3. z  and  moving axes
        angles are measured in mathematical sense (anticlockwise), i.e.
        Set(x, 0, 0) results in the following matrix
        | 1  0        0      |
        | 0  cos(x)  -sin(x) |
        | 0  sin(x)   cos(x) |
        
        (*this) = Rx*Ry*Rz =   

        |  c(y)c(z)               -c(y)s(z)                s(y)     |
        |  c(z)s(x)s(y)+c(x)s(z)   c(x)c(z)-s(x)s(y)s(z)  -c(y)s(x) |
        | -c(x)c(z)s(y)+s(x)s(z)   c(z)s(x)+c(x)s(y)s(z)   c(x)c(y) |

        with s(g) = sin(g), c(g) = cos(g), x = PhiX, y = PhiY and z = PhiZ

        @author frahm ,woelk */
    void SetXYZ(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
                ROTATION_MATRIX_TYPE PhiZ);
    inline void SetXYZ(Vector3<ROTATION_MATRIX_TYPE>& r)
    { SetXYZ(r[0], r[1], r[2]); }

    /** angles in rad, 
        either: order 1. z, 2. y, 3. x  and  moving axes 
        or: order 1. x, 2. y, 3. z  and  fixed  axes
        angles are measured in mathematical sense (anticlockwise), i.e.
        Set(x, 0, 0) results in the following matrix
        | 1  0        0      |
        | 0  cos(x)  -sin(x) |
        | 0  sin(x)   cos(x) |
        
        (*this) = Rz*Ry*Rx =   

        |  c(y)c(z)  s(x)s(y)c(z)-c(x)s(z)  c(x)s(y)c(z)+s(x)s(z) |
        |  c(y)s(z)  s(x)s(y)s(z)+c(x)c(z)  c(x)s(y)s(z)-s(x)s(z) |
        | -s(y)      s(x)c(y)               c(x)c(y)              |

        with s(g) = sin(g), c(g) = cos(g), x = PhiX, y = PhiY and z = PhiZ

        @author frahm ,woelk */
    void SetZYX(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
                ROTATION_MATRIX_TYPE PhiZ);

    inline void SetZYX(Vector3<ROTATION_MATRIX_TYPE>& r)
    { SetZYX(r[0], r[1], r[2]); }
 

    /** angles in rad, 
        either: order 1. y, 2. x, 3. z  and  moving axes 
        or: order 1. z, 2. x, 3. y  and  fixed  axes
        angles are measured in mathematical sense (anticlockwise), i.e.
        Set(x, 0, 0) results in the following matrix
        | 1  0        0      |
        | 0  cos(x)  -sin(x) |
        | 0  sin(x)   cos(x) |
        
        (*this) = Rz*Rx*Ry =   
        | c(y)c(z)-s(x)s(y)s(z), c(x)s(z), s(y)c(z)+s(x)c(y)s(z) |
        | s(z)c(y)+s(x)s(y)c(z), c(x)c(z),  s(y)*s(z)+s(x)-c(y)c(z) |
        | -c(x)s(y)             ,  s(x)  ,  c(x)c(y)              |
        
        with s(g) = sin(g), c(g) = cos(g), x = PhiX, y = PhiY and z = PhiZ

        @author haase */
    void SetZXY(ROTATION_MATRIX_TYPE PhiZ, ROTATION_MATRIX_TYPE PhiX, 
                ROTATION_MATRIX_TYPE PhiY);
    inline void SetZXY(Vector3<ROTATION_MATRIX_TYPE>& r)
    { SetZXY(r[0], r[1], r[2]); }

     /** angles in rad, 
         either: order 1. y, 2. x, 3. z  and  moving axes 
         or: order 1. z, 2. x, 3. y  and  fixed  axes 
         angles are measured in mathematical sense (anticlockwise), i.e.
         Set(x, 0, 0) results in the following matrix
         | 1  0        0      |
         | 0  cos(x)  sin(x) |
         | 0  -sin(x)   cos(x) |
         
         (*this) = Rz*Rx*Ry =   
         | c(y)c(z)+s(x)s(y)s(z), c(x)s(z), -s(y)c(z)+s(x)c(y)s(z) |
         | -s(z)c(y)+s(x)s(y)c(z), c(x)c(z),  -s(y)*-s(z)+s(x)c(y)c(z) |
         | c(x)s(y)             ,  -s(x)  ,  c(x)c(y)              |
         
         with s(g) = sin(g), c(g) = cos(g), x = PhiX, y = PhiY and z = PhiZ

         @author haase */
    void SetYXZ(ROTATION_MATRIX_TYPE PhiZ, ROTATION_MATRIX_TYPE PhiX, 
                ROTATION_MATRIX_TYPE PhiY);
    inline void SetYXZ(Vector3<ROTATION_MATRIX_TYPE>& r)
    { SetYXZ(r[0], r[1], r[2]); }
    

    /** initialize from vector and angle
        remark: Axis vector w must be an unit vector (||w||>1e-6) if the 
        angle is not zero, otherwise an exception is thrown
        @author evers, woelk */
    void Set(const Vector3<ROTATION_MATRIX_TYPE> &w, 
             const ROTATION_MATRIX_TYPE phi);


    /** initialize from vector * angle (axis _times_ angle !)
        @author evers */
    void SetFromAxisAngle(Vector3<ROTATION_MATRIX_TYPE> w);


    /* set rotation matrix from an orthogonalbasis given in world coordinate 
       system. (wcs).
       As R' is expressing the transformation from wcs to ics, 
       R is ics to wcs and hereby represent the ics base vectors expressed in 
       wcs
       the *rows* of R because the tranformation is the scalar product.        
       
       Assume xx,yy,zz are right hand orthogonal (rhs).
       E.g. the orthogonal image coord system base vectors expressed in world 
       coordinates (wcs).
       (inverse doesn't make sense because base would be Id., then).
       Normal vectors of length==1 are computed.
       // TODO warning if (xxx,yyy,zzz) are not orthogonalbasis.
       @author JW 09/2003, added excep */
    void SetFromOrthogonalBasis(const BIAS::Vector3<ROTATION_MATRIX_TYPE> &xxx,
                                const BIAS::Vector3<ROTATION_MATRIX_TYPE> &yyy,
                                const BIAS::Vector3<ROTATION_MATRIX_TYPE> &zzz);
    
    /** set rotation matrix from two vectors.
        form an orthonormalbasis.
        xh represent the first base vector and is left unchanged. 
        You can think of this routine as compute R from an image plane given 
        by two (usually orthogonal) base vectors.
        if the given base vecctors are not orthogonal xh is kept and yv is 
        orthogonalized appropriately.
        @author Jan Woetzel */
    void SetFromHV(const BIAS::Vector3<ROTATION_MATRIX_TYPE> & xh,
		   const BIAS::Vector3<ROTATION_MATRIX_TYPE> &vy);
    
    /** calculates the quaternion from a rotation matrix
        @author woelk 12 2003 */
    int GetQuaternion(Quaternion<ROTATION_MATRIX_TYPE> & quat) const ;
    
    /** Set RMatrix from a quternion 
        @author Daniel Grest, June2003*/
    int SetFromQuaternion(const Quaternion<ROTATION_MATRIX_TYPE> &q);
    
    /** Set Rotation matrix from orientation and up vector openGL conform, 
        similar to gluLookAt(). <br>
        openGL like means: if ori is (0,0,-1) and up:(0,1,0) the resulting
        matrix is the identity
        @author grest, Dec. 2005*/
    int SetFromOriUpGL(BIAS::Vector3<ROTATION_MATRIX_TYPE> ori,
                       BIAS::Vector3<ROTATION_MATRIX_TYPE> up);

    /** calculates an angle and a axis for the given rotation
        angle is returned in rad
        @author woelk 12 2003 */
    int GetRotationAxisAngle(Vector3<ROTATION_MATRIX_TYPE>& Axis, 
                             ROTATION_MATRIX_TYPE& Angle) const;

    /** interface for Axis component of GetRotationAxisAngle JW */
    BIAS::Vector3<ROTATION_MATRIX_TYPE> GetRotationAxis() const;

    /** interface for Angle component of GetRotationAxisAngle JW */
    ROTATION_MATRIX_TYPE GetRotationAngle() const;

    /** calculates a angle * axis for the given rotation
        @author evers  */
    int GetRotationAxisAngle(Vector3<ROTATION_MATRIX_TYPE>& w) const;

    /** not unique, has singularities at +- pi/2
        assumes (*this) = Rx * Ry * Rz,   i.e.
        either:  order 1. x, 2. y, 3. z  and  moving axes
        or:      order 1. z, 2. y, 3. x  and  fixed  axes
        @author woelk 01 2003 */
    int GetRotationAnglesXYZ(double& PhiX, double& PhiY, double& PhiZ);
    inline int GetRotationAnglesXYZ(Vector3<ROTATION_MATRIX_TYPE>& r)
    { return GetRotationAnglesXYZ(r[0], r[1], r[2]); }

    /** not unique, has singularities at +- pi/2
        assumes (*this) = Rz * Ry * Rx,   i.e.
        either:  order 1. x, 2. y, 3. z  and  fixed  axes
        or:      order 1. z, 2. y, 3. x  and  moving axes
        @author woelk 01 2003 */
    int GetRotationAnglesZYX(double& PhiX, double& PhiY, double& PhiZ);
    inline int GetRotationAnglesZYX(Vector3<ROTATION_MATRIX_TYPE>& r)
    { return GetRotationAnglesZYX(r[0], r[1], r[2]); }

    /** not unique, has singularities at +- pi/2
        rotation order ZXY = rotate around point/vector
        @author haase 2007 */
    int GetRotationAnglesZXY(double& PhiZ, double& PhiX, double& PhiY);
    inline int GetRotationAnglesZXY(Vector3<ROTATION_MATRIX_TYPE>& r)
    { return GetRotationAnglesZXY(r[2], r[0], r[1]); }

    /** not unique, has singularities at +- pi/2
        rotation order YXZ = rotate with moving axes
        @author haase 2007 */
    int GetRotationAnglesYXZ(double& PhiY, double& PhiX, double& PhiZ);
    inline int GetRotationAnglesYXZ(Vector3<ROTATION_MATRIX_TYPE>& r)
    { return GetRotationAnglesYXZ(r[1], r[0], r[2]); }
    
    // DLL HACK (JW)
    friend bool BIASGeometryBase_EXPORT 
    operator==(const RMatrixBase& left, const RMatrixBase& right);

  
    /** checks if this is a rotation matrix, i.e. 
        1. If the determinant is 1.0
        2. If the columns are orthonormal */
    bool Check(const double eps = 
                std::numeric_limits<double>::epsilon()) const;

    /** returns the reason for the check failure in a human readable string, 
        empty string when check succeeds */
    std::string GetCheckFailureReason(const double eps = 
                                      std::numeric_limits<double>::epsilon()) const;

    void EnforceConstraints();
    
    // zero matrix is not a valid rotation matrix
    void SetZero() { BIASABORT; }

  }; // class
} // namespace BIAS


#endif // __RMatrixBase_hh__
