/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
   @file TestRotationConversion.cpp
   @relates Quaternion
   @brief Test the conversion between rotation represenations
   @ingroup g_tests
   @author MIP
*/

#include <Base/Geometry/Quaternion.hh>
#include <Base/Math/Random.hh>
#include <Base/Common/CompareFloatingPoint.hh>

using namespace BIAS;
using namespace std;


Random ran;

Quaternion<double> GenerateRandomQuaternion();

void TestConversionQuaternionMatrix();

bool CompareQuaternions();

int main(int argc, char *argv[])
{
  TestConversionQuaternionMatrix();

  return 0;
}


Quaternion<double> GenerateRandomQuaternion()
{
  Quaternion<double> q;
  do {
    q.Set(ran.GetUniformDistributed(-1., 1.),
          ran.GetUniformDistributed(-1., 1.),
          ran.GetUniformDistributed(-1., 1.),
          ran.GetUniformDistributed(-1., 1.));
  } while (Equal(q.NormL2(), 0.));
  //cout << q << endl; 
  return q;
}

static double worst = -1.0;
bool CompareQuaternions(const Quaternion<double> &Q1, 
                        const Quaternion<double> &Q2)
{
  Quaternion<double> q1(Q1), q2(Q2); // de-const
  if (Equal(q1.NormL2(), 0.)) { BIASABORT; }
  if (Equal(q2.NormL2(), 0.)) { BIASABORT; }

  // multiplication with inverse must result in 0 0 0 1
  // or 0 0 0 -1
  q1.Mult(q2.Inverse());
  q1 /= q1.NormL1();

  const double eps = 5e-10;
  for (int j=0; j<3; j++){
    if (fabs(q1[j]-0.)>worst) worst = fabs(q1[j]-0.);
  }
  if (fabs(fabs(q1[3])-1.)>worst) worst = fabs(fabs(q1[3])-1.);
  

  if (!Equal(q1[0], 0.0, eps) || !Equal(q1[1], 0.0, eps) ||
      !Equal(q1[2], 0.0, eps) || !Equal(fabs(q1[3]), 1.0, eps)){
    BIASERR("inconsistency converting quaternion to rotation matrix and "
            "back to quaternion: \n  src*dst.Inverse(): "
            <<q1/q1.NormL2()<<endl<<Q1<<"\n"<<Q2<<endl);
    return false;
  }
  return true;
}

void TestConversionQuaternionMatrix()
{
  const int num = 10000;
  
  Quaternion<double> src, dst;
  RMatrixBase R;

  for (int i=0; i<num; i++){
    src = GenerateRandomQuaternion();

    if (R.SetFromQuaternion(src)!=0){ BIASABORT; }
    if (R.GetQuaternion(dst)!=0){ BIASABORT; }
    if (!CompareQuaternions(src, dst)){ 
      cout << "worst: "<<setprecision(30)<<worst<<endl;
      cerr << "i: "<<i<<endl<<src/src.NormL2()<<"\n"<<dst/dst.NormL2()<<endl;
      BIASABORT; }
    
  }
  cout << "worst: "<<setprecision(30)<<worst<<endl;

}
