#ifndef _TRIANGLE_HH__
#define _TRIANGLE_HH__

#include <Base/Geometry/HomgPoint2D.hh>


namespace BIAS {

  /**
   * @brief Represents a 2D triangle and allows some computations with it.
   *
   * @author rwulff 04/2010
   */
  class BIASGeometryBase_EXPORT Triangle {

    public:

    /**
     * Constructor.
     */
    Triangle(const HomgPoint2D &a, const HomgPoint2D &b, const HomgPoint2D &c);


    /**
     * @brief Checks if the given point is inside this triangle.
     *
     * @param p
     *        the point to check
     *
     * @return true iff the given point is inside this triangle
     */
    inline bool IsPointInTriangle(const HomgPoint2D &p) {
      return     ArePointsOnSameSide(p, a_, b_, c_)
              && ArePointsOnSameSide(p, b_, c_, a_)
              && ArePointsOnSameSide(p, c_, a_, b_);
    }


    /**
     * @brief Returns the barycentric coordinates of the given point with
     *        respect to this triangle.
     *
     * @param p
     *        the point
     * @param lambdaA
     *        the barycentric coordinate for the first corner of this triangle
     * @param lambdaB
     *        the barycentric coordinate for the second corner of this triangle
     * @param lambdaC
     *        the barycentric coordinate for the thrid corner of this triangle
     */
    void GetBarycentricCoords(const HomgPoint2D &p,
                              double &lambdaA,
                              double &lambdaB,
                              double &lambdaC);


    /**
     * @brief Sets the corners of this triangle.
     */
    inline void SetCorners(const HomgPoint2D &a,
                           const HomgPoint2D &b,
                           const HomgPoint2D &c) {
      a_ = a;
      b_ = b;
      c_ = c;
    }


    /**
     * @brief Returns the corners of this triangle.
     */
    inline void GetCorners(HomgPoint2D &a,
                           HomgPoint2D &b,
                           HomgPoint2D &c) {
      a = a_;
      b = b_;
      c = c_;
    }



    private:

    // the three corners of this triangle
    BIAS::HomgPoint2D a_, b_, c_;


    /**
     * Hidden default constructor.
     */
    Triangle() {};


    /**
     * @brief Helper method for IsPointInTriangle().
     *
     * Checks if p1 and p2 are on the same side of the vector (b - a).
     */
    inline bool ArePointsOnSameSide(HomgPoint2D p1, HomgPoint2D p2,
                                    HomgPoint2D a,  HomgPoint2D b) {
      const HomgPoint2D cp1 = (b - a).CrossProduct(p1 - a);
      const HomgPoint2D cp2 = (b - a).CrossProduct(p2 - a);

      return cp1.ScalarProduct(cp2) >= 0;
    }

  }; // end of class
} // end of namespace
#endif // _TRIANGLE_HH__