/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ImageCalc.hh"

using namespace std;
using namespace BIAS;


////////////////////////////////////////////////////////////////////////////////
// calculations between two images                                            //
////////////////////////////////////////////////////////////////////////////////


template<class T>
void BIAS::ImageCalc<T>::Add(const Image<T> &in1, const Image<T> &in2,
                             Image<T> &out) {
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount(),
           in1.GetStorageType());

  // get pointers to data
  const T *in1Data = in1.GetImageData();
  const T *in2Data = in2.GetImageData();
  T *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = in1Data[i] + in2Data[i];
  }
}


template<class T>
void BIAS::ImageCalc<T>::Subtract(const Image<T> &in1, const Image<T> &in2,
                                  Image<T> &out) {
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const T *in1Data = in1.GetImageData();
  const T *in2Data = in2.GetImageData();
  T *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = in1Data[i] - in2Data[i];
  }
}


template<class T>
void BIAS::ImageCalc<T>::Multiply(const Image<T> &in1, const Image<T> &in2,
                                  Image<T> &out) {
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const T *in1Data = in1.GetImageData();
  const T *in2Data = in2.GetImageData();
  T *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = in1Data[i] * in2Data[i];
  }
}


template<class T>
void BIAS::ImageCalc<T>::Divide(const Image<T> &in1, const Image<T> &in2,
                                Image<T> &out) {
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const T *in1Data = in1.GetImageData();
  const T *in2Data = in2.GetImageData();
  T *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    if (in2Data[i] != (T)0) {
      outData[i] = in1Data[i] / in2Data[i];
    }
    else {
      BIASWARN("prevented division by zero")
      outData[i] = (T)0;
    }
  }
}


////////////////////////////////////////////////////////////////////////////////
//calculations between one image and a scalar                                 //
////////////////////////////////////////////////////////////////////////////////


template <class T>
void BIAS::ImageCalc<T>::AddScalar(const Image<T> &in, T scalar,
                                   Image<T> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const T *inData = in.GetImageData();
  T *outData      = out.GetImageData();

  const unsigned int size =   in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = inData[i] + scalar;
  }
}


template <class T>
void BIAS::ImageCalc<T>::SubtractScalar(const Image<T> &in, T scalar,
                                        Image<T> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const T *inData = in.GetImageData();
  T *outData      = out.GetImageData();

  const unsigned int size =   in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = inData[i] - scalar;
  }
}


template <class T>
void BIAS::ImageCalc<T>::MultiplyScalar(const Image<T> &in, T scalar,
                                        Image<T> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const T *inData = in.GetImageData();
  T *outData      = out.GetImageData();

  const unsigned int size =   in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = inData[i] * scalar;
  }
}


template <class T>
void BIAS::ImageCalc<T>::DivideScalar(const Image<T> &in, T scalar,
                                      Image<T> &out) {
  BIASASSERT(scalar != (T)0)

  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const T *inData = in.GetImageData();
  T *outData      = out.GetImageData();

  const unsigned int size =   in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = inData[i] / scalar;
  }
}


////////////////////////////////////////////////////////////////////////////////
// calculations on a single image                                             //
////////////////////////////////////////////////////////////////////////////////


template <class T>
void BIAS::ImageCalc<T>::Log(const Image<T> &in, Image<T> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const T *inData  = in.GetImageData();
  T *outData       = out.GetImageData();

  const unsigned int size = in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = T(log(double(inData[i])));
  }
}


template <class T>
void BIAS::ImageCalc<T>::Log10(const Image<T> &in, Image<T> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const T *inData  = in.GetImageData();
  T *outData       = out.GetImageData();

  const unsigned int size = in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = T(log10(double(inData[i])));
  }
}


// explicit instantiation
#define INSTANCE_ImageCalc(type)\
template class BIASImageUtilsBase_EXPORT ImageCalc<type>;

INSTANCE_ImageCalc(unsigned char)
INSTANCE_ImageCalc(float)
#ifdef BUILD_IMAGE_INT
  INSTANCE_ImageCalc(int)
#endif
#ifdef BUILD_IMAGE_CHAR
  INSTANCE_ImageCalc(char)
#endif
#ifdef BUILD_IMAGE_SHORT
  INSTANCE_ImageCalc(short)
#endif
#ifdef BUILD_IMAGE_USHORT
  INSTANCE_ImageCalc(unsigned short)
#endif
#ifdef BUILD_IMAGE_DOUBLE
  INSTANCE_ImageCalc(double)
#endif
#ifdef BUILD_IMAGE_UINT
  INSTANCE_ImageCalc(unsigned int)
#endif
