/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Utils/Param.hh>
#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>

using namespace std;
using namespace BIAS;


namespace BIAS {

  /**
   * @brief Performs pixelwise calculations on images.
   *
   * @note Over- and underflows are not prevented. You may want to convert
   *       your 1-byte images (like char or unsigned char) to float.
   *
   * @note The output image is always initialised.
   *
   * Possible improvements:
   * - avoid over- and underflows by clipping to maximum or minimum value
   * - generalise for n input images
   * - implement for other storage types
   *
   * @author rwulff 03/2010
   */
  template<class T>
  class ImageCalc {

  public:

    ////////////////////////////////////////////////////////////////////////////
    // calculations between two images                                        //
    ////////////////////////////////////////////////////////////////////////////


    /**
     * @brief Computes the pixelwise sum of the input images.
     */
    static void Add(const Image<T> &in1, const Image<T> &in2,
                    Image<T> &out);


    /**
     * @brief Computes the pixelwise difference of the input images. The second
     *        image is subtracted from the first one.
     */
    static void Subtract(const Image<T> &in1, const Image<T> &in2,
                         Image<T> &out);


    /**
     * @brief Computes the pixelwise product of the input images.
     */
    static void Multiply(const Image<T> &in1, const Image<T> &in2,
                         Image<T> &out);


    /**
     * @brief Computes the pixelwise quotient of the input images. The first
     *        image is divided by the second one.
     *
     * @note In case of division by zero the output image's pixel will be set
     *       to 0 and a BIASWARN will be printed to stdout
     */
    static void Divide(const Image<T> &in1, const Image<T> &in2,
                       Image<T> &out);


    ////////////////////////////////////////////////////////////////////////////
    // calculations between one image and a scalar                            //
    ////////////////////////////////////////////////////////////////////////////


    /**
     * @brief Adds the scalar value to each pixel of the input image.
     */
    static void AddScalar(const Image<T> &in, T scalar,
                          Image<T> &out);


    /**
     * @brief Subtracts the scalar value from each pixel of the input image.
     */
    static void SubtractScalar(const Image<T> &in, T scalar,
                               Image<T> &out);


    /**
     * @brief Multiplies the scalar value to each pixel of the input image.
     */
    static void MultiplyScalar(const Image<T> &in, T scalar,
                               Image<T> &out);


    /**
     * @brief Divides each pixel from the input image by the scalar.
     *
     * @note The scalar must not be 0!
     */
    static void DivideScalar(const Image<T> &in, T scalar,
                             Image<T> &out);


    ////////////////////////////////////////////////////////////////////////////
    // calculations on a single image                                         //
    ////////////////////////////////////////////////////////////////////////////


    /**
     * @brief Computes the natural logarithm (base e) for each pixel.
     *
     * @note Each pixel's value will be cast to double for the computation and
     *       then back to it's original storage type.
     */
    static void Log(const Image<T> &in, Image<T> &out);


    /**
     * @brief Computes the common logarithm (base 10) for each pixel.
     *
     * @note Each pixel's value will be cast to double for the computation and
     *       then back to it's original storage type.
     */
    static void Log10(const Image<T> &in, Image<T> &out);

  };
}