/* This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel

   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.

   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


#include <Base/Common/W32Compat.hh>
#include "ImageDraw.hh"
#include <cmath>
#include <Base/ImageUtils/Bresenham.hh>
#include <Base/ImageUtils/BresenhamCircle.hh>
#include <Base/ImageUtils/BresenhamCircleEighth.hh>
#include <Base/Image/ColourRGB.hh>

#ifdef BIAS_HAVE_OPENCV
#  include <Base/Image/WrapBias2Ipl.hh>
#endif

using namespace BIAS;
using namespace std;

namespace BIAS {

  template <class StorageType>
  StorageType ImageDraw<StorageType>::
  GetContrastValue(const StorageType t)
  {
    BIASERR("Dont know what good contrast for "<<(double)t
            <<" would be for StorageType "
            <<PRINTTYPE(StorageType));
    StorageType thresh ;
    if (numeric_limits<StorageType>::is_signed){
      thresh = numeric_limits<StorageType>::min();
    } else {
      thresh = (StorageType)(( numeric_limits<StorageType>::max()-
                               numeric_limits<StorageType>::min() )/(StorageType)2.0);
    }
    return (t>thresh)?(numeric_limits<StorageType>::min()):
      (numeric_limits<StorageType>::max());
  }

  template <>
  float ImageDraw<float>::
  GetContrastValue(const float t)
  {
    return float((t>0.0)?-1.0:1.0);
  }

  template <>
  unsigned char ImageDraw<unsigned char>::
  GetContrastValue(const unsigned char t)
  {
    return (unsigned char)((t > 127) ? 0 : 255);
  }

  ////////////////////////////////////////////////////////////////////
  // rectangles
  ////////////////////////////////////////////////////////////////////

  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCorners(Image<StorageType>& im, const int aminx, const int aminy,
                   const  int amaxx, const  int amaxy,
                   const StorageType Value[])
  {
    const int minx = (aminx<amaxx)?(aminx):(amaxx);
    const int maxx = (amaxx>aminx)?(amaxx):(aminx);
    const int miny = (aminy<amaxy)?(aminy):(amaxy);
    const int maxy = (amaxy>aminy)?(amaxy):(aminy);
    const int w = im.GetWidth(), h=im.GetHeight(),
      cc = im.GetChannelCount();
    register StorageType **ida = im.GetImageDataArray();
    register int x, y, c, xp1, xp2;

    // top and bottom lines
    for (x = minx; x <= maxx; x++){
      if (x>=0 && x<w){
        xp1 = x * im.GetChannelCount();
        for (c=0; c<cc; c++){
          if (miny>=0 && miny<h) { ida[miny][xp1+c]=Value[c]; }
          if (maxy>=0 && maxy<h) { ida[maxy][xp1+c]=Value[c]; }
        }
      }
    }
    // left and right lines
    xp1=minx*im.GetChannelCount();
    xp2=maxx*im.GetChannelCount();
    for (y = miny; y <= maxy; y++){
      if (y>=0 && y<h) {
        for (c=0; c<cc; c++){
          if (minx>=0 && minx<w) { ida[y][xp1+c]=Value[c]; }
          if (maxx>=0 && maxx<w) { ida[y][xp2+c]=Value[c]; }
        }
      }
    }
    return (minx>=0 && miny>=0 && maxx<w && maxy<h)?(0):(-1);
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCorners(Image<StorageType>& im, const int minx, const int miny,
                   const int maxx, const int maxy, const StorageType value)
  {
    StorageType *val = new StorageType[im.GetChannelCount()];
    for (register int i=0; i<(int)im.GetChannelCount(); i++)
      val[i]=value;
    int res = RectangleCorners(im, minx, miny, maxx, maxy, val);
    delete[] val;
    return res;
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCorners(Image<StorageType>& im, const int aminx,
                   const int aminy, const int amaxx, const int amaxy)
  {
    const int minx = (aminx<amaxx)?(aminx):(amaxx);
    const int maxx = (amaxx>aminx)?(amaxx):(aminx);
    const int miny = (aminy<amaxy)?(aminy):(amaxy);
    const int maxy = (amaxy>aminy)?(amaxy):(aminy);
    const int w = im.GetWidth(), h=im.GetHeight(),
      cc = im.GetChannelCount();
    register StorageType **ida = im.GetImageDataArray();
    register int x, y, c, xp1, xp2;

    // top and bottom lines
    for (x = minx; x <= maxx; x++){
      if (x>=0 && x<w){
        xp1 = x * cc;
        for (c=0; c< cc; c++){
          if (miny>=0 && miny<h) {
            ida[miny][xp1+c]=GetContrastValue(ida[miny][xp1+c]);
          }
          if (maxy>=0 && maxy<h) {
            ida[maxy][xp1+c]=GetContrastValue(ida[maxy][xp1+c]);
          }
        }
      }
    }
    // left and right lines
    xp1=minx*im.GetChannelCount();
    xp2=maxx*im.GetChannelCount();
    for (y = miny; y <= maxy; y++){
      if (y>=0 && y<h) {
        for (c=0; c< cc; c++){
          if (minx>=0 && minx<w) {
            ida[y][xp1+c]=GetContrastValue(ida[y][xp1+c]);
          }
          if (maxx>=0 && maxx<w) {
            ida[y][xp2+c]=GetContrastValue(ida[y][xp2+c]);
          }
        }
      }
    }
    return (minx>=0 && miny>=0 && maxx<w && maxy<h)?(0):(-1);
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCornersGrey(Image<StorageType>& im, const int aminx,
                       const int aminy, const int amaxx,
                       const int amaxy, const StorageType GreyValue)
  {
    if (im.GetChannelCount()!=1){
      BEXCEPTION("ImageDraw::RectangleCornersGrey() is only for single "
                 <<"channel images");
    }
    const int minx = (aminx<amaxx)?(aminx):(amaxx);
    const int maxx = (amaxx>aminx)?(amaxx):(aminx);
    const int miny = (aminy<amaxy)?(aminy):(amaxy);
    const int maxy = (amaxy>aminy)?(amaxy):(aminy);
    const int w = im.GetWidth(), h=im.GetHeight();
    register StorageType **ida = im.GetImageDataArray();
    register int x, y;

    // top and bottom lines
    for (x = minx; x <= maxx; x++){
      if (x>=0 && x<w){
        if (miny>=0 && miny<h) { ida[miny][x]=GreyValue; }
        if (maxy>=0 && maxy<h) { ida[maxy][x]=GreyValue; }
      }
    }
    // left and right lines
    for (y = miny; y <= maxy; y++){
      if (y>=0 && y<h) {
        if (minx>=0 && minx<w) { ida[y][x]=GreyValue; }
        if (maxx>=0 && maxx<w) { ida[y][x]=GreyValue; }
      }
    }
    return (minx>=0 && miny>=0 && maxx<w && maxy<h)?(0):(-1);
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCornersGreyFill(Image<StorageType>& im, const int aminx,
                           const int aminy, const int amaxx,
                           const int amaxy,const StorageType GreyValue)
  {
    const int minx = (aminx<amaxx)?(aminx):(amaxx);
    const int maxx = (amaxx>aminx)?(amaxx):(aminx);
    const int miny = (aminy<amaxy)?(aminy):(amaxy);
    const int maxy = (amaxy>aminy)?(amaxy):(aminy);
    const int w = im.GetWidth(), h=im.GetHeight();
    const int mmaxx = (maxx>w)?(w):(maxx);
    const int mmaxy = (maxy>h)?(h):(maxy);

    register StorageType **ida = im.GetImageDataArray();
    register int x, y;

    for (y=miny; y<mmaxy; y++){
      for (x=minx; x<mmaxx; x++){
        ida[y][x] = GreyValue;
      }
    }

    return (minx>=0 && miny>=0 && maxx<w && maxy<h)?(0):(-1);
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCenter(Image<StorageType>& im, const int X,
                  const int Y, const int Size,
                  const StorageType Value[])
  {
    const int Halfsize = Size >> 1;
    if (im.GetChannelCount()==1){
      return RectangleCornersGrey(im, (X>Halfsize)?X-Halfsize:0,
                                  (Y>Halfsize)?Y-Halfsize:0,
                                  (X<(int)im.GetWidth()-Halfsize)?
                                  (X+Halfsize):((int)im.GetWidth()-1),
                                  (Y<(int)im.GetHeight()-Halfsize)?
                                  (Y+Halfsize):((int)im.GetHeight()-1),
                                  Value[0]);
    } else {
      return RectangleCorners(im, (X>Halfsize)?X-Halfsize:0,
                              (Y>Halfsize)?Y-Halfsize:0,
                              (X<(int)im.GetWidth()-Halfsize)?
                              (X+Halfsize):((int)im.GetWidth()-1),
                              (Y<(int)im.GetHeight()-Halfsize)?
                              (Y+Halfsize):((int)im.GetHeight()-1),
                              Value);
    }
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCenterGrey(Image<StorageType>& im, const int X,
                      const int Y, const int Size,
                      const StorageType Value)
  {
    const int Halfsize = Size >> 1;
    return RectangleCornersGrey(im, (X>Halfsize)?X-Halfsize:0,
                                (Y>Halfsize)?Y-Halfsize:0,
                                (X<(int)im.GetWidth()-Halfsize)?
                                (X+Halfsize):((int)im.GetWidth()-1),
                                (Y<(int)im.GetHeight()-Halfsize)?
                                (Y+Halfsize):((int)im.GetHeight()-1),
                                Value);
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCenterGreyFill(Image<StorageType>& im, const int X,
                          const int Y, const int Size,
                          const StorageType Value)
  {
    const int Halfsize = Size >> 1;
    return RectangleCornersGreyFill(im, (X>Halfsize)?X-Halfsize:0,
                                    (Y>Halfsize)?Y-Halfsize:0,
                                    (X<(int)im.GetWidth()-Halfsize)?
                                    (X+Halfsize):((int)im.GetWidth()-1),
                                    (Y<(int)im.GetHeight()-Halfsize)?
                                    (Y+Halfsize):((int)im.GetHeight()-1),
                                    Value);
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCenter(Image<StorageType>& im, const int x,
                  const int y, const int size)
  {
    const int Halfsize = size >> 1;
    return RectangleCorners(im, (x>Halfsize)?x-Halfsize:0,
                            (y>Halfsize)?y-Halfsize:0,
                            (x<(int)im.GetWidth()-Halfsize)?
                            (x+Halfsize):((int)im.GetWidth()-1),
                            (y<(int)im.GetHeight()-Halfsize)?
                            (y+Halfsize):((int)im.GetHeight()-1));
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  RectangleCenter(Image<StorageType>& im, const int coo[2],
                  const int size)
  { return RectangleCenter(im, coo[0], coo[1], size); }



  /////////////////////////////////////////////////////////////////////////
  // lines
  /////////////////////////////////////////////////////////////////////////

  template <class StorageType>
  int ImageDraw<StorageType>::
  Line(Image<StorageType>& im, const int start[2],
       const int end[2], const StorageType value[])
  {
    int res = 0;
    Bresenham bres(start, end);
    int next[2];
    next[0] = start[0];
    next[1] = start[1];
    const int w = im.GetWidth(), h = im.GetHeight();
    const unsigned cc = im.GetChannelCount();
    unsigned xcc, c;
    StorageType **ida = im.GetImageDataArray();
    do {
      if (next[0]<0 || next[0]>=w || next[1]<0 || next[1]>=h){
        res = -1;
        continue;
      }
      xcc = next[0]*cc;
      for (c=0; c<im.GetChannelCount(); c++)
        ida[next[1]][xcc+c] = value[c];
    } while (bres.GetNext(next));

    if (res!=0){
      BIASWARNONCE("ImageDraw<StorageType>::Line(): Line (partially) out of image");
    }
    return res;
  }


  template <class StorageType>
  int ImageDraw<StorageType>::
  Line(Image<StorageType>& im, const unsigned int start[2],
       const unsigned int end[2], const StorageType value[])
  {
    // check if conversioon from unsiged to int is valid
    if (start[0]>=(unsigned)numeric_limits<int>::max() ||
        start[1]>=(unsigned)numeric_limits<int>::max() ||
        end[0]>=(unsigned)numeric_limits<int>::max() ||
        end[1]>=(unsigned)numeric_limits<int>::max() ) {
      BIASERR("start or end value tro big");
      return -2;
    }
    int istart[] = { start[0], start[1] }, iend[] = { end[0], end[1] };
    return Line(im, istart, iend, value);
  }


  template <class StorageType>
  int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                   const int start[2],
                                   const int end[2],
                                   const StorageType value[],
                                   int width)
  {
    width -= 1;
    float m = fabs(static_cast<float>(end[1]-start[1]))/fabs(static_cast<float>(end[0]-start[0]));
//     cout<<" m "<<m<<endl;
    int offset[2] = {0,0};
    if(m<=1.0)
      offset[1] = 1;
    else
      offset[0] = 1;


    int startUsed[2];
    int endUsed[2];

    int next[2];
    for(int i=-width; i<=width; i++) {

      startUsed[0] = start[0]+ i*offset[0];
      startUsed[1] = start[1]+ i*offset[1];
      endUsed[0] = end[0]+ i*offset[0];
      endUsed[1] = end[1]+ i*offset[1];
//       cout<<" start "<<startUsed[0]<<", "<<startUsed[1]<<endl;
//        cout<<" end "<<endUsed[0]<<", "<<endUsed[1]<<endl;
      Bresenham bres(startUsed, endUsed);

      next[0]=startUsed[0];
      next[1]=startUsed[1];

      StorageType **ida=im.GetImageDataArray();
      do {
        if (!(next[0]<0 || next[0]>=int(im.GetWidth()) ||
              next[1]<0 || next[1]>=int(im.GetHeight()))) {
          for (unsigned int c=0; c<im.GetChannelCount(); c++)
          ida[next[1]][next[0]*im.GetChannelCount()+c]=value[c];
        }
      } while (bres.GetNext(next));
    }
    return 0;
  }

  template <class StorageType>
  int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                   const unsigned int start[2],
                                   const unsigned int end[2], StorageType value)
  {
#ifdef BIAS_DEBUG
    if (start[0]<0 || start[0]>=im.GetWidth() ||
        start[1]<0 || start[1]>=im.GetHeight() ||
        end[0]<0 || end[0]>=im.GetWidth() ||
        end[1]<0 || end[1]>=im.GetHeight()){
      BIASERR("coordinates out of bounds "<<start[0]<<" "<<start[1]<<" "
              <<end[0]<<" "<<end[1]);
      BIASABORT;
    }
#endif
    unsigned int next[2];
    Bresenham bres(start, end);

    next[0]=start[0];
    next[1]=start[1];

    StorageType **ida=im.GetImageDataArray();
    do {
      for (unsigned int c=0; c<im.GetChannelCount(); c++)
        ida[next[1]][next[0]*im.GetChannelCount()+c]=value;
    } while (bres.GetNext(next));
    return 0;
  }

  template <class StorageType>
  int ImageDraw<StorageType>::LineGrey(Image<StorageType>& im,
                                       const unsigned int start[2],
                                       const unsigned int end[2],
                                       StorageType value)
  {
#ifdef BIAS_DEBUG
    if (start[0]<0 || start[0]>=im.GetWidth() ||
        start[1]<0 || start[1]>=im.GetHeight() ||
        end[0]<0 || end[0]>=im.GetWidth() ||
        end[1]<0 || end[1]>=im.GetHeight()){
      BIASERR("coordinates out of bounds");
      BIASABORT;
    }
#endif
    unsigned int next[2];
    Bresenham bres(start, end);

    next[0]=start[0];
    next[1]=start[1];

    StorageType **ida=im.GetImageDataArray();
    do {
      ida[next[1]][next[0]]=value;
    } while (bres.GetNext(next));
    return 0;
  }


  template <class StorageType>
  int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                   const unsigned int start[2],
                                   const unsigned int end[2])
  {
#ifdef BIAS_DEBUG
    if (start[0]<0 || start[0]>=im.GetWidth() ||
        start[1]<0 || start[1]>=im.GetHeight() ||
        end[0]<0 || end[0]>=im.GetWidth() ||
        end[1]<0 || end[1]>=im.GetHeight()){
      BIASERR("coordinates out of bounds: "<<start[0]<<";"<<start[1]
              <<" to "<<end[0]<<";"<<end[1]);
      BIASABORT;
    }
#endif
    unsigned int next[2];
    Bresenham bres(start, end);

    next[0]=start[0];
    next[1]=start[1];

    StorageType **ida=im.GetImageDataArray();
    do {
      for (unsigned int c=0; c<im.GetChannelCount(); c++)
        ida[next[1]][next[0]*im.GetChannelCount()+c]=
          GetContrastValue(ida[next[1]][next[0]*im.GetChannelCount()+c]);
    } while (bres.GetNext(next));
    return 0;
  }


  template <class StorageType>
  int ImageDraw<StorageType>::LineGrey(Image<StorageType>& im,
                                       const unsigned int start[2],
                                       const unsigned int end[2])
  {
#ifdef BIAS_DEBUG
    if (start[0]<0 || start[0]>=im.GetWidth() ||
        start[1]<0 || start[1]>=im.GetHeight() ||
        end[0]<0 || end[0]>=im.GetWidth() ||
        end[1]<0 || end[1]>=im.GetHeight()){
      BIASERR("coordinates out of bounds");
      BIASABORT;
    }
    if (im.GetChannelCount() !=1){
      BIASERR("only for one channel images");
      BIASABORT;
    }
#endif
    unsigned int next[2];
    Bresenham bres(start, end);

    next[0]=start[0];
    next[1]=start[1];

    StorageType **ida=im.GetImageDataArray();
    do {
      ida[next[1]][next[0]]=GetContrastValue(ida[next[1]][next[0]]);
    } while (bres.GetNext(next));
    return 0;
  }



  template <class StorageType>
  int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                   const unsigned int StartX,
                                   const unsigned int StartY,
                                   const unsigned int EndX,
                                   const unsigned int EndY,
                                   const StorageType Value[])
  {
    unsigned int start[]={StartX, StartY}, end[]={EndX, EndY};
    return Line(im, start, end, Value);
  }


  template <class StorageType>
  int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                   const unsigned int StartX,
                                   const unsigned int StartY,
                                   const unsigned int EndX,
                                   const unsigned int EndY,
                                   const StorageType Value)
  {
    unsigned int start[]={StartX, StartY}, end[]={EndX, EndY};
    return Line(im, start, end, Value);
  }

  template <class StorageType>
  int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                   const unsigned int StartX,
                                   const unsigned int StartY,
                                   const unsigned int EndX,
                                   const unsigned int EndY)
  {
    unsigned int start[]={StartX, StartY}, end[]={EndX, EndY};
    return Line(im, start, end);
  }



  template <class StorageType>
  int ImageDraw<StorageType>::LineGrey(Image<StorageType>& im,
                                       const unsigned int StartX,
                                       const unsigned int StartY,
                                       const unsigned int EndX,
                                       const unsigned int EndY,
                                       const StorageType Value)
  {
    unsigned int start[]={StartX, StartY}, end[]={EndX, EndY};
    return LineGrey(im, start, end, Value);
  }


#define BIAS_DRAW_CROSS_PRODUCT(a, b, res)      \
  res[0] = a[1] * b[2] - a[2] * b[1];           \
  res[1] = a[2] * b[0] - a[0] * b[2];           \
  res[2] = a[0] * b[1] - a[1] * b[0];

#define BIAS_DRAW_PRECISION 1e-8

  template <class StorageType>
  int ImageDraw<StorageType>::
  Line(Image<StorageType>& im, const double start[2],
       const double end[2],
       const StorageType value[])
  {
    int w=im.GetWidth();
    int h=im.GetHeight();
    double wm=(double)(w-1);
    double hm=(double)(h-1);
    unsigned uistart[2], uiend[2];
    double s[]={start[0], start[1], 1.0}, e[]={end[0], end[1], 1.0};

    // the line is build by the cross product of the homogenous
    // start and end points
    double line[3];
    BIAS_DRAW_CROSS_PRODUCT(s, e, line);

    // now intersect the line with the image border lines
    // again by using the cross product
    double top[]={0, -1.0, 0.0}, bottom[]={0.0, -1.0, hm};
    double left[]={-1.0, 0.0, 0.0}, right[]={-1.0, 0.0, wm};
    double intersections[4][3]; // 0=left, 1=right, 2=top, 3=bottom
    BIAS_DRAW_CROSS_PRODUCT(line, left,   intersections[0]);
    BIAS_DRAW_CROSS_PRODUCT(line, right,  intersections[1]);
    BIAS_DRAW_CROSS_PRODUCT(line, top,    intersections[2]);
    BIAS_DRAW_CROSS_PRODUCT(line, bottom, intersections[3]);

    unsigned coo[4]={0,0,0,0};
    bool intersect[4];
    int intersectioncount=0;
    // deal with compiler precision
    for (unsigned int i=0; i<4; i++){
      if (fabs(intersections[i][0]) < BIAS_DRAW_PRECISION)
        intersections[i][0] = 0.0;
      if (fabs(intersections[i][1]) < BIAS_DRAW_PRECISION)
        intersections[i][1] = 0.0;
      if (fabs(intersections[i][0]-wm) < BIAS_DRAW_PRECISION)
        intersections[i][0] = wm;
      if (fabs(intersections[i][1]-hm) < BIAS_DRAW_PRECISION)
        intersections[i][1] = hm;
      if (fabs(intersections[i][2]) < BIAS_DRAW_PRECISION){
        intersections[i][2] = 0.0;
        intersect[i]=false;
      } else { // homogenize
        intersections[i][0]/=intersections[i][2];
        intersections[i][1]/=intersections[i][2];
        intersect[i]=true;
      }
    }

    // extract the intersection points into coo[]
    // left
    if (intersect[0] && intersections[0][1]>=0 && intersections[0][1]<=hm){
      intersectioncount++;
      coo[0] = 0;
      coo[1] = (unsigned int)rint(intersections[0][1]);
    }
    // right
    if (intersect[1] && intersections[1][1]>=0 && intersections[1][1]<=hm){
      if (intersectioncount==0){
        coo[0] = w- 1;
        coo[1] = (unsigned int)rint(intersections[1][1]);
      } else {
        coo[2] = w - 1;
        coo[3] = (unsigned int)rint(intersections[1][1]);
      }
      intersectioncount++;
    }
    // top
    if (intersect[2] && intersections[2][0]>=0 && intersections[2][0]<=wm){
      switch (intersectioncount){
      case 0:
        coo[0] = (unsigned int)rint(intersections[2][0]);
        coo[1] = 0;
        break;
      case 1:
        coo[2] = (unsigned int)rint(intersections[2][0]);
        coo[3] = 0;
        break;
      default:
        // do nothing, all intersections correctly calulated
        // because left and right intersection can not have a common point
        // cout << "found intersections "<<coo[0]<<", "<<coo[1]<<"  and  "<<coo[2]<<", "<<coo[3]<<"\nnew intersection "<<(unsigned int)rint(intersections[2][0])<<", 0\n";
        break;
      }
      intersectioncount++;
    }
    // bottom
    if (intersect[3] && intersections[3][0]>=0 && intersections[3][0]<=wm){
      switch (intersectioncount){
      case 0:
        coo[0] = (unsigned int)rint(intersections[3][0]);
        coo[1] = h-1;
        break;
      case 1:
        coo[2] = (unsigned int)rint(intersections[3][0]);
        coo[3] = h-1;
        break;
      case 2:
        // this can only happen if upper left or upper right corner of image
        // is intersection
        // It follows that the top corner is always stored in coo[2/3], because
        // interscetioncount was 1 when checking for interscetions with top edge
        coo[0] = (unsigned int)rint(intersections[3][0]);
        coo[1] = h-1;
        break;
      default:
        // do nothing, all intersections correctly calulated
        // cout << "found intersections "<<coo[0]<<", "<<coo[1]<<"  and  "<<coo[2]<<", "<<coo[3]<<"\nnew intersection "<<(unsigned int)rint(intersections[2][0])<<", 0\n";
        break;
      }
      intersectioncount++;
    }

    if (intersectioncount>1){

      // now draw the line
      if (start[0]>=0.0 && start[0]<=wm && start[1]>=0.0 && start[1]<=hm){
        // easy, start[0] in image and start[1] in image
        uistart[0]=(unsigned)rint(start[0]);
        uistart[1]=(unsigned)rint(start[1]);
      } else {
        // replace start point

        // take coo[0/1] if (end-coo[0/1]).ScalarProduct(end-start)==1.0
        double dex = end[0]-coo[0], dey = end[1]-coo[1];
        double dx = end[0]-start[0], dy = end[1]-start[1];
        if (dx*dex+dy*dey>0.0){ // angle < 90 degree take coo[0/1]
          uistart[0]=coo[0];
          uistart[1]=coo[1];
        } else {
          uistart[0]=coo[2];
          uistart[1]=coo[3];
        }

      }
      if (start[0]>=0.0 && start[0]<=wm && start[1]>=0.0 && start[1]<=hm){
        // easy, end[0] in image and end[1] in image
        uiend[0]=(unsigned)rint(end[0]);
        uiend[1]=(unsigned)rint(end[1]);
      } else {
        // replace end point

        // take coo[0/1] if (start-coo[0/1]).ScalarProduct(start-end)==1.0
        double dex = start[0]-coo[0], dey = start[1]-coo[1];
        double dx = start[0]-end[0], dy = start[1]-end[1];
        if (dx*dex+dy*dey>0.0){ // angle < 90 degree take coo[0/1]
          uistart[0]=coo[0];
          uistart[1]=coo[1];
        } else {
          uistart[0]=coo[2];
          uistart[1]=coo[3];
        }
      }
      return Line(im, uistart, uiend, value);
    }

    //cerr << "nothing to draw\n";
    return 0;
  }

#undef BIAS_DRAW_CROSS_PRODUCT
#undef BIAS_DRAW_PRECISION

  ////////////////////////////////////////////////////
  // interpolated lines
  ///////////////////////////////////////////////////////

  template <class StorageType>
  int ImageDraw<StorageType>::InterpolatedLine(Image<StorageType>& im,
                                               const int StartX,
                                               const int StartY,
                                               const int EndX,
                                               const int EndY,
                                               const StorageType Value[],
                                               const float Thickness,
                                               const float Opacity)
  {
    return InterpolatedLine(im,HomgPoint2D((float)StartX,(float)StartY,1.0),
                                HomgPoint2D((float)EndX,(float)EndY,1.0), Value,
                                Thickness,Opacity);
  }


  template <class StorageType>
  int ImageDraw<StorageType>::InterpolatedLine(Image<StorageType>& im,
                                               const HomgPoint2D Start,
                                               const HomgPoint2D End,
                                               const StorageType Value[],
                                               const float Thickness,
                                               const float Opacity)
  {


    StorageType **ida = im.GetImageDataArray();
    unsigned int cc=im.GetChannelCount();
    float opac = min(fabs(Opacity),1.0f);
    int returnValue = 0;

    // Set left and right line point for a better handling
    float LeftX, RightX, LeftY,RightY;

    if (Start[0] < End[0]){
      LeftX = (float)Start[0]; LeftY = (float)Start[1];
      RightX = (float)End[0]; RightY = (float)End[1];
    } else {
      LeftX = (float)End[0]; LeftY = (float)End[1];
      RightX = (float)Start[0]; RightY = (float)Start[1];
    }

    // Set the borders the loop has to clip to
    float MinY = (float)min(Start[1],End[1]) - Thickness;
    float MaxY = (float)max(Start[1],End[1]) + Thickness;

    //Set the the borders the rendering has to clip to:
    int ClipLeft = max((int)floor(LeftX-Thickness),0);
    int ClipRight = min((int)ceil(RightX+Thickness),(int)im.GetWidth()-1);

    float slope;
    if (RightX != LeftX)
       slope = (float)(RightY-LeftY) / (float)(RightX - LeftX);
    else
       slope = -10000.0;// This is a little bit dirty...

   float length = sqrt(((float)End[0]-(float)Start[0])*((float)End[0]-(float)Start[0])
                    +((float)End[1]-(float)Start[1])*((float)End[1]-(float)Start[1]));

    // First rotate the pixels by alpha (we just need y)
    float angle = atan2(RightY -  LeftY,RightX-LeftX);
    float sinAngle = sin(angle);
    float cosAngle = cos(angle);

    for (int xDrawPos = ClipLeft ; xDrawPos <= ClipRight ; xDrawPos++ ){
      // Set the y range to draw on
      float Y1 = LeftY +
      ((float)xDrawPos-LeftX-Thickness)*slope;
      float Y2 = LeftY +
      ((float)xDrawPos-LeftX+Thickness)*slope;
      float LowerY, UpperY;
      if (Y1 < Y2){
        LowerY = Y1 -Thickness;
        UpperY = Y2 + Thickness;
      } else {
        LowerY = Y2 -Thickness;
        UpperY = Y1 + Thickness;
      }


      LowerY = max(ceil(LowerY),ceil(MinY));
      UpperY = min(floor(UpperY),floor(MaxY));

      //Set the the borders the rendering has to clip to:
      int ClipUp = max((int)floor(LowerY-1),0);
      int ClipLow = min((int)ceil(UpperY)+1,(int)im.GetHeight()-1);


      for (int yDrawPos = ClipUp; yDrawPos <= ClipLow; yDrawPos++){
        // Rotate the point with the rotation that puts the line onto the y-axis:
        float xRotated = sinAngle*((float)yDrawPos-LeftY) + cosAngle*((float)xDrawPos-LeftX);
        float yRotated = -sinAngle*((float)xDrawPos-LeftX) + cosAngle*((float)yDrawPos-LeftY);
        float dist = fabs(yRotated);

         if (xRotated<0)
            dist = sqrt(yRotated*yRotated + xRotated*xRotated);

        if (fabs(xRotated)>=length)
          dist = sqrt(yRotated*yRotated + (fabs(xRotated)-length)*(fabs(xRotated)-length));

        if (dist < Thickness){
          dist = max(dist - Thickness + 1.0f,0.0f);
          float ratio = (1.0f - dist) * opac;
          for (unsigned int c = 0; c < cc; c++)
            ida[yDrawPos][xDrawPos*cc+c] = (StorageType)((1.0 - ratio)*(double)ida[yDrawPos][xDrawPos*cc+c] +
                                           ratio * (double)Value[c]);
        }
      }
    }

    return returnValue;
  }

  template <class StorageType>
  int ImageDraw<StorageType>::InterpolatedLineGrey(Image<StorageType>& im,
                                                   const HomgPoint2D Start,
                                                   const HomgPoint2D End,
                                                   const StorageType GreyValue,
                                                   const float Thickness ,
                                                   const float Opacity )
  {
    int channels = im.GetChannelCount();
    StorageType* color = new StorageType[channels];
    for (int c = 0; c < channels; c++)
      color[c] = GreyValue;
    int returnValue =  InterpolatedLine(im, Start,End,color,Thickness,Opacity);
    delete[] color;
    return returnValue;
  }

  template <class StorageType>
  int ImageDraw<StorageType>::InterpolatedLineGrey(Image<StorageType>& im,
                                                   const int StartX,
                                                   const int StartY,
                                                   const int EndX,
                                                   const int EndY,
                                                   const StorageType GreyValue,
                                                   const float Thickness,
                                                   const float Opacity)
  {
    return InterpolatedLineGrey(im, HomgPoint2D((float)StartX,(float)StartY,1.0),
                                     HomgPoint2D((float)EndX,(float)EndY,1.0),
                                     GreyValue,Thickness,Opacity);


  }

  ///////////////////////////////////////////////////////////
  // circles
  //////////////////////////////////////////////////////////

  template <class StorageType>
  int ImageDraw<StorageType>::CircleCenter(Image<StorageType>& im,
                                           unsigned int CenterX,
                                           unsigned int CenterY,
                                           unsigned int Radius,
                                           const StorageType Value[])
  {
    int res=0;
    BresenhamCircle bc;
    int center[2], next[]={0,0}, radius=(int)Radius;
    StorageType **ida=im.GetImageDataArray();
    unsigned int cc=im.GetChannelCount();
    const int height = (int)im.GetHeight();
    const int width = (int)im.GetWidth();

    register unsigned int c;

    center[0]=(int)CenterX;
    center[1]=(int)CenterY;

    bc.Init(center, radius);

    if (Value==NULL) {
      // no color supplied, use black and white dep. on background
      while (bc.GetNext(next)){
        if (next[1]>=0 && next[1]<height &&
            next[0]>=0 && next[0]<width) {
          for (c=0; c<cc; c++) {
            ida[next[1]][next[0]*cc+c]=
              GetContrastValue(ida[next[1]][next[0]*cc+c]);
          }
        }
      }
    } else {
      while (bc.GetNext(next)){
        if (next[1]>=0 && next[1]<height &&
            next[0]>=0 && next[0]<width) {
          for (c=0; c<cc; c++) {
            ida[next[1]][next[0]*cc+c]=Value[c];
          }
        }
      }
    }
    return res;
  }

  template <class StorageType>
  int ImageDraw<StorageType>::CircleCenterFilled(Image<StorageType>& im,
                                                 unsigned int CenterX,
                                                 unsigned int CenterY,
                                                 unsigned int Radius,
                                                 const StorageType Value[])
  {
    int res=0;
    BresenhamCircleEighth bc;
    int zerocenter[]={0,0};
    int center[2], next[2], radius=(int)Radius;
    unsigned int start[2], end[2];
    const int width = (int)im.GetWidth();
    const int height = (int)im.GetHeight();
    center[0]=(int)CenterX;
    center[1]=(int)CenterY;

    if (center[0]<=radius || center[1]<=radius ||
        center[0]+radius+1>=width || center[1]+radius+1>=height){
      return -1;
    }

    bc.Init(zerocenter, radius);

    while (bc.GetNext(next)){
      start[0]=center[0]+next[0];
      start[1]=center[1]+next[1];
      end[0]=center[0]-next[0];
      end[1]=center[1]+next[1];
      Line(im, start, end, Value);
      start[0]=center[0]+next[1];
      start[1]=center[1]+next[0];
      end[0]=center[0]-next[1];
      end[1]=center[1]+next[0];
      Line(im, start, end, Value);
      start[0]=center[0]+next[0];
      start[1]=center[1]-next[1];
      end[0]=center[0]-next[0];
      end[1]=center[1]-next[1];
      Line(im, start, end, Value);
      start[0]=center[0]+next[1];
      start[1]=center[1]-next[0];
      end[0]=center[0]-next[1];
      end[1]=center[1]-next[0];
      Line(im, start, end, Value);
    }

    return res;
  }


  template <class StorageType>
  int ImageDraw<StorageType>::InterpolatedCircleCenter(Image<StorageType>& im,
                                            HomgPoint2D Center,
                                            float Radius,
                                            const StorageType Value[],
                                            const float Thickness ,
                                            const float Opacity )
   {
     int res=0;
     StorageType **ida=im.GetImageDataArray();
     unsigned int cc=im.GetChannelCount();
     float opac = min(fabs(Opacity),1.0f);
     int leftBorder = (int)floor(Center[0] - Radius - Thickness);
     int rightBorder = (int)ceil(Center[0] + Radius+Thickness);
     int upperBorder = (int)floor(Center[1] - Radius - Thickness);
     int lowerBorder = (int)ceil(Center[1] + Radius + Thickness);

     // Clip to image borders
     leftBorder = max(0,leftBorder);
     rightBorder = min((int)im.GetWidth()-1,rightBorder);
     upperBorder = max(0,upperBorder);
     lowerBorder = min((int)im.GetHeight()-1,lowerBorder);

     for (int x = leftBorder; x <= rightBorder; x++)
       for (int y = upperBorder; y <= lowerBorder; y++){
         float dist = sqrt(((float)x-(float)Center[0])*((float)x-(float)Center[0])
                     +((float)y-(float)Center[1])*((float)y-(float)Center[1]));
         if (fabs(dist-Radius) < Thickness){
           dist = max(fabs(dist - Radius)- Thickness + 1.0f,0.0f);
           float ratio = (1.0f - dist) * opac;
           for (unsigned int c = 0; c < cc; c++)
              ida[y][x*cc+c] = (StorageType)((1.0 - ratio)*(double)ida[y][x*cc+c] +
                  ratio * (double)Value[c]);
         }
       }

     return res;
   }


  template <class StorageType>
  int ImageDraw<StorageType>::InterpolatedCircleCenter(Image<StorageType>& im,
                                            unsigned int CenterX,
                                            unsigned int CenterY,
                                            unsigned int Radius,
                                            const StorageType Value[],
                                            const float Thickness,
                                            const float Opacity)
   {
     return InterpolatedCircleCenter(im,HomgPoint2D(CenterX,CenterY,1.0),(float)Radius,Value,Thickness,Opacity);
   }


  template <class StorageType>
  int ImageDraw<StorageType>::Arrow(Image<StorageType>& im,
                                    const unsigned start[2],
                                    const unsigned mend[2],
                                    const unsigned length,
                                    const unsigned width,
                                    const StorageType value[])
  {
    int dx=(int)mend[0]-(int)start[0], dy=(int)mend[1]-(int)start[1];
    unsigned end[]={mend[0], mend[1]};
    double l = sqrt((double)dx*(double)dx+(double)dy*(double)dy);
    if (l<(double)length){
      unsigned num = (unsigned)ceil( (double)length / l );
      end[0] += dx * num;
      end[1] += dy * num;
    }

    float thick=2.0;

    //if (Line(im, start, end, value)!=0){
    //cout << "line: "<<start[0]<<"  "<<start[1]<<"  "<<end[0]<<"  "<<end[1]<<endl;
    if (Line(im, start[0], start[1], end[0], end[1], value, thick)!=0){
#ifdef BIAS_DEBUG
      BIASERR("error drawing line");
#endif
      return -1;
    }
    if (dx==0 && dy==0){
      // do not draw arrow, zero length and hence direction indetermined
      return -2;
    }

#ifdef BIAS_DEBUG
    if (sqrt(double(dx*dx+dy*dy))<length || sqrt(double(dx*dx+dy*dy))<width){
      BIASERR("funny arrow, head bigger than tail");
      // BIASABORT;
    }
#endif
    int next[2]={0,0};
    Bresenham bres(end, start), bres2;

    for (unsigned i=0; i<length; i++)
      bres.GetNext(next);

    int left[]={0,0}, right[]={0,0};
    int lend[2], rend[2];
    rend[0]=(int)next[0]-dy;
    rend[1]=(int)next[1]+dx;
    lend[0]=(int)next[0]+dy;
    lend[1]=(int)next[1]-dx;
    //cerr << "next: ("<<next[0]<<", "<<next[1]<<")  lend: ("<<lend[0]<<", "
    //     <<lend[1]<<")  rend: ("<<rend[0]<<", "<<rend[1]<<")\n";
    bres.Init(next, rend);
    bres2.Init(next, lend);
    for (unsigned i=0; i<width; i++){
      bres.GetNext(right);
      bres2.GetNext(left);
    }

    int res = 0;
    if (right[0]>=0 && right[1]<(int)im.GetWidth()){
      //cout << "right: "<<end[0]<<"  "<<end[1]<<"  "<<right[0]<<"  "<<right[1]<<endl;
      if (Line(im, end[0], end[1], (unsigned)(right[0]), (unsigned)(right[1]),
               value, thick)!=0){
        //if (Line(im, end, (unsigned *)right, value)!=0){
#ifdef BIAS_DEBUG
        BIASERR("error drawing line");
#endif
        return -1;
      }
    } else { res = -1; }

    if (left[0]>=0 && left[1]<(int)im.GetWidth()){
      //cout << "left: "<<end[0]<<"  "<<end[1]<<"  "<<left[0]<<"  "<<left[1]<<endl;
      if (Line(im, end[0], end[1], (unsigned)(left[0]), (unsigned)(left[1]),
               value, thick)!=0){
        //if (Line(im, end, (unsigned *)left, value)!=0){
#ifdef BIAS_DEBUG
        BIASERR("error drawing line");
#endif
        return -1;
      }
    } else { res = -1; }
    return res;
  }


  /// JW:
  template <class StorageType>
  int ImageDraw<StorageType>::Line(Image<StorageType>& im,
                                   const unsigned int StartX,
                                   const unsigned int StartY,
                                   const unsigned int EndX,
                                   const unsigned int EndY,
                                   const StorageType color[],
                                   const float thickness )
  {
#ifdef BIAS_HAVE_OPENCV
    int i_thickness=int(thickness);
    Line(im,
         StartX, StartY,
         EndX, EndY,
         ColourRGB<StorageType>(color[0],color[1], color[2]),
         i_thickness
         );
#else
    BIASERR("Please enable OpenCV to use Lines with thickness. (jw)");
    // workaround : draw multiple lines to get a thicker one:

    // direction of (original) line
    double dir[2]; //< direction od the line
    double ortho[2]; //< perpendicalur vector of line displacement for thickness
    dir[0] = EndX - StartX; //< dx
    dir[1] = EndY - StartY; //< dy

    // orthogonal offset is perpendicular to line direction:
    ortho[0] =  -dir[1];
    ortho[1] = dir[0]; //< y, x as perpendicular of dir
    double norm = sqrt(ortho[0]*ortho[0] + ortho[1]*ortho[1]);
    if (norm ==0) {
      ortho[0] = 1.0; ortho[1] = 0.0;
    } else {
      ortho[0] /= norm; // length one
      ortho[1] /= norm; // length one
    };

    // draw multiple parallel lines to emulate thickness
    // QUCKHACK. \todo implement Bresenham for thicked lines or even better with antialised lines. JW
    // \todo the line may actually be shifted for fractional part of thickness JW
    // it is not enogh to use 1.0 step...
    for (float l= float(-thickness/2.0); l <= float(thickness/2.0); l+=1.0) {
      //ImageDraw<StorageType::
      Line(im,
           (unsigned int)(StartX +l * ortho[0]),
           (unsigned int)(StartY +l * ortho[1]),

           (unsigned int)(EndX   +l * ortho[0] ),
           (unsigned int)(EndY   +l * ortho[1] ),
           color
           );
    };
#endif // OpenCV
    return 0;
  }

  template <class StorageType>
  int ImageDraw<StorageType>::Ellipse(Image<StorageType>& im,
                                      double center[2], double a[2],
                                      double b[2],
                                      const StorageType Value[])
  {
    // points on ellipse can be parametrized by angle t
    // (a*cos(t), b*sin(t)) if the ellipse-axes are parallel to coordinate axes

    const int w = im.GetWidth();
    const int h = im.GetHeight();
    StorageType **ida=im.GetImageDataArray();
    const int cc=im.GetChannelCount();

    double length_a = sqrt(a[0]*a[0]+a[1]*a[1]);
    double length_b = sqrt(b[0]*b[0]+b[1]*b[1]);

    // test for degenerate cases, where the ellipse
    // consists only of a single line
    if (length_a<1e-8){
      double start[2], end[2];
      start[0]=center[0]-b[0];
      start[1]=center[1]-b[1];
      end[0]=center[0]+b[0];
      end[1]=center[1]+b[1];
      return ImageDraw<StorageType>::Line(im, start, end, Value);
    }
    if (length_b<1e-8){
      double start[2], end[2];
      start[0]=center[0]-a[0];
      start[1]=center[1]-a[1];
      end[0]=center[0]+a[0];
      end[1]=center[1]+a[1];
      return ImageDraw<StorageType>::Line(im, start, end, Value);
    }

#ifdef BIAS_DEBUG
    // assume axes are perpendicular
    if ((a[0]*b[0]+a[1]*b[1])/(length_a*length_b)>1e-8){
      BIASERR("half axes of ellipse are not perpendicular: ("<<a[0]<<", "<<a[1]
              <<")   ("<<b[0]<<", "<<b[1]<<")\n");
      BIASASSERT((a[0]*b[0]+a[1]*b[1])/(length_a*length_b)<1e-8);
    }
#endif

    double dt;
    if (length_a>length_b){
      dt = atan2(1.0, length_a);
    } else {
      dt = atan2(1.0, length_b);
    }
    //cout << "dt: "<<dt<<endl;

    // angel of axis a with x axis
    double ang=atan2(a[1], a[0]);
    //cout << "ang: "<<ang*180.0/M_PI<<" deg"<<endl;
    // rotation with respect to coordinate system
    double ca=cos(ang), sa=sin(ang);

    double p[2], rp[2], ap[2];
    double t=0.0;
    int ix, iy;

    while (t<=M_PI_2){
      //cout << "t: "<<t<<endl;
      p[0] = length_a * cos(t);
      p[1] = length_b * sin(t);
      //cout << "p : "<<p[0]<<", "<<p[1]<<endl;

      ap[0] = p[0];
      ap[1] = p[1];
      // now rotate and translate the point
      rp[0] = ap[0]*ca - ap[1]*sa + center[0];
      rp[1] = ap[1]*ca + ap[0]*sa + center[1];
      //cout << "1 - rp : "<<rp[0]<<", "<<rp[1]<<endl;
      ix = (int)rint(rp[0]);
      iy = (int)rint(rp[1]);
      if (ix>=0 && ix<w  && iy>=0 && iy<h){
        for (int i=0; i<cc; i++)
          ida[iy][ix*cc+i]=Value[i];
      }

      // use 4-ways symmetry of ellipse
      ap[0] = -p[0];
      ap[1] = -p[1];
      // now rotate and translate the point
      rp[0] = ap[0]*ca - ap[1]*sa + center[0];
      rp[1] = ap[1]*ca + ap[0]*sa + center[1];
      //cout << "2 - rp : "<<rp[0]<<", "<<rp[1]<<endl;
      ix = (int)rint(rp[0]);
      iy = (int)rint(rp[1]);
      if (ix>=0 && ix<w  && iy>=0 && iy<h){
        for (int i=0; i<cc; i++)
          ida[iy][ix*cc+i]=Value[i];
      }

      // use 4-ways symmetry of ellipse
      ap[0] = p[0];
      ap[1] = -p[1];
      // now rotate and translate the point
      rp[0] = ap[0]*ca - ap[1]*sa + center[0];
      rp[1] = ap[1]*ca + ap[0]*sa + center[1];
      //cout << "3 - rp : "<<rp[0]<<", "<<rp[1]<<endl;
      ix = (int)rint(rp[0]);
      iy = (int)rint(rp[1]);
      if (ix>=0 && ix<w  && iy>=0 && iy<h){
        for (int i=0; i<cc; i++)
          ida[iy][ix*cc+i]=Value[i];
      }

      // use 4-ways symmetry of ellipse
      ap[0] = -p[0];
      ap[1] = p[1];
      // now rotate and translate the point
      rp[0] = ap[0]*ca - ap[1]*sa + center[0];
      rp[1] = ap[1]*ca + ap[0]*sa + center[1];
      //cout << "4 - rp : "<<rp[0]<<", "<<rp[1]<<endl;
      ix = (int)rint(rp[0]);
      iy = (int)rint(rp[1]);
      if (ix>=0 && ix<w  && iy>=0 && iy<h){
        for (int i=0; i<cc; i++)
          ida[iy][ix*cc+i]=Value[i];
      }

      t+=dt;
    }
    return 0;
  }


#ifdef BIAS_HAVE_OPENCV
  // JW
  template <class StorageType>
  void ImageDraw<StorageType>::Text(Image<StorageType>& dstImg,
                                    const std::string & message,
                                    const int & posX,
                                    const int & posY,
                                    const BIAS::ColourRGB<StorageType> &colorRGB,
                                    const int fontface,
                                    const double hscale,
                                    const double vscale,
                                    const double shear,
                                    const int thickness,
                                    const int linetype
                                    )
  {
    WrapBias2Ipl wrap( &dstImg );
    CvFont font;
    cvInitFont( &font,
                fontface,
                hscale,
                vscale,
                shear,
                thickness,
                linetype
                );
    cvPutText( wrap.p_imgIpl,
               message.c_str(),
               cvPoint(posX, posY),
               &font,
               cvScalar(colorRGB[0], colorRGB[1], colorRGB[2]) // no swap!
               );
  }

  // JW
  template <class StorageType>
  void ImageDraw<StorageType>::Circle(BIAS::Image<StorageType>& dstImg,
                                      const int & centerX,
                                      const int & centerY,
                                      const int & radius,
                                      const BIAS::ColourRGB<StorageType> &colorRGB,
                                      const int & thickness,
                                      const int & linetype,
                                      const int & shift
                                      )
  {
    WrapBias2Ipl wrap( &dstImg );
    cvCircle( wrap.p_imgIpl,
              cvPoint (centerX, centerY),
              radius,
              cvScalar(colorRGB[0], colorRGB[1], colorRGB[2] ),
              thickness,
              linetype,
              shift
              );
  }

  // JW
  template <class StorageType>
  void ImageDraw<StorageType>::Line(BIAS::Image<StorageType>& dstImg,
                                    const int & pt1X,
                                    const int & pt1Y,
                                    const int & pt2X,
                                    const int & pt2Y,
                                    const BIAS::ColourRGB<StorageType> &colorRGB,
                                    const int thickness,
                                    const int linetype,
                                    const int shift
                                    )
  {
    WrapBias2Ipl wrap( &dstImg );
    cvLine( wrap.p_imgIpl,
            cvPoint (pt1X, pt1Y),
            cvPoint (pt2X, pt2Y),
            cvScalar(colorRGB[0], colorRGB[1], colorRGB[2] ),
            thickness,
            linetype,
            shift
            );
  }

  template <class StorageType>
  void ImageDraw<StorageType>::
  Arrow(Image<StorageType>& im, const int startX,
        const int startY, const int endX, const int endY,
        const BIAS::ColourRGB<StorageType> &colorRGB,
        const int thickness)
  {
    WrapBias2Ipl wrap( &im );
    int dx=endX-startX, dy=endY-startY;
    if (dx==0 && dy==0){
      // do not draw arrow, zero length and hence direction indetermined
      return;
    }
    CvPoint start, end;
    start.x = startX;
    start.y = startY;
    end.x = endX;
    end.y = endY;

    CvScalar color = cvScalar(colorRGB[0], colorRGB[1], colorRGB[2]);
    double angle = atan2( (double)(start.y - end.y), (double)(start.x - end.x) );
    int shift = 0;

    // main line
    cvLine(wrap.p_imgIpl, start, end, color, thickness, CV_AA, shift);

    /* Now draw the tips of the arrow. I do some scaling so that the
     * tips look proportional to the main line of the arrow.   */
    start.x = (int) (end.x + 6 * cos(angle + M_PI / 4));
    start.y = (int) (end.y + 6 * sin(angle + M_PI / 4));
    cvLine(wrap.p_imgIpl, start, end, color, thickness, CV_AA, shift);

    start.x = (int) (end.x + 6 * cos(angle - M_PI / 4));
    start.y = (int) (end.y + 6 * sin(angle - M_PI / 4));
    cvLine(wrap.p_imgIpl, start, end, color, thickness, CV_AA, shift);

  }


  // JW
  template <class StorageType>
  void ImageDraw<StorageType>::Rectangle(BIAS::Image<StorageType>& dstImg,
                                         const int & pt1X,
                                         const int & pt1Y,
                                         const int & pt2X,
                                         const int & pt2Y,
                                         const BIAS::ColourRGB<StorageType> &colorRGB,
                                         const int thickness,
                                         const int linetype,
                                         const int shift
                                         )
  {
    WrapBias2Ipl wrap( &dstImg );
    cvRectangle(wrap.p_imgIpl,
                cvPoint (pt1X, pt1Y),
                cvPoint (pt2X, pt2Y),
                cvScalar(colorRGB[0], colorRGB[1], colorRGB[2] ),
                thickness,
                linetype,
                shift
                );
  }

  // JW
  template <class StorageType>
  void ImageDraw<StorageType>::Ellipse(BIAS::Image<StorageType>& dstImg,
                                       const int & centerX,
                                       const int & centerY,
                                       const int & axesX,
                                       const int & axesY,
                                       const double & angle,
                                       const double & start_angle,
                                       const double & end_angle,
                                       const BIAS::ColourRGB<StorageType> &colorRGB,
                                       const int & thickness,
                                       const int & linetype,
                                       const int & shift
                                       )
  {
    WrapBias2Ipl wrap( &dstImg );
    cvEllipse(wrap.p_imgIpl,
              cvPoint(centerX, centerY),
              cvSize (axesX, axesY),
              angle,
              start_angle,
              end_angle,
              cvScalar(colorRGB[0], colorRGB[1], colorRGB[2] ),
              thickness,
              linetype,
              shift );
  }


#endif //BIAS_HAVE_OPENCV

} // namespace BIAS


//
// solve explicit instantiation
//
#define INSTANCE_ImageDraw(type)\
template class BIASImageUtilsBase_EXPORT ImageDraw<type>;


// create instances
INSTANCE_ImageDraw(unsigned char)
INSTANCE_ImageDraw(float)
#ifdef BUILD_IMAGE_INT
INSTANCE_ImageDraw(int)
#endif
#ifdef BUILD_IMAGE_CHAR
INSTANCE_ImageDraw(char)
#endif
#ifdef BUILD_IMAGE_SHORT
INSTANCE_ImageDraw(short)
#endif
#if defined(BUILD_IMAGE_USHORT)
INSTANCE_ImageDraw(unsigned short)
#endif
#ifdef BUILD_IMAGE_DOUBLE
INSTANCE_ImageDraw(double)
#endif
#ifdef BUILD_IMAGE_UINT
INSTANCE_ImageDraw(unsigned int)
#endif
