#include<iostream>

#include <Base/Math/SparseUmfMatrix.hh>
#include <iostream>
#include <Base/Math/Random.hh>
#include <Base/Math/SparseMatrix.hh>
#include <MathAlgo/SVD.hh>

using namespace std;
using namespace BIAS;

SparseMatrix CreateMatrix();
SparseMatrix CreateIdentityMatrix();
void CreateUmfIdentity(SparseUmfMatrix& A);


/**
  @example ExampleSparseUmfMatrix2.cpp
  @class ExampleSparseUmfMatrix2
  @relates SparseMatrix, SparseUmfMatrix
  @brief   Example for sparse matrix with umfpack
  
  @see SparseMatrix.hh and umfpack userguide.pdf for more details on
       the sparse matrix methods

  @ingroup g_examples
  @author sedlazeck
  @date 06/2009
 */
int main(int argc, char* argv[])
{
	cout << "####################################################" << endl;
	cout << " SparseUmfMatrix example." << endl;
	cout << " Shows the usage of SparseUmfMatrix class together " << endl;
	cout << " with the SparseMatrix class. " << endl;
	cout << endl;

	cout << "---------------------------- Create SparseMatrix 'A'" << endl;
	// create a simple 4 by 4 matrix and print it
	SparseMatrix A = CreateMatrix();
	A.Print("A", cout);
	

	cout << "-------------------------- Get UMF representation" << endl;
	// We can always get a reference to the current SparseUmfMatrix structure.
	SparseUmfMatrix& A_umf = A.GetUmfRepresentation();

	// Getting the structure, will update the SparseUmfData to represent the same
	// matrix as the SparseMatrix does. However, we can start an update process
	// manually.
	if (!A.IsSparseUmfValid()){
		// ATTENTION: if usage of umf representation is disabled, so umf
		// representation is not updated.
		A.UpdateUmfRepresentation();
	}

	A_umf.Print(true);

	cout << "-------------------------------- Calculate B = A * A" << endl;
	// Now calculate B = A * A using the SparseMatrix interface
	SparseMatrix B = CreateIdentityMatrix();
	SparseUmfMatrix& B_umf = B.GetUmfRepresentation();

	A.Multiply(A, B);
	B.Print("B", cout);

	cout << "-------------------------- B in UMF representation" << endl;
	// Here we don't use 'B.GetUmfRepresentation()' to get the reference
	// to the SparseUmfMatrix, since we already have it. After the multiplication
	// is done, the data is synchronized.
	B_umf.Print(true);

	cout << "--------------- Change UMF representation to Identity" << endl;
	// We have the reference to the SparseUmfMatrix of A, so we can manipulate
	// the data outside of the SparseMatrix. In this example we change the
	// data to represent a 4 by 4 identity matrix.
  CreateUmfIdentity(A_umf);

  // Obviously the SparseMatrix A and its SparseUmfMatrix A_umf do represent
  // different data, but calling A.IsSparseUmfValid() will return true.
  // This is where care must be taken when changing umf data externally.
  // To synchronize both structures, we can now update the SparseMatrix
	A.UpdateFromUmfRepresentation();
	A.Print("A", cout);

	return 0;
}

SparseMatrix CreateMatrix()
{
	BIAS::Matrix<double> A(4, 4, MatrixIdentity);
	A[0][0] = 1;
	A[0][1] = 2;
	A[0][2] = 3;
	A[0][3] = 4;
	A[1][0] = 5;
	A[1][1] = 6;
	A[1][2] = 7;
	A[1][3] = 8;
	A[2][0] = 9;
	A[2][1] = 10;
	A[2][2] = 11;
	A[2][3] = 12;
	A[3][0] = 13;
	A[3][1] = 14;
	A[3][2] = 15;
	A[3][3] = 16;

	return SparseMatrix(A);
}

SparseMatrix CreateIdentityMatrix()
{
	BIAS::Matrix<double> A(4, 4, MatrixIdentity);
	return SparseMatrix(A);
}

void CreateUmfIdentity(SparseUmfMatrix& A)
{
	A.Reset(4, 4, 4, true);

	A.Entry(0, 0, 1.0);
	A.Entry(1, 1, 1.0);
	A.Entry(2, 2, 1.0);
	A.Entry(3, 3, 1.0);

	A.Compress();
}
