/*
 * SparseUmfMatrix.cpp
 *
 *  Created on: Aug 17, 2009
 *      Author: graefenstein
 */

#include <fstream>
#include <iostream>
#include <string>
#include <cstdlib>
#include "SparseUmfMatrix.hh"
#include <Base/Debug/Error.hh>

#include <Base/Common/BIASpragmaStart.hh>

#define MAX(x,y) ((x) > (y) ? (x) : (y))
#define MIN(x,y) ((x) < (y) ? (x) : (y))

using namespace BIAS;
using namespace std;

SparseUmfMatrix::SparseUmfMatrix() {
	m = n = nz = nzmax = 0;
	i = p = 0;
	x = 0;
}

SparseUmfMatrix::~SparseUmfMatrix() {
	if(i!=NULL) delete[] i;
	i = 0;
	if(p!=NULL) delete[] p;
	p = 0;
	if(x!=NULL) delete[] x;
	x = 0;
}

SparseUmfMatrix::SparseUmfMatrix(int m, int n, int nzmax, bool triplet) {
	this->m = m;
	this->n = n;
	this->nzmax = nzmax;
	this->nz = triplet ? 0 : -1;
	this->x = new double[nzmax];
	this->i = new int[nzmax];
	this->p = new int[triplet ? nzmax : n + 1];
}


SparseUmfMatrix& SparseUmfMatrix::operator=(const SparseUmfMatrix& S) {
 
  // copy members
  m = S.m;
  n = S.n;
  nzmax = S.nzmax;
  nz = S.nz;
  
  if(x!=NULL) delete[] x;
  if(i!=NULL) delete[] i;
  if(p!=NULL) delete[] p;
  
  // init and copy arrays
  x = new double[nzmax];
  i = new int[nzmax];
  int sizeP = nz;
  if(sizeP == -1) sizeP = n+1;
  p = new int[sizeP];
 
  for(int counter=0; counter < nzmax; counter++){
    x[counter] = S.x[counter];
    i[counter] = S.i[counter];
  }

  for(int counter = 0; counter < sizeP; counter++){
    p[counter] = S.p[counter];
  }
  
  return (*this);
}

double SparseUmfMatrix::CumulativeSum(int *p, int *c, int n) {
	if (!p || !c) {
		return (-1);
	}

	int nz = 0;
	double nz2 = 0; // same as nz but in double to avoid int overflow
	for (int i = 0; i < n; i++) {
		p[i] = nz;
		nz += c[i];
		nz2 += c[i];
		c[i] = p[i]; // also copy p[0..n-1] back into c[0..n-1]
	}

	p[n] = nz;
	return nz2; // return sum (c[0..n-1])
}

int SparseUmfMatrix::Scatter(SparseUmfMatrix& A, int j, double beta, int *w,
		double *x, int mark, SparseUmfMatrix& C, int nz) {
	if (!A.IsCompressed() || !C.IsCompressed()) {
		return -1;
	}

	for (int p = A.p[j]; p < A.p[j + 1]; p++) {
		int i = A.i[p];

		if (w[i] < mark) {
			w[i] = mark;
			C.i[nz++] = i;
			x[i] = beta * A.x[p];
		} else {
			x[i] += beta * A.x[p];
		}
	}
	return nz;
}

bool SparseUmfMatrix::Multiply(SparseUmfMatrix& B, SparseUmfMatrix& C) {
	if (!B.IsCompressed() || !IsCompressed()) {
		BIASERR("Multiplication cannot be performed on triplet form matrix!");
		//BIASABORT;
		return false;
	}
	if (this->n != B.m) {
		BIASERR("Unable to perform matrix multiplication. dimensions do not match.");
		//BIASABORT;
		return false;
	}

	// for better readability in this scope
	int Am = this->m;
	int An = this->n;
	int* Ap = this->p;

	int Cm = Am;
	int Cn = B.n;
	int Cnz = 0;

	// allocate workspace and fill with zeros
	int* w = new int[Cm];
	double* Cx = new double[Cm];
	fill_n(w, Cm, 0);
	fill_n(Cx, Cm, 0.0);

	// reset result
	C.Reset(Cm, Cn, Ap[An] + B.p[B.n], false);

	for (int j = 0; j < Cn; j++) {
		if (Cnz + Cm > C.nzmax) {
			C.Resize(C.nzmax * 2 + Cm);
		}

		C.p[j] = Cnz;
		for (int k = B.p[j]; k < B.p[j + 1]; k++) {
			Cnz = Scatter(*this, B.i[k], B.x[k], w, Cx, j + 1, C, Cnz);
		}
		for (int k = C.p[j]; k < Cnz; k++) {
			C.x[k] = Cx[C.i[k]];
		}
	}
	// store the number of zeros at C.p[C.n]
	// Then resize the matrix to get rid of the unused space
	C.p[C.n] = Cnz;
	C.Resize(Cnz);

	// release workspace and return
	delete[] w;
	delete[] Cx;
	return true;
}

bool SparseUmfMatrix::Transpose(SparseUmfMatrix& result) {
	if (!IsCompressed()) {
		BIASERR("Transpose cannot be performed on triplet form matrix!");
		//BIASABORT;
		return false;
	}

	// For better readability in this scope
	int Am = this->m;
	int An = this->n;
	int* Ap = this->p;
	int* Ai = this->i;
	double* Ax = this->x;
	int Anzmax = this->nzmax;

	// allocate result and workspace.
	result.Reset(An, Am, Anzmax, false);

	// allocate workspace
	int* w = new int[Am];
	fill_n(w, Am, 0);

	// Calculate how often a row occurs in this structure
	// (how much nonzero entries a row contains)
	for (int k = 0; k < Ap[An]; k++) {
		w[Ai[k]]++;
	}

	// Calculate the cumulative sum of the data in workspace
	// This creates the compressed columns for transposed result.
	CumulativeSum(result.p, w, Am);

	// Store the values of A at the transposed address in result.
	for (int j = 0; j < An; j++) {
		for (int k = Ap[j]; k < Ap[j + 1]; k++) {

			int q = w[Ai[k]]++;
			double v = Ax[k];

			result.i[q] = j;
			result.x[q] = v;
		}
	}

	// release workspace and return
	delete[] w;
	return true;
}

void SparseUmfMatrix::Print(bool brief) const {
	if (IsCompressed()) {
		cout << m << "-by-" << n << " Matrix In Compressed Column Form" << endl;
		cout << "NZMAX: " << nzmax << " NZ: " << p[n] << endl;
		if (p[n] <= 0) {
			return;
		}
		int count = 0;
		for (int j = 0; j < n; j++) {
			cout << "Column " << j << " locations " << p[j] << " to " << p[j + 1] - 1
					<< endl;
			for (int k = p[j]; k < p[j + 1]; k++) {
				cout << "    Row " << i[k] << " : " << x[k] << endl;
				if (brief && count > 20) {
					cout << "..." << endl;
					return;
				}
				count++;
			}
		}
	} else {
		cout << m << "-by-" << n << " Matrix In Triplet Form" << endl;
		cout << "NZMAX: " << nzmax << " NZ: " << nz << endl;
		cout << "i\tj\tx" << endl;
		for (int k = 0; k < nz; k++) {
			cout << i[k] << "\t" << p[k] << "\t" << x[k] << endl;
			if (brief && k > 20) {
				cout << "..." << endl;
				return;
			}
		}
	}
}

bool SparseUmfMatrix::Load(string path, SparseUmfMatrix& result) {
	const string SUPPORTEDHEADER =
			"%%MatrixMarket matrix coordinate real general";

	ifstream file;
	file.open(path.c_str());

	if (!file) {
		BIASERR("Unable to open file: " << path.c_str());
		//BIASABORT;
		return false;
	}

	// line buffer
	string line;

	// Read header and check if the format is supported
	std::getline(file, line);
	if (line != SUPPORTEDHEADER) {
		BIASERR("Tried to load unsupported 'MatrixMarket' storage format."
						<< "Supported format is: " << SUPPORTEDHEADER
						<< "But file contains  : " << line
						<< "See http://math.nist.gov/MatrixMarket/formats.html#MMformat");
		return 0;
	}

	// Skip all comments which start with '%'
	while (std::getline(file, line)) {
		if (line.substr(0, 1) != "%") {
			break;
		}
	}

	// first line after comments defines the matrix dimensions
	// resolve the definition
	if (sscanf(line.c_str(), "%d %d %d", &result.m, &result.n, &result.nzmax)
			!= 3) {
		BIASERR("Error reading matrix dimensions.");
		return 0;
	}


  if(result.x!=NULL) delete[] result.x;
  if(result.i!=NULL) delete[] result.i;
  if(result.p!=NULL) delete[] result.p;
	
	// allocate matrix result
	result.i = new int[result.nzmax];
	result.p = new int[result.nzmax];
	result.x = new double[result.nzmax];
	result.nz = 0;

	// read nz times an entry from file
	int posi, posj;
	double value;
	while (std::getline(file, line)) {
		if (sscanf(line.c_str(), "%d %d %lg", &posi, &posj, &value) != 3) {
			BIASERR("Error reading matrix entry.");
			return false;
		}
		if (!result.Entry(posi - 1, posj - 1, value)) {
			BIASERR("Error adding matrix entry.");
			return false;
		}
	}

	return true;
}

void SparseUmfMatrix::Reset(int m, int n, int nzmax, bool triplet) {
	// release old data
	if(p != NULL) delete[] p;
	if(i != NULL) delete[] i;
	if(x != NULL) delete[] x;

	// reinitialize
	this->m = m;
	this->n = n;
	this->nzmax = nzmax;
	this->nz = triplet ? 0 : -1;
	this->x = new double[nzmax];
	this->i = new int[nzmax];
	this->p = new int[triplet ? nzmax : n + 1];

	fill_n(x, nzmax, 0);
	fill_n(i, nzmax, 0);
	fill_n(p, triplet ? nzmax : n + 1, 0);
}

void SparseUmfMatrix::Resize(int newsize) {
	// allocate space with appropriate size.
	// if matrix is in compressed column form,
	// p must not be resized
	int* Tp = IsCompressed() ? 0 : new int[newsize];
	int* Ti = new int[newsize];
	double* Tx = new double[newsize];

	if (!IsCompressed()) {
		fill_n(Tp, nzmax, 0);
	}
	fill_n(Ti, newsize, 0);
	fill_n(Tx, newsize, 0);

	// copy old values
	for (int k = 0; k < MIN(nzmax, newsize); k++) {
		if (!IsCompressed()) {
			Tp[k] = p[k];
		}
		Ti[k] = i[k];
		Tx[k] = x[k];
	}

	// release old data
	if (!IsCompressed()){
		delete[] p;
	}
	delete[] i;
	delete[] x;

	// use new data instead
	if (!IsCompressed()) {
		p = Tp;
	}
	i = Ti;
	x = Tx;
	nzmax = newsize;
}

void SparseUmfMatrix::SetData(int m, int n, int nzmax, int nz, int*& p, int*& i, double*& x)
{
	this->m = m;
	this->n = n;
	this->nzmax = nzmax;
	this->nz = nz;

	delete[] this->p;
	delete[] this->i;
	delete[] this->x;

	this->p = p;
	this->i = i;
	this->x = x;
}

int SparseUmfMatrix::GetData(int*& p, int*& i, double*& x)
{
	p = this->p;
	i = this->i;
	x = this->x;

	if (!&p || !&i || !&x){
		return 0;
	}

	if(IsCompressed()){
		return this->p[this->n];
	} else {
		return nz;
	}
}

bool SparseUmfMatrix::Entry(int i, int j, double x) {
	if (IsCompressed()) {
		BIASERR("Entry to a compressed matrix is not supported!");
		//BIASABORT;
		return false;
	}

	// resize space, if space for nonzeros already full
	if (nz >= nzmax) {
		Resize(nzmax * 2);
	}

	// add entry and increment nz counter
	this->x[nz] = x;
	this->i[nz] = i;
	this->p[nz++] = j;

	// adjust matrix dimensions
	m = MAX(m, i+1);
	n = MAX(n, j+1);
	return true;
}

void SparseUmfMatrix::Compress() {
	if (IsCompressed()) {
		return;
	}

	// allocate workspace and initialize with 0
	int* w = new int[n];
	fill_n(w, n, 0);

	// allocate space for temporary result
	int* Cp = new int[n + 1];
	int* Ci = new int[nz];
	double* Cx = new double[nz];

	// count column occurance
	for (int k = 0; k < nz; k++) {
		w[p[k]]++;
	}

	// accumulate column occurance
	CumulativeSum(Cp, w, n);

	// set rowpointers and entries to result matrix
	for (int k = 0; k < nz; k++) {
		int j = w[p[k]]++;
		Ci[j] = i[k];
		Cx[j] = x[k];
	}

	// delete workspace and old data
	delete[] w;
	delete[] p;
	delete[] i;
	delete[] x;

	// use temporary result instead
	p = Cp;
	i = Ci;
	x = Cx;

	// matrix is now in compressed format
	nz = -1;
	

}

bool SparseUmfMatrix::IsCompressed() const {
	return nz < 0;
}

double SparseUmfMatrix::GetElement(int i, int j) const {
	if (i < 0 || i >= m || j < 0 || j >= n) {
		return 0;
	}

	if (IsCompressed()) {
		for (int k = p[j]; k < p[j + 1]; k++) {
			if (this->i[k] == i) {
				return x[k];
			}
		}
		return 0.0f;
	} else {
		for (int k = 0; k < nz; k++) {
			if (p[k] == j && this->i[k] == i) {
				return x[k];
			}
		}
	}
	return 0;
}

#include <Base/Common/BIASpragmaEnd.hh>
