/*
 * SparseUMFMatrix.hh
 *
 *  Created on: Aug 17, 2009
 *      Author: graefenstein
 */

#include <fstream>
#include <iostream>
#include <string>
#include <cstdlib>
#include "bias_config.h"

#ifndef SPARSEUMFMATRIX_HH_
#define SPARSEUMFMATRIX_HH_

namespace BIAS {
	  /**
	 * @brief		Sparse matrix in compressed column or in triplet form.
	 * @author	graefenstein
	 * @note  	The elements of this class may be used with the UMF Pack,
	 * 					but not the class itself.
	 *
	 *					example:
	 * 			 			Let A be a SparseUMFMatrix matrix. A call to the solve routine
	 * 						of the UMF Pack may look like this:
	 * 					  	'umfpack_di_solve(..., A->p, A->i, A->x, ...);'
	 *						Further A must be in compressed column form
	 *
	 * 					SparseUMFMatrix is in compressed column form when A->nz < 0,
	 * 					otherwise it is in triplet form and A->nz indicates
	 * 					the number of nonzero elements.
	 *
	 * @see	  	For more implementation details see
	 * 					'Direct Methods for Sparse Linear Systems Fundamentals of
	 * 					Algorithms'
	 */
	class BIASMathBase_EXPORT SparseUmfMatrix
	{
		int nzmax;	// maximum number of nonzero entries (allocated space)
		int m;			// number of rows
		int n;			// number of columns
		int* p;			//
		int* i;			//
		double* x;	// nonzero entries
		int nz;			// -1 when structure represents compressed column form
								// otherwise number of nonzeros in triple form

	public :
		/** @brief  Default constructor. This wont allocate the data arrays.
								Data arrays are set to null */
		SparseUmfMatrix();

		/** @brief  Destroys the data arrays. */
		~SparseUmfMatrix();

		/** @brief	Creates a new m-by-n sparse matrix with allocated space for
								'nzmax' nonzero entries
				@note		Creates a matrix in triplet form if 'triplet' is true,
								otherwise compressed column form matrix is created. The compressed
								form is preferred by most algorithms in this class. */
		SparseUmfMatrix(int m, int n, int nzmax, bool triplet);

		
    SparseUmfMatrix& operator=(const SparseUmfMatrix& S);
		
		/**	@brief	Calculates C = A * B, where A is the calling matrix.
				@return true, if calculation of A * B could be performed.
								false, if A or B is not in compressed column form or
								number of rows of B do not match the number columns of A. */
		bool Multiply(SparseUmfMatrix& B, SparseUmfMatrix& C);

		/** @brief	Transposes the rows and columns of a matrix
				@return true, if calculation could be performed.
								false, if the calling matrix is not in compressed column form. */
		bool Transpose(SparseUmfMatrix& result);

		/** @brief  Compresses the columns of a triple form matrix. Does nothing,
								if matrix is already in compressed column form. */
		void Compress();

		/** @brief	Loads a matrix from file.
				@return	true, if matrix could be loaded false otherwise.
				@note		The input file must contain a matrix in 'MatrixMarket' format.
								Loading is implemented only for real (not complex) matrices.
				@see		http://math.nist.gov/MatrixMarket/formats.html#MMformat */
		static bool Load(std::string path, SparseUmfMatrix& A);

		/** @brief	Prints the contents of a matrix to standard output.
				@note		If 'brief' is true, only first 20 entries are printed. */
		void Print(bool brief) const;

		/** @brief	Resets a matrix to the specified parameters
				@note		Old data will be deleted so all contents will be lost. */
		void Reset(int m, int n, int nzmax, bool triplet);

		/** @brief	Resizes the matrix to a specified size.
			  @note		This does not change the number of rows or columns,
								but reallocates space for nonzero entries. */
		void Resize(int newsize);

		/** @brief */
		void SetData(int m, int n, int nzmax, int nz, int*& p, int*& i, double*& x);

		/** @brief  Gets the data arrays.
				@return number of nonzero elements */
		int GetData(int*& p, int*& i, double*& x);

		/** @brief	Appends an entry to a triplet form matrix at position Aij.
							  Compressed column form is not supported here.
				@return true, if entry could be added, false otherwise.
				@note		This does not check if an entry at ij already exist. Adding an
								entry twice to same location, the matrix will contain
								two entries for same address. */
		bool Entry(int i, int j, double x);

		/** @brief	Calculates the cumulative sum of p
		            given the array c = {2, 3, 2, 1} of size n = 4, the result is
								p = {0, 2, 5, 7, 8} and c = {0, 2, 5, 7}, return value is 8
							 (mention the different sizes of p and c) 		 
				@param	- p: an array of size (n+1) where the cumulative sum is stored
								- c: an array of size (n) with values to sum up
								- n: number of elements in c.
	  */
		double CumulativeSum(int *p, int *c, int n);

		/** Helper function for multiply */
		int Scatter(SparseUmfMatrix& A, int j, double beta,
			int* w, double* x, int mark, SparseUmfMatrix& C, int nz);

		/** @brief  Checks whether this matrix is in compressed column form */
		bool IsCompressed() const;

		/** @brief	Gets an element it specified address.
			  @return Value at given address or 0 if address is invalid. */
		double GetElement(int i, int j) const;

		/** @brief	Gets a value indicating the number of rows of the matrix. */
		int GetRows() const { return m; }
		/** @brief	Gets a value indicating the number of columns of the matrix. */
		int GetCols() const { return n; }
	};
}
#endif /* SPARSEUMFMATRIX_HH_ */
