/* This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel
   
   
   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.
   
   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

/** @file TestSparseMatrix.cpp
  * @relates SparseMatrix, SparseUmfMatrix
    @brief Please add docu   
    @todo add docu
    @author MIP
*/

#include <iostream>
#include <Base/Math/Matrix.hh>
#include <Base/Math/SparseMatrix.hh>
#include <Base/Math/SparseUmfMatrix.hh>
#include <Base/Debug/TimeMeasure.hh>
#include <Base/Math/Random.hh>
#include <MathAlgo/SVD.hh>

using namespace BIAS;
using namespace std;

static bool DOCHECKDIMENSIONS = true;
static bool verbose = true;
#define PRINT(msg) if (verbose){ cout << msg << endl; }
#define ABS(value) ((value) < 0 ? -(value) : (value))

void CreateRandomDenseMatrix(Matrix<double>& dense, double threshold);
void CopyElements(Matrix<double>& A, Matrix<double>& B);
bool Check(string msg, Matrix<double>& A_dense, SparseMatrix& A_sparse, const SparseUmfMatrix& A_umf);
bool TestDenseToSparseToUmf();
bool TestUmfToSparseToDense();
bool TestTranspose(int m, int n, double threshold);
bool TestMultiply(int m, int n, double threshold);
bool TestAppend(int m, int n, double threshold);
bool TestSubmatrix(int m, int n, double threshold);
bool TestDiagonal(int m, int n, double threshold);
bool TestInvert();
bool TestPseudoInvert();
bool TestSolve();


int main(int argc, char *argv[])
{
	PRINT("############################################");
	bool ok = true;

	ok = ok && TestDenseToSparseToUmf();
	ok = ok && TestUmfToSparseToDense();

	ok = ok && TestTranspose(50, 50, 0.9);
	ok = ok && TestTranspose(100, 50, 0.9);

	ok = ok && TestMultiply(50, 50, 0.9);
	ok = ok && TestMultiply(100, 50, 0.9);

	ok = ok && TestAppend(50, 50, 0.9);
	ok = ok && TestAppend(100, 50, 0.9);

	ok = ok && TestSubmatrix(50, 50, 0.9);
	ok = ok && TestSubmatrix(100, 50, 0.9);

	ok = ok && TestDiagonal(50, 50, 0.9);

	ok = ok && TestSolve();
	ok = ok && TestInvert();

	ok = ok && TestPseudoInvert();
	
  return ok ? 0 : -1;
}

bool TestDenseToSparseToUmf()
{
	// Checks matrix conversion
	// from denste to sparse
	// from sparse to umf

	// create random dense matrix
	Matrix<double> A_dense(50, 50);
	CreateRandomDenseMatrix(A_dense, 0.9);

	// create sparse matrix from dense matrix
	SparseMatrix A_sparse(A_dense);

	// get umf representation. This updates the umf
	//representation from sparse matrix
	A_sparse.UpdateUmfRepresentation();
	const SparseUmfMatrix& A_umf = A_sparse.GetUmfRepresentation();

	// check all values
	return Check("DENSE TO SPARSE TO UMF CHECK", A_dense, A_sparse, A_umf);
}

bool TestUmfToSparseToDense()
{
	// Checks matrix conversion
	// from umf to sparse
	// from sparse to dense

	// allocate dense, sparse and umf matrix
	Matrix<double> A_dense(50, 50);
	SparseMatrix A_sparse = SparseMatrix(50, 50);
	A_sparse.UpdateUmfRepresentation();
	SparseUmfMatrix& A_umf = A_sparse.GetUmfRepresentation();

	// reset umf amtrix and fill with some values
	A_umf.Reset(4, 4, 4, true);
	double value = 1;
	for (int i = 0; i < 4; i++){
		for (int j = 0; j < 4; j++){
			A_umf.Entry(i, j, value++);
		}
	}
	// umf must be compressed before
	A_umf.Compress();

	// update sparse matrix from new umf matrix data
	A_sparse.UpdateFromUmfRepresentation();

	// get sparse matrix as dense matrix
	A_sparse.GetAsDense(0, 0, 4, 4, A_dense);

	// check all values
	return Check("UMF TO SPARSE TO DENSE CHECK", A_dense, A_sparse, A_umf);
}

bool TestAppend(int m, int n, double threshold)
{
	// Test: append matrix to right and to bottom
	PRINT("TEST APPEND MATRIX");
	PRINT("M: " << m << " N: " << n);

	// allocate dense matrix
	Matrix<double> A_dense(m, n);
	Matrix<double> WIDE_dense(m, n * 2);
	Matrix<double> HIGH_dense(m * 2, n);

	// create small random matrix
	CreateRandomDenseMatrix(A_dense, threshold);
	// create a wide matrix
	WIDE_dense.Set(0, 0, A_dense); // set left half of WIDE to A_
	WIDE_dense.Set(0, n, A_dense); // set right half of WIDE to A_
	// create a high matrix
	HIGH_dense.Set(0, 0, A_dense); // set top half of HIGH to A_
	HIGH_dense.Set(m, 0, A_dense); // set bottom half of HIGH to A_
	
	/************************* TEST append right ****************/
	SparseMatrix A_sparse = SparseMatrix(A_dense);
	SparseMatrix B_sparse = SparseMatrix(A_dense);
	A_sparse.AppendMatrixRight(B_sparse);
	A_sparse.UpdateUmfRepresentation();
	SparseUmfMatrix& A_umf = A_sparse.GetUmfRepresentation();
	if (!Check("Check after append matrix right", WIDE_dense, A_sparse, A_umf)){
		return false;
	}
	
	/************************* TEST append bottom ****************/
	B_sparse = SparseMatrix(A_dense);
	A_sparse = SparseMatrix(A_dense);
	
	A_sparse.UpdateUmfRepresentation();
	A_sparse.AppendMatrixBottom(B_sparse);
	
	A_sparse.UpdateUmfRepresentation();

  SparseUmfMatrix& A_umf2 = A_sparse.GetUmfRepresentation();

	if (!Check("Check after append matrix bottom", HIGH_dense, A_sparse, A_umf2)){
		return false;
	}
	
	return true;
}

bool TestSubmatrix(int m, int n, double threshold)
{
	// Test: set a sub matrix
	PRINT("TEST SUB MATRIX");
	PRINT("M: " << m << " N: " << n);

	// allocate dense matrix
	Matrix<double> A_dense(m, n);
	Matrix<double> BIG_dense(m * 2, n * 2);

	// create small random matrix
	CreateRandomDenseMatrix(A_dense, threshold);
	// create big matrix (twice as big as small matrix)
	BIG_dense.Set(0, 0, A_dense);
	BIG_dense.Set(m, n, A_dense);

	// create sparse matrix
	SparseMatrix A_sparse = SparseMatrix(0, 0);

	// set sub matrix
	A_sparse.SetSubMatrix(0, 0, A_dense, 0, 0, m, n);
	A_sparse.SetSubMatrix(m, n, A_dense, 0, 0, m, n);

	// get umf representation and check
	A_sparse.UpdateUmfRepresentation();
  SparseUmfMatrix& A_umf = A_sparse.GetUmfRepresentation();
	if (!Check("Check after set sub matrix", BIG_dense, A_sparse, A_umf)){
		return false;
	}

	return true;
}

bool TestTranspose(int m, int n, double threshold)
{
	// Test transpose algorithms of each matrix class
	PRINT("TEST TRANSPOSE");
	PRINT("M: " << m << " N: " << n);

	// create dense
	Matrix<double> A_dense(m, n);
	Matrix<double> B_dense(m, n);
	CreateRandomDenseMatrix(A_dense, threshold);
	CopyElements(A_dense, B_dense);

	// create sparse matrix
	SparseMatrix A_sparse(A_dense);
	SparseMatrix B_sparse(B_dense);

	// create umf matrix
	A_sparse.UpdateUmfRepresentation();
	B_sparse.UpdateUmfRepresentation();
	SparseUmfMatrix& A_umf = A_sparse.GetUmfRepresentation();
	SparseUmfMatrix& B_umf = B_sparse.GetUmfRepresentation();

	// disable the umf matrix at the sparse matrix objects, so sparse matrix
	// wont affect them now.

	if (!Check("Check after initialization", B_dense, B_sparse, B_umf)){
		return false;
	}

	/********************** TEST TRANSPOSE ***********************/
	// transpose dense
	B_dense = A_dense.Transpose();

	// transpose umf
	A_umf.Transpose(B_umf);

	// transpose sparse
	A_sparse.Transpose(B_sparse);

	if (!Check("Check after transpose", B_dense, B_sparse, B_umf)){
		return false;
	}

	return true;
}

bool TestMultiply(int m, int n, double threshold)
{
	// Test multiply algorithms of each matrix class
	PRINT("TEST MULTIPLY");
	PRINT("M: " << m << " N: " << n);

	// create dense
	Matrix<double> A_dense(m, n);
	Matrix<double> B_dense(n, m);
	Matrix<double> C_dense(m, m);
	CreateRandomDenseMatrix(A_dense, threshold);
	CreateRandomDenseMatrix(B_dense, threshold);

	// create sparse matrix
	SparseMatrix A_sparse(A_dense);
	SparseMatrix B_sparse(B_dense);
	SparseMatrix C_sparse(C_dense);

	// create umf matrix
	A_sparse.UpdateUmfRepresentation();
  B_sparse.UpdateUmfRepresentation();
	C_sparse.UpdateUmfRepresentation();
	SparseUmfMatrix& A_umf = A_sparse.GetUmfRepresentation();
	SparseUmfMatrix& B_umf = B_sparse.GetUmfRepresentation();
	SparseUmfMatrix& C_umf = C_sparse.GetUmfRepresentation();


	if (!Check("Check after initialization", B_dense, B_sparse, B_umf)){
		return false;
	}

	/********************** TEST MULTIPLY ***********************/

	// multiply dense
	A_dense.Mult(B_dense, C_dense);

	// multiply umf
	A_umf.Multiply(B_umf, C_umf);

	// multiply sparse
	A_sparse.Multiply(B_sparse, C_sparse);

	if (!Check("Check after multiplication", C_dense, C_sparse, C_umf)){
		return false;
	}

	return true;
}

bool TestDiagonal(int m, int n, double threshold)
{
	// Test add diagonal and multiply diagonal

	PRINT("TEST ADD TO DIAGONAL ");
	PRINT("M: " << m << " N: " << n);

	Matrix<double> A_dense(m, n);

	int value = 0;
	for (int i = 0; i < m; i++){
		for (int j = 0; j < n; j++){
			A_dense[i][j] = value;
			value++;
		}
	}

	SparseMatrix A_sparse = SparseMatrix(A_dense);
	A_sparse.UpdateUmfRepresentation();
	SparseUmfMatrix& A_umf = A_sparse.GetUmfRepresentation();

	/********************** TEST add to diagonal ***********************/
	for (int i = 0; i < (m < n ? m : n); i++){
		A_dense[i][i] = A_dense[i][i] + 10;
	}

	A_sparse.AddToDiagonal(10);
	A_sparse.UpdateUmfRepresentation();

	if (!Check("Check after add to diagonal", A_dense, A_sparse, A_umf)){
		return false;
	}

	/********************** TEST multiply diagonal by ***********************/
	for (int i = 0; i < (m < n ? m : n); i++){
		A_dense[i][i] = A_dense[i][i] * 10;
	}

	A_sparse.MultiplyDiagonalBy(10);
	A_sparse.UpdateUmfRepresentation();

	if (!Check("Check after multiply diagonal by", A_dense, A_sparse, A_umf)){
		return false;
	}

	return true;
}

bool TestInvert()
{
	int m = 5;
	int n = 5;
	PRINT("TEST INVERT");
	PRINT("M: " << m << " N: " << n);

	// create dense (hardcoded)
	Matrix<double> A_dense(m, n);
	Matrix<double> B_dense(n, m);
	Matrix<double> I_dense(m, m, MatrixIdentity);

	A_dense[0][0] = 2;
	A_dense[0][1] = -1;
	A_dense[0][2] = -5;
	A_dense[0][3] = -3;
	A_dense[0][4] = 5;

	A_dense[1][0] = -4;
	A_dense[1][1] = 2;
	A_dense[1][2] = -5;
	A_dense[1][3] = 1;
	A_dense[1][4] = -1;

	A_dense[2][0] = 3;
	A_dense[2][1] = -1;
	A_dense[2][2] = 2;
	A_dense[2][3] = -1;
	A_dense[2][4] = 4;

	A_dense[3][0] = 0;
	A_dense[3][1] = -1;
	A_dense[3][2] = 0;
	A_dense[3][3] = -2;
	A_dense[3][4] = -1;

	A_dense[4][0] = 2;
	A_dense[4][1] = 0;
	A_dense[4][2] = 2;
	A_dense[4][3] = 0;
	A_dense[4][4] = -2;

	// create sparse matrix
	SparseMatrix A_sparse(A_dense);
	SparseMatrix B_sparse(B_dense);
	SparseMatrix I_sparse(I_dense);

	// create umf matrix
	B_sparse.UpdateUmfRepresentation();
	I_sparse.UpdateUmfRepresentation();
	SparseUmfMatrix& B_umf = B_sparse.GetUmfRepresentation();
	SparseUmfMatrix& I_umf = I_sparse.GetUmfRepresentation();

	// invert to dense matrix
	A_sparse.Invert(B_dense);
	// invert to sparse matrix
	A_sparse.Invert(B_sparse);
	// update umf representation
	B_sparse.UpdateUmfRepresentation();

	if (!Check("Check after invert", B_dense, B_sparse, B_umf)){
		return false;
	}

	// multiply A*B where B is A^-1
	// result must be identity matrix
	A_sparse.Multiply(B_sparse, I_sparse);

	if (!Check("Check for identity", I_dense, I_sparse, I_umf)){
		return false;
	}

	return true;
}

bool TestPseudoInvert(){
  
  int m = 10;
  int n = 5;
  PRINT("TEST PSEUDO-INVERT");
  PRINT("M: " << m << " N: " << n);

  // create dense A
  Matrix<double> A_dense(m, n);

  Random randomizer;
  
  for(int i=0; i < m; i++){
    for(int j=0; j < n; j++){
      A_dense[i][j] = randomizer.GetUniformDistributed(-10.0, 10.0);
    }
  }
  
  // create sparse matrix
  SparseMatrix A_sparse(A_dense);

 
  // compute pseudo inverse of all of them
  Matrix<double> A_dense_inv;
  SVD svd;
  A_dense_inv = svd.Invert(A_dense);
  
  SparseMatrix A_sparse_inv(n, m);
  A_sparse.PseudoInverse(A_sparse_inv);

  A_sparse_inv.UpdateUmfRepresentation();
  SparseUmfMatrix& A_umf_inv = A_sparse_inv.GetUmfRepresentation();
  if (!Check("Check after pseudo-invert", A_dense_inv, A_sparse_inv, A_umf_inv)){
    return false;
  }

  return true;

}

bool TestSolve()
{
	// small linear test
	// solution must be [1;2;3;4;5]

	int m = 5;
	int n = 5;
	PRINT("TEST SOLVE");
	PRINT("M: " << m << " N: " << n);

	// create dense ( hardcoded)
	Matrix<double> A_dense(m, n);

	A_dense[0][0] = 2;
	A_dense[0][1] = 3;
	A_dense[0][2] = 0;
	A_dense[0][3] = 0;
	A_dense[0][4] = 0;

	A_dense[1][0] = 3;
	A_dense[1][1] = 0;
	A_dense[1][2] = 4;
	A_dense[1][3] = 0;
	A_dense[1][4] = 6;

	A_dense[2][0] = 0;
	A_dense[2][1] = -1;
	A_dense[2][2] = -3;
	A_dense[2][3] = 2;
	A_dense[2][4] = 0;

	A_dense[3][0] = 0;
	A_dense[3][1] = 0;
	A_dense[3][2] = 1;
	A_dense[3][3] = 0;
	A_dense[3][4] = 0;

	A_dense[4][0] = 0;
	A_dense[4][1] = 4;
	A_dense[4][2] = 2;
	A_dense[4][3] = 0;
	A_dense[4][4] = 1;

	// create b vector
	Vector<double> b(5);
	b[0] = 8;
	b[1] = 45;
	b[2] = -3;
	b[3] = 3;
	b[4] = 19;

	// allocate result
	Vector<double> result(5);

	// create sparse matrix
	SparseMatrix A_sparse(A_dense);

	// solve
	A_sparse.Solve(b, result);

	PRINT("CHECK RESULT");
	PRINT("Result is:      " << result);
	PRINT("Result must be: [ 1 2 3 4 5 ]");
	if (result.Size() != 5){
		PRINT("RESULT SIZE NOT OK")
		PRINT("Size is " << result.Size());
		return false;
	}
	PRINT("RESULT SIZE OK");

	double eps = 1e-12;
	for (int i = 0; i < 5; i++){
		double currentValue = result[i];
		double expectedValue = i+1;
		double delta = currentValue - expectedValue;
		if (ABS(delta) > eps){
			PRINT("RESULT NOT OK");
			PRINT("Entry " << i << " is " << currentValue << " but must be " << expectedValue);
			return false;
		}
	}
	PRINT("RESULT VALUES OK");
	return true;
}

void CreateRandomDenseMatrix(Matrix<double>& dense, double threshold)
{
	Random rand;
	double randElem;
	for(int i=0; i < (int)dense.GetRows(); i++){
	  for(int j=0; j < (int)dense.GetCols(); j++){
	    randElem = rand.GetNormalDistributed(0.0, 1.0);
	    if(randElem > threshold){
	    	dense[i][j] = i*1000+j;//randElem * 100.0; 
	    } else {
	    	dense[i][j] = 0.0;
	    }
	  }
	}
}

void CopyElements(Matrix<double>& A, Matrix<double>& B)
{
	for(int i=0; i < (int)A.GetRows(); i++){
	  for(int j=0; j < (int)A.GetCols(); j++){
	  	B[i][j] = A[i][j];
	  }
	}
}

bool Check(string msg, Matrix<double>& A_dense, SparseMatrix& A_sparse, const SparseUmfMatrix& A_umf)
{
	PRINT(msg);
	PRINT("Dense is  :" << A_dense.GetRows() << " by " << A_dense.GetCols());
	PRINT("Sparse is :" << A_sparse.GetRowsNum() << " by " << A_sparse.GetColsNum());
	PRINT("Umf is    :" << A_umf.GetRows() << " by " << A_umf.GetCols());


	if (DOCHECKDIMENSIONS){
		PRINT("CHECK DIMENSIONS!");
		if ((int)A_dense.GetRows() != (int)A_umf.GetRows() ||
				(int)A_dense.GetCols() != (int)A_umf.GetCols() ||
				(int)A_dense.GetRows() != (int)A_sparse.GetRowsNum() ||
				(int)A_dense.GetCols() != (int)A_sparse.GetColsNum()){
			PRINT("Dimensions do not match!");
			return false;
		} else {
			PRINT("CHECK OK!");
		}
	} else {
		PRINT("DIMENSION CHECK IS OFF!");
	}


	PRINT("CHECK ENTRIES!");
	int count = 0;
	double eps = 1e-12;
	for (int i = 0; i < (int) A_dense.GetRows(); i++) {
		for (int j = 0; j < (int) A_dense.GetCols(); j++) {
			double valDense = A_dense[i][j];
			double valUmf = A_umf.GetElement(i, j);
			double valSparse = A_sparse.GetElement(i, j);

			if ((ABS(valDense - valUmf) > eps) ||
					(ABS(valDense - valSparse) > eps)) {
				PRINT("Entries i:" << i << " j:" << j << " do not match.");
				PRINT("Dense is  :" << valDense);
				PRINT("Umf is    :" << valUmf);
				PRINT("Sparse is :" << valSparse);

				return false;
			}

			count++;
		}
	}
	PRINT(count << " ENTRIES OK!" << endl);
	return true;
}
