# ADD_TESTS_EXPECTED_COMPILER_ERRORS(SOURCE_FILE PREFIX)
#    adds test(s) checking for expected compile time errors.
#    Each block with an expected compile time error must be bracketed by a 
#    preprocessor conditional checking for a variable whose name starts
#    with prefix and whose last part is a unique description for the
#    expected error. For a given source file foo.cpp: 
#
#    class my_class {
#     public:
#       my_class(int i) : data_(i) {};
#     private:
#       int data_;
#       my_class();
#    };
#
#    int main(int argc, char *argv[]){      
#    #ifdef MY_PREFIX_PrivateDefaultConstructor
#       my_class foo; // error: default constructor of my_class is private
#    #endif
#
#       my_class foo2(7); // this constructor is allowed
#
#    #ifdef MY_PREFIX_AssignDouble
#        foo2 = 7.; // cannot assign double to 
#    #endif
#    }
#
#    the statement 
#    INCLUDE(AddTestsExpectedCompilerErrors)
#    ADD_TESTS_EXPECTED_COMPILER_ERRORS(foo.cpp MY_PREFIX) in the 
#    CMakeLists.txt would add two tests with names
#    - PrivateDefaultConstructor
#    - AssignDouble
#    to ctest. Each test tries to compile the source (foo.cpp) using the 
#    std. c++ compiler where the additional command line option 
#     -DMY_PREFIX_PrivateDefaultConstructor is given
#    

MACRO(ADD_TESTS_EXPECTED_COMPILER_ERRORS SOURCE_FILE PREFIX)

  # extract the base of the filename
  STRING(REGEX REPLACE "[.][a-zA-Z0-9_]+$" "" SOURCE_BASE ${SOURCE_FILE})

  # read the source file and look for ifdef PREFIX
  FILE(READ ${SOURCE_FILE} MY_SOURCE)
  STRING(REGEX MATCHALL "ifdef[ ]+${PREFIX}[a-zA-Z0-9_]+" 
    COMP_DEFINES ${MY_SOURCE})
  #MESSAGE("COMP_DEFINES = ${COMP_DEFINES}\n")
  #MESSAGE("Adding tests for expected compile time errors in ${MY_SOURCE_FILE} ...\n")

  # get the include dirs and definitions
  GET_DIRECTORY_PROPERTY(MY_INCLUDE_DIRS INCLUDE_DIRECTORIES)
  GET_DIRECTORY_PROPERTY(MY_DEFS DEFINITIONS)
  # prepend -I to each include directory
  FOREACH(INC_DIR ${MY_INCLUDE_DIRS})
    IF(WIN32)
      LIST(APPEND MY_INC_DIRS "/I ${INC_DIR}")
    ELSE(WIN32)
      LIST(APPEND MY_INC_DIRS "-I${INC_DIR}")
    ENDIF(WIN32)
  ENDFOREACH(INC_DIR)

  SET(MY_ARGS "${CMAKE_CXX_FLAGS} ${MY_INC_DIRS} ${MY_DEFS}")
  SEPARATE_ARGUMENTS(MY_ARGS)
  
  # for each define, add a compile test, it should fail
  FOREACH(MY_COMP_DEFINE ${COMP_DEFINES})
    # remove ifdef from MY_COMP_DEFINE
    STRING(REGEX REPLACE "ifdef[ ]+" "" MY_COMP_DEFINE ${MY_COMP_DEFINE})
    # remove prefix and extract unque name 
    STRING(REGEX REPLACE "${PREFIX}" "" COMP_ERR_NAME ${MY_COMP_DEFINE})

    IF(WIN32)
     FIND_PACKAGE(CLEXE)
     ADD_TEST(${SOURCE_BASE}_${COMP_ERR_NAME} ${CMAKE_CXX_COMPILER} ${MY_ARGS} /D${MY_COMP_DEFINE} /Fo${CMAKE_CURRENT_BINARY_DIR} ${CMAKE_CURRENT_SOURCE_DIR}/${SOURCE_FILE})
    ELSE(WIN32)
      ADD_TEST(${SOURCE_BASE}_${COMP_ERR_NAME} ${CMAKE_CXX_COMPILER} ${MY_ARGS} -D${MY_COMP_DEFINE} -o ${CMAKE_CURRENT_BINARY_DIR}/ThisObjectFileShouldNeverExist.o -c ${CMAKE_CURRENT_SOURCE_DIR}/${SOURCE_FILE})
   ENDIF(WIN32)
    SET_TESTS_PROPERTIES(${SOURCE_BASE}_${COMP_ERR_NAME} 
      PROPERTIES WILL_FAIL TRUE)

  ENDFOREACH(MY_COMP_DEFINE)
ENDMACRO(ADD_TESTS_EXPECTED_COMPILER_ERRORS)
