# - This module determines if BIAS is available and sets a number of 
# useful variables
# See http://www.mip.informatik.uni-kiel.de
# ------------------------------------------------------- 
# This code sets the following variables:
#
# BIAS_FOUND            = system has BIAS lib
#
# BIAS_LIBRARIES        = full path to the libraries
#                         on Unix/Linux with additional linker flags from "bias-config --libs" 
#                         if (deprecated) shell script is used.
#
# BIAS_CXX_FLAGS        = compiler flags for BIAS, essentially "`bias-config --cxxflags`" if shell script is used
#
# BIAS_INCLUDE_DIR      = where to find headers 
#
# BIAS_LINK_DIRECTORIES = link directories, useful for rpath on Unix
# BIAS_EXE_LINKER_FLAGS = linker flags (may contain rpath on Unix)
#
# BIAS_CONFIG_CMAKE_SCRIPT   = contains used BIASConfig.cmake path if PackageConfig (1) is used.
# BIAS_BIASCONFIG_EXECUTABLE = (DEPRECATED) contains used path of shell script bias-config (2) is used 
#                              (instead of BIASConfig.cmake)
#
#
# OPTIONS
# FIND_BIAS_SKIP = if true: Skip this find script completely.
#                  Required for toplevel projects including BIAS in source tree 
#                  together with subprojects using FIND_PACKAGE(BIAS). Avoids 
#                  inconsistent BIAS link directories. 
# -------------------------------------------------------
#
# typical USAGE in user projects CMakeLists.txt: 
#
# FIND_PACKAGE(BIAS REQUIRED)
# IF (BIAS_FOUND)
#   MESSAGE("including ${BIAS_USE_FILE}")
#   INCLUDE(${BIAS_USE_FILE})
# ELSE(BIAS_FOUND)
#   MESSAGE(FATAL_ERROR "BIAS was not found. Please set BIAS_DIR pointing to BIASConfig.cmake build dir.")
# ENDIF (BIAS_FOUND)
# ------------------------------------------------------- 
# This code sets the following variables: 
#
# BIAS_FOUND = 
# BIAS_CONFIG_CMAKE_SCRIPT = 
# BIAS_DIR =  
# BIAS_USE_FILE
# BIAS_INCLUDE_DIR
# BIAS_LINK_DIRECTORIES = 
# BIAS_BUILD_SETTINGS_FILE
# BIAS_BUILD_SHARED_LIBS
# BIAS_SOURCE_DIR
# BIAS_LIBRARIES = 
# .... and many more ... 
# ------------------------------------------------------- 
# original author Jan Woetzel <http://www.mip.informatik.uni-kiel.de/~jw> 
# 2007/10 deprecation cleanups by Felix Woelk woelk@vision-n.de
# 2008/02 FIND_BIAS_SKIP option added Jan Woetzel

MACRO(MSG _MSG)
MESSAGE(STATUS 
    "${CMAKE_CURRENT_LIST_FILE}(${CMAKE_CURRENT_LIST_LINE}): ${_MSG}")
ENDMACRO(MSG)

IF (NOT FIND_BIAS_SKIP)

  # This is the workaround for FIND_PACKAGE which couldn't be used here 
  # because CMAKE_MODULE_PATH has higher priority than BIASConfig.cmake path. 
  FIND_FILE(BIAS_CONFIG_CMAKE_SCRIPT BIASConfig.cmake 
    ${BIAS_ROOT_DIR}
    $ENV{BIAS_ROOT_DIR}
    ${BIAS_DIR}
    $ENV{BIAS_DIR}
    "${CMAKE_CURRENT_BINARY_DIR}/../BIAS"
    )

  MSG("Using BIAS_CONFIG_CMAKE_SCRIPT=${BIAS_CONFIG_CMAKE_SCRIPT}")

  IF(BIAS_CONFIG_CMAKE_SCRIPT)

    # OK, PackageConfig found. Works on Unix and Windows
    INCLUDE(${BIAS_CONFIG_CMAKE_SCRIPT})
    SET(BIAS_FOUND 1)
    # update BIAS_DIR to match BIASConfig.cmake, convert to fwd. slashes
    GET_FILENAME_COMPONENT(BIAS_DIR ${BIAS_CONFIG_CMAKE_SCRIPT} PATH)
  
  ELSE(BIAS_CONFIG_CMAKE_SCRIPT)

    # make FIND_PACKAGE friendly
    IF(NOT BIAS_FIND_QUIETLY)
      IF(BIAS_FIND_REQUIRED)
	MESSAGE(FATAL_ERROR
          "BIAS required, Please select BIAS_CONFIG_CMAKE_SCRIPT by hand or guide cmake with BIAS_DIR.")
      ELSE(BIAS_FIND_REQUIRED)
	MESSAGE(STATUS 
          "ERROR: BIAS was not found. Please select BIAS_CONFIG_CMAKE_SCRIPT by hand or guide with BIAS_DIR.")
      ENDIF(BIAS_FIND_REQUIRED)
    ENDIF(NOT BIAS_FIND_QUIETLY)

  ENDIF(BIAS_CONFIG_CMAKE_SCRIPT)

ELSE(NOT FIND_BIAS_SKIP)

  IF(NOT BIAS_FIND_QUIETLY)
    MSG("Skipped by FIND_BIAS_SKIP")
  ENDIF(NOT BIAS_FIND_QUIETLY)

ENDIF(NOT FIND_BIAS_SKIP)
