#ifndef __BLOBDETECTOR_CCA_HH__
#define __BLOBDETECTOR_CCA_HH__
#include <vector>
#include <Base/Image/Image.hh>
#include <Base/Math/Vector2.hh>
#include <Filter/Label.hh>
#include "BlobDetectorBase.hh"

namespace BIAS {

  /** @class BlobDetectorCCA
      Does a Connected Component Analyis and gives information about regions
      in an image, like centroid and bounding boxes.
      Geometry
      useful after morphological operators like Erosion or Dilation
      The input image has to be segmented in only a few grey values,
      best is black or white.
      @ingroup g_feature
      neighborhood 8 so far not implemented
      @author grest, Aug. 2004, ischiller 08/08
  */

  template <class StorageType>
  class BIASFeatureDetector_EXPORT BlobDetectorCCA :
    public BIAS::Label,public BlobDetectorBase<StorageType> {
  public:

    BlobDetectorCCA();
    virtual ~BlobDetectorCCA();

     /** 
        Does a Connected Component Analyis (4-neighborhood)
        and deletes all regions, whose sizes are smaller 
        than p percent of the image.<br>
        Use SetPercent and SetNeighborhoud4 before
        If percent is negative, only the two largest regions are not deleted.
        While there is no difference between fore- and background (see 
        BIAS::Label), so the largest region is usually the background.
        One channel images only !<br>              
      
        @author Daniel Grest, Aug. 2004
    */
    int Detect(Image<StorageType>& image,
               std::vector<BIAS::BIASBlob>& blobs);
      
    /** call this afer Process() to get the medians of all regions.
        Calculates the medians of each region, sorted by size of the region,
        largest region first.
        returns 0, in case of no error    
        @author Daniel Grest, June. 2003
    */
    int GetMedians(std::vector<BIAS::Vector2<unsigned int> > &medians);
                 
    /** call this afer Process() to get the centroids of all regions.
        Calculates the centroids of each region, sorted by size of the region,
        largest region first.
        returns 0, in case of no error    
        @author Daniel Grest, June. 2003
    */
    int GetCentroids(std::vector<BIAS::Vector2<double> > &centroids);

    
    /**
       @brief
    */
    void SetPercent(float percent){
      fPercent_ = percent;
    }

    /**
       @brief If neighborHood4, then a max of 255 regions is possible, if there are 
       more it results in undefined behaviour (probably a segfault).
       Best is to delete small regions before, e.g. with Erosion3Fast or 
       DeleteSingles() from BIAS::Morphology              
    */
    void SetNeighborhoud4(bool neighborhoud4){
      bNeighborHood4_ = neighborhoud4;
    }
    
  protected:  
    /**
       init is called before processing the first image
    */
    void Init_(const BIAS::Image<StorageType> &img);
   
    int Process_(BIAS::Image<StorageType> &img, float percent,
                 bool neighborHood4=true);  

    bool bIsInited_;
    float fPercent_;
    bool bNeighborHood4_;
 
    int Process8_(BIAS::Image<StorageType> &img, float percent);
    std::vector<std::vector<BIAS::Vector2<int> > > allBlobs_;


  };//class
}//namespace



#endif //_BLOBDETECTOR_CCA_HH__
