/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __CheckerboardDetector_hh__
#define __CheckerboardDetector_hh__

#include <bias_config.h>
#ifndef BIAS_HAVE_OPENCV
#  error Please recompile BIAS with USE_OPENCV to use this code. 
#endif 

#include <Base/Image/ImageBase.hh>
#include <vector>
#include <cv.h>

namespace BIAS {

  /// helper vector of coordinates
  typedef std::vector< std::pair<float,float> > t_CoordVecf;

  /**
     @class CheckerboardDetector
     @brief Detects a checkerboard pattern in an image (typically 7x4 inner
            corners, asymmetric layout recommended).
     @note  Based on OpenCV implementation. 
     @author Jan Woetzel 
     @date 2006 
     @ingroup g_feature     
  */
  class BIASFeatureDetector_EXPORT CheckerboardDetector {

  public:

    /** @brief Detects a (recommended asymmetric) checkerboard pattern.
        @return 0 for success,
                <0 for error (e.g. could not detect all inner corners).
     */
    static int Compute(BIAS::Image<unsigned char> &img,
                       const unsigned int &xCorners,
                       const unsigned int & yCorners,
                       t_CoordVecf &coords,
                       const bool &saveDebugImage = false);

    /** @brief Takes number of inner corners on x axis and y axis, 
               and number of images to build an array of corners,
               and draws computed corners into CV image.
        @note  Assumes all corners to lie on z = 0 plane.
     */
    static void DrawPoints(CvArr *img,
                           CvPoint2D32f *ptArray,
                           //t_CoordVecf & coords,
                           int numPts,
                           const int xCorners,
                           const int yCorners,
                           const bool DRAW_IDEAL_OVERLAY = true,
                           const bool DRAW_AXES = true,
                           const bool DRAW_POINT_LABEL = true);

    /* @brief Set flags for OpenCV checkerboard corner detection algorithm.
       Valid flags are CV_CALIB_CB_ADAPTIVE_THRESH, CV_CALIB_CB_NORMALIZE_IMAGE,
       CV_CALIB_CB_FILTER_QUADS or combinations of these flags.
    */
    static void SetCornerDetectorFlags(int flags) {
      cornerDetectorFlags_ = flags;
    };

    /* @brief Set max. number of iterations for subpixel corner refinement. */
    static void SetSubPixelIterations(int maxIters) {
      subPixelIterations_ = maxIters;
    };

    /* @brief Set accuracy to achieve in subpixel corner refinement. */
    static void SetSubPixelAccuracy(double accuracy) {
      subPixelAccuracy_ = accuracy;
    };

  private:

    // Flags for checkerboard corner detection algorithm (OpenCV)
    static int cornerDetectorFlags_;
    
    // Max. number of iterations for subpixel corner refinement
    static int subPixelIterations_;
    
    // Accuracy to achieve in subpixel corner refinement
    static double subPixelAccuracy_;

  };
}

#endif //__CheckerboardDetector_hh__
