/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef __BIAS_DISTTRANS_HH__
#define __BIAS_DISTTRANS_HH__
#include <Base/Common/BIASpragmaStart.hh>

// std includes
#include <vector>

// BIAS includes
#include <Base/Image/Image.hh>
#include <Base/Math/Vector2.hh>
#include "FilterBase.hh"
#include "FilterNToN.hh"

namespace BIAS {
  /** @class DistTransform
      @brief Class for calculating the distance transform.

      Generates an image where for all pixels the distance to the step of
      pixels with zero value in the input image is calculated. The Borgefor
      distance is used to measure distances (diagonal step has distance 4,
      horizontal/vertical step has distance 3).
      
      @author Dennis Herzog, Jan. 2005, edited by grest
  */
  template <class InputStorageType, class OutputStorageType> 
  class BIASFilter_EXPORT DistTransform
    : public FilterNToN<InputStorageType, OutputStorageType>
  {

  public:     
    DistTransform();
    DistTransform(const DistTransform<InputStorageType, OutputStorageType>& other);
    virtual ~DistTransform() {}

    /** @brief Calculates the distanceImage from source image src.
        Converts the given image to a distance transformed image (the Borgefor 
        distance is used, that means one step in diagonal direction has a
        distance of 4 other steps have distances of 3); <br>
        The pixel set, to which the distance is calculated
        has to be marked as zero valued pixels, e.g. grey value zero or 
        SetZeroLevel().
        @author herzog
        @param img input and output image
    */   
    virtual int Filter(const Image<InputStorageType>& src,
                       Image<OutputStorageType>& dst);

    virtual int FilterInt(const Image<InputStorageType>& src, 
                          Image<OutputStorageType>& dst);

    virtual int FilterFloat(const Image<InputStorageType>& src, 
                            Image<OutputStorageType>& dst);

    virtual FilterNToN<InputStorageType, OutputStorageType>* Clone() const
    { return new DistTransform<InputStorageType, OutputStorageType>(*this); }

    void SetZeroLevel(InputStorageType value) { zeroLevel_=value;}

  protected:
    void MakeDistanceImage_(BIAS::Image<OutputStorageType>& img);  
    void InitDistanceImage_(BIAS::Image<OutputStorageType>& img);
    InputStorageType zeroLevel_;

    virtual void GetBordersValid_(int &border_x, int& border_y) const;
  };

}

#include <Base/Common/BIASpragmaEnd.hh>

#endif
