#ifndef __AutoControl_HH__
#define __AutoControl_HH__

#include "ControlInterface.hh"
#include "TrackballControl.hh"
#include <GLviewer/RenderContextBase.hh>
#include <Base/Math/Vector3.hh>

/** 
  \class AutoControl
  \author cmenk
	\ingroup g_glviewer
  \brief class for camera movement in autopilot mode.
*/

namespace BIAS {
  class BIASGLviewer_EXPORT AutoControl : public ControlInterface
  {    
  public:
    AutoControl();

    /**
    * Initialisizes the autopilot.
    */
    void Init();

    /**
    * Adds a defined pose to the autopilot at the end of PoseList_ F1
    */
    void AddPose(BIAS::Vector3<double> C, BIAS::Vector3<double> PoI,
                 BIAS::Vector3<double> up);

    /**
    * Adds the current pose of ControldObject to the autopilot at the
    * end of PoseList_ F1
    */
    void AddPose();

    /**
    * Goes to the next calculated pose. F2
    */
    int NextStep();

    /**
    * Goes to the defined step and sector.
    returns 0 for sucess, -1 if step or sector exceeds max
    */
    int GoTo(unsigned int step, unsigned int sector);

    /**
    * Goes to the defined pose.
    returns 0 for sucess, -1 if step or sector exceeds max
    */
    int GoToPose(int pose);

    unsigned int GetStep(){return Step_;};

    unsigned int GetSector(){return Sector_;};
    unsigned int GetNumSectors(){return PoseList_.size()-1;};

    /**
    * Sets the autopilot to start point. F3
    */
    void Rewind();

    /**
    * Deletes the last added pose and resets the autopilot F4
    */
    void DeleteLast();

    /**
    * Deletes the defined pose and resets the autopilot F4
    */
    void DeletePose(int pose);

    /**
    * Deletes all poses in PoseList_ and resets autopilot.
    */
    void DeleteAll();

    /**
    * Handles special key events.
    */
    bool SpecialKeyPressed(int key);
               
    void SetContext(RenderContextBase* context);
  
    int SaveToFile(const std::string &file);

    int LoadFromFile(const std::string &file);

    unsigned int GetNumberOfSteps(){return NumberOfSteps_;};
    void SetNumberOfSteps(unsigned int n){ NumberOfSteps_ = n;};

    void Activate() {Active_ = true;};
    void Deactivate() {Active_ = false;};
    bool IsActive() {return Active_;};
    bool IsInitialized() {return PoseList_.size()>0;};
    bool TimerExpired();

  protected:
    struct CameraPose
    {
      BIAS::Vector3<double> C;
      BIAS::Vector3<double> PoI;
      BIAS::Vector3<double> up; 
    }; 
    
    std::vector<CameraPose> PoseList_;
    RenderContextBase* Context_;
    
    BIAS::Vector3<double> VecC_;
    BIAS::Vector3<double> VecUp_;
    BIAS::Vector3<double> VecPoI_;
    double StepSizeC_;
    double StepSizeUp_;
    double StepSizePoI_;
    
    unsigned int NumberOfSteps_; // steps for every sector
    unsigned int Step_;          
    unsigned int Sector_;
    int InitializedSector_;

    bool Active_;
  };
}//end of namespace
#endif
