#ifndef __GLPROJECTIONPARAMETERSINTERFACE_hh__
#define __GLPROJECTIONPARAMETERSINTERFACE_hh__

// includes
#include <bias_config.h>

#ifndef BIAS_HAVE_OPENGL
#  error You need BIAS with USE_OpenGL Please recompile BIAS.
#endif // BUILD_GLviewer

#include <Base/Geometry/HomgPoint3D.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Geometry/Quaternion.hh>
#include <Geometry/RMatrix.hh>
#include <Geometry/ProjectionParametersBase.hh>

#include <OpenGLFramework/Base/glfRenderTarget.hh>
#include <OpenGLFramework/Base/glfFramebufferObject.hh>
#include <GLviewer/Scenes/SceneBGImage.hh>
#include <GLviewer/OutputAttachments/GLProjectionOutputAttachment.hh>

/** Read as frequency/granularity BASIC/MEDIUM/HIGH granularity or frequency */
#define GLP_ACTIVITY_BASIC  1
#define GLP_ACTIVITY_MEDIUM 2
#define GLP_ACTIVITY_HIGH   4
#define GLP_ACTIVITY_FULL   7
#define GLP_STATE_BASIC     8
#define GLP_STATE_MEDIUM    16
#define GLP_STATE_HIGH      32
#define GLP_STATE_FULL      63
#define GLP_PROJECTION      1<<17
#define GLP_VIEWPORT        1<<18

namespace BIAS {

  /** @class GLProjectionParametersInterface
      @brief Abstract interface class to handle changes in rendering
      parameters by controllers and in rendering contextes
      @ingroup g_glviewer
      @author ischiller 04/2009 */
  class BIASGLviewer_EXPORT GLProjectionParametersInterface {
  public:

   typedef enum {
      /** Do not do anything! Do not render either! */
      Deny=0,
      /** rescale internal projection parameters preserving aspect ratio and
          scale by integer values only. This can lead to two black borders: 
          right and bottom */
      AutoRescaleParams,
      /** Render part of the image or the whole image, does not change
          projection parameters.
          drawingUpperLeft_ are the projection image coordinates which will
          appear in the rendered image in the upper left corner. */
      AutoCutImage,
      /** Calculate new simple perspective camera parameters keeping the
          current field of view in "y direction". */
      AutoSimplePerspectiveCam,
      AdaptToCroppedImage,
      /** rescale internal projection parameters preeerving aspectration but 
          use non-integer values.
          This should result in only one black border.*/
      AutoRescaleSimple
    } ViewingReshapeBehaviour;

    virtual BIAS::ProjectionParametersBase* 
      GetMyselfAsProjectionParametersBase() const =0;

    virtual void SetDebugLevel(const long int lv)=0;
    virtual void AddDebugLevel(const long int lv)=0;

    virtual int SetExtrinsics(const BIAS::HomgPoint3D& C,
                              const BIAS::Quaternion<double>& Q)=0;
    virtual int SetExtrinsics(const BIAS::Vector3<double>& C,
                              const BIAS::Quaternion<double>& Q)=0;

    virtual int SetExtrinsics(const BIAS::HomgPoint3D& C,
                              const BIAS::RMatrix& R)=0;
    virtual int SetExtrinsics(const BIAS::Vector3<double>& C,
                              const BIAS::RMatrix& R)=0;

    virtual int SetExtrinsics(const BIAS::Vector3< double >& C,
                              const BIAS::Vector3< double >&	scenePoint,
                              const BIAS::Vector3< double >&	up)=0;

    //Getter
    virtual int GetExtrinsics(BIAS::Vector3<double>& C,
                              BIAS::Vector3<double>& up)const =0;
    virtual int GetExtrinsics(BIAS::Vector3<double>& C,
                              BIAS::RMatrix & R)const =0;


    virtual BIAS::HomgPoint3D UnProject(const unsigned int x, 
                                        const unsigned int y) = 0;
    virtual BIAS::HomgPoint3D UnProject(BIAS::HomgPoint2D& point2D, 
                                        double depth) = 0;
    virtual BIAS::HomgPoint2D Project(BIAS::HomgPoint3D &point) = 0;

    virtual bool PoseValid() = 0;

    /** Takes x, y as window coordinates, checks whether the coordinates
        lie in the rendered image (in the viewrange of projection). If so
        \returns the coordinates relative to the upper left image corner of the
        used viewport and true. else coordintes are unchanged and return value 
        is false. */
    virtual bool GetViewportCoordinates(int& x, int& y)=0;

    /** Setting the near and far z-Clipping planes. * @{ */
    virtual void  SetZClippingPlanes  (float n, float f)=0;
    virtual void  SetNearClippingPlane(float n)=0;
    virtual void  SetFarClippingPlane (float f)=0;
    virtual float GetNearClippingPlane()=0;
    virtual float GetFarClippingPlane ()=0;
    /** @} */

#ifdef BIAS_HAVE_XML2
    virtual int XMLRead (const std::string& Filename)=0;
    virtual int XMLWrite(const std::string& Filename) const =0;
    virtual int Load    (const std::string& filename)=0;
    virtual int Save    (const std::string& filename) const =0;
#endif
    virtual void SetRenderTarget(BIAS::glfRenderTarget* RT)=0;
    virtual BIAS::glfRenderTarget* GetRenderTarget()=0;

    virtual void SetClearColor(const BIAS::Vector4<float>& cc)=0;
    virtual void SetClearDepth(const float d)=0;

    virtual BIAS::Vector4<float> GetClearColor()=0;
    virtual float GetClearDepth()=0;


    virtual void SwitchToOffscreenRendering(bool offscreen = true)=0;
    virtual void AppendOutputAttachment(int val, std::vector<int> args
                                     = std::vector<int>())=0;
    virtual BIAS::GLProjectionOutputAttachment *GetAttachmentByType(int type)=0;
    virtual void SetAutoReshapeBehaviour(ViewingReshapeBehaviour newBehaviour)=0;

    //obtain rendered images and parts of it
    virtual double GetDepth(const unsigned int x, const unsigned int y) = 0;
    virtual double GetDepth(const BIAS::HomgPoint2D& p) =0;

    virtual int GetDepthMap(BIAS::Image<float>& depthMap, 
                            const float invalid=0.0)=0;
    virtual void TranslateZToDepth(BIAS::Image<float>& ztodepth,
                                   const float invalid = 0.0,
                                   const bool flip = true) const = 0;
    virtual void TranslateDepthToZ(BIAS::Image<float>& depthtoz,
                                   const float invalid = 0.0,
                                   const bool flip = true) const = 0;
    virtual void TranslateZToMetricZ(BIAS::Image<float>& depthMap,
                                     const float invalid = 0.0,
                                     const bool flip = true) const = 0;
    virtual int GetZBuffer(BIAS::Image<float>& zBuffer)=0;
    virtual int GetMetricZMap(BIAS::Image<float>& metricZMap,
                              const float invalid = 0.0) =0;
    virtual void GetImage(BIAS::Image<unsigned char>& image)=0;
    virtual void GetGreyImage(BIAS::Image<unsigned char>& image)=0;

    //Draw function
    virtual int Draw(std::vector<BIAS::SceneBase *>& scenes,
                     BIAS::SceneBGImage* backgroundImageScene = NULL,
                     BIAS::glfFramebufferObject* theTarget = NULL)=0;
  protected:
    bool offscreenRenderingEnabled_;

  private:

  }; // class GLProjectionParametersInterface

} // namespace

#endif //__GLPROJECTIONPARAMETERSINTERFACE_hh__
