#ifndef __RENDER_CONTEXT_BASE_HH__
#define __RENDER_CONTEXT_BASE_HH__

#include <bias_config.h>

#ifndef BIAS_HAVE_OPENGL
#  error You need BIAS with USE_OpenGL Please recompile BIAS.
#endif // BUILD_GLviewer

#include <vector>

#include <Base/Common/W32Compat.hh>
#include <Base/Image/Image.hh>

#include <GLviewer/Scenes/SceneBase.hh>
#include <GLviewer/GLProjectionParametersInterface.hh>
#include <GLviewer/Controller/ControlInterface.hh>
#include <GLviewer/View.hh>

#ifdef BIAS_HAVE_PTHREADS
#include <Utils/MutexSingleton.hh>
#endif

namespace BIAS {
  class BIASGLviewer_EXPORT SceneBGImage;

  /** \class RenderContextBase 
      \brief Base for all classes creating interface 
      between GL and  "window manager".
      new version without glprojection
      \ingroup g_glviewer
  */
  class BIASGLviewer_EXPORT RenderContextBase {
  public:
    RenderContextBase();
    virtual ~RenderContextBase() {}

    void PrintGLStatus();

    /** Simple rendering routine.
     * Assumes that all viewport settings are good and renders scene.
     * \returns 0 all good.
     */
    virtual int Render();

    virtual void MakeGLContextCurrent() = 0;

#ifdef BIAS_HAVE_PTHREADS
    int LockContextSwitch();
    int ReleaseContextSwitch();
#endif

    /** Demands a rendering from the Context;
     */
    virtual void Redisplay() = 0;

    virtual int GetImage(BIAS::Image<unsigned char>& Image);
    virtual int GetGreyImage(BIAS::Image<unsigned char>& Image);
    virtual int GetDepthMap(BIAS::Image<float>& Image,
                            const float invalid = 0.0);
    virtual int GetZBuffer(BIAS::Image<float>& Image);
    virtual int GetMetricZMap(BIAS::Image<float>& Image,
                              const float invalid = 0.0);
    
    /**
     * \attention Do not append scene of type SceneBGImage here, for correct
     * background image rendering rather use SetBackgroundImageScene().
     */
    void AppendScene(SceneBase* scene);
    /** @brief removes scene with the given pointer from the vector of scenes */
    void RemoveScene(SceneBase* scene);
    /**
     * Inserts or deletes a scene at the specified position
     * in std::vector<SceneBase*> scenes_.
     */
    void InsertSceneAt(SceneBase* scene, unsigned int index);
    void DeleteSceneAt(unsigned int index);
    
    void SetGLProjectionParametersInterface(GLProjectionParametersInterface* newCamera);  
    void SetControl(BIAS::ControlInterface* newControl);

    /**
     * Deactivate background image rendering pass by passing NULL here,
     * or deactivating scene itself. 
     * First choice will speed up rendering more (if noticeable anyway).
     */
    void SetBackgroundImageScene(SceneBGImage* backgroundImageScene);
    
    void DeleteScenes();
    void ClearScenes();
    void DeleteCamera() {delete camera_; camera_ = NULL;}
    void DeleteControler() {delete control_; control_ = NULL;}
    void DeleteController() {delete control_; control_ = NULL;}
    void DeleteBackgroundImageScene();
    
    void AppendAdditionalListener(ControlInterface* listener) 
    {if(listener!=NULL) additionalListeners_.push_back(listener); }
    void ClearAdditionalListeners() {additionalListeners_.clear(); }
    void RemoveAdditionalListener(ControlInterface& listener);

    /**
     * Turn AutoClipping on, now the near and far clipping plane is automatically
     * derived and set.
     */
    void SetAutoClipping(bool on) { AutoClipping_ = on;}
    
    /**
       @brief sets minimum znear for autoclipping
       if autoclpiing is enabled the z-near and z-far plane are computed using
       the boudingspheres of the scenes. SetAutoClipMinZNear defines the minimal
       used znear value.
    */
    void SetAutoClipMinZNear(double znear)
    {
      AutoMinZNear_ = znear;
    };
    
    /**
     * Returns the center of all bounding boxes of appended scenes
     * and the radius which include all scenes. Return true iff
     * at least one scene with a bounding box was found, else false
     */
    bool GetSceneCenter(BIAS::Vector3<double>& center,
                        BIAS::Vector3<double>& radius);
    
    //control implementation:
    /** . 
     * Control is implemented the following way:       cout<<"F1"<<endl;
     * All input is passed to control_ first, if this is not handling 
     * the input, the information is passed to all scenes_. One can think of
     * extending this to different behaviour models, like
     * not passing anything to scenes or always passing input to scenes.
     * Distingushing between the scenes (exluding scene of information)
     * must be handled on scene side.
     *
     * \return true if controler handled input
     * @{
     */
    bool LeftMouseDoubleClicked(int x, int y,int m);
    bool LeftMouseSingleClicked(int x, int y);
    bool LeftMouseDown(int x, int y);
    bool LeftMouseUp(int x, int y);
    bool LeftMouseMoved(int x1, int x2, int y1, int y2);
    bool RightMouseSingleClicked(int x, int y); 
    bool RightMouseDoubleClicked(int x, int y);
    bool RightMouseMoved(int x1, int x2, int y1, int y2);
    bool MiddleMouseMoved(int x1, int x2, int y1, int y2);
    bool LeftAndRightMouseMoved(int x1, int x2, int y1, int y2);
    bool MouseLeftWindow();
    bool SpecialKeyPressed(int key);
    bool StandardKeyPressed(unsigned char ,int ,int );
    /** Always informs both camera and scenes! */
    bool TimerExpired();
    bool WindowReshape(int width, int height);
    int  GetSceneCount();
    bool MouseWheelUsed(double stepsize);
    bool MouseMoved(int x, int y);
    /**
     * @}
     */

    // activate timer with intervall in ms
    virtual void SetTimer(unsigned int interval) =0;
    virtual void DeactivateTimer() =0;

    const GLProjectionParametersInterface* GetCamera() {return camera_;}
  
    inline void SetClearColor(const BIAS::Vector4<float>& cc)
    { 
      ClearColor_ = cc; 
      if (camera_) camera_->SetClearColor(ClearColor_); 
      
    };

    bool SetBlending(bool blend=true);
    bool SetBackfaceCulling(bool cull=true);

    void CenterViewport(bool center = true);

    void AppendView(View* view);
    
    inline void GlobalBoundingBoxSwitch(bool theswitch) {
      for (unsigned int i=0; i<scenes_.size(); i++) {
        scenes_[i]->SetRenderBoundingBox(theswitch);
      }
    };

    const std::vector<SceneBase*>& GetScenes() const {
      return scenes_;
    };

    void EnableStencilUsage(bool enable);

  protected:    
 
    std::vector<SceneBase*> scenes_;
    SceneBGImage* backgroundImageScene_;
    ControlInterface* control_;
    GLProjectionParametersInterface* camera_;    
    unsigned int TimerInterval_; //0 = disabled
    bool AutoClipping_;
    double AutoMinZNear_;
    BIAS::Vector4<float> ClearColor_;

    /** Listening for special keys only */
    std::vector<ControlInterface*> additionalListeners_;

    /** called once when rendering starts */
    void Init_();
    bool initialized_;
    /** Default: true */
    bool blendingAllowed_;
    bool culling_;
    bool centerViewport_;
    bool UseStencil_;

#ifdef BIAS_HAVE_PTHREADS
  private:
    static BIAS::MutexSingleton *contextLock_;
#endif //BIAS_HAVE_PTHREADS

    std::vector<View*> views_;
  };

}//end namespace
#endif
