#ifndef __TriangleMeshSplatRendering_hh__
#define __TriangleMeshSplatRendering_hh__

#include <bias_config.h>

#include <GLviewer/GLProjectionParametersInterface.hh>
#include <OpenGLFramework/Base/glfShaderProgram.hh>
#include <OpenGLFramework/Base/glfTexture2D.hh>
#include <OpenGLFramework/Base/glfFramebufferObject.hh>
#include <OpenGLFramework/Base/glfScreen.hh>
#include <Utils/ThreeDOut.hh>
#include <GLviewer/Scenes/SceneBase.hh>
#include <vector>

namespace BIAS {

  /**
   * @class TriangleMeshSplatRendering
   * @brief Scene for rendering multiple triangle meshes generated from depth
   * maps. Overlapping regions of the meshes will be blended, i.e. a
   * per-pixel average is computed in screen space.
   * @ingroup g_glviewer
   * @author jkollmann/ischiller
   */
  class BIASGLviewer_EXPORT TriangleMeshSplatRendering {
  public:
    TriangleMeshSplatRendering();
    ~TriangleMeshSplatRendering();

    /**
     * Sets the resolution of the internal framebuffer object.
     */
    void SetInternalResolution(int width, int height);

    /**
     * Sets the maximum depth difference at which geometry is considered
     * to represent the same surface and is blended.
     */
    void SetDepthTolerance(float depthTolerance);

    void SetCamera(BIAS::GLProjectionParametersInterface* camera);

    /**
     * Set depth tolerance based on fraction of bounding box of tdo objects.
     */
    void ComputeDepthTolerance(float fraction = 0.01f);

    /**
     * Returns the current depth tolerance.
     */
    float GetDepthTolerance() const { return depthTolerance_; }

    /**
     * Adds an arbitrary triangle ThreeDOut object to the scene.
     * The tdo must have been allocated with 'new'. The scene takes
     * responsibility to delete the tdo.
     * Note: Only the indexed face sets are rendered.
     */
    void AddThreeDOut(BIAS::ThreeDOut* tdo);
    void ClearThreeDOuts();
    /**
     *@brief main splatting routine
     */
    void Splat();

    void GetBoundingBox(BIAS::Vector3<double>& min,
                        BIAS::Vector3<double>& max);
  private:

    // configuration
    int         internalWidth_, internalHeight_;
    float       depthTolerance_;
    bool        enableBlending_;

    bool        mustCreateShaders_;
    bool        mustCreateFrameBuffer_;

    void CreateShaders();
    void CreateFrameBuffer(int width, int height);

    BIAS::glfShaderProgram      depthPassProgram_;
    BIAS::glfShaderProgram      normalizationPassProgram_;

    BIAS::glfTexture2D          colorBuffer_;
    BIAS::glfTexture2D          depthBuffer_;
    BIAS::glfFramebufferObject  fbo_;
    BIAS::GLProjectionParametersInterface* camera_;
    std::vector<BIAS::ThreeDOut*> tdos_;
  };

}

#endif // __TriangleMeshSplatRendering_hh__

