#include "AbsoluteOrientationRANSAC.hh"
#include <Base/Math/Matrix3x3.hh>
#include <Base/Math/Matrix.hh>
#include <Base/Math/Vector.hh>
#include <Base/Geometry/Quaternion.hh>
#include <MathAlgo/SVD.hh>
#include <iostream>

#ifdef BIAS_HAVE_GSL
// Include for InvChiSquareCulmProbFun()
#  include <MathAlgo/GSL.hh>
#endif

using namespace std;
using namespace BIAS;

AbsoluteOrientationRANSAC::AbsoluteOrientationRANSAC()
  : sampleSize_(4), numSamples_(100), maxError_(0.01),
    minRatio_(0.75), refine_(true), greedy_(false),
    useWeightedTargetFunction_(false)
{
#ifndef BIAS_DEBUG
  randomizer_.Reset();
#endif
}

AbsoluteOrientationRANSAC::~AbsoluteOrientationRANSAC()
{
}

void AbsoluteOrientationRANSAC::
SetSignificance(double significance)
{
  if (significance <= 0.0 || significance >= 1.0) {
    BIASERR("Invalid significance level " << significance <<
            " given (must be between 0.0 and 1.0)!");
  } else {
#ifdef BIAS_HAVE_GSL
    maxError_ = InvChiSquareCulmProbFun(1.0-significance, 3); // 3 dof
#else
    BIASWARNONCE("No GSL support, using fixed significance level of 0.05!");
    significance = 0.05;
    maxError_ = 5.9915;
#endif
  }
}

int AbsoluteOrientationRANSAC::
Compute_(const std::vector< BIAS::Vector3<double> > &X,
         const std::vector< BIAS::Vector3<double> > &Y,
         const std::vector< BIAS::Matrix3x3<double> > &CovX,
         const std::vector< BIAS::Matrix3x3<double> > &CovY,
         BIAS::RMatrix &dR, BIAS::Vector3<double> &dt,
         double &ds, bool enforceScale,
         std::vector<bool> &inliers, std::vector<double> &errors,
         double &errorSum, const std::vector<double> &w)
{
  const unsigned int num = X.size();
  const bool useCovs = (CovX.size() > 0 || CovY.size() > 0);
  const bool useWeights = (w.size() > 0);

  BIASASSERT(num >= sampleSize_);
  BIASASSERT(X.size() == num && Y.size() == num);
  BIASASSERT(!useCovs || (CovX.size() == num && CovY.size() == num));
  BIASASSERT(!useWeights || w.size() == num);
#ifdef BIAS_DEBUG
  if (useWeights && useCovs) {
    BIASERR("Weights and covariances should NOT be used at the same time!");
    BIASABORT;
  }
#endif

  int res = 0;
  std::vector<int> samples(0);
  std::vector< Vector3<double> > sampleX(0);
  std::vector< Vector3<double> > sampleY(0);
  std::vector< Matrix3x3<double> > sampleCovX(0);
  std::vector< Matrix3x3<double> > sampleCovY(0);
  std::vector<double> samplew(0);
  BIAS::RMatrix actdR, bestdR;
  BIAS::Vector3<double> actdt, bestdt;
  double actds, bestds = 0.0;
  double actScore = 0.0, bestScore = 0.0;
  double actErrorSum = 0.0, bestErrorSum = 0.0;
  int actInlierCount = -1, bestInlierCount = -1;
  std::vector<bool> actInliers;
  std::vector<double> actErrors;
  bool init = false;
  bool done = false;
  
  // Evaluate each sample solution and find best solution

  for (unsigned int k = 0; k < numSamples_ && !done; k++) {

    res = CreateRandomSample_(sampleSize_, X, Y, CovX, CovY, w,
                              samples, sampleX, sampleY,
                              sampleCovX, sampleCovY, samplew);
    BIASASSERT(res == (int)sampleSize_);

    actdR = dR; actdt = dt; actds = ds;
    res = solver_.Compute_(sampleX, sampleY, actdR, actdt, actds,
                           enforceScale, samplew);
    if (res == 0) {
      actInlierCount = EvaluateInliers_(actInliers, actErrors, actErrorSum,
                                        X, Y, CovX, CovY, actdR, actdt, actds, w);
      if (useWeightedTargetFunction_) {
        actScore = 0.0;
        for (register unsigned int i = 0; i < num; i++) {
          if (actInliers[i] && actErrors[i] <= maxError_) {
            actScore += fabs(maxError_-actErrors[i])/maxError_;
          }
        }
      } else {
        actScore = (double)actInlierCount;
      }
      if (!init || actScore > bestScore) {
        bestScore = actScore;
        bestInlierCount = actInlierCount;
        bestErrorSum = actErrorSum;
        bestdR = actdR;
        bestdt = actdt;
        bestds = actds;
        init = true;
        done = (greedy_ && bestInlierCount/(double)num >= minRatio_);
      }
    }
                        
  }

  if (init && bestInlierCount >= (int)sampleSize_) {
        // && bestInlierCount/(double)num >= minRatio_) {

    // Refine best solution optionally
    if (refine_) {
      EvaluateInliers_(actInliers, actErrors, actErrorSum,
                       X, Y, CovX, CovY, bestdR, bestdt, bestds, w);
      samples.clear();
      sampleX.clear();
      sampleY.clear();
      samples.reserve(bestInlierCount);
      sampleX.reserve(bestInlierCount);
      sampleY.reserve(bestInlierCount);
      if (useCovs) {
        sampleCovX.clear();
        sampleCovY.clear();
        sampleCovX.reserve(bestInlierCount);
        sampleCovY.reserve(bestInlierCount);
      }
      if (useWeights) {
        samplew.clear();
        samplew.reserve(bestInlierCount);
      }
      for (register unsigned int i = 0; i < num; i++) {
        if (actInliers[i]) {
          samples.push_back(i);
          sampleX.push_back(X[i]);
          sampleY.push_back(Y[i]);
          if (useCovs) {
            sampleCovX.push_back(CovX[i]);
            sampleCovY.push_back(CovY[i]);
          }
          if (useWeights) {
            samplew.push_back(w[i]);
          }
        }
      }
      BIASASSERT((int)samples.size() == bestInlierCount);
      dR = bestdR; dt = bestdt; ds = bestds;
      res = solver_.Compute_(sampleX, sampleY, dR, dt, ds,
                             enforceScale, samplew);
      if (res == 0) {
        return EvaluateInliers_(inliers, errors, errorSum,
                                X, Y, CovX, CovY, dR, dt, ds, w);
      } else {
        BIASWARN("Refinement of RANSAC solution failed!");
        return -1;
      }
    }

  } else {

    dR = bestdR; dt = bestdt; ds = bestds;
    return EvaluateInliers_(inliers, errors, errorSum,
                            X, Y, CovX, CovY, dR, dt, ds, w);

  }

  BIASWARN("RANSAC failed, found no solution with enough inliers "
           << "(found " << bestInlierCount << ", needed min. "
           << sampleSize_ << " inliers)!");
  return -1;
}

int AbsoluteOrientationRANSAC::
ComputeRotation(const std::vector< BIAS::Vector3<double> > &X,
                const std::vector< BIAS::Vector3<double> > &Y,
                BIAS::RMatrix &dR, std::vector<bool> &inliers,
                std::vector<double> &errors, double &errorSum,
                const std::vector<double> &w)
{
#ifdef BIAS_DEBUG
  const unsigned int num = X.size();
  const bool useWeights = (w.size() > 0);
#endif //BIAS_DEBUG

  BIASASSERT(num >= sampleSize_);
  BIASASSERT(X.size() == num && Y.size() == num);
  BIASASSERT(!useWeights || w.size() == num);

  int res = 0;

  BIASWARN("RANSAC computation of rotation only is not implemented yet!");
  BIASABORT;

  return res;
}

double AbsoluteOrientationRANSAC::
GetResidualErrors(std::vector<double> &errors,
                  const std::vector< BIAS::Vector3<double> > &X,
                  const std::vector< BIAS::Vector3<double> > &Y,
                  const BIAS::RMatrix &dR, const BIAS::Vector3<double> &dt,
                  const double &ds,  const std::vector<bool> &inliers,
                  const std::vector<double> &w)
{
  const unsigned int num = X.size();
#ifdef BIAS_DEBUG
  const bool useWeights = (w.size() > 0);
#endif //BIAS_DEBUG
  const bool useInliers = (inliers.size() > 0);

  BIASASSERT(X.size() == num && Y.size() == num);
  BIASASSERT(!useInliers || inliers.size() == num);
  BIASASSERT(!useWeights || w.size() == num);

  // Compute residual error of inliers only
  if (useInliers) {
    double errorSum = 0.0;
    solver_.GetResidualErrors(errors, X, Y, dR, dt, ds, w);
    for (register unsigned int i = 0; i < num; i++) {
      if (inliers[i]) {
        errorSum += fabs(errors[i]);
      }
    }
    return errorSum;
  } else {
    return solver_.GetResidualErrors(errors, X, Y, dR, dt, ds, w);
  }
}

double AbsoluteOrientationRANSAC::
GetResidualErrors(std::vector<double> &errors,
                  const std::vector< BIAS::Vector3<double> > &X,
                  const std::vector< BIAS::Vector3<double> > &Y,
                  const std::vector< BIAS::Matrix3x3<double> > &CovX,
                  const std::vector< BIAS::Matrix3x3<double> > &CovY,
                  const BIAS::RMatrix &dR, const BIAS::Vector3<double> &dt,
                  const double &ds,  const std::vector<bool> &inliers)
{
  const unsigned int num = X.size();
  const bool useInliers = (inliers.size() > 0);

  BIASASSERT(X.size() == num && Y.size() == num);
  BIASASSERT(CovX.size() == num && CovY.size() == num);
  BIASASSERT(!useInliers || inliers.size() == num);

  // Compute residual error of inliers only
  if (useInliers) {
    double errorSum = 0.0;
    solver_.GetResidualErrors(errors, X, Y, CovX, CovY, dR, dt, ds);
    for (register unsigned int i = 0; i < num; i++) {
      if (inliers[i]) {
        errorSum += fabs(errors[i]);
      }
    }
    return errorSum;
  } else {
    return solver_.GetResidualErrors(errors, X, Y, CovX, CovY, dR, dt, ds);
  }
}

int AbsoluteOrientationRANSAC::
CreateRandomSample_(unsigned int samplesize,
                    const std::vector< Vector3<double> > &X,
                    const std::vector< Vector3<double> > &Y,
                    const std::vector< Matrix3x3<double> > &CovX,
                    const std::vector< Matrix3x3<double> > &CovY,
                    const std::vector<double> &w,
                    std::vector<int> &samples,
                    std::vector< Vector3<double> > &sampleX,
                    std::vector< Vector3<double> > &sampleY,
                    std::vector< Matrix3x3<double> > &sampleCovX,
                    std::vector< Matrix3x3<double> > &sampleCovY,
                    std::vector<double> &samplew)
{
  const unsigned int num = X.size();
  const bool useCovs = (CovX.size() > 0 || CovY.size() > 0);
  const bool useWeights = (w.size() > 0);

  BIASASSERT(X.size() == num && Y.size() == num);
  BIASASSERT(!useCovs || (CovX.size() == num && CovY.size() == num));
  BIASASSERT(!useWeights || w.size() == num);
  BIASASSERT(num >= samplesize);

  samples.clear();
  sampleX.clear();
  sampleY.clear();
  samples.reserve(samplesize);
  sampleX.reserve(samplesize);
  sampleY.reserve(samplesize);
  if (useCovs) {
    sampleCovX.clear();
    sampleCovY.clear();
    sampleCovX.reserve(samplesize);
    sampleCovY.reserve(samplesize);
  }
  if (useWeights) {
    samplew.clear();
    samplew.reserve(samplesize);
  }
  int r;
  std::vector<bool> selected(num, false);
  for (register unsigned int i = 0; i < samplesize; i++) {
    r = randomizer_.GetUniformDistributedInt(0, int(num)-1);
    for (register unsigned int j = 0; selected[r] && j < num; j++) {
      r = (r+1) % num;
    }
    selected[r] = true;
    samples.push_back(r);
    sampleX.push_back(X[r]);
    sampleY.push_back(Y[r]);
    if (useCovs) {
      sampleCovX.push_back(CovX[r]);
      sampleCovY.push_back(CovY[r]);
    }
    if (useWeights) {
      samplew.push_back(w[r]);
    }
  }
  return samples.size();
}

int AbsoluteOrientationRANSAC::
EvaluateInliers_(std::vector<bool> &inliers, 
                 std::vector<double> &errors, double &errorSum,
                 const std::vector< Vector3<double> > &X,
                 const std::vector< Vector3<double> > &Y,
                 const std::vector< Matrix3x3<double> > &CovX,
                 const std::vector< Matrix3x3<double> > &CovY,
                 const RMatrix &dR, const Vector3<double> &dt, const double &ds,
                 const std::vector<double> &w)
{
  const unsigned int num = X.size();
  const bool useCovs = (CovX.size() > 0 || CovY.size() > 0);
#ifdef BIAS_DEBUG
  const bool useWeights = (w.size() > 0);
#endif //BIAS_DEBUG
  BIASASSERT(X.size() == num && Y.size() == num);
  BIASASSERT(!useCovs || (CovX.size() == num && CovY.size() == num));
  BIASASSERT(!useWeights || w.size() == num);
#ifdef BIAS_DEBUG
  if (useWeights && useCovs) {
    BIASERR("Weights and covariances should NOT be used at the same time!");
    BIASABORT;
  }
#endif

  // Compute residual errors for all correspondences
  if (useCovs) {
    solver_.GetResidualErrors(errors, X, Y, CovX, CovY, dR, dt, ds);
  } else {
    solver_.GetResidualErrors(errors, X, Y, dR, dt, ds, w);
  }

  // Compute inliers and squared error sum for inliers only
  inliers.resize(num);
  errorSum = 0.0;
  int inlierCount = 0;
  double error;
  for (register unsigned int i = 0; i < num; i++) {
    error = fabs(errors[i]);
    if (error > maxError_) {
      inliers[i] = false;
    } else {
      inliers[i] = true;
      inlierCount++;
      errorSum += error;
    }
  }
  return inlierCount;
}
