/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersPerspectiveDepth_hh__
#define __BIAS_ProjectionParametersPerspectiveDepth_hh__

#include "bias_config.h"
#include <MathAlgo/Interpolator.hh>
#include <Base/Common/BIASpragmaStart.hh>
#include <Geometry/ProjectionParametersPerspective.hh>

namespace BIAS {

static unsigned nrOfSplineParameters_ =12;
static unsigned nrOfReflectivityParameters_=3;
 
  enum BIAS_TOF_DEPTH_ERROR_MODEL
    {
      DEPTH_MODEL_POLYNOM = 0,   // standard depth error model
      DEPTH_MODEL_LINEAR  = 1,   // Linear depth error model
      DEPTH_MODEL_SPLINE  = 2,    // Splines depth error model
      DEPTH_MODEL_NONE  = 3    // No depth deviation
    };

  /** @class ProjectionParametersPerspectiveDepth
   *  @ingroup g_geometry
   *  @brief additional depth calibration parameters for a perspective depth camera
   */
  class BIASGeometry_EXPORT ProjectionParametersPerspectiveDepth
    : public  ProjectionParametersPerspective {
  public:

    ProjectionParametersPerspectiveDepth(const unsigned int width = 0,
                                         const unsigned int height = 0);     

    ProjectionParametersPerspectiveDepth(BIAS_TOF_DEPTH_ERROR_MODEL model);
    
    ProjectionParametersPerspectiveDepth(const ProjectionParametersPerspectiveDepth & P)
      {
        Init();
        *this = P;
      };

    ProjectionParametersPerspectiveDepth(const ProjectionParametersPerspective & P)
      {
        Init();
        *this = P;
      };

    ProjectionParametersPerspectiveDepth&
      operator=(const ProjectionParametersPerspectiveDepth & P){
      ProjectionParametersPerspective::operator=(P);
      DepthCalibrationParameters_ = P.DepthCalibrationParameters_;
      listSplineKnotPoints_ = P.listSplineKnotPoints_;
      DepthCalibrationModel_= P.DepthCalibrationModel_;
   
      ReflectivityCalibrationParameters_ = P.ReflectivityCalibrationParameters_;
      depthIntNormValues_ = P.depthIntNormValues_;
      minNormIntCoeffs_ = P.minNormIntCoeffs_;
      maxNormIntCoeffs_= P.maxNormIntCoeffs_;
      radiusIntNormCoeffs_ = P.radiusIntNormCoeffs_;
      minDepth_ = P.minDepth_;
      maxDepth_ = P.maxDepth_;
      if(DepthCalibrationModel_ == DEPTH_MODEL_SPLINE)
        PrepareSpline_();
      return *this;
    };
 
    /** @brief Init function called by constructors       
     */
    void Init();

    ProjectionParametersPerspectiveDepth&
      operator=(const ProjectionParametersPerspective & P){
      ProjectionParametersPerspective::operator=(P);
      DepthCalibrationParameters_.resize(6, 0.0);
      nrOfReflectivityParameters_ = 3;
      DepthCalibrationModel_=DEPTH_MODEL_POLYNOM;
      // 3 for Reflectivity polynom f(), 2 for radial g()
      ReflectivityCalibrationParameters_.resize(nrOfReflectivityParameters_, 0.0);      
      minDepth_ = 0.0;
      maxDepth_ = 15000.0;
      return *this;
    };

    virtual ~ProjectionParametersPerspectiveDepth(){
      DepthCalibrationParameters_.clear();
      listSplineKnotPoints_.clear();
      ReflectivityCalibrationParameters_.clear();
      depthIntNormValues_.clear();
      minNormIntCoeffs_.clear();
      maxNormIntCoeffs_.clear();
      radiusIntNormCoeffs_.clear();
    };


    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersPerspectiveDepth* Clone() const {
      return new ProjectionParametersPerspectiveDepth(*this);
    }

    void SetDepthCalibrationModel(BIAS_TOF_DEPTH_ERROR_MODEL model) {
      DepthCalibrationModel_= model;
      switch(model){
      case DEPTH_MODEL_POLYNOM:
        DepthCalibrationParameters_.resize(6, 0.0);
        break;
      case DEPTH_MODEL_LINEAR:
        DepthCalibrationParameters_.resize(2, 0.0);
        break;
      case DEPTH_MODEL_SPLINE:
        // 12 control points for spline
        DepthCalibrationParameters_.resize(nrOfSplineParameters_, 0.0); 
        listSplineKnotPoints_.resize(nrOfSplineParameters_, 0.0);
        
        //set first to minDepth_ 
        for(unsigned i=0;i<nrOfSplineParameters_-1;i++){
          listSplineKnotPoints_[i]=DepthCalibrationParameters_[i]=
            (minDepth_ + double(i)*(maxDepth_-minDepth_)/double(nrOfSplineParameters_-3));    
        }
        //set last to maxDepth_
        listSplineKnotPoints_[nrOfSplineParameters_-1]=
          DepthCalibrationParameters_[nrOfSplineParameters_-1] = maxDepth_;
        break;
      case DEPTH_MODEL_NONE:
      default:
        break;
      }
    }

    void SetDepthCalibrationParameters (const std::vector<double>& params,
                                        float minDepth=0.0, 
                                        float maxDepth=15000.0);

    void SetDepthCalibrationParametersSpline (const std::vector<double>& knots,
                                        const std::vector<double>& controls,
                                        float minDepth=0.0, 
                                        float maxDepth=15000.0);

    BIAS_TOF_DEPTH_ERROR_MODEL GetDepthCalibrationModel(){
      return DepthCalibrationModel_;
    }

    void GetDepthCalibrationParameters(std::vector<double> &params){
      params.clear();           
      for(unsigned i=0;i<DepthCalibrationParameters_.size();i++){
        params.push_back(DepthCalibrationParameters_[i]);
      }
    }

    void GetDepthCalibrationSplineKnots(std::vector<double> &knots){
      knots.clear();           
      for(unsigned i=0;i<listSplineKnotPoints_.size();i++){
        knots.push_back(listSplineKnotPoints_[i]);
      }
    }

    void GetMinMaxDepth(float& minDepth, 
                        float& maxDepth){
      minDepth = minDepth_;
      maxDepth = maxDepth_;
    }

    /**
     * \brief sets the reflectivity calibration parameters and Fit polynomials 
     * to radial and reflectivity normalization parameters
     * \return 0 on sucess, < 0 if polynomial fitting failes      
     */
    int SetReflectivityNormalizationParameters(const std::vector<double>& depthValues,
                                             const std::vector<double>& minInt, 
                                             const std::vector<double>& maxInt,
                                             const std::vector<double>& radiusValues,
                                             const std::vector<double>& radiusIntensities);

    void SetReflectivityCalibrationParameters (const std::vector<double>& params);

    void GetReflectivityCalibrationParameters(std::vector<double> &params){
      params.clear();
      for(unsigned i=0;i<ReflectivityCalibrationParameters_.size();i++)
        params.push_back(ReflectivityCalibrationParameters_[i]);
    }

    /* @brief Distort depth value d at image position pos */
    void DistortDepth (HomgPoint2D const &pos, float &d);

    /* @brief UnDistort depth value d at image position pos */
    void UnDistortDepth (HomgPoint2D const &pos, float &d);

    /** Undistorts the passed depth map in place.
     * Uses method UnDistortDepth.
     * \returns 0 if everything went smoothly.
     * \returns -1 if parameter's size didn't fit the input image's size.
     **/
    int UnDistortDepthMapIP(BIAS::Image<float>& depthMap) ;

    /* @brief Distort Reflectivity value i at image position pos */
    void DistortReflectivity(float &depth,float &Reflectivity,float &radius);

    /* @brief UnDistort Reflectivity value i at image position pos */
    void UnDistortReflectivity(float &depth,float &Reflectivity,float &radius);

    int DistortReflectivityDepthIP(BIAS::Image<float>& depthMap, 
                                BIAS::Image<float>& ReflectivityImage,
                                float principleX,
                                float principleY);
    int UnDistortReflectivityDepthIP(BIAS::Image<float>& depthMap, 
                                  BIAS::Image<float>& ReflectivityImage,
                                  float principleX,
                                  float principleY);


    void SetDepthDistortionToZero();
    void SetReflectivityDepthDistortionToZero();
    bool HasDepthDistortion() const;
    bool HasReflectivityDistortion() const;

    static unsigned GetNrOfDepthErrormodelParameters(BIAS_TOF_DEPTH_ERROR_MODEL model);
    static unsigned GetNrOfReflectivityParameters();


#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject);
#endif

    friend std::ostream& operator<<(std::ostream &os,
                                    const ProjectionParametersPerspectiveDepth& p);

    
  protected:
   
    float minDepth_,maxDepth_;
    std::vector<double> depthIntNormValues_,minNormIntCoeffs_,maxNormIntCoeffs_;
    std::vector<double> radiusIntNormValues_, radiusIntNormCoeffs_;
    void PrepareSpline_();
    bool bSplinePrepared_;
    BIAS::Interpolator  splineInterpolator_;
    BIAS::Interpolator  splineInterpolatorUndist_;
    std::vector<double> listSplineKnotPoints_;
    std::vector<double> DepthCalibrationParameters_;
    std::vector<double> ReflectivityCalibrationParameters_;
    BIAS_TOF_DEPTH_ERROR_MODEL DepthCalibrationModel_;
  };

  /////////////// inline code ///////////////////////

  inline std::ostream& operator<<(std::ostream &os,
                                  const ProjectionParametersPerspectiveDepth& p)
  {
    ProjectionParametersPerspective ppp = p;
    os << ppp << std::endl;
    os << "ProjectionParametersPerspectiveDepth:" << std::endl;
    switch(p.DepthCalibrationModel_){
    case DEPTH_MODEL_POLYNOM:
      os << "Model:"<<"Polynom"<<std::endl;
      break;
    case DEPTH_MODEL_LINEAR:
      os << "Model:"<<"Linear"<<std::endl;
      break;
    case DEPTH_MODEL_SPLINE:
      os << "Model:"<<"Spline"<<std::endl;
      break;
    case DEPTH_MODEL_NONE:
      os << "Model:"<<"None"<<std::endl;
      break;
    }

    if(p.DepthCalibrationModel_!=DEPTH_MODEL_NONE){
      os<<"MinDepth:"<<p.minDepth_<<" MaxDepth_:"<<p.maxDepth_<<std::endl;
      if(p.DepthCalibrationParameters_.size()>0)
        os<<"DepthCalibration:"<<std::endl;
      for (unsigned int i=0; i<p.DepthCalibrationParameters_.size(); i++) {
        os << p.DepthCalibrationParameters_[i]<<" ";
      }
      os << std::endl;
      if(p.listSplineKnotPoints_.size()>0 && p.DepthCalibrationModel_==DEPTH_MODEL_SPLINE)
        os<<"Spline Knot Points:"<<std::endl;
      for (unsigned int i=0; i<p.listSplineKnotPoints_.size(); i++) {
      	os << p.listSplineKnotPoints_[i]<<" ";
      }
      os << std::endl;
    }
    if(p.ReflectivityCalibrationParameters_.size()>0){
    	os<<"Reflectivity Calibration:"<<std::endl;
    	for (unsigned int i=0; i<p.ReflectivityCalibrationParameters_.size(); i++) {
    		os << p.ReflectivityCalibrationParameters_[i]<<" ";
    	}
    	os << std::endl;
    }
    if(p.minNormIntCoeffs_.size()>0){
    	os<<"Minimum Reflectivity Polynom :"<<std::endl;
    	for (unsigned int i=0; i<p.minNormIntCoeffs_.size(); i++) {        
    		os << "cmin" << i<<": ";
    		os << p.minNormIntCoeffs_[i]<<" ";
    	}
    	os << std::endl;
    }
    if(p.maxNormIntCoeffs_.size()>0){
    	os<<"Maximum Reflectivity Polynom :"<<std::endl;
    	for (unsigned int i=0; i<p.maxNormIntCoeffs_.size(); i++) {       
    		os << "cmax" << i<<": ";
    		os << p.maxNormIntCoeffs_[i]<<" ";
    	}
    	os << std::endl;
    }
    if(p.radiusIntNormCoeffs_.size()>0){
    	os<<"Radial Vignetting Polynom :"<<std::endl;
    	for (unsigned int i=0; i<p.radiusIntNormCoeffs_.size(); i++) {
    		os << "crad" << i<<": ";
    		os<< p.radiusIntNormCoeffs_[i]<<" ";
    	}
    	os << std::endl;
    }

    return os;
  }



} // end namespace

#include <Base/Common/BIASpragmaEnd.hh>

#endif

