/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersSpherical_hh__
#define __BIAS_ProjectionParametersSpherical_hh__
#include "bias_config.h"

#include <Base/Common/BIASpragmaStart.hh>
#include <Geometry/ProjectionParametersBase.hh>
#include <Geometry/ProjectionParametersPerspective.hh>
#include <MathAlgo/Interpolator.hh>

namespace BIAS {

 /** @class ProjectionParametersSpherical
  *  @ingroup g_geometry
     @brief camera parameters which define the mapping between rays in the
     camera coordinate system and pixels in the image as well as external pose
     @author koeser/evers/streckel
 */
  class BIASGeometry_EXPORT ProjectionParametersSpherical:
    public  ProjectionParametersBase {
  public:

    ProjectionParametersSpherical(const unsigned int width = 0,
                                  const unsigned int height = 0)
      : ProjectionParametersBase(width, height)
    {
      radius_=0;
      MaxCamAngle_=0;
      interpolatorInit_ = false;
    };

    ProjectionParametersSpherical(const double radius,
                                  const double maxAngle,
                                  const unsigned int width,
                                  const unsigned int height)
      : ProjectionParametersBase(width, height)
    {
      radius_=radius;
      MaxCamAngle_=maxAngle;
      interpolatorInit_ = false;
    }

    ProjectionParametersSpherical(const ProjectionParametersSpherical & P){
      *this = P;
    };

    ProjectionParametersSpherical&
    operator=(const ProjectionParametersSpherical & P){
      ProjectionParametersBase::operator=(P);
      radius_=P.radius_;
      MaxCamAngle_=P.MaxCamAngle_;
      interpolatorAngle_ = P.interpolatorAngle_;
      interpolatorAngleBack_ = P.interpolatorAngleBack_;
      interpolatorInit_ = P.interpolatorInit_;
      return *this;
    };

    virtual ~ProjectionParametersSpherical(){};

    /** @brief call this to start a run at the outer boundary of an image.
     *
     *  You get the first boundary position in it, e.g. (0,0) for a perspective
     *  image. Feed this into GetNextBorderPixel to get the next position.
     *  Special implmentation here, because we need to run around the FOV
     *  circle. NOT YET IMPLEMENTED, PLEASE DO !
     */
    virtual void GetFirstBorderPixel(PixelIterator& it);

    /** @brief call this iteratively to run at the outer boundary of an image.
     *
     *  All returned coordinates must have valid local rays in the camera
     *  coordinate system (e.g. in fisheye cams, we run at the fov circle) and
     *  must be in the image. Two subsequent coordinates must not be more
     *  distant than 1 pixel. NOT YET IMPLEMENTED, PLEASE DO !
     *  @param it must be initialized by GetFirstBorderPixel
     *  @return false if the returned position closes the loop around the image
     */
    virtual bool GetNextBorderPixel(PixelIterator& it);


    virtual const bool
    DoIntrinsicsDiffer(const ProjectionParametersBase* p) const;

    virtual bool Distort(BIAS::HomgPoint2D& point2d) const;

    virtual bool Undistort(BIAS::HomgPoint2D& point2d) const;



    bool DoesPointProjectIntoImageLocal(const BIAS::Vector3<double>& localX,
                                        HomgPoint2D& x,
                                        bool IgnoreDistortion = false) const;

    /** @brief calculates the projection of a point
        in the camera coordinate system
        to a pixel in the image plane of the camera

        In the simplest case perspective pinhole projection x = K * y
        where y is the projection of X using  y = (R^T | -R^T C) X
    */
    virtual HomgPoint2D ProjectLocal(const Vector3<double>& point,
                                     bool IgnoreDistortion = false) const;

    /** @brief calculates the projection of a point in the local camera
        coordinate system to a pixel in the image plane of the camera. The
        only case when this function may not compute the 2d point is when the
        camera center and the 3d point coincide. This case must be indicated
        by a negative return value. In all other cases, a 2d point must be
        computed, *particularily* when the 3d point is behind the camera
        or when the resulting 2d point is at infinity.

        In the simplest case perspective pinhole projection x = K * point
        where point is transformed of X using  point = (R^T | -R^T C) X
        @author woelk 08/2008 (c) www.vision-n.de */
    virtual int ProjectLocal(const Vector3<double>& point, HomgPoint2D &p2d,
                             bool IgnoreDistortion = false) const;

    /** @brief map points from image onto unit diameter image plane in 3D.
     * Chosen is the spherical image plane with radius of one from the
     * image center.
     */
    HomgPoint3D UnProjectToImagePlane(const HomgPoint2D& pos,
                                      const double& depth = 1.0,
                                      bool IgnoreDistortion = false) const;


    /** @brief calculates the viewing ray from the camera center (in the
        camera coordinate system) which belongs to the image position pos
        e.g. (0,0,1) means optical axis
    */
    virtual Vector3<double>
    UnProjectLocal(const HomgPoint2D& pos,
                   bool IgnoreDistortion = false) const;

    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersSpherical* Clone() const {
      return new ProjectionParametersSpherical(*this);
    };

   inline double GetRadius() const{
     return radius_;
    }

    inline void SetRadius(double r) {
      radius_ = r;
    }

    inline double GetMaxCamAngle() const {
      return MaxCamAngle_;
    }

    int SetUndistortion(const std::vector<double>& AngleCorrX,
                        const std::vector<double>& AngleCorrY,
                        const double radius);

    /** @deprecated, use vector<double> interface below */
    inline int InitAngleCorrFromPoly(double MaxCamAngle,double acCoeff0,
                                     double acCoeff1, double acCoeff2,
                                     double acCoeff3, double acCoeff4) {
      std::vector<double> coefficients;
      coefficients.reserve(5);
      coefficients.push_back(acCoeff0);
      coefficients.push_back(acCoeff1);
      coefficients.push_back(acCoeff2);
      coefficients.push_back(acCoeff3);
      coefficients.push_back(acCoeff4);
      return InitAngleCorrFromPoly(MaxCamAngle, coefficients);
    }

    /** @brief inits object from polynomial coefficients of undistortion
        (radius -> z) or distortion (theta -> radius) function  */
    int InitAngleCorrFromPoly(const double& MaxCamAngle,
                              const std::vector<double>& coefficients,
                              bool mapsRadiusToZ = true);

    /** @brief Adapt internal parameters to resampled image.
        @param ratio 2.0 = downsample by 2, 0.5 = upsample by 2
        @param offset Offset for principal point (applied before down/upsampling!)
        \attention Delivers only correct results if ratio is chosen
        to rescale image size to integral image dimensions
        in both directions. Image size is rounded!
     */
    virtual void Rescale(double ratio, const double offset = 0.0) {
      ProjectionParametersBase::Rescale(ratio, offset);
      radius_ /= ratio;
    }

    virtual void Rescale(unsigned int width, unsigned int height){
      unsigned int oldwidth = width_;
      ProjectionParametersBase::Rescale(width, height);
      radius_ /= ((double)oldwidth/(double)width_);
    }

    virtual double ViewDifference(const ProjectionParametersBase* pPPB) const;

    /** @brief Returns a fake KMatrix for the fisheye camera.

        Uses MaxAngle_ to limit maxangle, then calls PPB::GetFakeK.
        The FoV of the KMatrix is specified by maxangle (PI*80/180 for 160deg)
        or the max fisheye angle, if this is <160deg.
        @param resolution The resolution parameter influences the
        Resolution of the perspective image that is represented by the camera.
        0 - The perspective image has the same size as the fisheye image
        1 - The perspective image has the same resolution as the fisheye
        image in the center (highest resolution)
        2 - The perspective image has the same resolution as the fisheye
        image near the border (lowest resolution)
        @param imgsize returns the image size
        @param maxangle maximum theta in rad, e.g. (PI*80/180 for 160deg fov)
        @author streckel 05/2006
    */
    virtual BIAS::KMatrix GetFakeKMatrix(double& imgsize, int resolution=0,
                                         const double& maxangle = 1.4) const;

    /** @brief stupid reimplementation since this class does not inherit
        function from base class (gcc) */
    virtual BIAS::KMatrix GetFakeKMatrix(int resolution=0,
                                         const double& maxangle = 1.4) const {
      return ProjectionParametersBase::GetFakeKMatrix(resolution, maxangle);
    }

    /** Calculates perspective camera with specified viewing angle and
     * optical axis aligned with the optical ray corresponding to viewCenter
     * in spherical camera.
     * @param viewCenter pixel position in sptherical image that is the
     * center of the perspective cutout.
     * @param perspHalfViewingAngle half field of view of the cutout in Degree
     * @pp returning cutout parameters
     * \returns false if perspective cut out could not be calculated.
     * @author streckel 08/2006
     */
    bool GetPerspectiveCutOutParameters(const BIAS::HomgPoint2D& viewCenter,
                                        const double perspHalfViewingAngleDEG,
                                        ProjectionParametersPerspective& pp)
      const;

    /** @brief calculates a 3D point in
        a local camera coordinate system specified by camSystem, which
        belongs to the image position pos in cam with distance depth to the
        camera center cam. Overload to habe return value (0,0,0)
        for undefined or out of image radius pixels.
        @author streckel  */
    virtual Vector3<double> UnProjectToPointLocal(const HomgPoint2D& pos,
                                                  const double& depth,
                                                  bool IgnoreDistortion=false)
      const;


    /** computes polynomial according to scaramuzza toolbox
     *
     *  undistortion polynomial is a function z(radius) which typically
     *  has a local extremum at radius=0 and therefore cannot have a
     *  linear monomial
     *  @author koeser 04/2007 */
    int EstimateUndistortionPolynomial(std::vector<double>& coefficients,
                                       const unsigned int degree = 4,
                                       bool LinearMonomialIsZero=true) const;

    /** computes polynomial according to scaramuzza toolbox
     *
     *  the distortion polynomial is a function radius(theta) which is
     *  usually close to linear for fisheye cameras and SHOULD therefore have
     *  a linear monomial
     *  @author koeser 04/2007 */
    int EstimateDistortionPolynomial(std::vector<double>& coefficients,
                                     const unsigned int degree = 4,
                                     bool LinearMonomialIsZero=false) const;

    void GetAngleCorrCoefficients(std::vector<double> &coeffs){
      coeffs.clear();
      for(unsigned i=0;i<acCoeff_.size();i++)
        coeffs.push_back(acCoeff_[i]);
    }

    inline void GetAngleCorr(std::vector<double>& AngleCorrX, std::vector<double>& AngleCorrY ){
      interpolatorAngle_.GetKnotPoints(AngleCorrX); // AngleCorrX
      interpolatorAngle_.GetControlPoints(AngleCorrY); // AngleCorrY
    }

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject);
#endif

    friend std::ostream& operator<<(std::ostream &os,
                                    const ProjectionParametersSpherical& p);
  protected:

    /** @brief Transforms a coordinate pair of image coordinates x, y in the
        uncalibrated spherical image, to a calibrated pair of polar coordinates
        phi, theta.
        @author Birger Streckel */
    int TransfCoordImage2Sphere_ (const HomgPoint2D& Source,
                                 double &theta, double &phi) const;

    /** @brief Transforms a coordinate pair of image coordinates x, y in the
        calibrated spherical image, to a calibrated pair of polar coordinates
        phi, theta.
        @author Birger Streckel */
    int TransfCoordCalibImage2Sphere_ (const HomgPoint2D& Source,
                                      double &theta, double &phi) const;

    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in the uncalibrated
        spherical image.
        @author Birger Streckel */
    int TransfCoordSphere2Image_ (double theta, double phi,
                                 HomgPoint2D &Source) const;

    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in the uncalibrated
        spherical image.
        @author Birger Streckel */
    int TransfCoordRotateSphere2Image_ (double theta, double phi,
                                       double viewCenterX,
                                       double viewCenterY,
                                       HomgPoint2D &Source) const;

    /** @brief Transforms a calibrated pair of polar coordinates phi, theta
        to a coordinate pair of image coordinates x, y in an equidistant
        spherical image.
        @author Birger Streckel */
    int TransfCoordSphere2CalibImage_ (double theta, double phi,
                                      HomgPoint2D &Source) const;

    /** @brief theta and phi are periodic in 2*M_PI, if they are out of a
        suitable range, try to find best equivalent angle by adding 2*l*M_PI
        @return 0 on success, <0 indicates that at least one angle is invalid
        @author koeser 10/2004 */
    int EnforceNormalRange_(double& theta, double& phi) const;



    // image radius if image is spherical
    double radius_;

    // Maximum image angle of the fisheye camera
    double MaxCamAngle_;
    // these are the polynom coefficients
    std::vector<double>  acCoeff_;

    // Spline interpolators for image distortion
    // x : angle in an undistorted image
    // y : distorted angle
    // Example: x=pi, y =.8*pi means that the image at angle pi is
    // distorted with factor 0.8
    Interpolator interpolatorAngle_;

    // back-transformation
    Interpolator interpolatorAngleBack_;

    // determine if interpolator has been initialized
    bool interpolatorInit_;

  };

  /////////////// inline code ///////////////////////
  inline std::ostream& operator<<(std::ostream &os,
                                    const ProjectionParametersSpherical& p)
  {
    os<<"ProjectionParametersSpherical:"<<std::endl;
    os<<"Radius: "<<p.radius_<<std::endl;
    os<<"MaxCamAngle: "<<p.MaxCamAngle_<<std::endl;
    os<<"Coefficients:"<<std::endl;
    std::vector<double> controlpoints, knotpoints;
    p.interpolatorAngle_.GetControlPoints(controlpoints);
    p.interpolatorAngle_.GetKnotPoints(knotpoints);
    for (unsigned int i=0; i<controlpoints.size(); i++) {
      os<< knotpoints[i]<<" "<< controlpoints[i]<<std::endl;
    }
    return os;
  }

} // end namespace


#include <Base/Common/BIASpragmaEnd.hh>

#endif

