/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "RMatrix.hh"
#include <MathAlgo/SVD.hh>
#include <MathAlgo/Minpack.hh> 
#include <Base/Geometry/Quaternion.hh>

using namespace BIAS;
using namespace std;




RMatrix::~RMatrix()
{}

RMatrix::RMatrix(const TNT::Matrix<ROTATION_MATRIX_TYPE>& mat)
  : RMatrixBase(mat)
{
  if ((mat.num_rows() != 3) || (mat.num_cols() != 3)) {
    BIASERR("wrong matrix size "<<mat.num_rows()<<"x"
           << mat.num_cols());
  }
}

int RMatrix::EnforceRConstraints()
{
  // enforce real rotation matrix using SVD
  // by setting all singular values to unity 
  SVD RotSVD(Matrix<ROTATION_MATRIX_TYPE>(*this));
  *this = RotSVD.GetU() * RotSVD.GetVT();

  // enforce right-handed system (det = +1)
  if (this->GetDeterminant()<0.0) *this *= -1.0;
  return 0;
}



// -------------------- numerical L-M error computation ---------------------
vector<RMatrix> const *pGLOBALvecR = NULL;
double GLOBALbestErrorSoFar = DBL_MAX;
Quaternion<double> GLOBALbestQuaternionSoFar(0,0,0,1);

void ErrorFunction(long int * /*numerrors*/, long int *numvars, double *x,
                   double *fvec, long int * /*iflag*/) 
{
  BIASASSERT(numvars!=NULL);
  BIASASSERT(*numvars==4);

  // create rotation matrix vector from parameterization
  Quaternion<double> q(x[0],x[1],x[2],x[3]);
  q.Normalize();
  RMatrixBase R;
  R.SetFromQuaternion(q);
 

  // compute residuum
  double error = 0, err;
  for (unsigned int j=0; j<pGLOBALvecR->size(); j++) {
    for (unsigned int row=0; row<3; row++) {
      for (unsigned int col=0; col<3; col++) {
        err = fvec[9*j+3*col+row]   
          = R[row][col] - (*pGLOBALvecR)[j][row][col];
        error += err*err;
      }
    }
  }

  // in case lm runs out of iterations, remember best in a global variable
  if (error<GLOBALbestErrorSoFar) {
    GLOBALbestErrorSoFar = error;
    GLOBALbestQuaternionSoFar = q;
  }
  cout<<error<<" ";
}
// ---------------------------------------------------------------------------







RMatrix RMatrix::GetAverageRMatrix(const std::vector<RMatrix>& vecR,
                                   bool RefineRotation) const {
  BIASASSERT(!vecR.empty());
  BIASWARN("untested code so far !");
  RMatrix Result = vecR[0];
  Vector3<double> Axis(0,0,0);
  double angle(0.0);
  double weight(1.0);
  // for the first two matrices we have RDiff = vecR[0] * vecR[1].Transpose();
  // Therefore Result = sqrt(Rdiff) * vecR[1] = sqrt(Rdiff^T) * vecR[1]
  // is an average orientation, where
  // square root is realized by axis-angle decomposition (sqrt = half angle)
  for (unsigned int i=1; i<vecR.size(); i++) {
    // after the first averaging, our result becomes more reliable, so
    // use a higher weight each time
    // this is equivalent to not using sqrt = ^0.5 but ^weight
    weight = 1.0 / double(i+1);
    RMatrix RDiff = Result * vecR[i].Transpose();
    RDiff.GetRotationAxisAngle(Axis, angle);
    // compute transpose of diff by -1 and go only by weight into this new dir
    angle *= -1.0 * weight;
    RDiff.Set(Axis, angle);
    Result = RDiff * Result;
    Result.EnforceRConstraints();
  }

  // now if desired, take this as astart value for nonlinear optimization 
  if (RefineRotation) {
    pGLOBALvecR = &vecR;
    GLOBALbestErrorSoFar = DBL_MAX;
    long int num_params = 4;
    long int num_errors = 9 * vecR.size();
    long int max_error_calls = 1000;
    double epsilon_for_jacobian_approximation = 0.02;
    
    // set up parametrization
    Quaternion<double> q(0,0,0,1), qres(0,0,0,1);
    Result.GetQuaternion(q);
    Vector<double> qvec(4),qresvec(4);
    for (unsigned int i=0; i<4; i++)  qvec[i] = q[i];
    
    long int LMResult = 
      LevenbergMarquardtExtended(&ErrorFunction, num_errors, num_params, 
                                 qvec, qresvec, 
                                 epsilon_for_jacobian_approximation, 1e-10, 
                                 max_error_calls);
    if ((LMResult != LM_OK) && (LMResult != LM_TOO_MANY_IT)) {
      BIASERR("Error in Levenberg-Marquardt-Optimization. Error code is "
              <<LMResult);
    } else {
      cout<<"minimization terminated with code "<<LMResult<<endl;
      // dirty non-threadsafe workaround in case LM makes too many iters:
      Result.SetFromQuaternion(GLOBALbestQuaternionSoFar);
      cout<<"best error so far is "<<GLOBALbestErrorSoFar<<endl;
    }
  }

  return Result;
}
