/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __TrifocalTensor_hh__
#define __TrifocalTensor_hh__

#include <Base/Geometry/TrifocalTensorBase.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Math/Matrix4x4.hh>
#include <Geometry/PMatrix.hh>
#include <Geometry/FMatrix.hh>

namespace BIAS {

  class BIASGeometry_EXPORT TrifocalTensor : public TrifocalTensorBase
  {
  public:
    TrifocalTensor();
    ~TrifocalTensor();
    TrifocalTensor(BIAS::PMatrix &P1, BIAS::PMatrix &P2, BIAS::PMatrix &P3);
    TrifocalTensor(BIAS::PMatrix &P2, BIAS::PMatrix &P3);
    void Init();

    /** Composes the trifocal tensor from three 3x4 camera matrices.
        @author Matthias Dunda */
    void Compose(BIAS::PMatrix &P1, BIAS::PMatrix &P2, BIAS::PMatrix &P3);
    
    /** Composes the trifocal tensor from two 3x4 camera matrices,
        assuming that P1=[I | 0].
        @author Matthias Dunda */
    void Compose(BIAS::PMatrix &P2, BIAS::PMatrix &P3);

    /** @return the canonical H matrix that is computed when composing
        from three camera-matrices with P1 != [I|0]
        @author mdunda 01 2004 */
    inline BIAS::Matrix4x4<PMATRIX_TYPE> GetCanonicalH() const {return CanH_;}

    // epipoles -----------------------------------------------------------
    /** @return epipole in the second view according to the first
        @param  ForceReCalc forces a recalculation although a value is cached
        @author Matthias Dunda */
    BIAS::HomgPoint2D GetEpipole21( bool ForceReCalc=false );

    /** @return epipole in the third view according to the first
        @param  ForceReCalc forces a recalculation although a value is cached
        @author Matthias Dunda */
    BIAS::HomgPoint2D GetEpipole31( bool ForceReCalc=false );

    void GetEpipolarLines(Vector3<TRIFOCALTENSOR_TYPE> &Point1,
                          HomgLine2D &Line2,
                          HomgLine2D &Line3);

    // F-matrices ---------------------------------------------------------
    /** @return the F-matrix between the first and the second view
        @param  ForceReCalc forces a recalculation although a value is cached
        @author Matthias Dunda */
    BIAS::FMatrix GetFMatrix21( bool ForceReCalc=false );
    /** @return the F-matrix between the first and the third view
        @param  ForceReCalc forces a recalculation although a value is cached
        @author Matthias Dunda */
    BIAS::FMatrix GetFMatrix31( bool ForceReCalc=false );

    // P-matrices ---------------------------------------------------------
    /** @return the P-matrix for the second view (assuming P1=[I|0])
        @param  ForceReCalc forces a recalculation although a value is cached
        @author Matthias Dunda */
    BIAS::PMatrix GetPMatrix2( bool ForceReCalc=false );
    /** @return the P-matrix for the third view (assuming P1=[I|0])
        @param  ForceReCalc forces a recalculation although a value is cached
        @author Matthias Dunda */
    BIAS::PMatrix GetPMatrix3( bool ForceReCalc=false );

    // Transfer -----------------------------------------------------------
    /** @return the transferred point in the third view given points
        in the first and second view
        @author mdunda */
    HomgPoint2D TransferView3(Vector3<TRIFOCALTENSOR_TYPE> &x1,
                              Vector3<TRIFOCALTENSOR_TYPE> &x2);

    TrifocalTensor& operator=(const TrifocalTensor& t);

  protected:
    BIAS::Matrix4x4<PMATRIX_TYPE> CanH_;
    BIAS::HomgPoint2D *Epipole21_;
    BIAS::HomgPoint2D *Epipole31_;
    void ComputeEpipoles();

  }; // class

} // namespace


#endif /* __TrifocalTensor_hh__ */

