#include <Base/Common/BIASpragma.hh>
#include <Gui/DrawTextWx.hh>
#include <Base/Image/Image.hh>
#include <Base/Image/ColourRGB.hh>
#include <Gui/StringConv.hh>

#include <wx/wx.h>

using namespace BIAS;
using namespace std;

template <class StorageType>
int DrawTextWx<StorageType>::
GetTextExtend(const std::string &text,
              unsigned &width, unsigned &height,
              const unsigned point_size)
{
  wxBitmap bmp(10, 10);
  wxMemoryDC dc(bmp);
  if (!dc.IsOk()){
    BIASERR("DrawTextWx::Text(): wxMemoryDC is not OK\n");
    return DTWX_WX_ERROR;
  }
  wxFont font = dc.GetFont();
  font.SetPointSize(point_size);
  dc.SetFont(font);

  int res = DTWX_OK;
  wxString str = AsciiToWx(text);
#ifdef WIN32
  wxCoord x,y;
  dc.GetTextExtent(str,&x,&y);
  width = x;
  height =y;
#else  
  wxSize size = dc.GetTextExtent(str);
  width = size.GetWidth();
  height = size.GetHeight();
#endif
  return res;
}


template <class StorageType>
int DrawTextWx<StorageType>::
Text(Image<StorageType> &im, const std::string &text,
     const unsigned &pos_x, const unsigned &pos_y, 
     const ColourRGB<StorageType> &foreground_color,
     const ColourRGB<StorageType> &background_color,
     const unsigned point_size,
     const enum EVerticalAlign vertical_align,
     const enum EHorizontalAlign horizontal_align)
{
  if (im.GetChannelCount()!=3){
    BIASERR("DrawTextWx::Text(): Cannot draw text in single channel image.\n");
    return DTWX_INVALID_ARGUMENT;
  }
  if (im.GetColorModel()!=ImageBase::CM_RGB){
    BIASERR("DrawTextWx::Text(): Works only with RGB images.\n"); 
    return DTWX_INVALID_ARGUMENT;
  }

  const bool static_data = true;
  wxImage wximg(im.GetWidth(), im.GetHeight(),
                im.GetImageData(), static_data);
  if (!wximg.IsOk()){
    BIASERR("DrawTextWx::Text(): wxImage is not OK\n");
    return DTWX_WX_ERROR;
  }
  // the following line core dumps in console applications, because wx is
  // not competely initialized. It is absolutly necessary to derive from 
  // the class wxApp and implement teh function OnInit() to use this class
  wxBitmap bmp(wximg); 
  if (!bmp.IsOk()){
    BIASERR("DrawTextWx::Text(): wxBitmap is not OK\n");
    return DTWX_WX_ERROR;
  }
  wxMemoryDC dc(bmp);
  if (!dc.IsOk()){
    BIASERR("DrawTextWx::Text(): wxMemoryDC is not OK\n");
    return DTWX_WX_ERROR;
  }

  dc.SetTextForeground(wxColor(foreground_color[0],
                               foreground_color[1],
                               foreground_color[2]));
  dc.SetTextBackground(wxColour(background_color[0],
                                background_color[1],
                                background_color[2]));

  wxFont font = dc.GetFont();
  font.SetPointSize(point_size);
  dc.SetFont(font);

  int res = DTWX_OK;
  wxString str = AsciiToWx(text);
#ifdef WIN32
  wxCoord w,h;
  dc.GetTextExtent(str,&w,&h);
  wxSize size(w,h);
#else
  wxSize size = dc.GetTextExtent(str);
#endif
  int px = 0, py = 0;
  switch (vertical_align){
  case V_ALIGN_TOP: py = (int)pos_y; break;
  case V_ALIGN_CENTER: py = (int)pos_y - size.GetHeight()/2; break;
  case V_ALIGN_BOTTOM: py = (int)pos_y - size.GetHeight(); break;
  default: cerr << "unkown vertical alignment: "<<vertical_align<<endl; 
    return DTWX_INVALID_ARGUMENT; break;
  } 
  if (py<0 || py+size.GetHeight()>=(int)im.GetHeight()){ 
    res = DTWX_OUT_OF_IMAGE; 
  }
  switch (horizontal_align){
  case H_ALIGN_LEFT: px = (int)pos_x; break;
  case H_ALIGN_CENTER: px = (int)pos_x - size.GetWidth()/2; break;
  case H_ALIGN_RIGHT: px = (int)pos_x - size.GetWidth(); break;
  default: cerr << "unkown horizontal alignment: "<<horizontal_align<<endl; 
    return DTWX_INVALID_ARGUMENT; break;
  } 
  if (px<0 || px+size.GetWidth()>=(int)im.GetWidth()){ 
    res = DTWX_OUT_OF_IMAGE; 
  }
  dc.DrawText(str, px, py);

  dc.SelectObject(wxNullBitmap);
  
  wxImage dst_img = bmp.ConvertToImage();
  
  const unsigned img_size = im.GetSizeByte();
  memcpy(im.GetImageData(), dst_img.GetData(), img_size); 

  return res;
}


///////////////////////////////////
// instantiation
///////////////////////////////////

#define INST(type) template class BIASGui_EXPORT DrawTextWx<type>

// create instances
INST(unsigned char);
// INST(float);
// #ifdef BUILD_IMAGE_INT
// INST(int);
// #endif
// #ifdef BUILD_IMAGE_CHAR
// INST(char);
// #endif
// #ifdef BUILD_IMAGE_SHORT
// INST(short);
// #endif
// #if defined(BUILD_IMAGE_USHORT)
// INST(unsigned short);
// #endif
// #ifdef BUILD_IMAGE_DOUBLE
// INST(double);
// #endif
// #ifdef BUILD_IMAGE_UINT
// INST(unsigned int);
// #endif

#undef INST
