/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Gui/FilterDialogBase.hh>
#include <Gui/FilterDialogPreviewWindow.hh>
#include <Gui/ImageCanvas.hh>
#include <Gui/StringConv.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageIO.hh>
#include <wx/filedlg.h>

using namespace BIAS;

DEFINE_EVENT_TYPE(wxEVT_FILTER_SAVE)

template <class InputST, class OutputST>
FilterDialogBase<InputST, OutputST>
::FilterDialogBase(const Image<InputST>& origImage,
				   wxWindow* parent,
				   wxWindowID id,
				   const wxString& title,
				   const wxString& name,
				   const wxPoint& pos,
				   const wxSize& size,
				   long style)
				   : wxFrame(parent, id, title, pos, size, style, name),
				   OrigImg(origImage),
				   FilteredImage(Image<OutputST>()),   // at first empty but filled below
				   SavedImagePath(wxEmptyString),
				   MainSizer(new wxBoxSizer(wxVERTICAL)),
				   ParamSizer(new wxBoxSizer(wxHORIZONTAL)),
				   ButtonSizer(new wxBoxSizer(wxHORIZONTAL)),  
				   ImageIsFiltered(false)
{
	SaveButton = new wxButton(this,
		wxID_OK,
		wxT("Save"),
		wxDefaultPosition,
		wxDefaultSize,
		0);
	FilterButton = new wxButton(this,
		BIAS::ID_BTN_FILTER_FDB,
		wxT("Filter"),
		wxDefaultPosition,
		wxDefaultSize,
		0);
	PreviewButton = new wxButton(this,
		BIAS::ID_BTN_PREVIEW_FDB,
		wxT("Preview"),
		wxDefaultPosition,
		wxDefaultSize,
		0);
	ResetButton = new wxButton(this,
		BIAS::ID_BTN_RESET_FDB,
		wxT("Reset"),
		wxDefaultPosition,
		wxDefaultSize,
		0);
	CancelButton = new wxButton(this,
		wxID_CANCEL,
		wxT("Cancel"),
		wxDefaultPosition,
		wxDefaultSize,
		0);
	Line = new wxStaticLine(this,
		wxID_STATIC,
		wxDefaultPosition,
		wxDefaultSize,
		wxLI_HORIZONTAL);

	CopyOriginalImage(); // now fill the filtered image
	SetSizer(MainSizer);

	MainSizer->Add(ParamSizer, 0, wxALIGN_CENTER_HORIZONTAL | wxALL, 5);
	MainSizer->Add(Line, 0, wxGROW | wxALL, 5);
	MainSizer->Add(ButtonSizer, 0, wxALIGN_CENTER_HORIZONTAL | wxALL, 5);

	// buttons
	ButtonSizer->Add(SaveButton, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);
	ButtonSizer->Add(FilterButton, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);
	ButtonSizer->Add(PreviewButton, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);
	ButtonSizer->Add(ResetButton, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);
	ButtonSizer->Add(CancelButton, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

	// connect events to handlers
	// Note: Filter Button is not connected here, because implementing
	// class should do it.
	Connect(wxID_CANCEL,
		wxEVT_COMMAND_BUTTON_CLICKED,
		wxCommandEventHandler(FilterDialogBase::OnCancelButton));
	Connect(wxID_OK,
		wxEVT_COMMAND_BUTTON_CLICKED,
		wxCommandEventHandler(FilterDialogBase::SaveFilteredImage));
	Connect(BIAS::ID_BTN_PREVIEW_FDB,
		wxEVT_COMMAND_BUTTON_CLICKED,
		wxCommandEventHandler(FilterDialogBase::OnPreviewButton));
	Connect(BIAS::ID_BTN_RESET_FDB,
		wxEVT_COMMAND_BUTTON_CLICKED,
		wxCommandEventHandler(FilterDialogBase::OnResetButton));
}


template <class InputST, class OutputST>
FilterDialogBase<InputST, OutputST>::~FilterDialogBase() { }


template <class InputST, class OutputST>
wxBoxSizer* FilterDialogBase<InputST, OutputST>::GetMainSizer() const
{
	return this->MainSizer;
}


template <class InputST, class OutputST>
wxBoxSizer* FilterDialogBase<InputST, OutputST>::GetParamSizer() const
{
	return this->ParamSizer;
}


template <class InputST, class OutputST>
wxBoxSizer* FilterDialogBase<InputST, OutputST>::GetButtonSizer() const
{
	return this->ButtonSizer;
}


template <class InputST, class OutputST>
Image<InputST> FilterDialogBase<InputST, OutputST>::GetOriginalImage() const
{
	return this->OrigImg;
}


template <class InputST, class OutputST>
Image<OutputST> FilterDialogBase<InputST, OutputST>::GetFilteredImage() const
{
	return this->FilteredImage;
}


template <class InputST, class OutputST>
bool FilterDialogBase<InputST, OutputST>::GetImageIsFiltered() const
{
	return this->ImageIsFiltered;
}


template <class InputST, class OutputST>
wxString FilterDialogBase<InputST, OutputST>::GetSavedImagePath() const
{
	return this->SavedImagePath;
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>
::SetOriginalImage(const Image<InputST>& img)
{
	this->OrigImg = img;
	this->CopyOriginalImage(); // update also the filtered image
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>
::SetFilteredImage(const Image<OutputST>& img)
{
	this->FilteredImage = img;
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>::SetImageIsFiltered(const bool filtered)
{
	this->ImageIsFiltered = filtered;
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>::CopyOriginalImage()
{
	if (typeid(InputST) != typeid(OutputST)) {

		int notOk;

		if (typeid(unsigned char) == typeid(OutputST)) {
			notOk = ImageConvert::ConvertST(this->OrigImg,
				this->FilteredImage,
				ImageBase::ST_unsignedchar);
			if (notOk) {
				BIASERR("FilterDialogBase::CopyOriginalImage: Error Converting StorageType");
			}

		} else if (typeid(float) == typeid(OutputST)) {
			notOk = ImageConvert::ConvertST(this->OrigImg,
				this->FilteredImage,
				ImageBase::ST_float);
			if (notOk) {
				BIASERR("FilterDialogBase::CopyOriginalImage: Error Converting StorageType");
			}

		} else {
			BIASERR("FilterDialogBase::CopyOriginalImage: StorageType not supported yet");
		}

	} else {
		this->FilteredImage = this->OrigImg;
	}

}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>::SaveFilteredImage(wxCommandEvent& WXUNUSED(event))
{
	wxFileDialog dlg(this,
		wxT("Where To Save?"),
		this->SavedImagePath,
		wxT("FilteredImage.mip"),
		wxT("*.*"),
		wxSAVE | wxOVERWRITE_PROMPT);
	wxString path = wxEmptyString;
	if (wxID_OK == dlg.ShowModal()) {
		this->SavedImagePath = dlg.GetDirectory();
		path = dlg.GetPath();
		if (0 != ImageIO::Save(WxToAscii(path), this->FilteredImage)) {
			BIASERR("FilterDialogBase::SaveFilteredImage: Saving Error");
		} else {
			wxCommandEvent event( wxEVT_FILTER_SAVE, GetId() );
			event.SetEventObject( this );
			event.SetString( path );
			// Send it
			GetEventHandler()->ProcessEvent( event );
		}
	}
}


template<class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>::EnableSaveButton(bool enable) const
{
	this->SaveButton->Enable(enable);
}


template<class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>::EnableFilterButton(bool enable) const
{
	this->FilterButton->Enable(enable);
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>::EnablePreviewButton(bool enable) const
{
	this->PreviewButton->Enable(enable);
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>::EnableResetButton(bool enable) const
{
	this->ResetButton->Enable(enable);
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>::EnableCancelButton(bool enable) const
{
	this->CancelButton->Enable(enable);
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>::RemoveLine()
{
	this->MainSizer->Detach(this->Line);
}

/******************************************************************************/
/* ============================== protected ================================= */
/******************************************************************************/

template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>
::OnPreviewButton(wxCommandEvent& WXUNUSED(event))
{
	FilterDialogPreviewWindow<OutputST>* PreviewWin;
	PreviewWin = new FilterDialogPreviewWindow<OutputST>(this->FilteredImage,
		this,
		wxT("Preview Window"),
		BIAS::ID_WIN_PREVIEW);
	PreviewWin->Show(true);
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>
::OnResetButton(wxCommandEvent& WXUNUSED(event))
{
	if (this->ImageIsFiltered) {
		this->ImageIsFiltered = false;
		this->CopyOriginalImage();
	}
}


template <class InputST, class OutputST>
void FilterDialogBase<InputST, OutputST>
::OnCancelButton(wxCommandEvent& WXUNUSED(event))
{
	this->Destroy();
}


/* =================== supported storagetype combinations =================== */
template class FilterDialogBase<unsigned char, unsigned char>;
template class FilterDialogBase<unsigned char, float>;
template class FilterDialogBase<float, unsigned char>;
template class FilterDialogBase<float, float>;
