/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __FILTER_DIALOG_BASE_HH_
#define __FILTER_DIALOG_BASE_HH_

#include <Base/Common/BIASpragmaStart.hh>
#include <Base/Image/Image.hh>
#include <wx/wx.h>
#include <wx/frame.h>
#include <wx/button.h>
#include <wx/sizer.h>
#include <wx/statline.h>
#include <Base/Common/BIASpragmaEnd.hh>

DECLARE_EVENT_TYPE(wxEVT_FILTER_SAVE, -1)

namespace BIAS {

	/**
	* @class FilterDialogBase
	* @ingroup g_gui
	* @brief Base class for all filter dialogs
	*
	* this class is a simple container with some gui elements and it provides some
	* very basic funtionality. It doesn't contain a Filter because it is intended
	* that subclasses can take whatever Filter they want.
	*
	* @author jfernand 08/09
	*/

	template <class InputST, class OutputST>
	class BIASGui_EXPORT FilterDialogBase : public wxFrame
	{

	public:

		FilterDialogBase(const Image<InputST>& origImg,
			wxWindow* parent,
			wxWindowID id,
			const wxString& title,
			const wxString& name = wxT("Filter Dialog Base"),
			const wxPoint& pos = wxDefaultPosition,
			const wxSize& size = wxDefaultSize,
			long style = wxDEFAULT_FRAME_STYLE
			& ~ (wxRESIZE_BORDER
			| wxRESIZE_BOX
			| wxMAXIMIZE_BOX));
		virtual ~FilterDialogBase();

		/**
		* Orientation: wxVERTICAL
		* @return pointer to the main sizer of this window
		*/
		virtual wxBoxSizer* GetMainSizer() const;

		/**
		* It's purpose is to hold few elements like spinctrl/slider/etc for parameter
		* inputs. If you want elements to be added vertically, then add another sizer
		* in this sizer with your flavor.
		*
		* Orientation: wxHORIZONTAL
		*
		* @return pointer to the sizer (wxHORIZONTAL).
		*/
		virtual wxBoxSizer* GetParamSizer() const;

		/**
		* @return pointer to the sizer which holds the buttons
		* Orientation: wxHORIZONTAL
		*/
		virtual wxBoxSizer* GetButtonSizer() const;

		/**
		* @return original (NEVER filtered!!!) image
		*/
		Image<InputST> GetOriginalImage() const;

		/**
		* @return filtered image if it was succesfully filtered
		*/
		Image<OutputST> GetFilteredImage() const;

		/**
		* @return TRUE if image was filtered
		*/
		virtual bool GetImageIsFiltered() const;

		/**
		* @return the directory in which the filtered image was saved
		*/
		virtual wxString GetSavedImagePath() const;

		/**
		* Sets the source image for filter
		*/
		virtual void SetOriginalImage(const Image<InputST>& img);

		/**
		* Sets the filtered image. Used by derived classes for replacing the old
		* filtered image with that from the successful filter procedure.
		*/
		virtual void SetFilteredImage(const Image<OutputST>& img);

		/**
		* Sets if image was filtered or not.
		*/
		virtual void SetImageIsFiltered(const bool filtered);

		/**
		* Copy the OrigImg's content to the FilteredImage
		*/
		virtual void CopyOriginalImage();

		/**
		* Save the image to the disk
		*/
		void SaveFilteredImage(wxCommandEvent&  WXUNUSED(event));

		virtual void EnableSaveButton(bool enable = true) const;
		virtual void EnableFilterButton(bool enable = true) const;
		virtual void EnablePreviewButton(bool enable = true) const;
		virtual void EnableResetButton(bool enable = true) const;
		virtual void EnableCancelButton(bool enable = true) const;

		/**
		* Just for optical reason, if you don't want that line, just remove it
		*/
		virtual void RemoveLine();


	protected:

		/**
		* The save logic should be implemented here
		*/
		//virtual void OnSaveButton(wxCommandEvent& WXUNUSED(event)) = 0;

		/**
		* The filter logic should be implemented here
		*/
		virtual void OnFilterButton(wxCommandEvent& WXUNUSED(event)) = 0;

		/**
		* Shows the filtered image
		*/
		virtual void OnPreviewButton(wxCommandEvent& WXUNUSED(event));

		/**
		* Reset the original image but it does not reset the given parameter(s)
		*/
		virtual void OnResetButton(wxCommandEvent& WXUNUSED(event));

		/**
		* Quits the dialog
		*/
		virtual void OnCancelButton(wxCommandEvent& WXUNUSED(event));


	private:

		Image<InputST> OrigImg;

		/**
		* The filtered image. StorageType conversion between InputST and OutputST
		* is handled automatically within this base class.
		*/
		Image<OutputST> FilteredImage;

		/**
		* the location of the last saved image else empty
		*/
		wxString SavedImagePath;

		/**
		* The dialog's main sizer. It contains the ParamSizer and the ButtonSizer
		* Orientation: wxVERTICAL
		*/
		wxBoxSizer* MainSizer;

		/**
		* It's intention is to hold elements for parameter inputs. This should
		* hold spinctrl/slider/etc... but in the base class it has no content.
		* Orientation: wxHORIZONTAL
		*/
		wxBoxSizer* ParamSizer;

		/**
		* The sizer which holds the buttons below
		* Orientation: wxHORIZONTAL
		*/
		wxBoxSizer* ButtonSizer;

		wxButton* SaveButton;
		wxButton* FilterButton;
		wxButton* PreviewButton;
		wxButton* ResetButton;
		wxButton* CancelButton;

		wxStaticLine* Line;

		/**
		* An indicator if the image was filtered.
		*/
		bool ImageIsFiltered;

	};


	/**
	* all non-standard IDs related to this dialog -- classes inherited from this
	* included and all classes that use this class -- should be entered here to
	* decrease conflict possibilities.
	*/
	enum FILTER_DIALOG_IDs {

		ID_WIN_GAUSS = 1934572,
		ID_WIN_MEAN,
		ID_WIN_GRADIENT_SOBEL_3X3,
		ID_WIN_MEDIAN,
		ID_WIN_RESCALE,
		ID_WIN_CANNY_EDGE,
		ID_WIN_PREVIEW,

		ID_PANEL_DOUBLE_VAL,

		ID_WIN_BVWX_MAINFRAME,
		ID_WIN_IMAGECANVAS_UC,
		ID_WIN_IMAGECANVAS_FL,

		ID_BTN_FILTER_FDB,
		ID_BTN_PREVIEW_FDB,
		ID_BTN_RESET_FDB,

		ID_WIN_FILTER
	};


} // end namespace BIAS

#endif // __FILTER_DIALOG_BASE_HH_
