#ifndef __ImageCanvas_hh__
#define __ImageCanvas_hh__

#include <bias_config.h>

#ifndef BIAS_HAVE_WXWIDGETS
#  error BIAS_HAVE_WXWIDGETS not defined but required by ImageCanvas. Please enable BIAS_HAVE_WXWIDGETS in CMake configure step.
#endif

#include <Base/Debug/Debug.hh>
#include <Base/Image/Image.hh>

#include <Base/Common/BIASpragmaStart.hh>
#include <wx/wx.h>
//#include <wx/textctrl.h>
#include <wx/dcbuffer.h>
#include <wx/thread.h>
#include <Base/Common/BIASpragmaEnd.hh>

#define D_IC_IM (1<<0)


namespace BIAS{

  class ZoomFrame;
  class HistoFrame;
  template <class T> class PyramidImage;


  /** @class ImageCanvas
      @ingroup g_gui
      @brief display image in wx application, provides zoom and investigation
      functionality
      @author woelk */
  class BIASGui_EXPORT ImageCanvas 
    : public wxScrolledWindow, public BIAS::Debug
  {
  public:
    ImageCanvas(wxWindow* parent, wxStatusBar* sbar=NULL,
                int sbarid=0, wxWindowID id = -1,
                const wxPoint& pos = wxDefaultPosition,
                const wxSize& size = wxSize(320, 240),
                long style = ( (wxHSCROLL | wxVSCROLL) & wxTRANSPARENT),
                const wxString& name = wxT("ScrolledWindow"));
    ~ImageCanvas();


    void Show(BIAS::Image<unsigned char>& im, std::string name);
    void Show(BIAS::ImageBase& origim, BIAS::Image<unsigned char>& scaledrgbim,
              std::string name);
    void Show(BIAS::PyramidImage<unsigned char>& pim, std::string name);

    //returns the position of the histogramm window
    wxPoint GetHistoPos();
    //returns the size of the histogramm window
    wxSize GetHistoSize();
    //The Method that shows Histogramm-Window if it's not visible yet
    void ShowHistogramm(wxPoint pos  = wxPoint(10,10),
                        wxSize  size = wxSize(100,100));
    void ShowHistogramm(ImageBase& image,
                        wxPoint pos  = wxPoint(10,10),
                        wxSize  size = wxSize(100,100));

    // to let user access classic hide/show function
    virtual bool Show(bool show=true) {return wxScrolledWindow::Show(show);};

    inline BIAS::ImageBase *GetOrigImage() { return &origim_; }
    //return the 3 channel image which is displays on screen
    inline BIAS::Image<unsigned char>* GetImageShown(){ return &rgbim_;}

    bool GetMousePos(int &x, int &y);

    inline void SetDefaultZoom(unsigned z)
    { DefaultZoom_=z; }

    /**
     * Canvas now uses double buffering to avoid flickering, so it should be ok
     * to always clear the canvas.
     */
    inline void SetClearDrawingContext(bool clear)
    { ClearDrawingContext_=clear; }

    inline void SetShowScaled(bool on){ShowScaled_ = on;};
    inline bool GetShowScaled(){return ShowScaled_;};

    /** Sets the size of the scaled image. Only used if show scaled is on. */
    inline void SetScaledImageSize(int width, int height)
    {
      ScaledImageSizeX_ = width;
      ScaledImageSizeY_ = height;
    }

    /**
      Set the quality used for image size rescaling
      false: Uses the normal default scaling method of pixel replication
      true: Uses bicubic and box averaging resampling methods for upsampling 
      and downsampling respectively
    */
    inline void SetHighScaleQuality(bool on)
    {
      if(on) //it's on!      
        ScaleQuality_ = wxIMAGE_QUALITY_HIGH; 
      else
        ScaleQuality_ = wxIMAGE_QUALITY_NORMAL;
    }

    bool HistogrammIsShown();
    void KeepScrollPosition(bool flag);

    /**
      Set the colo(u)r which is used to draw the roi.
    */
    void SetRoiColor(unsigned char red,unsigned char green,
                     unsigned char blue, unsigned alpha = 128);    
    inline void DrawRoi(bool enable){ DrawRoi_ = enable; Refresh();};
    inline bool IsRoiDrawn(){return DrawRoi_; };
 // raises or lowers the zoom window if it is shown
    void RaiseZoomWindow(const bool raise_or_lower); 

  protected:
    void InitHisto(wxPoint pos  = wxPoint(10,10),
                   wxSize  size = wxSize(100,100));
    void UpdateStatusBar(unsigned x, unsigned y);
    void Zoom(unsigned x, unsigned y);

    void OnPaint(wxPaintEvent& event);

    void OnErase(wxEraseEvent& event);

    void OnSize(wxSizeEvent& event);
    void OnMouseMove(wxMouseEvent &event);
    void OnMouseLeave(wxMouseEvent &event);
    void OnMouseWheel(wxMouseEvent &event);
    void OnRightMouseButton(wxMouseEvent &event);
    void OnLeftMouseButton(wxMouseEvent &event);
    void OnKeyPress(wxKeyEvent &event);

    void PaintRoiOnDc_(wxBufferedPaintDC& dc);

    wxBitmap *Bitmap_;
    wxImage *wximage_;
    wxStatusBar *SBar_;
    int SBarID_;
    int SBarWidths_[2];

    ZoomFrame *MyZoomWin_;
    HistoFrame *MyHistoWin_;
    unsigned width_, height_;
    unsigned lastwidth_, lastheight_;
    int mousex_, mousey_;
    BIAS::ImageBase origim_;
    BIAS::Image<unsigned char> rgbim_;
    wxBitmap *RoiBitmap_;
    wxImage  *wxRoiImage_;
    BIAS::Image<unsigned char> roimask_;

    bool DrawRoi_;

    // for scaled-down display
    bool ShowScaled_;
    int ScaledImageSizeX_,ScaledImageSizeY_;
    int ScaleQuality_;

    wxMutex wxbitmap_mutex_, origim_mutex_;

    unsigned DefaultZoom_;

    bool KeepScrollPosition_;

    // Canvas now uses double buffering to avoid flickering, so it should be ok
    // to always clear the canvas.
    bool ClearDrawingContext_;

    std::string ImageName_;
    wxString Title_;

    DECLARE_EVENT_TABLE()
      }; // ImageCanvas



} // namespace


#endif // __ImageCanvas_hh__
