#ifndef __ImageCanvasCheckerboardExtract_hh__
#define __ImageCanvasCheckerboardExtract_hh__

#include "ImageCanvas.hh"
#include <Gui/StringConv.hh>
#include <Base/Math/Vector2.hh>
#include <Base/Geometry/HomgPoint2D.hh>

namespace BIAS {
  
  /** @class ICECCallBackInterface
      @ingroup g_gui
      @brief callback interface for ImageCanvasCheckerBoardExtract

      instructions:
      derive your class e.g. MyFrame from ICECallBackInterface
      and override the method CheckerboardSelected(...)to react to a finished 
      selection of a Checkerboard
      
      also declare a pointer to ImageCanvasCheckerBoardExtract, e.g.
      ImageCanvasCheckerBoardExtract *_ic;
      
      to select The Checkerboard corners then call:
      _ic->SelectCheckerBoardCorners(*this);
      where *this is a reference to a ICECCallBackInterface
      (if your class is derived from ICECCallBackInterface this is *this)
      
      @author ischiller 07/2007*/
  class BIASGui_EXPORT ICECCallbackInterface{
  public:
    ICECCallbackInterface();
    virtual ~ICECCallbackInterface();
    /// this function should be overloaded
    virtual void CheckerboardSelected(std::vector<HomgPoint2D > &corners, 
                                      bool bOnlyFourPointsGiven, bool valid=true);
  };

  /** @class ImageCanvasCheckerboardExtract
      @ingroup g_gui
      @brief provides functionality for selection of checkerboard patterns           
      @author schiller 2007/08 */
  class BIASGui_EXPORT ImageCanvasCheckerboardExtract : public ImageCanvas
  {
  public:
    ImageCanvasCheckerboardExtract(wxWindow* parent, wxStatusBar* sbar=NULL,
                                   int sbarid=0, wxWindowID id = -1, 
                                   const wxPoint& pos = wxDefaultPosition, 
                                   const wxSize& size = wxSize(640, 480), 
                                   long style = wxHSCROLL | wxVSCROLL |wxFULL_REPAINT_ON_RESIZE, 
                                   const wxString& name = AsciiToWx("CheckerboardExtract"));
    ~ImageCanvasCheckerboardExtract();
    
    /** Call SelectCheckerboardCorners with an object derived from ICECallbackInterface.
        The function CheckerboardSelected() should be overloaded in this object.
        After rectangle selection is finished (4 clicks with left mouse,
        right button cancels) the function obj.CheckerboardSelected() is called
        with either 4 coordinates or xCorners*yCorners coordinates.
        @author ischiller 07/2007 */
    void SelectCheckerboardCorners(ICECCallbackInterface& obj,
                                   bool allCorners,
                                   bool autoInvalidate=false);   

    /** Call DisplayCheckerboardCorners with an object derived from ICECallbackInterface.
        The function CheckerboardSelected() should be overloaded in this object.       
        @author ischiller 07/2007 */
    void DisplayCheckerboardCorners( std::vector<BIAS::Vector2<double> > &corners,
                                    double rescaleFactor=1.0);
    
    void DisplayCheckerboardCorners( std::vector< BIAS::HomgPoint2D > &corners,
                                    double rescaleFactor=1.0);


    void ManuallySetSelectionMode(bool SelectAllCorners){
      bEnforceAllCorners_ = SelectAllCorners;
      for(unsigned i=0;i<allCheckerboardCorners_.size();i++)
        allCheckerboardCorners_[i] = HomgPoint2D(0.0,0.0,1.0);
      NumLeftClicks_=0;
    };

    void SetPatternParameters(unsigned int xCorners, unsigned int yCorners);
    void EnableCornerSelection(bool selectable=true);
    void OnLeftMouseButton(wxMouseEvent &event);
    void OnMouseMove(wxMouseEvent &event);
    void OnRightMouseButton(wxMouseEvent &event);
    void OnPaint(wxPaintEvent& event);
  
  protected:
    void DrawCheckerboardCorners_(double rescaleFactor = 1.0);
    void RefreshDrawIm_();

  protected:
  	bool         bSelectable_;
    bool         bEnforceAllCorners_;
    bool         SelectCheckerboard_;
    bool         DisplayCheckerboard_;
    unsigned int xCorners_;
    unsigned int yCorners_;
    unsigned int NumLeftClicks_;   

    ICECCallbackInterface *CallbackObj_;
      
    std::vector<BIAS::HomgPoint2D > allCheckerboardCorners_;   

    DECLARE_EVENT_TABLE()
  }; // class 


} // namespace



#endif // __ImageCanvasCheckerboardExtract_hh__
