#ifndef __ScaledImageCanvas_hh__
#define __ScaledImageCanvas_hh__

#include <bias_config.h>

#ifndef BIAS_HAVE_WXWIDGETS
#  error BIAS_HAVE_WXWIDGETS not defined but required by ScaledImageCanvas. Please enable BIAS_HAVE_WXWIDGETS in CMake configure step.
#endif

#include <Base/Image/Image.hh>
#include <Gui/ImageCanvas.hh>
#include <Base/Common/BIASpragmaStart.hh>
#include <wx/wx.h>
#include <wx/spinctrl.h>
#include <Base/Common/BIASpragmaEnd.hh>

namespace BIAS {
  
  template <class T> class PyramidImage;

  /** @class ScaledImageCanvas
      @ingroup g_gui
      @brief display float images in wx application, provides zoom, scale,
      shift  and investigation functionality
      @author woelk */
  class BIASGui_EXPORT ScaledImageCanvas 
    : public wxPanel
  {
  public:
    ScaledImageCanvas(wxWindow* parent, wxStatusBar* sbar=NULL,
                      int sbarid=0, wxWindowID id = -1,
                      const wxPoint& pos = wxDefaultPosition,
                      const wxSize& size = wxDefaultSize,
                      long style = wxTAB_TRAVERSAL,
                      const wxString& name = wxT("ScaledScrolledWindow"));
    ~ScaledImageCanvas();

    void Show(BIAS::ImageBase& im, std::string name);

    void Show(BIAS::PyramidImage<float>& pim, std::string name);

    // to let user access classic hide/show function
    virtual bool Show(bool show=true) {return MiaScWi_->Show(show);};
    //inline void DrawRoi(bool enable){ DrawRoi_ = enable; Refresh();};
    inline void SetScaleRange(double min, double max)
    { BIASASSERT(max>min); MinScale_=min; MaxScale_=max; }

    inline void SetOffsetRange(double min, double max)
    { BIASASSERT(max>min); MinOffSet_=min; MaxOffSet_=max; }

    int SetScale(double scale);
    int SetScale(double scale, bool update);

    int SetOffset(double offset);
    int SetOffset(double offset, bool update);

    inline void SetDefaultZoom(unsigned z)
    { MiaScWi_->SetDefaultZoom(z); }

    /**
      Set the quality used for image size rescaling
      false: Uses the normal default scaling method of pixel replication
      true: Uses bicubic (SLOW!!) and box averaging resampling 
      methods for upsampling and downsampling respectively      
    */
    inline void SetHighScaleQuality(bool on)
    {
      if(MiaScWi_)
      MiaScWi_->SetHighScaleQuality(on);
    }

    /** @return  (pointer) to class member ImageCanvas used for display JW */
    ImageCanvas* GetImageCanvas();

    /** @return datat image origim_ class member JW */
    BIAS::ImageBase & GetOrigimg();

    /** @return the image that is shown in the canvas JF */
    BIAS::ImageBase& GetShownImage();

    /** pointer to sbar to be used JW */
    void SetStatusBarPtr(wxStatusBar* p_sbar);

    /** display text to sbar if one is assigned */
    void SetStatusText(const std::string & s);

    /** recalculate scale and shift for each new frame or not ?*/
    void SetAutoadaptScale(const bool b){AutoadaptScale_ = b;}
    
    /** set ranges so that they would look beautiful **/
    void SetNiceRanges(double scale_factor,
                       double offset_factor);

    /** set ranges of sliders **/
    void SetRanges(double value_sc_double_l,
                   double value_sc_double_r,
                   double value_of_double_l,
                   double value_of_double_r);

    void GetScaleParameters(bool   &Auto,
                              double &value_sc_double_l,
                              double &value_sc_double_r,
                              double &value_of_double_l,
                              double &value_of_double_r,
                              double &scale,
                              double &offset,
                              int    &value_sc_steps,
                              int    &value_of_steps);
   
    void SetScaleParameters(bool   Auto,
                            double value_sc_double_l,
                            double value_sc_double_r,
                            double value_of_double_l,
                            double value_of_double_r,
                            double scale,
                            double offset,
                            int    value_sc_steps,
                            int    value_of_steps);
   

    void UpdateHistogramm(ImageBase image);
   
    void KeepScrollPosition(bool flag);
  // raises or lowers the zoom window if it is shown
    void RaiseZoomWindow(const bool raise_or_lower); 


  protected:
    void OnScaleOffset       (wxScrollEvent & event);
    void OnScaleOffsetSteps  (wxSpinEvent & event);
    void OnScaleOffsetBorders(wxCommandEvent & event);

    void OnAutoScaleOffset   (wxCommandEvent& event);
    void OnHisto             (wxCommandEvent& event);
    void OnMouseWheel(wxMouseEvent &event);
    
    void UpdateScaleOffset();
    void UpdateScaleOffsetBorders();
    void AutoScale_();

    wxCheckBox*   Check_Auto_Sc_Of_;
    wxCheckBox*   Check_Histo_;

    //Controls to change Scale-Slider Parameters
    wxStaticText* Scale_Text_;

    wxTextCtrl*   Scale_L_;
    wxTextCtrl*   Scale_R_;
    wxSpinCtrl*   Scale_Steps_;

    double        scale_l_old;
    int             scale_old;
    double        scale_r_old;

    //Controls to change Offset-Slider Parameters
    wxStaticText* Offset_Text_;

    wxTextCtrl*   Offset_L_;
    wxTextCtrl*   Offset_R_;
    wxSpinCtrl*   Offset_Steps_;

    double        offset_l_old;
    int             offset_old;
    double        offset_r_old;

    //Sizes of text/spin/checkboxes
    wxSize box_size;
    wxSize box_size_steps;
    wxSize box_size_checks;

    //Count of steps in sliders
    int    slider_sc_digits;
    int    slider_of_digits;

    //Old Values of sliders, text/spin/checkboxes
    wxString scale_border_l_value;
    wxString scale_border_r_value;

    wxString offset_border_l_value;
    wxString offset_border_r_value;


    wxString string_base;

    // data member
    ImageCanvas *MiaScWi_; // true data image
    wxSlider *ScaleSL_, *OffsSL_;
    BIAS::ImageBase origim_;
    BIAS::Image<float> scaledf_;
    BIAS::Image<unsigned char> rgbim_, ucim_;
    std::string ImageName_;
    double MinScale_, MaxScale_, MinOffSet_, MaxOffSet_;
    bool DrawRoi_;
    // pointer to statusbar to use if any
    wxStatusBar * p_StatBar;
    bool AutoadaptScale_;
    DECLARE_EVENT_TABLE()
  }; // ScaledImageCanvas

} // namespace


#endif // __ScaledImageCanvas_hh__
