#ifndef __IMAGEPACKAGE_HH__
#define __IMAGEPACKAGE_HH__

#include <bias_config.h>

#include <Base/Image/Image.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageIO.hh>
#include <Image/Camera.hh>
#include <Utils/IOUtils.hh>
#include <Image/BackwardMapping.hh>
#include <Image/ProjectionMapping.hh>

namespace BIAS {

  /**
   * An Image Package is multiple .mip images in one file. This class has methods for reading and writing.
   * @author fkellner, 2008
   */
  class BIASImage_EXPORT ImagePackage {
  public:

    class BIASImage_EXPORT IPImageHeader {
    public:
      unsigned int len_filename;
      unsigned int len_metadata;
      unsigned int len_pixeldata;
      friend BIASImage_EXPORT std::ostream& operator<<(std::ostream& os, const IPImageHeader &ipih) {
        os.write(reinterpret_cast<const char*>(&(ipih.len_filename)), sizeof(unsigned int));
        os.write(reinterpret_cast<const char*>(&(ipih.len_metadata)), sizeof(unsigned int));
        os.write(reinterpret_cast<const char*>(&(ipih.len_pixeldata)), sizeof(unsigned int));
        return os;
      }
      friend BIASImage_EXPORT std::istream& operator>>(std::istream& is, IPImageHeader& ipih) {
        is.read((char*)&(ipih.len_filename), sizeof(unsigned int));
        is.read((char*)&(ipih.len_metadata), sizeof(unsigned int));
        is.read((char*)&(ipih.len_pixeldata), sizeof(unsigned int));
        return is;
      }
    };

    struct IPFileInfo {
      unsigned char fileNum;
      unsigned long int offset;
      std::string fileName;
    };

    /**
     * An Image Package is much like an uncompressed zip file of mip images. The Package files are split at 1GB, where
     * the files are named Images0001.pack, Images0002.pack etc. Each image inside a package can have its own filename, so
     * writing the files to disk is made easy. Images in a package can be of any mixed size, storage type, but will always
     * be saved uncompressed.
     */
    ImagePackage();

    ~ImagePackage();

    /**
     * Create image Package file
     * @param num used for numbering of .pack files, only used internally.
     */
    int Create(const std::string filename, unsigned int num=1);

    /**
     * Adding an image with empty filename will use default name
     */
    int AddImage(const ImageBase image, const std::string filename="");

    /** open a package
     * @return 0 on success
     */
    int Open(const std::string filename);

    /** get image by given index number
     * @return 0 on success
     */
    int GetImage(const int imageNum, ImageBase &image, std::string &filename);

    /** get next image, call subsequently to walk through package file
     * @return 0 if image exists
     */
    int GetNextImage(ImageBase &image, std::string &filename);

    /** get previous image, call subsequently to walk through package file
     * @return 0 if image exists
     */
    int GetPrevImage(ImageBase &image, std::string &filename);

    /**
     * helper to write image with index imageNum, empty filename will use the one stored in package file
     */
    int WriteSingleImage(const unsigned int imageNum, const std::string filename="");

    /**
     * write all Images from the package. filename can be used to overwrite the ones from the package file (numbers will be added)
     */
    int WriteAllImages(const std::string filename="");

    /**
     * close package
     */
    int Close();
    
    bool IsOpen() {
      return bOpen_;
    }

  private:

    int currentImg_;
    int currentFile_;
    unsigned int numImages_;
    unsigned int fileNum_;
    std::string filename_;
    bool bOpen_;
    bool bReadMode_;

    std::vector<std::ifstream*> ifs_;
    std::ofstream ofs_;

    std::vector<IPFileInfo> fileInfos_;


  };

} // namespace

#endif // __IMAGEPACKAGE_H_
