/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __UnVignette_hh__
#define __UnVignette_hh__

#include <Base/Common/BIASpragmaStart.hh>

#include <Base/Image/Image.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Geometry/KMatrix.hh>
#include <MathAlgo/Interpolator.hh>
#include <Image/CameraParam.hh>
#include <Base/Math/Matrix.hh>
#include <Base/Math/Vector.hh>

#include <vector>
#include <string>

#define UNDIST_EPS 1e-9

namespace BIAS{

  enum BIASImage_EXPORT INTERPOLATION_METHOD
  {
    LEAST_SQUARES =1,
    SPLINE_WITH_LUT,
    SPLINE_WITHOUT_LUT
  };

  
  /** @class UnVignette
      @ingroup g_image
      @brief Unvignette algorithm for images. Measure the vignetting of images 
      and correct the vignetting of images

      @author ischiller
      @date   02/2006
  */
  class BIASImage_EXPORT UnVignette{

  public:
    UnVignette();
    UnVignette(CameraParam &cParam, INTERPOLATION_METHOD method = LEAST_SQUARES, bool bUseLuT = true);

    ~UnVignette();
    
    inline bool IsInitialized(){return bIsInitialized_;};

    void Compute(BIAS::Image<unsigned char> &Image);

    /**
      Init the Unvignette algorithm
      give ControlPoints in image as offset from Principal Point in pixel
      give percentage values in second vector
      vectors have to be of same size
      @author ischiller
      @date 02/06
    */
    int Init(const std::vector<double> ControlPoints,
             std::vector<double> PercentageValues,
             unsigned int dImageWidth , unsigned int dImageHeight,
             int dPPX, int dPPY,INTERPOLATION_METHOD method = LEAST_SQUARES,bool bUseLuT = true);
    /**
      Init the Unvignette algorithm with CameraParam object
      CameraParameter objcect has to contain ControlPoints and PercentageValues
      vectors have to be of same size
      @author ischiller
      @date 02/06
    */
    int Init(BIAS::CameraParam cParam, INTERPOLATION_METHOD method = LEAST_SQUARES, bool bUseLuT = true);

  private: 
    INTERPOLATION_METHOD  dInterPolationMethod_;
    double                CalcRadius_(double x, double y);
    int                   PrepareLuImage_();
    int                   PrepareLeastSquaresMin_();
    int                   LeastSquares_(double &res, double radius);
    int                   Init_();
  private:   
    std::vector<double> ControlPoints_;
    std::vector<double> PercentageValues_;

    BIAS::Matrix<double>  A_; //matrix for least squares computation (measured percentages)
    BIAS::Vector<double>  Y_; //vector with measurement points
    BIAS::Vector<double>  X_; //result vector of least squares polynomial fitting
   
    bool                  bLeastSquaresPrepared_;                  


    double                dWidthHalfMPPX_;
    double                dHeightHalfMPPY_;
    BIAS::Image<float>    LuImage_;
   
    bool                  bUseLuT_;
  
    double                PrinciplePointX_;
    double                PrinciplePointY_;
    unsigned int          dWidth_;
    unsigned int          dHeight_;
    unsigned int          dChannelCount_;
   
    bool                  bIsInitialized_;
    BIAS::Interpolator    InterpolatorIllu_;

    // illumination vignette correction
    // x is distance in pixel from principal point
    // y is illumination correction factor
    std::vector<int>      IlluCorrX_;
    std::vector<double>   IlluCorrY_;
  };
}

#include <Base/Common/BIASpragmaEnd.hh>

#endif
