/* This file is part of the BIAS library (Basic ImageAlgorithmS).
   
   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel
   
   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.
   
   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
#ifndef __TrackerBaseInterface__hh__
#define __TrackerBaseInterface__hh__

#include <Base/Common/BIASpragmaStart.hh>

#include <Base/Debug/Debug.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <Base/Math/Matrix2x2.hh>
#include <Base/Image/Image.hh>
#include <MathAlgo/Median1D.hh>
#include <Filter/FilterMask.hh>

#include <vector>

#define KLT_TYPE double

#define D_TRACKERB_INIT        0x00000001
#define D_TRACKERB_KLT         0x00000002
#define D_TRACKERB_BILINEAR    0x00000004
#define D_TRACKERB_INPUT       0x00000008
#define D_TRACKERB_REJECT      0x00000010
#define D_TRACKERB_AFFINE      0x00000020

#define TRACKERBASE_INVALID_PIXEL -1e10

namespace BIAS {

  enum ERejectionType { RT_Invalid=-1, RT_MAD, RT_X84, RT_X84M };

  enum ETrackerResult {
    TRACKER_SUCCESS        =  0, ///< success (error < maxerror)
    TRACKER_NOSTRUCTURE    = -1, ///< no spatial structure is present
    TRACKER_OUTOFIMAGE     = -2, ///< point slid out of image
    TRACKER_TOOMANYITER    = -3, ///< maxiter is reached and error is above maxerror
    TRACKER_OUTOFROI       = -4, ///< a point lies outside of valid region in images
    TRACKER_MEANABSDIF     = -5, ///< point was rejected by MAD criterion
    TRACKER_IGNOREPOINT    = -6, ///< input point was at infinity
    TRACKER_X84            = -7, ///< point is rejected by X84/X84M
    TRACKER_ODDAFFINEWARP  = -8, ///< affine warp seems degenerated
    TRACKER_ERRORINCREASED =  1  ///< error increased from last iteration 
  };

  std::ostream& operator<<(std::ostream& os, const enum ETrackerResult& res);

  std::ostream& operator<<(std::ostream& os, const enum ERejectionType& rt);

  /** @class TrackerBaseInterface
      @ingroup g_tracker
      @brief  Base class for the different tracking algorithms, defining the interfaces
      for the tracking functions. Also stores some basic options and datas, 
      i.e. the pointer to the image datas. See ExampleTrackerBaseInterface for brief 
      exemplary usage.
      @author woelk 09/2004, kollmann 07/2006 */
  template <class StorageType>
  class BIASMatcher2D_EXPORT TrackerBaseInterface : public Debug
  {
  public :
    TrackerBaseInterface();
    virtual ~TrackerBaseInterface();

    /** @brief Prepare for tracking with prefiltered images.

    The ROI must be set correctly in gradient images (representing the 
    gradient filter mask size). 
    For good results im1 and im2 should be lowpass filtered images.

    @param im1      (low-pass filtered) image 1
    @param im2      (low-pass filtered) image 2
    @param gradx1   horizontal gradient image of im1
    @param grady1   vertical gradient image of im1
    @param gradx2   horizontal gradient image of im2
    @param grady2   vertical gradient image of im2 */
    void Init(Image<StorageType>& im1, Image<StorageType>& im2, 
              Image<StorageType>& gradx1, Image<StorageType>& grady1, 
              Image<StorageType>& gradx2, Image<StorageType>& grady2);

    /** @brief Prepare for tracking with on-demand filtering.

    Lowpass and gradient filtering is done only where needed, using the
    given filter masks. All filter masks must have odd size.

    @param im1              (low-pass filtered) image 1
    @param im2              (low-pass filtered) image 2
    @param lowpass_mask     filter mask for lowpass filtering
    @param gradx_mask       filter mask for horizontal gradient computation
    @param grady_mask       filter mask for vertical gradient computation */
    void Init(Image<StorageType>& im1, Image<StorageType>& im2,
              FilterMask& lowpass_mask, FilterMask& gradx_mask, 
              FilterMask& grady_mask);

    /** @brief Calculates correspondence from image1 to image2. The two images
        and corresponding gradients must specified using the function Init()
        before calling this function.

        The algorithm searches for best correspondence to point p1 from image1 
        in image2. The search starts at p2 in image 2.

        All homogenous points must be homogenized (i.e. w=1.0).
        
        The algorithms either terminates if the displacement resulting from
        one iteration drops below *_Maxerror or the maximum number of 
        iterations *_MaxIterations is reached.

        @param[in]  p1           Point in image1
        @param[in]  p2           Algorithm starts searching at this point in image2
        @param[out] p2tracked    Best matching point in image2
        @param[out] error        Change in displacement calculated in last iteration
        @param[out] iter         Number of iterations used
        @param[out] residuumMAD  Mean absolute grey value difference  between p1 in
        image 1 and p2tracked in image 2 over the tracking
        support window
        @param[out] residuumMSD  Squared grey value difference between p1 in image 1
        and p2tracked in image 2 summed over the tracking
        support window
        @param[out] cov          Approximation to the covarinace matrix image1 and
        result in image2 over the tracking window (= sum of
        absolute difference divided by number of pixels in
        tracking value)
        @param[in]  AffinePred   Predicted affine matrix (only used in affine tracking)
        @param[out] AffineResult Result of affine tracking (only used in affine tracking)
        @return                  See TRACKER_* */
    int Track(HomgPoint2D& p1, HomgPoint2D& p2, HomgPoint2D& p2tracked, 
              KLT_TYPE& error, int& iter, KLT_TYPE& residuumMAD,
              KLT_TYPE& residuumMSD, Matrix<KLT_TYPE>& cov,
              const Matrix2x2<KLT_TYPE>& AffinePred =
              Matrix2x2<KLT_TYPE>(MatrixIdentity),
              Matrix2x2<KLT_TYPE>* AffineResult = NULL);

    /** @brief Same as Track(), but a correspondence is searched for every 
        point in the vector p1. */
    void Track(std::vector<HomgPoint2D>& p1, 
               std::vector<HomgPoint2D>& p2, 
               std::vector<HomgPoint2D>& p2tracked, 
               std::vector<KLT_TYPE>& error, 
               std::vector<int>& numiter,
               std::vector<KLT_TYPE>& residuiMAD, 
               std::vector<KLT_TYPE>& residuiMSD, 
               std::vector<int>& res,
               std::vector<Matrix<KLT_TYPE> >& cov,
               const std::vector<Matrix2x2<KLT_TYPE> >& AffinePred =
               std::vector<Matrix2x2<KLT_TYPE> >(),
               std::vector<Matrix2x2<KLT_TYPE> >* AffineResult = NULL);

    inline void SetHalfWinSize(const int hws)              
    { _HalfWinSize=hws;  _WinSize=(hws<<1)+1; }
    inline void SetMaxIterations(const int maxiter)        
    { _MaxIterations=maxiter; }
    inline void SetMaxError(const KLT_TYPE maxerr)         
    { _MaxError=maxerr; }

    /** !!!
        Also used for X84M as maximal residuum. That is features with a 
        residuum <= 'maxres' will be accept automatically!
        !!! 
    **/
    inline void SetMaxResiduumMAD(const KLT_TYPE maxres)   
    { _MaxResiduumMAD=maxres; }
    inline void SetRejectionType(const int rejection_type) 
    { _RejectionType = rejection_type; }
    

    /** enable brightness variance and offset invariant computation */
    virtual void SetAffineBrightnessInvariance(bool bi) { 
      _AffineBrightnessInvariance = bi;
    }

    // these are needed in the highlevel Tracker class
    inline int GetHalfWinSize()   { return _HalfWinSize; }
    inline int GetMaxIterations() { return _MaxIterations; }
    inline KLT_TYPE GetMaxError() { return _MaxError; }
    inline int GetRejectionType() { return _RejectionType; }

    virtual void SetWeightMatrix(const BIAS::Matrix<KLT_TYPE>& weight) {}
    
  protected:

    ///////////////////////////////////////////////////////////////////////////
    //  params
    ///////////////////////////////////////////////////////////////////////////
    
    int       _HalfWinSize;    ///< use support window of size 2*hws+1
    int       _MaxIterations;  ///< iteration stops after _MaxIterations
    /// iteration stops if position refinement is less than *_MaxError
    KLT_TYPE  _MaxError;       
    KLT_TYPE  _MaxResiduumMAD; ///< outlier rejection via MAD and
                                 // acception via X84M respectively
    int       _RejectionType; ///< The rejection type: RT_MAD,RT_X84 or RT_X84M


    /// compute brightness "invariant"
    bool _AffineBrightnessInvariance;
    
    /// mean and std deviation of image patch
    KLT_TYPE mean1_, mean2_, dev1_, dev2_;

    ///////////////////////////////////////////////////////////////////////////

    // pointers to images
    Image<StorageType> *_im1, *_gradim1x, *_gradim1y;
    Image<StorageType> *_im2, *_gradim2x, *_gradim2y;

    // filtered images are computed when set to true
    bool _ComputeFilteredImages;

    // lowpass filter computation type
    enum LowpassFilter
      {
        LOWPASS_BINOMIAL3X3,
        LOWPASS_FILTERMASK,
        LOWPASS_FILTERMASK_SEPARABLE
      };

    // gradient filter computation type
    enum GradientFilter
      {
        GRAD_SOBEL3X3,
        GRAD_FILTERMASK,
        GRAD_FILTERMASK_SEPARABLE
      };
    
    LowpassFilter  _LowpassFilter;
    GradientFilter _GradXFilter, _GradYFilter;
    FilterMask     _LowpassMask, _GradXMask, _GradYMask;
    KLT_TYPE       _LowpassMaskSum, _GradXMaskSum, _GradYMaskSum;
    int            _LowpassWindowExtraSize;
    
    // the image borders modulo filter and support windows or ROI
    // for image 1 and image 2
    int _maxx1, _maxy1, _minx1, _miny1;
    int _maxx2, _maxy2, _minx2, _miny2;
    
    // remember last support window size (for recalculation of gaussian weight)
    int _LastHalfWinSize, _WinSize;
    
    // used in iteration
    Matrix2x2<KLT_TYPE> _G, _Ginv;
    Vector2<KLT_TYPE> _e, _d, _disp;
    Matrix<KLT_TYPE> _gx, _gy, _bl1, _bl2, _gx1, _gy1, _gx2, _gy2, _gsx, _gsy;
    KLT_TYPE _gxx, _gxy, _gyy, _idet;
    Matrix<KLT_TYPE> _weight; // the weight matrix used in TrackWeighted_
    
    ///////////////////////////////////////////////////////////////////
    //  tracking
    ///////////////////////////////////////////////////////////////////

    /** @brief Interface of all tracking algorithms implemented in
        derived classes. */
    virtual int Track_(Vector2<KLT_TYPE>& p1, Vector2<KLT_TYPE>& p2,
                       Vector2<KLT_TYPE>& result, KLT_TYPE& error, 
                       int &iter,
                       const Matrix2x2<KLT_TYPE>& AffinePred,
                       Matrix2x2<KLT_TYPE>& AffineResult) = 0;

    ///////////////////////////////////////////////////////////////////
    //  calculation of cov, sad and ssd
    ///////////////////////////////////////////////////////////////////

    /** Uses _bl1, _bl2, _gxx, _gxy and _gyy to calculate the residui
        and covariance matrix and must hence be called immediatly after the 
        Track_() function call!!
        @author woelk */
    virtual void EvaluateResult_(KLT_TYPE& mad,KLT_TYPE& msd, 
                                 Matrix<KLT_TYPE>& cov);

    ///////////////////////////////////////////////////////////////////
    //  outlier rejection
    ///////////////////////////////////////////////////////////////////

    /** @author woelk */
    int RejectMAD_(std::vector<KLT_TYPE>& sad, std::vector<int>& res);

    /** @author woelk */
    int RejectX84_(std::vector<KLT_TYPE>& ssd, std::vector<KLT_TYPE>& sad,
                   std::vector<int>& res);

    ///////////////////////////////////////////////////////////////////
    //  helper functions
    ///////////////////////////////////////////////////////////////////
    
    /** resizes the internal matrices _gx1, _gy1, _bl1, _gx2, _gy2 and _bl2
        @author woelk */
    inline void Resize_(const int halfwinsize);

    ////////////////////////////////////////////////////////////////////
    //  bilinear interpolation functions 
    ////////////////////////////////////////////////////////////////////

    KLT_TYPE ComputeMaskSum(const FilterMask& mask);

    /** The lowpass filter functions apply a filter to a window of
        the source image and store the result in a matrix. */
    //@{

    /// Lowpass filtering using a kernel filter mask.
    inline void FilterLowpass_ByMask(int minX, int minY,
                                     int rows, int cols,
                                     const Image<StorageType>& image,
                                     Matrix<KLT_TYPE>& out,
                                     const FilterMask& mask,
                                     KLT_TYPE maskSum);

    /// Lowpass filtering using a separable filter mask.
    inline void FilterLowpass_BySeparableMask(int minX, int minY,
                                              int rows, int cols,
                                              const Image<StorageType>& image,
                                              Matrix<KLT_TYPE>& out,
                                              const FilterMask& mask,
                                              KLT_TYPE maskSum);

    /// Optimized lowpass filtering using the binomial 3x3 filter.
    inline void FilterLowpass_Binomial3x3(int minX, int minY,
                                          int rows, int cols,
                                          const Image<StorageType>& image,
                                          Matrix<KLT_TYPE>& out);
    //@}

    /** The gradient filter functions apply a filter to an image window
        stored in a matrix and store the result in another matrix. */
    //@{

    /// Optimized gradient X sobel3x3 filter.
    inline void Filter_GradXSobel3x3(const Matrix<KLT_TYPE>& in,
                                     Matrix<KLT_TYPE>& out);

    /// Optimized gradient Y sobel3x3 filter.
    inline void Filter_GradYSobel3x3(const Matrix<KLT_TYPE>& in,
                                     Matrix<KLT_TYPE>& out);

    /// Generic gradient filter.
    inline void Filter_ByMask(const Matrix<KLT_TYPE>& in,
                              Matrix<KLT_TYPE>& out,
                              const FilterMask& mask,
                              KLT_TYPE maskSum);

    /// Generic gradient filter using separable filter mask.
    inline void Filter_BySeparableMask(const Matrix<KLT_TYPE>& in,
                                       Matrix<KLT_TYPE>& out,
                                       const FilterMask& mask,
                                       KLT_TYPE maskSum);

    //@}

    /** Fills the matrices _gx1, _gy1 and _bl1 by interpolating gradients
        and lowpass filtered image at positions between (x-hws, y-hws) and
        (x+hws, y+hws). */
    inline void BilinearRegion1_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Fills the matrices _gx2, _gy2 and _bl2 by interpolating gradients
        and lowpass filtered image at positions between (x-hws, y-hws) and
        (x+hws, y+hws). */
    inline void BilinearRegion2_(KLT_TYPE x, KLT_TYPE y, int hws);

    /** Bilinear filtering using images as source (for prefiltered images) */
    inline KLT_TYPE 
    BilinearRegionFromImages_(KLT_TYPE x, KLT_TYPE y, int hws,
                              Image<StorageType>& image,
                              Image<StorageType>& gradx,
                              Image<StorageType>& grady,
                              Matrix<KLT_TYPE>& bl,
                              Matrix<KLT_TYPE>& gx,
                              Matrix<KLT_TYPE>& gy);

    /** Bilinear filtering using matrices as source */
    inline KLT_TYPE 
    BilinearRegionFromMatrices_(KLT_TYPE x, KLT_TYPE y, int hws,
                                const Matrix<KLT_TYPE>& image,
                                const Matrix<KLT_TYPE>& gradx,
                                const Matrix<KLT_TYPE>& grady,
                                Matrix<KLT_TYPE>& bl,
                                Matrix<KLT_TYPE>& gx,
                                Matrix<KLT_TYPE>& gy);

    /** bring region to mean zero and stddev 1 for brightness invariance 
        @return scale of region which has been applied */
    inline KLT_TYPE NormalizeRegion_(Matrix<KLT_TYPE>& bl,
                                     Matrix<KLT_TYPE>& gx,
                                     Matrix<KLT_TYPE>& gy,
                                     const KLT_TYPE&  min_value = -1e9);
    
  }; // class

  ////////////////////////////////////////////////////////////////////
  //  inline functions
  ////////////////////////////////////////////////////////////////////

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  Resize_(const int halfwinsize)
  {
    _LastHalfWinSize=halfwinsize;
    _WinSize=(halfwinsize<<1)+1;
    _gx1.newsize(_WinSize, _WinSize);
    _gy1.newsize(_WinSize, _WinSize);
    _gx2.newsize(_WinSize, _WinSize);
    _gy2.newsize(_WinSize, _WinSize);
    _gsx.newsize(_WinSize, _WinSize);
    _gsy.newsize(_WinSize, _WinSize);
    _bl1.newsize(_WinSize, _WinSize);
    _bl2.newsize(_WinSize, _WinSize);
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  FilterLowpass_ByMask(int minX, int minY, int rows, int cols,
                       const Image<StorageType>& image,
                       Matrix<KLT_TYPE>& out,
                       const FilterMask& mask,
                       KLT_TYPE maskSum)
  {
    BIASASSERT(!mask.IsSeparable());
    
    const Matrix<FM_FLOAT>& kernel = *mask.GetKernelf();

    int borderRows = (kernel.GetRows()-1) / 2;
    int borderCols = (kernel.GetCols()-1) / 2;
  
    out.newsize(rows, cols);

    const StorageType** ida = image.GetImageDataArray();

    for (int i = 0; i < rows; i++)
      {
        for (int j = 0; j < cols; j++)
          {
            out[i][j] = 0;
            for (int k = 0; k < (int)kernel.GetRows(); k++)
              {
                for (int l = 0; l < (int)kernel.GetCols(); l++)
                  {
                    out[i][j] +=
                      kernel[k][l] * ida[i+k-borderRows+minY][j+l-borderCols+minX];
                  }
              }
            out[i][j] /= maskSum;
          }
      }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  FilterLowpass_BySeparableMask(int minX, int minY, int rows, int cols,
                                const Image<StorageType>& image,
                                Matrix<KLT_TYPE>& out,
                                const FilterMask& mask,
                                KLT_TYPE maskSum)
  {
    BIASASSERT(mask.IsSeparable());
    
    const Vector<FM_FLOAT>& vert = *mask.GetSepfv();
    const Vector<FM_FLOAT>& horz = *mask.GetSepfh();
    
    int borderRows = (vert.Size()-1) / 2;
    int borderCols = (horz.Size()-1) / 2;
  
    out.newsize(rows, cols);

    const StorageType** ida = image.GetImageDataArray();

    for (int i = 0; i < rows; i++)
      {
        for (int j = 0; j < cols; j++)
          {
            out[i][j] = 0;
            for (int k = 0; k < (int)vert.Size(); k++)
              {
                for (int l = 0; l < (int)horz.Size(); l++)
                  {
                    out[i][j] +=
                      vert[k] * horz[l]
                      * ida[i+k-borderRows+minY][j+l-borderCols+minX];
                  }
              }
            out[i][j] /= maskSum;
          }
      }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  FilterLowpass_Binomial3x3(int minX, int minY, int rows, int cols,
                            const Image<StorageType>& image,
                            Matrix<KLT_TYPE>& out)
  {
    out.newsize(rows, cols);

    const StorageType** ida = image.GetImageDataArray();

    for (int i = 0; i < rows; i++)
      {
        for (int j = 0; j < cols; j++)
          {
            int x = j + minX;
            int y = i + minY;

            /* 1 2 1
               2 4 2
               1 2 1 */
            out[i][j] =
              ida[y-1][x-1] + 2*ida[y-1][x] +   ida[y-1][x+1]
              +2*ida[y  ][x-1] + 4*ida[y  ][x] + 2*ida[y  ][x+1]
              +  ida[y+1][x-1] + 2*ida[y+1][x] +   ida[y+1][x+1];
            out[i][j] /= 16.0;
          }
      }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  Filter_GradXSobel3x3(const Matrix<KLT_TYPE>& in, Matrix<KLT_TYPE>& out)
  {
    int rows = in.GetRows();
    int cols = in.GetCols();
    
    out.newsize(rows, cols);

    for (int i = 1; i < rows-1; i++)
      {
        for (int j = 1; j < cols-1; j++)
          {
            /* -1 0 1
               -2 0 2
               -1 0 1 */
            out[i][j] =
              -  in[i-1][j-1] +   in[i-1][j+1]  // row 0
              -2*in[i  ][j-1] + 2*in[i  ][j+1]  // row 1
              -  in[i+1][j-1] +   in[i+1][j+1]; // row 2
            out[i][j] /= 8.0;
          }
      }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  Filter_GradYSobel3x3(const Matrix<KLT_TYPE>& in, Matrix<KLT_TYPE>& out)
  {
    int rows = in.GetRows();
    int cols = in.GetCols();
  
    out.newsize(rows, cols);

    for (int i = 1; i < rows-1; i++)
      {
        for (int j = 1; j < cols-1; j++)
          {
            /* -1 -2 -1
               0  0  0
               1  2  1 */
            out[i][j] = 
              -in[i-1][j-1] - 2*in[i-1][j] - in[i-1][j+1]  // row 0
              +in[i+1][j-1] + 2*in[i+1][j] + in[i+1][j+1]; // row 2
            out[i][j] /= (StorageType)8.0;
          }
      }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  Filter_ByMask(const Matrix<KLT_TYPE>& in,
                Matrix<KLT_TYPE>& out,
                const FilterMask& mask,
                KLT_TYPE maskSum)
  {
    BIASASSERT(!mask.IsSeparable());
    
    const Matrix<FM_FLOAT>& kernel = *mask.GetKernelf();

    int borderRows = (kernel.GetRows()-1) / 2;
    int borderCols = (kernel.GetCols()-1) / 2;
  
    int rows = in.GetRows();
    int cols = in.GetCols();
  
    out.newsize(rows, cols);

    for (int i = borderRows; i < rows-borderRows; i++)
      {
        for (int j = borderCols; j < cols-borderCols; j++)
          {
            out[i][j] = 0;
            for (int k = 0; k < (int)kernel.GetRows(); k++)
              {
                for (int l = 0; l < (int)kernel.GetCols(); l++)
                  {
                    out[i][j] += kernel[k][l] * in[i+k-borderRows][j+l-borderCols];
                  }
              }
            out[i][j] /= maskSum;
          }
      }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  Filter_BySeparableMask(const Matrix<KLT_TYPE>& in,
                         Matrix<KLT_TYPE>& out,
                         const FilterMask& mask,
                         KLT_TYPE maskSum)
  {
    BIASASSERT(mask.IsSeparable());
    
    const Vector<FM_FLOAT>& vert = *mask.GetSepfv();
    const Vector<FM_FLOAT>& horz = *mask.GetSepfh();
    
    int borderRows = (vert.Size()-1) / 2;
    int borderCols = (horz.Size()-1) / 2;

    int rows = in.GetRows();
    int cols = in.GetCols();
  
    out.newsize(rows, cols);

    for (int i = borderRows; i < rows-borderRows; i++)
      {
        for (int j = borderCols; j < cols-borderCols; j++)
          {
            out[i][j] = 0;
            for (int k = 0; k < (int)vert.Size(); k++)
              {
                for (int l = 0; l < (int)horz.Size(); l++)
                  {
                    out[i][j] +=
                      vert[k] * horz[l]
                      * in[i+k-borderRows][j+l-borderCols];
                  }
              }
            out[i][j] /= maskSum;
          }
      }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  BilinearRegion1_(KLT_TYPE x, KLT_TYPE y, int hws)
  {
    if (_ComputeFilteredImages)
      {
        static Matrix<KLT_TYPE> lowpass;
        int lowpassHWS = hws + _LowpassWindowExtraSize;
        int lowpassWS = 2*lowpassHWS+2;
        int lowpassMinX = (int)floor(x) - lowpassHWS;
        int lowpassMinY = (int)floor(y) - lowpassHWS;

        switch (_LowpassFilter)
          {
          case LOWPASS_BINOMIAL3X3:
            FilterLowpass_Binomial3x3(lowpassMinX, lowpassMinY,
                                      lowpassWS, lowpassWS,
                                      *_im1, lowpass);
            break;
          case LOWPASS_FILTERMASK:
            FilterLowpass_ByMask(lowpassMinX, lowpassMinY,
                                 lowpassWS, lowpassWS,
                                 *_im1, lowpass,
                                 _LowpassMask, _LowpassMaskSum);
            break;
          case LOWPASS_FILTERMASK_SEPARABLE:
            FilterLowpass_BySeparableMask(lowpassMinX, lowpassMinY,
                                          lowpassWS, lowpassWS,
                                          *_im1, lowpass,
                                          _LowpassMask, _LowpassMaskSum);
            break;
          }

        static Matrix<KLT_TYPE> gradx;
        switch (_GradXFilter)
          {
          case GRAD_SOBEL3X3:
            Filter_GradXSobel3x3(lowpass, gradx);
            break;
          case GRAD_FILTERMASK:
            Filter_ByMask(lowpass, gradx, _GradXMask, _GradXMaskSum);
            break;
          case GRAD_FILTERMASK_SEPARABLE:
            Filter_BySeparableMask(lowpass, gradx, _GradXMask, _GradXMaskSum);
            break;
          }

        static Matrix<KLT_TYPE> grady;
        switch (_GradYFilter)
          {
          case GRAD_SOBEL3X3:
            Filter_GradYSobel3x3(lowpass, grady);
            break;
          case GRAD_FILTERMASK:
            Filter_ByMask(lowpass, grady, _GradYMask, _GradYMaskSum);
            break;
          case GRAD_FILTERMASK_SEPARABLE:
            Filter_BySeparableMask(lowpass, grady, _GradYMask, _GradYMaskSum);
            break;
          }
      
        dev1_ = this->BilinearRegionFromMatrices_(x - lowpassMinX,
                                                  y - lowpassMinY,
                                                  hws,
                                                  lowpass, gradx, grady,
                                                  _bl1, _gx1, _gy1);
      }
    else
      {
        dev1_ = this->BilinearRegionFromImages_(x, y, hws,
                                                *_im1, *_gradim1x, *_gradim1y,
                                                _bl1, _gx1, _gy1);
      }
  }

  template <class StorageType>
  inline void TrackerBaseInterface<StorageType>::
  BilinearRegion2_(KLT_TYPE x, KLT_TYPE y, int hws)
  {

    if (_ComputeFilteredImages)
      {
        static Matrix<KLT_TYPE> lowpass;
        int lowpassHWS = hws + _LowpassWindowExtraSize;
        int lowpassWS = 2*lowpassHWS+2;
        int lowpassMinX = (int)floor(x) - lowpassHWS;
        int lowpassMinY = (int)floor(y) - lowpassHWS;

        switch (_LowpassFilter)
          {
          case LOWPASS_BINOMIAL3X3:
            FilterLowpass_Binomial3x3(lowpassMinX, lowpassMinY,
                                      lowpassWS, lowpassWS,
                                      *_im2, lowpass);
            break;
          case LOWPASS_FILTERMASK:
            FilterLowpass_ByMask(lowpassMinX, lowpassMinY,
                                 lowpassWS, lowpassWS,
                                 *_im2, lowpass,
                                 _LowpassMask, _LowpassMaskSum);
            break;
          case LOWPASS_FILTERMASK_SEPARABLE:
            FilterLowpass_BySeparableMask(lowpassMinX, lowpassMinY,
                                          lowpassWS, lowpassWS,
                                          *_im2, lowpass,
                                          _LowpassMask, _LowpassMaskSum);
            break;
          }

        static Matrix<KLT_TYPE> gradx;
        switch (_GradXFilter)
          {
          case GRAD_SOBEL3X3:
            Filter_GradXSobel3x3(lowpass, gradx);
            break;
          case GRAD_FILTERMASK:
            Filter_ByMask(lowpass, gradx, _GradXMask, _GradXMaskSum);
            break;
          case GRAD_FILTERMASK_SEPARABLE:
            Filter_BySeparableMask(lowpass, gradx, _GradXMask, _GradXMaskSum);
            break;
          }

        static Matrix<KLT_TYPE> grady;
        switch (_GradYFilter)
          {
          case GRAD_SOBEL3X3:
            Filter_GradYSobel3x3(lowpass, grady);
            break;
          case GRAD_FILTERMASK:
            Filter_ByMask(lowpass, grady, _GradYMask, _GradYMaskSum);
            break;
          case GRAD_FILTERMASK_SEPARABLE:
            Filter_BySeparableMask(lowpass, grady, _GradYMask, _GradYMaskSum);
            break;
          }
      
        dev2_ = this->BilinearRegionFromMatrices_(x - lowpassMinX,
                                                  y - lowpassMinY,
                                                  hws,
                                                  lowpass, gradx, grady,
                                                  _bl2, _gx2, _gy2);
      } else {
        dev2_ = this->BilinearRegionFromImages_(x, y, hws,
                                                *_im2, *_gradim2x, *_gradim2y,
                                                _bl2, _gx2, _gy2);
      }
  }

  template <class StorageType>
  inline KLT_TYPE TrackerBaseInterface<StorageType>::
  BilinearRegionFromImages_(KLT_TYPE x, KLT_TYPE y, int hws,
                            Image<StorageType>& image,
                            Image<StorageType>& gradx,
                            Image<StorageType>& grady,
                            Matrix<KLT_TYPE>& bl,
                            Matrix<KLT_TYPE>& gx,
                            Matrix<KLT_TYPE>& gy)  {
    StorageType** image_ida = image.GetImageDataArray();
    StorageType** gradx_ida = gradx.GetImageDataArray();
    StorageType** grady_ida = grady.GetImageDataArray();

    int x_floor = (int)floor(x);
    int y_floor = (int)floor(y);
    int x_floor_end = x_floor+hws+1;
    int y_floor_end = y_floor+hws+1;
    KLT_TYPE brdy = y - (KLT_TYPE)(y_floor);
    KLT_TYPE brdx = x - (KLT_TYPE)(x_floor);
    KLT_TYPE brdxy = brdx*brdy;

    int rf, rc, cf, cc, r, c;
    for (rf=y_floor-hws, rc=rf+1, r=0; rf<y_floor_end; rf++, rc++, r++) {
      for (cf=x_floor-hws, cc=cf+1, c=0; cf<x_floor_end; cf++, cc++, c++) {
        KLT_TYPE ul, ur, ll, lr;
	
        ul = (KLT_TYPE)image_ida[rf][cf];
        ur = (KLT_TYPE)image_ida[rf][cc];
        ll = (KLT_TYPE)image_ida[rc][cf];
        lr = (KLT_TYPE)image_ida[rc][cc];
        bl[r][c] = (ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(gradx_ida[rf][cf]);
        ur = (KLT_TYPE)(gradx_ida[rf][cc]);
        ll = (KLT_TYPE)(gradx_ida[rc][cf]);
        lr = (KLT_TYPE)(gradx_ida[rc][cc]);
        gx[r][c] = (ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(grady_ida[rf][cf]);
        ur = (KLT_TYPE)(grady_ida[rf][cc]);
        ll = (KLT_TYPE)(grady_ida[rc][cf]);
        lr = (KLT_TYPE)(grady_ida[rc][cc]);
        gy[r][c] = (ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
      }
    }
    if (_AffineBrightnessInvariance) return NormalizeRegion_(bl, gx, gy);
    return 1.0;
  }

  template <class StorageType>
  inline KLT_TYPE TrackerBaseInterface<StorageType>::
  BilinearRegionFromMatrices_(KLT_TYPE x, KLT_TYPE y, int hws,
                              const Matrix<KLT_TYPE>& image,
                              const Matrix<KLT_TYPE>& gradx,
                              const Matrix<KLT_TYPE>& grady,
                              Matrix<KLT_TYPE>& bl,
                              Matrix<KLT_TYPE>& gx,
                              Matrix<KLT_TYPE>& gy) {
    int x_floor = (int)floor(x);
    int y_floor = (int)floor(y);
    int x_floor_end = x_floor+hws+1;
    int y_floor_end = y_floor+hws+1;
    KLT_TYPE brdy = y - (KLT_TYPE)(y_floor);
    KLT_TYPE brdx = x - (KLT_TYPE)(x_floor);
    KLT_TYPE brdxy = brdx*brdy;

    int rf, rc, cf, cc, r, c;
    for (rf=y_floor-hws, rc=rf+1, r=0; rf<y_floor_end; rf++, rc++, r++) {
      for (cf=x_floor-hws, cc=cf+1, c=0; cf<x_floor_end; cf++, cc++, c++) {
        KLT_TYPE ul, ur, ll, lr;
	
        ul = (KLT_TYPE)image[rf][cf];
        ur = (KLT_TYPE)image[rf][cc];
        ll = (KLT_TYPE)image[rc][cf];
        lr = (KLT_TYPE)image[rc][cc];
        bl[r][c] = (ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(gradx[rf][cf]);
        ur = (KLT_TYPE)(gradx[rf][cc]);
        ll = (KLT_TYPE)(gradx[rc][cf]);
        lr = (KLT_TYPE)(gradx[rc][cc]);
        gx[r][c] = (ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
        ul = (KLT_TYPE)(grady[rf][cf]);
        ur = (KLT_TYPE)(grady[rf][cc]);
        ll = (KLT_TYPE)(grady[rc][cf]);
        lr = (KLT_TYPE)(grady[rc][cc]);
        gy[r][c] = (ul + brdy*(ll - ul) + brdx*(ur - ul) + 
                    brdxy*(ul + lr - ll - ur));
      }
    }
    if (_AffineBrightnessInvariance) return NormalizeRegion_(bl, gx, gy);
    return 1.0;
  }

  template <class StorageType>
  inline KLT_TYPE TrackerBaseInterface<StorageType>::
  NormalizeRegion_(Matrix<KLT_TYPE>& bl,
                   Matrix<KLT_TYPE>& gx,
                   Matrix<KLT_TYPE>& gy,
                   const KLT_TYPE&  min_value)  {

    // mean value of patch
    KLT_TYPE mean = 0;
 
    int matrixsize = bl.num_cols()*bl.num_rows();
    // number of pixels above a certain threshold (for mask tracker)
    int validpixels = 0;
    
    // compute mean on not-black pixels
    KLT_TYPE *pbl = bl.GetData()-1, *pblend = bl.GetData() + matrixsize;
    while (++pbl<pblend) {
      if (*pbl > min_value) {
        mean += *pbl;
        validpixels++;
      } else *pbl = min_value;
    }
    // normalize mean for stddev computation
    if (validpixels>0) mean *= 1.0 / KLT_TYPE(validpixels);


    // estimate stddev
    KLT_TYPE stddev = 0.0;
    pbl = bl.GetData()-1;
    while (++pbl<pblend) {
      if (*pbl > min_value) {
        stddev += (*pbl - mean)*(*pbl - mean);
      }
    }
    if (fabs(stddev)>1e-10) { 
      stddev = sqrt(KLT_TYPE(validpixels)/stddev);
    } else {
      stddev = 1.0;
    }

    // normalize image and gradient values such that new image patch has 
    // zero mean and stddev 1
    KLT_TYPE *pgx = gx.GetData(),*pgy = gy.GetData();
    pbl = bl.GetData()-1;
    while (++pbl<pblend) {
      if (*pbl > min_value) {
        *pbl  = (*pbl-mean) * stddev;
        *pgx *= stddev;
        *pgy *= stddev;
      } else {
        *pgx = *pgy = 0;
        *pbl = TRACKERBASE_INVALID_PIXEL;
      }
      pgx++; pgy++;
    }
    return stddev;
  }


} // namespace

#include <Base/Common/BIASpragmaEnd.hh>

#endif // __TrackerBaseInterface__hh__
