/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



/** @example ExampleRANSAC.cpp
    @class ExampleRANSAC
    @relates RANSAC
    @ingroup g_examples    
    @brief  example for using deriving from BIAS::RANSAC for own purposes
    @author MIP
*/


#include <Base/Common/BIASpragma.hh>

#include "Base/Math/Random.hh"
#include <vector>
#include <iostream>
#include <iomanip>
#include <float.h>
#include "ExampleRANSAC_double.hh"

using namespace BIAS;
using namespace std;


/* input : inlier   percentage of inlier
           count    number of data generated
           mean     mean ofthe generated data
           sigma    standard deviation of the generated data
   output: data     the generated data */
void GenerateData(unsigned int count, double mean, double sigma, 
                  double inlier_fraction, vector<double>& data, 
                  Random& ran)
{
  if (data.size()!=count) data.resize(count);

  double outlier_num;

  for (unsigned int i=0; i<count; i++){
    outlier_num=ran.GetUniformDistributed(0.0, 1.0);
    if (outlier_num>inlier_fraction){ // generate a outlier
      data[i]=ran.GetNormalDistributed(mean, sigma)+
        ran.GetUniformDistributed(3*sigma, 30.0*sigma);
    } else { // generate an inlier
      data[i]=ran.GetNormalDistributed(mean, sigma);
    }
  }
}


//int main(int argc, char *argv[])
int main()
{
  Random ran;
  int res =0;
  vector<vector<double> > data_array;
  vector<double> data;
  double min=-100.0, max=100.0;
  double minsigma=0.0, maxsigma=5.0;
  vector<double> mean;
  vector<double> sigma;
  double inlier_fraction=0.7;
  vector<double> expected_inlier_fraction;
  unsigned int data_length=1000;
  unsigned int sample_size=2;
  bool refine_solution=true;
  double inlier_distance_threshold;
  double solution;
  vector<bool> inliers;
  bool debug=true;
  doubleRANSAC dr;
  double mean_error=0.0;
  double solution_error_threshold;

  
  if (debug){
    dr.AddDebugLevel("D_RANSAC_SAMPLES");
    dr.AddDebugLevel("D_RANSAC_SOLVE_MASTER");
    dr.AddDebugLevel("D_DOUBLE_RANSAC_SOLUTION");
  }

   unsigned int count=1000;
   if (debug) { count=1; data_length=100; }

  // do this first because Random is used in doubleRANSAC and 
  // therefor srand is called
  data_array.resize(count);
  mean.resize(count);
  sigma.resize(count);
  expected_inlier_fraction.resize(count);
  for (unsigned int i=0; i<count; i++){
    mean[i]=ran.GetUniformDistributed(min, max);
    sigma[i]=ran.GetUniformDistributed(minsigma, maxsigma);
    
    expected_inlier_fraction[i]=inlier_fraction+
      ran.GetNormalDistributed(0.0, (1.0-inlier_fraction)/3.0);
    if (expected_inlier_fraction[i]>1.0) expected_inlier_fraction[i]=1.0;
    GenerateData(data_length, mean[i], sigma[i], inlier_fraction, 
                 data_array[i], ran);
  }

  for (unsigned int i=0; i<count; i++){
    inlier_distance_threshold=3.0*sigma[i];
    solution_error_threshold=1.5*sigma[i];
    //expected_inlier_fraction[i]=0.5;
    
    if (debug) {
      for (unsigned int j=0; j<data_length; j++)
        cout << data_array[i][j] << "  ";
      cout << endl;
      cout <<setw(5)<<i<<":\tmean: "<<mean[i]<<"\tsigma: "<<sigma[i]<<endl;
      cout << "inlier_distance_threshold = "<<inlier_distance_threshold<<endl;
      cout << "solution_error_threshold = "<<solution_error_threshold<<endl;
      cout << "expected_inlier_fraction " << expected_inlier_fraction[i]<<endl;
    }
  
    dr.Init(data_array[i], data_length, inlier_distance_threshold, 
            refine_solution, sample_size, expected_inlier_fraction[i], 
            solution_error_threshold);

    if (dr.SolveMaster(expected_inlier_fraction[i], solution, inliers)!=0){
      BIASERR("error runnign SolveMaster");
      cerr << "true solution          \t" << mean[i] <<"  \t"
           << "sigma: "<<sigma[i]<<"    \t"
           << "expected_inlier_fraction: "<<expected_inlier_fraction[i]<<endl;
    } else {
      cerr << "true solution          \t" << mean[i] <<"  \t"
           << "sigma: "<<sigma[i]<<"    \t"
           << "expected_inlier_fraction: "<<expected_inlier_fraction[i]<<endl
           << "estimated solution     \t" << solution <<"  \t"
           << "error: "<<fabs(solution-mean[i])<<"    \t"
           << "expected_inlier_fraction: "<<expected_inlier_fraction[i]<<endl;
      if (solution!=solution) exit(-3);
      mean_error+=(sigma[i]!=0.0)?(fabs(solution-mean[i])/sigma[i]):
        (fabs(solution-mean[i])/DBL_MIN);
    }
  }
  mean_error/=(double)count;

  cerr << "\n mean error: "<<mean_error<<" / sigma"<<endl;
  return res;
}
