/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



  /**  This is an example file on how to use the RANSAC algorithm.
      It tries to find a dominant plane in a number of 3d points.

      RANSAC:
      In GetSampleSolutions, 3 of the points are picked and a plane is
      constructed from them.

      In EvaluateSolution, it is checked, howe many points are "inliers" to
      that plane.

      In RefineSiolution, least squares is done on all inliers to get a
      better plane.

      You have to set the inlier threshold (max distance to plane to accept
      a point as being on the plane) and the fraction of valid 3d points,
      which defines the number of samples performed by RANSAC.

      When you call SolveMaster then, you will get the best Plane according
      to your input.     
  */

#include <Base/Common/BIASpragma.hh>

#include "Base/Math/Random.hh"
#include <vector>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <float.h>
#include "ExampleRANSACPlane.hh"


using namespace BIAS;
using namespace std;

#define NUM_POINTS 20

void readFile(char *filename, std::vector<HomgPoint3D> &points)
{
  ifstream infile(filename);
  if (!infile) {
    BIASERR("couldnt open:"<<filename);
    return;
  }
  char c;
  HomgPoint3D point(0,0,0,1);
  while (!infile.eof()) {
    infile >> point[0] >> point[1]>>point[2]>>c;
    if (!infile.eof()) {
      points.push_back(point);
      cout <<" reading point "<<point<<endl;
    }
  }
}


void GeneratePoints(std::vector<HomgPoint3D>& points){

  // generate points on plane z=2
  Random randomizer;
  double range = 10.0;
  HomgPoint3D point3D;
  double noise = 0.5;
  for(unsigned int i=0; i < NUM_POINTS; i++){
    point3D[0] = randomizer.GetUniformDistributed(-range, range);
    point3D[1] = randomizer.GetUniformDistributed(-range, range);
    point3D[2] = 2.0 + randomizer.GetUniformDistributed(-noise, noise);;
    point3D[3] = 1.0;
    points.push_back(point3D);
    cout << "point " << i << " " << point3D << endl;
  }

}

int main(int argc, char *argv[])
{
  int res =0;
  // inlier_fraction is passed onto SolveMaster and
  // contains the assumed fraction of inliers in the data
  // this parameter is used to determine the number of samples drawn
  // and is therefore important, unless the number of samples is set
  // explicitly
  double inlier_fraction=0.1;
  // determines if promising solutions are to be refined based on all inliers or not
  bool refine_solution=true;
  // max distance between point and plane, for point to be classified as inlier
  double  maxsquareddistance = 0.0001;

  HomgPlane3D solution;
  vector<bool> inliers;
  bool debug=true;
  PlaneRANSAC dr;


  vector<HomgPoint3D> points;
  if (argc<2) {
    cout << "Generating some random noisy points for testing... " << endl;
    GeneratePoints(points);
  } else {
    readFile(argv[1], points);
  }

  if (debug){
    dr.AddDebugLevel("D_RANSAC_SAMPLES");
    dr.AddDebugLevel("D_RANSAC_SOLVE_MASTER");
    dr.AddDebugLevel("D_DOUBLE_RANSAC_SOLUTION");
  }



  dr.Init(points, points.size(), maxsquareddistance, refine_solution, 3,
          inlier_fraction, -1.0);

  dr.SolveMaster(inlier_fraction, solution, inliers);

  cout << "inlier/outlier classification " << endl;
  for(unsigned int i=0; i < inliers.size(); i++){
    cout << "point " << i << " is inlier (1)/ outlier(0) " << inliers[i] << endl;
  }
  cout <<"solution is "<<solution<<endl;

  return res;
}
