/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** @example ExampleRANSACPlane.cpp
    @class ExampleRANSACPlane
    @relates RANSAC
    @ingroup g_examples    
    @brief  this is a class evaluating the mean of the double data assuming some outlier
        using the RANSAC algorithm. In order for the RANSAC algorithm to work, the methods GetSampleSolution,
        EvaluateSolution, and Refine solution need to be implemented
        
      This is an example file on how to use the RANSAC algorithm.
      It tries to find a dominant plane in a number of 3d points.

      RANSAC:
      In GetSampleSolutions, 3 of the points are picked and a plane is
      constructed from them.

      In EvaluateSolution, it is checked, howe many points are "inliers" to
      that plane.

      In RefineSiolution, least squares is done on all inliers to get a
      better plane.

      You have to set the inlier threshold (max distance to plane to accept
      a point as being on the plane) and the fraction of valid 3d points,
      which defines the number of samples performed by RANSAC.

      When you call SolveMaster then, you will get the best Plane according
      to your input.     
    @author MIP
*/



#ifndef __ExampleRANSAC_Plane_hh__
#define __ExampleRANSAC_Plane_hh__

#include <Base/Common/BIASpragmaStart.hh>


#include "MathAlgo/RANSAC.hh"
#include "MathAlgo/SVD.hh"
#include "Base/Geometry/HomgPoint3D.hh"
#include "Base/Geometry/HomgPlane3D.hh"

#include <string>

namespace BIAS {

  class PlaneRANSAC : public RANSAC<HomgPlane3D>
  {
    /**this is a class evaluating the mean of the
        double data assuming some outlier
        using the RANSAC algorithm
        In order for the RANSAC algorithm to work, the methods GetSampleSolution,
        EvaluateSolution, and Refine solution need to be implemented . */
  public:
    inline PlaneRANSAC();

    inline ~PlaneRANSAC();

    /**
     * init function calling the regular init function from RANSAC. After that, the
     * other parameters are set.
     * @attention The RANSAC init method resets a lot of parameters to default values.
     * For example explicitSampleCount.
     * So setting these before calling init will have not effect.
     */
    inline void Init(std::vector<HomgPoint3D>& data, unsigned int data_length,
                     double inlier_distance_threshold, bool refine_solution,
                     unsigned int sample_size, double expected_inlier_fraction,
                     double solution_error_threshold);

    /** takes sample and computes solution for this sample. This is usually done
     *  in an easy way, e.g. linearly. Sometimes solutions based on a minimum
     *  sample size are not unambiguous as for example in case of the 7 point
     *  algorithm for essential matrix estimation. In those cases all possible
     *  solutions have to be returned in the solutions vector.
     */
    inline int GetSampleSolutions(std::vector<unsigned int> &which_samples,
                                  std::vector<HomgPlane3D> &solutions);

    /**
     * The solution denoted by solution needs to be ranked. This is done based
     * on the number of inliers found for this solution and based on the accuracy
     * reached on these inliers. So some kind of measurement is needed. Here the measurement
     * is the absolute distance between the homogeneous 3D points and the plane. If
     * such a distance is below the threshold, and inlier has been found. The
     * average distance over all inliers is the evaluate_score. While the number
     * of inliers is entered in accept_count. For more details refer to
     * documentation of this method in RANSAC.hh.
     */
    inline bool EvaluateSolution(const HomgPlane3D& solution,
                                 bool existing_solution_flag,
                                 std::vector<bool>& inlier,
                                 int& accept_count,
                                 double& evaluate_score,
                                 bool& ok_to_terminate_flag);

    /** RefineSolution is only called, if the RefineSolutionFlag is set to true.
     * In any case, it is only applied to solutions, that have been found to be good,
     * indicated by the return value of EvaluateSolution.
     *
     */
    inline bool RefineSolution(std::vector<unsigned int>& which_samples,
                               HomgPlane3D& solution,
                               std::vector<bool>& new_inliers);

  protected:
    std::vector<HomgPoint3D> _vData;
    /// data  closer to solution than _dInlierDistanceThreshold will be
    /// regarded as inlying
    double _dInlierDistanceThreshold;
    /// if a sample yields more than _dExpectedInlierFraction inliers and
    /// the evaluate_score is lower than _dSolutionErrorThreshold,
    /// RANSACing will stop
    double _dExpectedInlierFraction;
    /// if a sample yields more than _dExpectedInlierFraction inliers and
    /// the evaluate_score is lower than _dSolutionErrorThreshold,
    /// RANSACing will stop
    double _dSolutionErrorThreshold;
  };

  PlaneRANSAC::PlaneRANSAC() : RANSAC<HomgPlane3D>()
  {}

  PlaneRANSAC::~PlaneRANSAC()
  {}

  void PlaneRANSAC::Init(std::vector<HomgPoint3D>& data,
                          unsigned int data_length,
                          double inlier_distance_threshold,
                          bool refine_solution, unsigned int sample_size,
                          double expected_inlier_fraction,
                          double solution_error_threshold)
  {
    RANSAC<HomgPlane3D>::Init(sample_size, 3, refine_solution);
    _vData=data;
    _uiDataSize=data_length;
    _dInlierDistanceThreshold=inlier_distance_threshold;
    _dExpectedInlierFraction=expected_inlier_fraction;
    _dSolutionErrorThreshold=solution_error_threshold;
  }

  int
  PlaneRANSAC::GetSampleSolutions(std::vector<unsigned int> &which_samples,
                                   std::vector<HomgPlane3D> &solutions)
  {
    solutions.clear();
    // computing a plane out of 3 3D points is really easy!
    solutions.push_back(HomgPlane3D(_vData[which_samples[0]],
                                    _vData[which_samples[1]],
                                    _vData[which_samples[2]]));

    // and there is only one solution in this case

    return solutions.size();
  }

  bool PlaneRANSAC::EvaluateSolution(const HomgPlane3D& solution,
                                      bool existing_solution_flag,
                                      std::vector<bool>& inlier,
                                      int& accept_count,
                                      double& evaluate_score,
                                      bool& ok_to_terminate_flag)
  {
    bool good_flag=true;
    inlier.clear();
    inlier.resize(_uiDataSize);

    int my_accept_count = 0;
    evaluate_score = 0.0;

    for (unsigned int i=0; i<_uiDataSize; i++){
      inlier[i] = (solution.DistanceSquared(_vData[i]) <
                   _dInlierDistanceThreshold);
      if (inlier[i]) {
        my_accept_count++;
        evaluate_score += solution.DistanceSquared(_vData[i]);
    //    if (existing_solution_flag && my_accept_count>accept_count) break;
      }
    }

    // determine average score
    if (my_accept_count!=0) {
      evaluate_score/=my_accept_count;
    } else {
      evaluate_score=DBL_MAX;
      good_flag=false;
    }

    // set new accept count
    accept_count=my_accept_count;

    // ok_to_terminate_flag is important to set, if greedy evaluation is needed.
    // If set to true, the algorithm will stop after this sample
    ok_to_terminate_flag=
      ((((double)accept_count/(double)_uiDataSize) > _dExpectedInlierFraction)
      && (evaluate_score < _dSolutionErrorThreshold));

    std::string bo=(ok_to_terminate_flag)?("true"):("false");
    std::cout << "terminating ? " << bo << std::endl;

    // this return value decides if the solution will be considered further by refining it.
    return good_flag;
  }

  bool PlaneRANSAC::RefineSolution(std::vector<unsigned int>& which_samples,
                                    HomgPlane3D& solution,
                                    std::vector<bool>& new_inliers
                                    )
  {
    BIAS::Matrix<double>  A;

    // fill the 3d  points rowwise into A using the scheme
    const unsigned int nr_of_points = which_samples.size();
    //std::cout<<"RefineSolution has "<<nr_of_points<<" inliers"<<std::endl;
    A.newsize( nr_of_points, 4 );

    for (unsigned int i=0; i<nr_of_points; i++) {
        //        -1            X_i             Y_i             Z_i
      A[i][0] = _vData[which_samples[i]][0];
      A[i][1] = _vData[which_samples[i]][1];
      A[i][2] = _vData[which_samples[i]][2];
      A[i][3] = _vData[which_samples[i]][3];
    }
    A = A.Transpose()*A;
    // get a solution (linear fit) of this equation system by its SVD
    SVD svd(A , 1e-5, false);

    // some information about svd results
  // cout<<"U = " << svd.GetU() <<endl;
  // cout<<"S = " << svd.GetS() <<endl;
  // cout<<"V = " << svd.GetV() <<endl;
  // cout<<"Nullspace dim = " << svd.NullspaceDim() <<endl;;

    Vector4<double> regplane;
    if (svd.Rank()<3) return false;
    regplane = svd.GetNullvector(); // get the last Nullvector
    solution = HomgPlane3D(regplane);
    // cout <<"solution: " << solution << endl;


    return true;
  }

} // namespace

#include <Base/Common/BIASpragmaEnd.hh>


#endif // __ExampleRANSAC_double_hh__
