/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __Median1D_hh__
#define __Median1D_hh__

#include <bias_config.h>

#include <Base/Common/BIASpragmaStart.hh>
#include <Base/Debug/Debug.hh>

#include <vector>

namespace BIAS {

  /** @class Median1D
      @ingroup g_mathalgo
      @brief computes the median of a vector
      @author woelk 06/2005 */
  template <class DataType>
  class  Median1D : public Debug // No BIASMathAlgo_EXPORT here, all code is in header file !
  {
  public:
    Median1D();

    Median1D(std::vector<DataType>& vec);

    Median1D(const std::vector<DataType>& vec);

    ~Median1D();

    /** computation of the median using a quicksort approach
        @author woelk 06/2005 */
    void Compute(std::vector<DataType>& vec);

    void Compute(const std::vector<DataType>& vec);

    

    inline DataType GetMedian()
    { return _Median; }

    /** Compute and return median  of absolute differeces to median
        @authoer ask woelk
    */
    DataType GetX84(std::vector<DataType>& vec, DataType median);

  protected:
    DataType _Median;
  
  }; // class 

  /////////////////////////////////////////////////////////////////////////
  // implementation
  /////////////////////////////////////////////////////////////////////////

  template<class DataType>
  Median1D<DataType>::Median1D()
  {}

  template<class DataType>
  Median1D<DataType>::Median1D(std::vector<DataType>& vec)
  {
  
    Compute(vec);
  }

  template<class DataType>
  Median1D<DataType>::Median1D(const std::vector<DataType>& vec)
  {
  
    Compute(vec);
  }

  template<class DataType>
  Median1D<DataType>::~Median1D()
  {}

  template<class DataType>
  void Median1D<DataType>::Compute(const std::vector<DataType>& vec)
  {
    std::vector<DataType> tmp=vec;
    Compute(tmp);
  }

  /** This Quickselect routine is based on the algorithm described in
      "Numerical recipies in C", Second Edition,
      Cambridge University Press, 1992, Section 8.5, ISBN 0-521-43108-5 */

#define BIAS_SWAP(a,b) { t=(a); (a)=(b); (b)=t; }

  template<class DataType>
  void Median1D<DataType>::Compute(std::vector<DataType>& vec)
  {
    if (vec.size()==0){
      BEXCEPTION("You are trying to compute the median of no values!");
      _Median = DataType(); // undefined
      return;
    }

    int low, high;
    int med;
    int mid, ll, hh;
    int n=vec.size();
    register DataType t;
    //this is catched above (NAN isn't defined everywhere)!!!
    //   if (vec.empty()) {
    //     BIASERR("cannot compute median of empty vector");
    //     _Median = (DataType)NAN;
    //     return;
    //   }
    low = 0 ; high = n-1 ; med = (low + high) / 2;
    for (;;) {
      if (high <= low) { /* One element only */
        _Median = vec[med];
        return;
      }

      if (high == low + 1) {  /* Two elements only */
        if (vec[low] > vec[high])
          BIAS_SWAP(vec[low], vec[high]) ;
        _Median = vec[med];
        return;
      }

      /* Find med of low, mid and high items; swap into position low */
      mid = (low + high) / 2;
      if (vec[mid] > vec[high]) BIAS_SWAP(vec[mid], vec[high]) ;
      if (vec[low] > vec[high]) BIAS_SWAP(vec[low], vec[high]) ;
      if (vec[mid] > vec[low])  BIAS_SWAP(vec[mid], vec[low]) ;

      /* Swap low item (now in position mid) into position (low+1) */
      BIAS_SWAP(vec[mid], vec[low+1]) ;

      /* Nibble from each end towards mid, swapping items when stuck */
      ll = low + 1;
      hh = high;
      for (;;) {
        do ll++; while (vec[low] > vec[ll]) ;
        do hh--; while (vec[hh]  > vec[low]) ;

        if (hh < ll)
          break;

        BIAS_SWAP(vec[ll], vec[hh]) ;
      }

      /* Swap mid item (in position low) back into correct position */
      BIAS_SWAP(vec[low], vec[hh]) ;

      /* Re-set active partition */
      if (hh <= med)
        low = ll;
      if (hh >= med)
        high = hh - 1;
    }
  
  }




  template<class DataType>
  DataType Median1D<DataType>::GetX84(std::vector<DataType>& vec, DataType median)
  {
    if (vec.size()==0)
      BEXCEPTION("Can not compute X84 on empty data");
    unsigned int n=vec.size();
  
    std::vector<DataType> diff;
    diff.resize(n);

    for (unsigned int i=0;i<n;i++)
      diff[i] = (vec[i]>median)?(vec[i]-median):(median-vec[i]);

    Median1D<DataType> subfilter(diff);
    return   subfilter.GetMedian();

  }

#undef BIAS_SWAP


} // namespace


#include <Base/Common/BIASpragmaEnd.hh>

#endif // __Median1D_hh__

