/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Minpack.hh"
#include <Base/Math/tnt/fortran.h>
#include <Base/Common/CompareFloatingPoint.hh>

#ifdef __cplusplus 
extern "C" 
{
#endif
  /* hybrd and hybrd1 solve a system of n non-linear equations using the
     powell hybrid method. Jacobian is aproximated numerical */
  void hybrd_(void (*fcn)(long int *n, double *x, double *fvec, long int *iflag), 
              cfi_ n, fda_ x, fda_ fvec, cfd_ xtol, cfi_ maxfev,
              cfi_ ml, cfi_ mu, cfd_ epsfcn, fda_ diag, cfi_ mode, cfd_ factor,
              cfi_ nprint, fi_ info, fi_ nfev, fda_ fjac, cfi_ ldfjac, fda_ r,
              cfi_ lr, fda_ qtf, fda_ wa1, fda_ wa2, fda_ wa3, fda_ wa4);

  void hybrd1_(void (*fcn)(long int *n, double *x, double *fvec, long int *iflag), 
               cfi_ n, fda_ x, fda_ fvec, cfd_ tol, fi_ info,
               fda_ wa, cfi_ lwa);

  /* lmder and lmder1 solve a system of n non-linear equations with m 
     parameters using the Levenberg-Marquardt algorithm. A least squares 
     solution is used. Jacobian is given analytically */
  void lmder_(void (*fcn)(long int *m, long int *n, double *x, double *fvec, 
                          double *fjac, long int *ldfjac, long int *iflag), 
              cfi_ m, cfi_ n, fda_ x, fda_ fvec, fda_ fjac, 
              cfi_ ldfjac, cfd_ ftol, cfd_ xtol, cfd_ gtol, cfi_ maxfev, 
              fda_ diag, cfi_ mode, cfd_ factor, cfi_ nprint, fi_ info, 
              fi_ nfev, fi_ njev, fia_ ipvt, fda_ qtf, fda_ wa1, fda_ wa2, 
              fda_ wa3, fda_ wa4);


  void lmder1_(void (*fcn)(long int *m, long int *n, double *x, double *fvec, 
                           double *fjac, long int *ldfjac, long int *iflag), 
               cfi_ m, cfi_ n, fda_ x, fda_ fvec, fda_ fjac, 
               cfi_ ldfjac, cfd_ tol, fi_ info, fia_ ipvt, fda_ wa, cfi_ lwa);

  /* lmdif and lmdif1 solve a system of n non-linear equations with m 
     parameters using the Levenberg-Marquardt algorithm. A least squares 
     solution is used. Jacobian is aproximated numerical */
  void lmdif_(void (*fcn)(long int *m, long int *n, double *x, double *fvec, long int *iflag),
              cfi_ m, cfi_ n, fda_ x, fda_ fvec, cfd_ ftol, cfd_ xtol, 
              cfd_ gtol, cfi_ maxfev, cfd_ epsfcn, fda_ diag, cfi_ mode,
              cfd_ factor, cfi_ nprint, fi_ info, fi_ nfev, fda_ fjac,
              cfi_ ldfjac, fia_ ipvt, fda_ qtf, fda_ wa1, fda_ wa2, fda_ wa3,
              fda_ wa4);

  void lmdif1_(void (*fcn)(long int *m, long int *n, double *x, double *fvec, 
                           long int *iflag), cfi_ m, cfi_ n, fda_ x, fda_ fvec,
               cfd_ tol, fi_ info, fia_ iwa, fda_ wa, cfi_ lwa, cfi_ maxfev);

  int chkder_(long int *m, long int *n, double *x, double *fvec, double *fjac, 
              long int *ldfjac, double *xp, double *fvecp, long int *mode, 
              double *err);

  int fdjac2_(void (*fcn)(long int *m, long int *n, double *x, double *fvec, 
                          long int *iflag), long int *m, long int *n, double *x, double *fvec,
              double *fjac, long int *ldfjac, long int *iflag, double *eps, double *wa);
#ifdef __cplusplus 
}
#endif

namespace BIAS{
  using namespace std;

  long int Powell(void (*fun)(long int *n, double *x, double *fvec, long int *iflag), 
             Vector<double>& InitialGuess, Vector<double>& Result, 
             double Tolerance)
  {
    long int info;
    long int n = InitialGuess.Size();
    long int lwa=(n*(3*n+13))/2;
    double *wa=new double[lwa];
    Result=InitialGuess;
    Vector<double> FuncVec(n);

    hybrd1_(fun, &n, Result.GetData(), FuncVec.GetData(), &Tolerance, 
            &info, wa, &lwa);

    delete[] wa;
    return (info==1)?(0):(info==0?-1:info);
  }

  long int PowellExtended(void (*fun)(long int *n, double *x, double *fvec, long int *iflag), 
                     Vector<double>& InitialGuess, Vector<double>& Result, 
                     double EpsFcn, double Tolerance)
  {
    long int info;
    long int n=InitialGuess.Size();
    double *wa=new double[4*n];
    Result=InitialGuess;
    Vector<double> FuncVec(n);
    long int maxfev = 200*(n + 1);
    long int ml = n - 1;
    long int mu = n - 1;
    long int mode = 2;
    long int lr = (n*(n + 1))/2;
    double factor=1.0;
    long int nprint = 0;
    double *diag=new double[n];
    for (long int i=0; i<n; i++) diag[i]=1.0;
    long int nfev;
    double *fjac=new double[n*n];
    long int ldfjac=n;
    double *r=new double[lr];
    double *qtf=new double[n];

    hybrd_(fun, &n, Result.GetData(), FuncVec.GetData(), &Tolerance, 
           &maxfev, &ml, &mu, &EpsFcn, diag, &mode, &factor, &nprint, 
           &info, &nfev, fjac, &ldfjac, r, &lr, qtf, &wa[0], &wa[n], 
           &wa[2*n], &wa[3*n]);
    
    delete[] wa;
    delete[] fjac;
    delete[] r;
    delete[] qtf;
    delete[] diag;
 
    return (info==1)?(0):(info==0?-1:info);
  }

  long int LevenbergMarquardt(void (*fun)(long int *m, long int *n, double *x, 
                                     double *fvec, double *fjac, 
                                     long int *ldfjac, long int *iflag), 
                         long int m, long int n, Vector<double>& InitialGuess, 
                         Vector<double>& Result, double Tolerance)
  {
#ifdef BIAS_DEBUG
    if (m<n) {
      BIASERR("dimension of function output is bigger than input "
              <<m<<" "<<n);
    }
#endif
    long int info;
    Result=InitialGuess;
    Vector<double> FuncVec(m);
    double *fjac=new double[m*n];
    long int ldfjac=n;
    long int *ipvt=new long int[n];
    long int lwa = 5*n+m;
    double *wa=new double[lwa];

    lmder1_(fun, &m, &n, Result.GetData(), FuncVec.GetData(), fjac, 
            &ldfjac, &Tolerance, &info, ipvt, wa, &lwa);

    delete[] fjac;
    delete[] ipvt;
    delete[] wa;

    cout << "lmder1_ Tolerance "<<Tolerance<<endl;
    cout << "lmder1_ returned "<<info<<endl;

    return (info>0 && info<5)?(0):(info==0?-1:info);
  }

  long int LevenbergMarquardtExtended(void (*fun)(long int *m, long int *n, double *x, 
                                             double *fvec, double *fjac, 
                                             long int *ldfjac, long int *iflag), 
                                 long int m, long int n, Vector<double>& InitialGuess, 
                                 Vector<double>& Result, double Tolerance,
                                 long int MaxIter)
  {
#ifdef BIAS_DEBUG
    if (m<n) {
      BIASERR("dimension of function output is bigger than input "
              <<m<<" "<<n);
      BIASASSERT(m>=n);
    }
#endif
    long int info;
    Result=InitialGuess;
    Vector<double> FuncVec(m);
    //    long int maxfev = 200*(n + 1);
    long int maxfev = MaxIter;
    // termination occurs when chaneg in residuum is less than ftol
    double ftol = Tolerance;
    // termination occurs when change in solution vector is less than
    // xtol
    double xtol = Tolerance;
    double gtol = 0.0;
    long int mode = 1;
    long int nprint = 0;
    double *diag=new double[n];
    double factor=1.0;
    long int *iptv=new long int[n];
    double *qtf=new double[n];
    double *wa=new double[m+3*n];
    long int nfev, njfev;
    double *fjac=new double[m*n];
    long int ldfjac=m;

    lmder_(fun, &m, &n, Result.GetData(), FuncVec.GetData(), fjac, &ldfjac, 
           &ftol, &xtol, &gtol, &maxfev, diag, &mode, &factor, &nprint, &info, 
           &nfev, &njfev, iptv, qtf, &wa[0], &wa[n], &wa[2*n], &wa[3*n]);
    
    delete[] diag;
    delete[] iptv;
    delete[] qtf;
    delete[] wa;
    delete[] fjac;
    
    //    cout << "lmder_ Tolerance "<<Tolerance<<endl;
    //    cout << "lmder_ returned "<<info<<endl;
    if (info==0){
      BIASERR("improper input parameter");
    }
    return (info>0 && info<4)?(0):(info==0?-1:info);
  }

  long int LevenbergMarquardt(void (*fun)(long int *m, long int *n, double *x, 
                                     double *fvec, long int *iflag), 
                         long int m, long int n, Vector<double>& InitialGuess, 
                         Vector<double>& Result, double Tolerance, long int MaxIter)
  {
#ifdef BIAS_DEBUG
    if (m<n) {
      BIASERR("dimension of function output is bigger than input "
              <<m<<" "<<n);
    }
#endif
    long int info;
    Result=InitialGuess;
    Vector<double> FuncVec(m);
    long int *iwa=new long int[n];
    long int lwa = m*n+5*n+m;
    double *wa=new double[lwa];
    
    lmdif1_(fun, &m, &n, Result.GetData(), FuncVec.GetData(), &Tolerance, 
            &info, iwa, wa, &lwa, &MaxIter);

    delete[] iwa;
    delete[] wa;

    return (info>0 && info<5)?(0):(info==0?-1:info);
  }


  long int LevenbergMarquardtExtended(void (*fun)(long int *m, long int *n, double *x, 
                                             double *fvec, long int *iflag), 
                                 long int num_errors, long int num_vars, 
                                 Vector<double>& InitialGuess, 
                                 Vector<double>& Result, double EpsFcn,
                                 double Tolerance, long int MaxIter, double* fjac, 
                                 long int* ipvt, double *SumOfSquaredErrors)
  {
#ifdef BIAS_DEBUG
    if (num_errors<num_vars) {
      BIASERR("number of params is bigger than number of error functions "
              <<num_errors<<" "<<num_vars);
    }
#endif
    long int info;
    Result=InitialGuess;
    Vector<double> FuncVec(num_errors);
    long int maxfev;
    if ( MaxIter==LM_DEF_MAX_ITER )
      maxfev = 200*(num_vars + 1);
    else
      maxfev = MaxIter;

    double ftol = Tolerance;
    double xtol = Tolerance;
    //    double gtol = 0.0;
    double gtol = Tolerance;
    long int mode = 1;
    long int nprint = 0;
    double *diag=new double[num_vars];
    double factor=1.0;
    bool ExhaustiveData = 
      (fjac!=NULL) && (ipvt!=NULL) && (SumOfSquaredErrors!=NULL);
    if (!ExhaustiveData) {
      fjac = new double[num_vars*num_errors];
      ipvt = new long int[num_vars];
    }


    const long int ldfjac  = num_errors;
    double *qtf = new double[num_vars];
    double *wa  = new double[num_errors+3*num_vars];
    long int nfev;

    lmdif_(fun, &num_errors, &num_vars, Result.GetData(), 
           FuncVec.GetData(), &ftol, &xtol, 
           &gtol, &maxfev, &EpsFcn, diag, &mode, &factor, &nprint, &info, 
           &nfev, fjac, &ldfjac, ipvt, qtf, &wa[0], &wa[num_vars], 
           &wa[2*num_vars], &wa[3*num_vars]);

    delete[] diag;
    if (!ExhaustiveData) {
      delete[] fjac;
      fjac = NULL;
      delete[] ipvt;
      ipvt = NULL;
    } else {
      *SumOfSquaredErrors = 0;
      for (long int i=0; i<num_vars; i++) *SumOfSquaredErrors += qtf[i]*qtf[i];
    }
   
    delete[] qtf;
    delete[] wa;

    return (info>0 && info<5)?(0):(info==0?-1:info);
  }


  long int ComputeCovariance(long int NumErrors, long int NumParams, double *Jac, 
                        long int *Permutation, double &SumOfSquaredErrors, 
                        Matrix<double>& Cov)
  {
    Matrix<double>mJac(NumParams, NumErrors);
    Vector<long int> mperm(NumParams);
    memcpy(mJac.GetData(), Jac, NumErrors*NumParams*sizeof(double));
    memcpy(mperm.GetData(), Permutation, NumParams*sizeof(long int));
    return ComputeCovariance(NumErrors, NumParams, mJac, mperm, 
                             SumOfSquaredErrors, Cov);
  }


  long int ComputeCovariance(const long int NumErrors, const long int NumParams, 
                        const Matrix<double> &Jac, 
                        const Vector<long int>& Permutation, 
                        const double &SumOfSquaredErrors, 
                        Matrix<double>& Cov)
  {
    // enforce correct size
    Cov.newsize(NumParams, NumParams);
 
    // buffer for intermediate results 
    Matrix<double> TMP(NumParams, NumParams), P(NumParams, NumParams);
   
    // Get R out of LM-Jac
    // citing LM-documentation (where ipvt is Permutation in our code):
    //
    // "fjac is an output m by n array. the upper n by n submatrix 
    //  of fjac contains an upper triangular matrix r with 
    //  diagonal elements of nonincreasing magnitude such that 
    //
    //                t     t           t 
    //               p *(jac *jac)*p = r *r, 
    //
    //  where p is a permutation matrix and jac is the final
    //  calculated jacobian. column j of p is column ipvt(j)
    //  (see below) of the identity matrix. the lower trapezoidal
    //  part of fjac contains information generated during
    //  the computation of r."
    int row, column, k;
    for (row = 0; row < NumParams; row++) 
      for (column = 0; column < NumParams; column++) 
        Cov[row][column] = (column<row) ? 0.0 : Jac[column][row];
  
    //cout << "1: Cov: "<<Cov<<endl;

    // invert R (triangular)
    double s;
    for (row = 0; row < NumParams; row++) {
      // invert main diagonal
      Cov[row][row] = (Equal(Cov[row][row], 0.0)) ? 0.0 : 1.0 / Cov[row][row];
      // upper right triang part
      for (column = row + 1; column < NumParams; column++) {
        s = 0;
        for (k = row; k < column; k++) {
          s -= Cov[row][k] * Cov[k][column];
        }
        Cov[row][column] = 
          (Equal(Cov[row][column], 0.0)) ? 0.0 : s / Cov[column][column];
      }
    }

    //cout << "2: Cov: "<<Cov<<endl;

    // calculation of R^-1 R^-T 
    for (row = 0; row < NumParams; row++)
      for (column = 0; column < NumParams; column++) {
        TMP[row][column] = 0;
        for (k = 0; k < NumParams; k++) {
          TMP[row][column] += Cov[row][k] * Cov[column][k];
        }
      }

    //cout << "2: TMP: "<<TMP<<endl;

    // Cov = TMP
    memcpy(Cov.GetData(), TMP.GetData(), 
           NumParams * NumParams * sizeof(double) );

    //cout << "3: Cov: "<<Cov<<endl;

    // construction of permutation P
    P.SetZero();
    for (row = 0; row < NumParams; row++) {
      P[Permutation[row]-1][row] = 1;
    }

    //cout << "3: P: "<<P<<endl;

    // compute (R^-1 R^-T) * P^T 
    for (row = 0; row < NumParams; row++)
      for (column = 0; column < NumParams; column++) {
        TMP[row][column] = 0;
        for (k = 0; k < NumParams; k++) {
          TMP[row][column] += Cov[row][k] * P[column][k];
        }
      }

    //cout << "3: TMP: "<<TMP<<endl;

    double sigma0squared;
    if( NumErrors > NumParams ) {
      sigma0squared = SumOfSquaredErrors / (double)(NumErrors - NumParams);
    } else {
      BIASERR("Covariance approximation not valid, NumErrors=NumParams");
      sigma0squared = SumOfSquaredErrors;
      BIASABORT;
      return -1;
    }
    //cout <<"Sigma0squared is "<<sigma0squared<<endl;


    // Cov = P *  (R^-1 R^-T * P^T)
    P.Mult(TMP, Cov);
    //cout << "4: Cov: "<<Cov<<endl;
    Cov.MultiplyIP(sigma0squared);
    //cout << "5: Cov: "<<Cov<<endl;
    return 0;
  }
  
  void CheckJacobian(void (*fcn)(long int *m, long int *n, double *x, double *fvec, 
                                 double *fjac, long int *ldfjac, long int *iflag), 
                     const long int NumErrors, const Vector<double> &x, 
                     Vector<double> &err)
  {
    long int m = NumErrors;
    long int n = (long int)x.size();
    long int ldfjac = n;
    Vector<double> mx = x, xp(n);
    Matrix<double> fjac(NumErrors, n);
    Vector<double> fvec(NumErrors), fvecp(NumErrors);
    err.newsize(NumErrors);
    
    long int mode = 1;
    long int iflag = 1;
    // compute xp
    chkder_(&m, &n, mx.GetData(), fvec.GetData(), fjac.GetData(), &ldfjac, 
            xp.GetData(), fvecp.GetData(), &mode, err.GetData());

    // evaluate the function at x and put into fvec
    fcn(&m, &n, mx.GetData(), fvec.GetData(), fjac.GetData(), &ldfjac, &iflag);

    // evaluate the function at xp and put into fvecp
    fcn(&m, &n, xp.GetData(), fvecp.GetData(), fjac.GetData(), &ldfjac,&iflag);

    // compute the jacobian at x and put into fjac
    iflag = 2;
    fcn(&m, &n, mx.GetData(), fvec.GetData(), fjac.GetData(), &ldfjac, &iflag);

    // now compare jacobian and numerical computed jacobian
    mode = 2;
    chkder_(&m, &n, mx.GetData(), fvec.GetData(), fjac.GetData(), &ldfjac, 
            xp.GetData(), fvecp.GetData(), &mode, err.GetData());

  }

  void ComputeJacobian(void (*fcn)(long int *m, long int *n, double *x, double *fvec, 
                                   long int *iflag), const int NumErrors, 
                       const Vector<double> &x, Matrix<double>& fjac, 
                       const double gradientEpsilon)
  {
    long int m = NumErrors;
    long int n = (long int)x.size();
    // the leading dimension in bias matrix data is the number of columns
    long int ldfjac = m;
    Vector<double> mx = x;
    Matrix<double> mfjac(n, m);
    Vector<double> fvec(NumErrors);
    long int iflag = 0;
    double eps = gradientEpsilon;
    double *wa = new double[m];
    BIASASSERT(wa);
    // fill fvec
    fcn(&m, &n, mx.GetData(), fvec.GetData(), &iflag);
    
    fdjac2_(fcn, &m, &n, mx.GetData(), fvec.GetData(), mfjac.GetData(),
            &ldfjac, &iflag, &eps, wa);
    fjac = mfjac.Transpose();
    delete[] wa;
  }


} // namespace
