/*
 * NurbsSurface.cpp
 *
 *  Created on: Nov 18, 2009
 *      Author: bruenger
 */

#include "NurbsSurface.hh"
#include <iostream>
#ifdef BIAS_HAVE_XML2
#include <Base/Common/XMLIO.hh>
#endif
#include <Base/Math/Random.hh>

using namespace std;
using namespace BIAS;

NurbsSurface::NurbsSurface() {
	initialized_ = false;
}

NurbsSurface::~NurbsSurface() {
}


void NurbsSurface::Init(int dimu, int dimv, int order)
{
	dimu_ = dimu;
	dimv_ = dimv;
	order_ = order;

	if(dimu_ < order_ || dimv_ < order_){
		cout << "NurbsSurface::Init - The Dimensions have to be greater then the order" << endl;
		BIASABORT;
	}

	knotsu_.resize(dimu_ + order_ + 1);
	knotsv_.resize(dimv_ + order_ + 1);

	int i;

	ctrlpoints_.resize(dimu_);
	for (i = 0; i < dimu_; i++){
		ctrlpoints_[i].resize(dimv_);
	}

	initialized_ = true;

}

void NurbsSurface::SetKnotsToDefault()
{
	if(!initialized_){
		cout << "NurbsSurface::SetKnotsToDefault - Please initialize the Surface first" << endl;
		BIASABORT;
	}
	float tmp;
	for(int i = 0; i < (dimu_ + order_ + 1); i++){
		if(i < order_ + 1){
			tmp = (float)0.0;
		}else if(i > (dimu_ - 1)){
			tmp = (float)1.0;
		}else{
			tmp = (float)(1.0/(dimu_ - order_))*(float)(i-order_);
		}
		SetKnotu(i,tmp);
	}

	for(int i = 0; i < (dimv_ + order_ + 1); i++){
		if(i < order_ + 1){
			tmp = (float)0.0;
		}else if(i > (dimv_ - 1)){
			tmp = (float)1.0;
		}else{
			tmp = (float)(1.0/(dimv_ - order_))*(float)(i-order_);
		}
		SetKnotv(i,tmp);
	}
}

void NurbsSurface::SetCTRLPointsToRandom()
{
	if(!initialized_){
		cout << "NurbsSurface::SetCTRLPointsToRandom - Please initialize the Surface first" << endl;
		BIASABORT;
	}

	Random r;
	r.Reset();
	SetKnotsToDefault();

	for(int x = 0; x < dimu_; x++){
		for(int y = 0; y < dimv_; y++){
			Vector3<float> tmp;
			tmp.Set((float)x*5,(float)y*5,(float)r.GetUniformDistributed(0.0, 10.0));
			SetCtrlPoint(x,y,tmp);
		}
	}

}


void NurbsSurface::SetCtrlPoint(int u, int v, Vector3<float> &pos)
{
	if(!initialized_){
		cout << "NurbsSurface::SetCtrlPoint - Please initialize the Surface first" << endl;
		BIASABORT;
	}

	ctrlpoints_[u][v] = pos;
}


Vector3<float> NurbsSurface::GetCtrlPoint(int u, int v)
{
	if(!initialized_){
		cout << "NurbsSurface::GetCtrlPoint - Please initialize the Surface first" << endl;
		BIASABORT;
	}

	return ctrlpoints_[u][v];
}

void NurbsSurface::SetKnotu(int i, float &val)
{
	if(!initialized_){
		cout << "NurbsSurface::SetKnot - Please initialize the Surface first" << endl;
		BIASABORT;
	}

	knotsu_[i] = val;
}


float NurbsSurface::GetKnotu(int i)
{
	if(!initialized_){
		cout << "NurbsSurface::GetKnot - Please initialize the Surface first" << endl;
		BIASABORT;
	}

	return knotsu_[i];
}

void NurbsSurface::SetKnotv(int i, float &val)
{
	if(!initialized_){
		cout << "NurbsSurface::SetKnot - Please initialize the Surface first" << endl;
		BIASABORT;
	}

	knotsv_[i] = val;
}


float NurbsSurface::GetKnotv(int i)
{
	if(!initialized_){
		cout << "NurbsSurface::GetKnot - Please initialize the Surface first" << endl;
		BIASABORT;
	}

	return knotsv_[i];
}

#ifdef BIAS_HAVE_XML2
void NurbsSurface::LoadFromXML(std::string xmlfile)
{
	if(initialized_){
		cout << "NurbsSurface::LoadFromXML - Your initialization will be overwritten by the XML" << endl;
	}

	if (xmlfile == "") {
		cout << "NurbsSurface::LoadFromXML - No XML-File given" << endl;
		BIASABORT;
	}

	XMLIO myXML;
	xmlNodePtr rootNode;
	xmlNodePtr currentNode;
	xmlNodePtr ctrlpointsNode;

	rootNode = myXML.read(xmlfile);

	if (rootNode == NULL) {
		cout << "NurbsSurface::LoadFromXML - Error while parsing XML-File" << endl;
		BIASABORT;
	}

	currentNode = myXML.getChild(rootNode, "dimu");
	dimu_ = myXML.getNodeContentInt(currentNode);
	currentNode = myXML.getChild(rootNode, "dimv");
	dimv_ = myXML.getNodeContentInt(currentNode);
	currentNode = myXML.getChild(rootNode, "order");
	order_ = myXML.getNodeContentInt(currentNode);


	if(dimu_ <= order_ || dimv_ <= order_){
		cout << "NurbsSurface::LoadFromXML - The Dimensions have to be greater than the order" << endl;
		BIASABORT;
	}

	initialized_ = true;

	int i,j;

	knotsu_.resize(dimu_ + order_ + 1);
	knotsv_.resize(dimv_ + order_ + 1);

	ctrlpointsNode = myXML.getChild(rootNode, "knotsu");
	xmlNodePtr currentCtrlPoint = myXML.getFirstChild(ctrlpointsNode);

	float k;

	for (i = 0; i < dimu_ + order_ + 1; i++){
		k = (float)myXML.getAttributeValueDouble(currentCtrlPoint,"value");
		SetKnotu(i, k);
		currentCtrlPoint = myXML.getNextChild();
	}

	ctrlpointsNode = myXML.getChild(rootNode, "knotsv");
	currentCtrlPoint = myXML.getFirstChild(ctrlpointsNode);

	for (i = 0; i < dimv_ + order_ + 1; i++){
		k = (float)myXML.getAttributeValueDouble(currentCtrlPoint,"value");
		SetKnotv(i, k);
		currentCtrlPoint = myXML.getNextChild();
	}

	ctrlpoints_.resize(dimu_);
	for (i = 0; i < dimu_; i++){
		ctrlpoints_[i].resize(dimv_);
	}


	ctrlpointsNode = myXML.getChild(rootNode, "ctrlpoints");
	currentCtrlPoint = myXML.getFirstChild(ctrlpointsNode);

	Vector3<float> vec;

	for (i = 0; i < dimu_; i++){
		for (j = 0; j < dimv_; j++){
			vec.Set((float)myXML.getAttributeValueDouble(currentCtrlPoint,"x"),(float)myXML.getAttributeValueDouble(currentCtrlPoint,"y"),(float)myXML.getAttributeValueDouble(currentCtrlPoint,"z"));
			SetCtrlPoint(i,j,vec);
			currentCtrlPoint = myXML.getNextChild();
		}
	}

}

#endif

int NurbsSurface::FindSpan(int n, int p, float u, std::vector<float> U){
	if(u == U[n+1]) return n;
	int low = p;
	int high = n+1;
	int mid = (low+high)/2;
	while(u < U[mid] || u >= U[mid+1]){
		if(u < U[mid])
			high = mid;
		else
			low = mid;
		mid= (low+high)/2;
	}
	return mid;
}

std::vector<float> NurbsSurface::BasisFuns(int i, float u, int p, std::vector<float> U){
	std::vector<float> N, left, right;
	N.resize(p+1);
	left.resize(p+1);
	right.resize(p+1);
	float temp, saved;
	N[0] = 1.0f;
	for(int j = 1; j <= p; j++){
		left[j] = u-U[i+1-j];
		right[j] = U[i+j]-u;
		saved = 0.0f;
		for(int r = 0; r < j; r++){
			temp = N[r]/(right[r+1]+left[j-r]);
			N[r] = saved+right[r+1]*temp;
			saved = left[j-r]*temp;
		}
		N[j] = saved;
	}
	return N;
}

std::vector<std::vector<float> >  NurbsSurface::DersBasisFuns(int i, float u, int p, int n, std::vector<float> U){
	std::vector<std::vector<float> > ndu, ders, a;
	std::vector<float> left, right;
	ndu.resize(p+1);
	ders.resize(p+1);
	a.resize(p+1);
	for (int x = 0; x < p+1; x++){
		ndu[x].resize(p+1);
		ders[x].resize(p+1);
		a[x].resize(p+1);
	}
	left.resize(p+1);
	right.resize(p+1);
	float temp, saved, d;
	int s1,s2,rk,pk,j1,j2,j,k,r;//,l;
	ndu[0][0] = 1.0f;
	for(j = 1; j <= p; j++){
		left[j] = u-U[i+1-j];
		right[j] = U[i+j]-u;
		saved = 0.0f;
		for(int r = 0; r < j; r++){
			ndu[j][r] = right[r+1] + left[j-r];
			temp = ndu[r][j-1] / ndu[j][r];

			ndu[r][j] = saved+right[r+1]*temp;
			saved = left[j-r]*temp;
		}
		ndu[j][j] = saved;
	}

	for(j = 0; j <= p; j++)
		ders[0][j] = ndu[j][p];

	for(r = 0; r <= p; r++){
		s1 = 0;
		s2 = 1;
		a[0][0] = 1.0f;
		for(k = 1; k <= n; k++){
			d = 0.0f;
			rk = r-k;
			pk = p-k;
			if(r >= k){
				a[s2][0] = a[s1][0] / ndu[pk+1][rk];
				d = a[s2][0] * ndu[rk][pk];
			}

			if(rk >= -1){
				j1 = 1;
			}else{
				j1 = -rk;
			}

			if(r-1 <= pk){
				j2 = k-1;
			}else{
				j2 = p-r;
			}

			for(j = j1; j <= j2; j++){
				a[s2][j] = (a[s1][j] - a[s1][j-1]) / ndu[pk+1][rk+j];
				d += a[s2][j] * ndu[rk+j][pk];
			}

			if(r <= pk){
				a[s2][k] = -a[s1][k-1] / ndu[pk+1][r];
				d += a[s2][k] * ndu[r][pk];
			}
			ders[k][r] = d;
			j = s1;
			s1 = s2;
			s2 = j;
		}
	}
	r = p;
	for(k = 1; k <= n; k++){
		for(j = 0; j <= p; j++){
			ders[k][j] *= r;
		}
		r *= (p-k);
	}
	return ders;
}

Vector3<float> NurbsSurface::GetSurfacePoint(float u, float v){
	int uspan = FindSpan(dimu_,order_,u,knotsu_);
	std::vector<float> Nu = BasisFuns(uspan,u,order_,knotsu_);
	int vspan = FindSpan(dimv_,order_,v,knotsv_);
	std::vector<float> Nv = BasisFuns(vspan,v,order_,knotsv_);
	int uind = uspan - order_;
	Vector3<float> temp, S;
	S.SetZero();
	for(int l = 0; l <= order_; l++){
		temp.SetZero();
		int vind = vspan-order_+l;
		for(int k = 0; k <= order_; k++){
			temp[0] += Nu[k] * GetCtrlPoint(uind+k,vind).GetData()[0];
			temp[1] += Nu[k] * GetCtrlPoint(uind+k,vind).GetData()[1];
			temp[2] += Nu[k] * GetCtrlPoint(uind+k,vind).GetData()[2];
		}
		S.Set(S.GetData()[0] + Nv[l]*temp[0],S.GetData()[1] + Nv[l]*temp[1],S.GetData()[2] + Nv[l]*temp[2]);
	}
	return S;
}

std::vector<std::vector<Vector3<float> > > NurbsSurface::GetSurfaceDerivs(float u, float v, int d){
	std::vector<std::vector<Vector3<float> > > SKL;
	SKL.resize(d+1);
	for (int i = 0; i <= d; i++){
		SKL[i].resize(d+1);
	}
	int du,dv,dd,k,l,s,r;
	if(d < order_){
		du = d;
	}else{
		du = order_;
	}
	for(k = order_+1; k <= d; k++)
		for(l = 0; l <= d-k; l++)
			SKL[k][l].SetZero();
	if(d < order_){
		dv = d;
	}else{
		dv = order_;
	}
	for(l = order_+1; l <= d; l++)
		for(k = 0; k <= d-l; k++)
			SKL[k][l].SetZero();
	int uspan = FindSpan(dimu_,order_,u,knotsu_);
	std::vector<std::vector<float> > Nu = DersBasisFuns(uspan,u,order_,du,knotsu_);
	int vspan = FindSpan(dimv_,order_,v,knotsv_);
	std::vector<std::vector<float> > Nv = DersBasisFuns(vspan,v,order_,dv,knotsv_);
	std::vector<Vector3<float> > temp;
	temp.resize(order_+1);
	for(k = 0; k <= du; k++){
		for(s = 0; s <= order_; s++){
			temp[s].SetZero();
			for(r = 0; r <= order_; r++){
				temp[s][0] += Nu[k][r] * GetCtrlPoint(uspan-order_+r,vspan-order_+s).GetData()[0];
				temp[s][1] += Nu[k][r] * GetCtrlPoint(uspan-order_+r,vspan-order_+s).GetData()[1];
				temp[s][2] += Nu[k][r] * GetCtrlPoint(uspan-order_+r,vspan-order_+s).GetData()[2];
			}
		}
		if((d-k) < dv){
			dd = (d-k);
		}else{
			dd = dv;
		}
		for(l = 0; l <= dd; l++){
			SKL[k][l].SetZero();
			for(s = 0; s <= order_; s++){
				SKL[k][l][0] += Nv[l][s] * temp[s][0];
				SKL[k][l][1] += Nv[l][s] * temp[s][1];
				SKL[k][l][2] += Nv[l][s] * temp[s][2];
			}
		}
	}
	return SKL;
}
