/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
   @file TestLaguerre.cpp
   @brief Test Laguerre
   @relates Laguerre
   @ingroup g_tests
   @author MIP
*/


#include "Base/Debug/Error.hh"
#include <MathAlgo/Laguerre.hh>
#include <Base/Common/CompareFloatingPoint.hh>
#include <iostream>
#include <stdlib.h>
using namespace BIAS;
using namespace std;


double frand(double min = -10.0 , double max = 10.0)
{
  double tmin = min;
  double range = double(max) - tmin;
   double rnd  = double(rand()) / double(RAND_MAX);
 
  return(rnd*range + tmin);
}

//bitmask looks like 00001111 
unsigned int GetNumberOfBinaryOnes(unsigned int val,
                                   unsigned int mask)
{
  unsigned int bitmask = mask;
  unsigned int temp    = val;
  unsigned int count   = 0;
  while((bitmask)!=0)  //dont use the full size of unsigned int
  {
    if((temp&1)==1)
    {
      count++;
    }
    temp    = temp>>1;
    bitmask = bitmask>>1;
  }
  return count;
}

void GetIndexListFromBinaryZeros(unsigned int val,
                                 unsigned int mask,
                                 std::vector<unsigned int>& indices)
{
  indices.clear();
  unsigned int bitmask = mask;
  unsigned int temp    = val;
  unsigned int count   = 0;
  while((bitmask)!=0)  //dont use the full size of unsigned int
  {
    if((temp&1)==0)
    {
      indices.push_back(count);
    }
    temp    = temp>>1;
    bitmask = bitmask>>1;
    count++;
  }
}

// create a random
void CreateRandomPolynomial(unsigned int degree,
                            vector< complex<double> >& coeffs,
                            vector< complex<double> >& roots)
{
  BIASASSERT(degree>0);
  roots.clear();
  coeffs.clear();
  std::vector<unsigned int> tempIDs;
  unsigned int coeffidx ;
  //create random roots ... (x-z0)(x-z1)(x-z2)...(x-zdegree)
  for(unsigned int r=0; r<degree; r++)
  {
    roots.push_back(frand());
  }
   coeffs.resize(degree+1,complex<double>(0,0));  //2
//   //create coeffs for random roots
  unsigned int maxiter = 1<<(degree); 
  unsigned int bitmask = maxiter - 1;   // 00111
  for(unsigned int c=0; c<maxiter; c++)
  {
    GetIndexListFromBinaryZeros(c,bitmask,tempIDs);
     coeffidx = GetNumberOfBinaryOnes(c,bitmask);
     complex<double> someroots(1,0);
    for(unsigned int r=0;r<tempIDs.size();r++)
    {
      someroots = someroots * (-roots[tempIDs[r]]);
    }
 //    cout<<c<<"  "<<"coeffidx: "<<coeffidx<<endl;
     BIASASSERT(coeffidx<coeffs.size());
     coeffs[coeffidx]= coeffs[coeffidx] + someroots;
  }

//   for(unsigned int c=0;c<coeffs.size(); c++)
//   {
//     cout<<coeffs[c]<<" ";
//   }
  
}

void CreateRandomPolynomialRealSq(vector< complex<double> >& coeffs,
                                  vector< complex<double> >& roots)
{
  coeffs.clear();
  roots.clear();
  for(unsigned r=0; r<2; r++)
  {
    roots.push_back(complex<double>(frand(),0));
  }
  //compute coeffs for given roots
  coeffs.push_back(complex<double>(roots[0].real()*roots[1].real(),0));
  coeffs.push_back(complex<double>(-(roots[1].real()+roots[0].real()),0));
  coeffs.push_back(complex<double>(1,0));
  
  //cout<<coeffs[0]<<" + "<<coeffs[1]<<"*x +"<<coeffs[2]<<"*x^2 "<<endl;
}

void CreateRandomPolynomialReal(unsigned int degree,
                                vector< complex<double> >& coeffs)
{
  BIASASSERT(degree>0);
  coeffs.clear();
  //create random coeffcicients
  for(unsigned r=0; r<degree+1; r++)
  {
    coeffs.push_back(complex<double>(frand(),0));
  }
}

bool CheckRoots(vector< complex<double> >& coeffs,
                vector< complex<double> >& roots)
{
  for(unsigned int roo=0;roo<roots.size(); roo++)
  {
    complex<double> polyval(0,0);
    for(unsigned int i=0;i<coeffs.size(); i++)
    {
      polyval = polyval + coeffs[i] * pow(roots[roo],i);
    }
   
    if((!Equal(polyval.real(),0.0,10e-5))||(!Equal(polyval.imag(),0.0,10e-5)))
    {
      cout<<"degree: "<<coeffs.size()-1<<endl; cout<<"p("<<roots[roo]<<")="<<polyval<<endl;
      return false;
    }
  }
  return true;
}

bool Equal(const complex<double>& a, 
           const complex<double>& b)
{
  return((Equal(a.real(),b.real(),10e-6))&&
        ((Equal(a.imag(),b.imag(),10e-6))));
}

bool Compare(vector< complex<double> >& v1,
             vector< complex<double> >& v2)
{
  vector< complex<double> > v2copy=v2;
  if(v1.size()!=v2copy.size())
  { 
    BIASERR("size mismatch!");
  //  return false;
  }
  
  for(unsigned int i=0; i<v1.size(); i++)
  {
    //    cout<<"comparing: "<<v1[i]<<" and "<<v2[i]<<endl;
    //order of vectors may differ
    vector< complex<double> >::iterator it;
    for(it=v2copy.begin(); it!=v2copy.end(); it++)
    {
      if(Equal(v1[i],*it))
      {
        v2copy.erase(it);
        break;
      }
    }
  }
  if(v2copy.size()!=0)
  {
    cout<<"----------------------------------------------------"<<endl;
    for(unsigned int i=0; i<v1.size(); i++)
    {
      cout<<v1[i]<<" , ";
    }
    cout<<"rest:"<<endl;
    for(unsigned int i=0; i<v2copy.size(); i++)
    {
      cout<<v2copy[i]<<" , ";
    }
    return false;
  }
  return true;
}


int main(int argc, char *argv[])
{
  srand(49152);
  LaguerreSolver l;
  vector< complex<double> > realcoeff;
  vector< complex<double> > roots;
  vector< complex<double> > expected_roots;

  
  for(unsigned int run=0; run<500; run++)
  {
    cout<<"#"<<run<<" - degree: ";
    int deg = rand()%8+2;
    cout<<deg<<endl;
    CreateRandomPolynomial(deg,realcoeff,expected_roots);
	
 //   BIASASSERT(CheckRoots(realcoeff,expected_roots));
    l.Solve(realcoeff, roots);
    if(!CheckRoots(realcoeff,roots))
    {
      BIASERR("this roots don't seem to be roots ");
      return -1;
    }
    for (unsigned int ii=0;ii<roots.size();ii++)
      cout <<roots[ii]<<"   ";
    cout <<endl;
    if(!Compare(roots,expected_roots))
    {
      BIASERR("test tailed");
      return -2;
    }
  }
  
//   //test if expected roots are returned
//   for(unsigned run=0; run<100; run++)
//   {
//     cout<<run<<", ";
//     CreateRandomPolynomialRealSq(realcoeff,expected_roots);
//     BIASASSERT(CheckRoots(realcoeff,expected_roots));
//     l.Solve(realcoeff, roots);
//     BIASASSERT(CheckRoots(realcoeff,roots));
//     if(!Compare(roots,expected_roots))
//     {
//       BIASERR("test tailed");
//       return -2;
//     }
//   }
//   
//   //test with random coeffss
//   for(unsigned run=0; run<100; run++)
//   {
//     unsigned int deg = 4 ;
//     CreateRandomPolynomialReal(deg,realcoeff);
//     l.Solve(realcoeff, roots);
//     if(!CheckRoots(realcoeff,roots))
//     {
//       BIASERR("random test "<<run<<" failed.");
//       return -1;
//     }
//   }
//   
cout<<"-----------------------------------------------------------------"<<endl;
cout<<"this was a triumph!"<<endl;
  
  return 0;
}

