/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __glfException_hh__
#define __glfException_hh__

#include "glfCommon.hh"
#include <string>
#include <sstream>

namespace BIAS {

  /**
   * Utility macro for throwing a glfException.
   * Use like BIASERR:
   * <code>GLF_THROW_EXCEPTION("xyz" << i << "123");</code>
   */
  #define GLF_THROW_EXCEPTION(MESSAGE) \
    { \
      std::stringstream msg; \
      msg << MESSAGE; \
      throw glfException(__FILE__, __LINE__, msg.str()); \
    }

  /**
   * Throws a glfException if an OpenGL error occured.
   */
  #define GLF_THROW_ON_OPENGL_ERROR \
    { \
      GLenum err = glGetError(); \
      if (err != GL_NO_ERROR) { \
        GLF_THROW_EXCEPTION("OpenGL Error: " << gluErrorString(err)); \
      } \
    }

#ifdef BIAS_DEBUG
#define GLF_THROW_ON_OPENGL_ERROR_DEBUG GLF_THROW_ON_OPENGL_ERROR
#else
#define GLF_THROW_ON_OPENGL_ERROR_DEBUG 
#endif

  /**
   * Throws an glfException if an OpenGL error occured.
   */
  #define GLF_THROW_ON_OPENGL_ERROR_INFO(MESSAGE) \
    { \
      GLenum err = glGetError(); \
      if (err != GL_NO_ERROR) { \
        GLF_THROW_EXCEPTION("["<<MESSAGE<<"] OpenGL Error: " << gluErrorString(err)); \
      } \
    }

  /**
   * \brief Exception class used for run-time errors in the OpenGLFramework.
   * \ingroup g_openglframework	
   * \author jkollmann
   */
  class BIASOpenGLFramework_EXPORT glfException
  {
  public:

    /**
     * Exception constructor.
     * \param fileName File name of source code file where the error occured.
     * \param lineNumber Line number in source code file where the error occured.
     * \param message Description of the error.
     */
    glfException(const std::string& fileName,
                 int lineNumber,
                 const std::string& message);

    /**
     * Returns the source code file name where the error occured.
     */
    const std::string& GetFileName() const;

    /**
     * Returns the line number in the source code file where the error occured.
     */
    int GetLineNumber() const;

    /**
     * Returns the description of the error including the file name and line number
     * where the error occured.
     */
    const std::string& GetMessageString() const;

  private:
    std::string fileName_;
    int lineNumber_;
    std::string message_;
  };
  
} // namespace BIAS

#endif // __glfException_hh__
