/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __glfFramebufferObject_hh__
#define __glfFramebufferObject_hh__

#include "glfTexture.hh"
#include "glfRenderbuffer.hh"
#include "glfRenderTarget.hh"
#include "glfException.hh"
#include <vector>

namespace BIAS {

  /**
   * \brief Framebuffer object.
   * \ingroup g_openglframework
   * \author jkollmann
   */
  class BIASOpenGLFramework_EXPORT glfFramebufferObject : public glfRenderTarget {
  public:
    glfFramebufferObject();
    ~glfFramebufferObject();

    /**
     * Creates the framebuffer object. Before it can be used, you must
     * attach one or several textures and check glfFramebufferObject::IsComplete.
     */
    void Create();

    /**
     * Attaches a texture to an attachment point which can be
     *  - GL_COLOR_ATTACHMENT0_EXT .. GL_COLOR_ATTACHMENT<MAX_COLOR_ATTACH>_EXT
     *  - GL_DEPTH_ATTACHMENT_EXT
     *  - GL_STENCIL_ATTACHMENT_EXT
     *
     * If the texture is a cube map, then cubeMapSide must be one of the following
     * values:
     * GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
     * GL_TEXTURE_CUBE_MAP_POSITIVE_Y, GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
     * GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
     * 
     */
    void AttachTexture(const glfTexture& texture,
                       GLenum attachmentPoint,
                       int attachedMipMapLevel = 0,
                       int zSlice = 0,
                       GLenum cubeMapSide = 0);

    /**
     * Attaches a renderbuffer to the framebuffer object.
     * See glfFramebufferObject::AttachTexture for valid attachment points.
     */
    void AttachRenderbuffer(const glfRenderbuffer& renderbuffer,
                            GLenum attachmentPoint);

    /**
     * Removes an attachment from the framebuffer object.
     */
    void ReleaseAttachment(GLenum attachmentPoint);

    void SetDrawBuffers(const std::vector<GLenum>& drawbuffers);
    void ResetDrawBuffers();

    /**
     * Checks whether the framebuffer object is supported in its current state.
     * If not, an exception is thrown. This method should be called before the
     * framebuffer object is used as a render target.
     */
    void CheckComplete() const;

    // glfRenderTarget interface
    virtual void Bind() const;

    static GLint MaxRenderTargets();

    /** 
     * Blits the attachments from source to target, requires EXT_framebuffer_blit extension.
     * Copied content is only the color attachment.
     * \param source is the source of the blit operation, i.e. the content copied.
     * \param target is the target of the blit operation.
     * \param colorFilter is either GL_LINEAR or GL_NEAREST.
     **/
    static void BlitColorBuffer(glfFramebufferObject* source, glfFramebufferObject* target, 
                                    GLenum colorFilter);

    /** 
     * Blits the attachments from source to target, requires EXT_framebuffer_blit extension.
     * \param source is the source of the blit operation, i.e. the content copied.
     * \param target is the target of the blit operation.
     * \param bufferMask is the mask identifying the buffers that are copied, 
     *    and can consists of GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT and GL_STENCIL_BUFFER_BIT.
     * \attention the filter operation applied is GL_NEAREST.
     **/
    static void Blit(glfFramebufferObject* source, 
                     glfFramebufferObject* target, 
                     GLbitfield bufferMask);

    void GetSize(GLint& width, GLint& height);


  private:
    /** 
     * Blits the attachments from source to target, requires EXT_framebuffer_blit extension.
     * \param source is the source of the blit operation, i.e. the content copied.
     * \param target is the target of the blit operation.
     * \param bufferMask is the mask identifying the buffers that are copied, 
     *    and can consists of GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT and GL_STENCIL_BUFFER_BIT.
     * \param colorFilter is either GL_LINEAR or GL_NEAREST, GL_LINEAR is only possible with GL_COLOR_BUFFER_BIT.
     **/
    static void Blit_(glfFramebufferObject* source, 
                      glfFramebufferObject* target, 
                      GLbitfield bufferMask,
                      GLenum colorFilter);
    GLuint id_;
    /** Defines the number of set draw buffers for multiple render targets**/
    GLsizei numSimultaneousDrawBuffers_;
    GLenum* listOfSimultaneousDrawBuffers_;
    bool FramebufferCreated_;

    /** An intuitive assignment operator should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this operator cannot be used.
     */
    glfFramebufferObject& operator=(const glfFramebufferObject& fbb);
    
    /** An intuitive copy constructor should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this constructor cannot be used.
     */
    glfFramebufferObject(const glfFramebufferObject& fbo);    


  };
  
} // namespace BIAS

#endif // __glfFramebufferObject_hh__
