/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <Base/Image/ImageBase.hh>
#include "glfRenderbuffer.hh"
#include "glfException.hh"
#include "glfFormatDB.hh"


using namespace BIAS;

glfRenderbuffer::glfRenderbuffer()
{
  id_ = 0;
}

glfRenderbuffer::~glfRenderbuffer()
{
  if (id_ != 0) {
    glDeleteRenderbuffersEXT(1, &id_);
  }
}

void glfRenderbuffer::Create(GLenum internalFormat, int width, int height, GLsizei samples)
{
  if(!GLEW_EXT_framebuffer_object) {
    GLF_THROW_EXCEPTION("GLEW_EXT_framebuffer_object:objects of this type can only be used"
                        " with EXT_framebuffer_object OpenGL extension");
  }

  if(id_ == 0) {
    glGenRenderbuffersEXT(1, &id_);
  }

  Bind();
  if(samples > 0 && !GLEW_EXT_framebuffer_multisample) {
    BIASWARN("Multisampling is not supported.");
  }
  if(samples==0 || !GLEW_EXT_framebuffer_multisample) {
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
                             internalFormat,
                             width, height);
  } else {
    GLsizei maxSamples;
    glGetIntegerv(GL_MAX_SAMPLES_EXT, &maxSamples);
    if(maxSamples<samples) {
      samples = maxSamples;
      BIASWARN("requested number samples are not support, reducing to "<<samples<<" samples");
    }
    glRenderbufferStorageMultisampleEXT(GL_RENDERBUFFER_EXT,
                                        samples,
                                        internalFormat,
                                        width, height);
  }

  GLF_THROW_ON_OPENGL_ERROR;
}

void glfRenderbuffer::Bind() const
{
  BIASASSERT(id_ != 0);
  glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, id_);
  GLF_THROW_ON_OPENGL_ERROR;
}

GLuint glfRenderbuffer::GetRenderbufferID() const
{
  return id_;
}

void glfRenderbuffer::
GetSize(GLint& width, GLint& height) {
  Bind();

  glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_WIDTH_EXT, &width);
  glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_HEIGHT_EXT, &height);

  }

void glfRenderbuffer::
CopyToImage(ImageBase& image) {
  Bind();

  GLint numSamples;
  glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, 
                                  GL_RENDERBUFFER_SAMPLES_EXT, 
                                  &numSamples);
  if(numSamples>0) {
    GLF_THROW_EXCEPTION("num samples must be zero to read out buffer.");    
  }

  GLint internalFormat;
  glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, 
                                  GL_RENDERBUFFER_INTERNAL_FORMAT_EXT, 
                                  &internalFormat);
  int width, height;
  GetSize(width, height);

  GLenum format;
  int numChannels;
  glfFormatDB::GetFormatForInternalFormat(internalFormat, format, numChannels);
  ImageBase::EStorageType storageType = image.GetStorageType();
  if (long(image.GetWidth()) != width && long(image.GetHeight()) != height &&
      long(image.GetChannelCount()) != numChannels)
    {
      image.Init(width, height, numChannels, storageType);
    }

  // get OpenGL pixel type
  GLenum type = glfFormatDB::GetPixelTypeForStorageType(image.GetStorageType());

  glPixelStorei(GL_PACK_ALIGNMENT, 1);

  glReadPixels( 0, 0, width, height, format, type, image.GetImageData());
  GLF_THROW_ON_OPENGL_ERROR;
}
