/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __glfShader_hh__
#define __glfShader_hh__

#include "glfCommon.hh"
#include <string>

namespace BIAS {

  /**
   * \brief A GLSL vertex shader or fragment shader, which must be linked
   * in a shader program.
   * \ingroup g_openglframework
   * \author jkollmann
   */
  class BIASOpenGLFramework_EXPORT glfShader {
  public:
    glfShader();
    ~glfShader();

    /**
     * Creates the shader without the GLSL source code, use other
     * Create() methods to upload source afterwards.
     * \param type e.g. GL_VERTEX_SHADER or GL_FRAGMENT_SHADER.
     * \param sourceCode The GLSL source.
     */
    void Create(GLenum type);

    /**
     * Creates the shader from the given GLSL source code.
     * \param type e.g. GL_VERTEX_SHADER or GL_FRAGMENT_SHADER.
     * \param sourceCode The GLSL source.
     */
    void Create(GLenum type, const std::string& sourceCode);

    /**
     * Creates the shader from GLSL source code loaded from a file.
     * \param type Either GL_VERTEX_SHADER or GL_FRAGMENT_SHADER.
     * \param fileName Name of file to load the GLSL source from.
     */
    void CreateFromFile(GLenum type, const std::string& fileName);


    /**
     * Returns the info log containing information about the compilation
     * of the shader.
     * \note If compiling the shader fails, the info log is also contained in
     * the message of the thrown exception.
     */
    std::string GetInfoLog() const;

    /**
     * Returns the OpenGL id of the shader.
     */
    GLuint GetShaderID() const;

    /**
     * Returns the OpenGL id of the shader.
     */
    inline GLenum GetType() {return type_;};

  private:
    GLuint id_;
    GLenum type_;

    /** An intuitive assignment operator should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this operator cannot be used.
     */
    glfShader& operator=(const glfShader& b);
    
    /** An intuitive copy constructor should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this constructor cannot be used.
     */
    glfShader(const glfShader& b);    
  };
}

#endif // __glfShader_hh__
