/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C)  2008, 2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __glfTexture_hh__
#define __glfTexture_hh__

#include "glfCommon.hh"
#include <Base/Image/ImageBase.hh>
#include <map>

namespace BIAS {

  /**
   * \brief Base class for textures. When creating a texture, call:
   * 1. Create()
   * 2. Set texture parameters.
   * 3. Upload...(...) method of a derived class.
   * \ingroup g_openglframework
   * \author jkollmann
   */
  class BIASOpenGLFramework_EXPORT glfTexture {
  public:
    virtual ~glfTexture();

    /**
     * Creates the texture but does not upload any data yet.
     */
    void Create();

    /**
     * Sets the minifying function.
     * Valid values are: GL_NEAREST, GL_LINEAR, GL_NEAREST_MIPMAP_NEAREST,
     * GL_LINEAR_MIPMAP_NEAREST, GL_NEAREST_MIPMAP_LINEAR,
     * GL_LINEAR_MIPMAP_LINEAR.
     */
    void SetMinFilter(GLenum minFilter);

    /**
     * Sets the magnification function.
     * Valid values are: GL_NEAREST, GL_LINEAR.
     */
    void SetMagFilter(GLenum magFilter);

    /**
     * Sets the wrapping mode for the 1st texture coordinate.
     * Valid values are: GL_CLAMP, GL_CLAMP_TO_BORDER, GL_CLAMP_TO_EDGE,
     * GL_MIRRORED_REPEAT, GL_REPEAT.
     */
    void SetWrapS(GLenum wrapS);

    /**
     * Sets the wrapping mode for the 2nd texture coordinate.
     * \see TextureParameters::SetWrapS.
     */
    void SetWrapT(GLenum wrapT);

    /**
     * Sets the wrapping mode for the 3rd texture coordinate.
     * \see TextureParameters::SetWrapS.
     */
    void SetWrapR(GLenum wrapR);

    /**
     * Sets whether to automatically generate mipmaps when the first mipmap is uploaded.
     */
    void SetGenerateMipMap(bool generateMipMap);

    /**
     * Sets the minimum (base) mip map level to be used.
     */
    void SetBaseLevel(int baseLevel);

    /**
     * Sets the maximum mip map level to be used.
     */
    void SetMaxLevel(int maxLevel);

    /**
     * Sets the border color for the texture to be used.
     */
    void SetBorderColor(GLfloat r, GLfloat g, GLfloat b, GLfloat a);

    /**
     * Sets whether to enlarge each texture dimension to the next power of two,
     * if non-power-of-two textures are not supported. Otherwise, uploading
     * a non-power-of-two texture will cause an exception to be thrown, if the
     * hardware does not support it. The default is 'false'.
     * Must be called before any data is uploaded to the texture.
     */
    void EnlargeToPowerOfTwoIfRequired(bool enlargeToPotIfRequired = true);

    /**
     * Explicitly generates the mipmaps of the texture in hardware.
     * This can be used, when the texture was generated by rendering to a
     * framebuffer object.
     */
    void GenerateMipMap();

    /**
     * Binds the texture.
     * \attention For internal usage inside OpenGLFramework Base library only.
     */
    void Bind() const;

    /**
     * Binds the texture.
     * Set texture active. Use this for multitexturing shaders without the use of a glfbatch.
     * \attention For internal usage inside OpenGLFramework Base library only.
     */
    void BindTU() const;

    /**
     * Binds the texture.
     * Sets the texture active with GL_TEXTURE0+number, before binding
     * \attention For internal usage inside OpenGLFramework Base library only.
     */
    void BindTU(int number) const;

    /**
     * Returns the OpenGL texture id.
     * \attention For internal usage inside OpenGLFramework Base library only.
     */
    GLuint GetTextureID() const { return id_; }

    /**
     * Returns the OpenGL texture target.
     * \attention For internal usage inside OpenGLFramework Base library only.
     */
    GLenum GetTextureTarget() const { return target_; }

    /**
     * Returns the OpenGL texture number.
     * \attention For internal usage inside OpenGLFramework Base library only.
     */
    inline GLint GetTextureNr() { return textureNr_; }

    /**
     * Sets the OpenGL texture number (GL_TEXTURE0 - GL_TEXTURE[MaxTextureUnit-1]).
     * \attention For internal usage inside OpenGLFramework Base library only.
     */
    void SetTextureNr(GLint textureNr);

    /**
     * Sets the OpenGL texture number (0 - MaxTextureUnit-1).
     */
    void SetTextureNrInUniformFormat(int textureNr);

    /**
     * Returns the OpenGL texture number in uniform format or -1 for error.
     * \attention For internal usage inside OpenGLFramework Base library only.
     *
     * Intended usage:
     *
     * renderedTexture.SetTextureNr(GL_TEXTURE5);
     *
     * renderedTexture2.SetTextureNr(GL_TEXTURE6);
     *
     * shaderProg.SetUniform("renderedTexture", renderedTexture.GetTextureNrInUniformFormat());
     *
     * shaderProg.SetUniform("renderedTexture2", renderedTexture2.GetTextureNrInUniformFormat());
     */
    inline int GetTextureNrInUniformFormat() { return textureNr_ - GL_TEXTURE0;}

    /** Swap the GL object IDs.
     * Will throw an exception if one texture has id 0 or if textures have different targets.
     * \author bartczak 03/2009
     **/
    friend BIASOpenGLFramework_EXPORT void SwapGLObjects(glfTexture& A, glfTexture& B);

  protected:
    glfTexture(GLenum target);

    /**
     * \name Utility functions used by derived classes.
     */
    //@{

    /**
     * Returns the smallest power of two that is greater than <code>i</code>.
     */
    static int NextPowerOfTwo(int i);

    /**
     * Computes a pixel format that can be used with <code>glGetTexImage</code>
     * for an internal format of a texture. Also returns the number of channels.
     * For example: If internalFormat = GL_R3_G3_B2, then format = GL_RGB
     * and numChannels = 3.
     * \deprecated use glfFormatDB
     */
//     static void GetFormatForInternalFormat(GLenum internalFormat,
//                                            GLenum& format,
//                                            int& numChannels);

    /**
     * Returns the OpenGL pixel type (GL_UNSIGNED_BYTE etc.) for a BIAS storage type.
     * \deprecated use glfFormatDB
     */
//     static GLenum GetPixelTypeForStorageType(ImageBase::EStorageType storageType);

    

    /** Returns the OpenGL pixel format (GL_RGB etc.) for a BIAS color model.
     * \deprecated use glfFormatDB
     */
//     static GLenum GetFormatForColorModel(ImageBase::EColorModel colorModel);

    
    /* Returns an OpenGL internal pixel format for a BIAS storage type and color model.
     * \deprecated use glfFormatDB
     */
//     static GLenum ProposeInternalFormat(ImageBase::EStorageType storageType,
//                                         ImageBase::EColorModel colorModel);
    
    /**
     * cp from glfBatch
     */
    static int GetMaxSupportedTextureUnits();

    //@}

    GLuint id_;
    GLenum target_;
    GLint textureNr_;
    int MAX_TEXTURE_UNITS_;
    bool enlargeToPot_;

  private:
    /** An intuitive assignment operator should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this operator cannot be used.
     */
    glfTexture& operator=(const glfTexture& b);

    /** An intuitive copy constructor should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this constructor cannot be used.
     */
    glfTexture(const glfTexture& t);
  };

  BIASOpenGLFramework_EXPORT void SwapGLObjects(glfTexture& A, glfTexture& B);

} // namespace BIAS

#endif // __glfTexture_hh__
