/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C)  2008, 2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __glfVertexBuffer_hh__
#define __glfVertexBuffer_hh__

#include "glfVertexFormat.hh"

namespace BIAS {

  /**
   * \brief A vertex buffer contains an array of vertices that can be used for
   * rendering. If possible, the buffer is stored in the video memory.   
   * \ingroup g_openglframework   
   * \author jkollmann
   */
  class BIASOpenGLFramework_EXPORT glfVertexBuffer {
  public:
    glfVertexBuffer();
    ~glfVertexBuffer();

    /**
     * Creates the vertex buffer for the given number of vertices and vertex format.
     * If the given data is non-<code>NULL</code>, it will be uploaded to the
     * vertex buffer. It must contain <code>numVertices</code> vertices of the
     * given <code>format</code>. The size of the given data buffer must be
     * <code>numVertices * format.GetVertexSize()</code> bytes.
     */
    void Create(int numVertices, const glfVertexFormat& format, 
                const void* data = NULL, bool useBufferIfPossible = true);

    /**
     * Secure convenience wrapper for other glfVertexBuffer::Create method.
     * T is the vertex structure in C++ that is described by <code>format</code>.
     */
    template<class T>
    inline void Create(const std::vector<T>& vertices, const glfVertexFormat& format) {
      BIASASSERT(sizeof(T) == format.GetVertexSize());
      Create((int)vertices.size(), format, &vertices[0]);
    }

    /**
     * Maps the vertex buffer to system memory, so it can be modified after creation.
     * \attention Access is write-only!
     */
    void* Map();

    /**
     * Unmaps the vertex buffer. Must be called after glfVertexBuffer::Map and before
     * the vertex buffer is used for rendering again.
     */
    void Unmap();

    /**
     * Returns the number of vertices in the vertex buffer.
     */
    int GetNumVertices() const;

    /**
     * Binds the vertex buffer in OpenGL.
     */
    void Bind() const;

  private:
    glfVertexFormat format_;
    int numVertices_;
    bool useBuffer_;
    GLuint bufferID_;
    bool mapped_;
    // std::vector<GLubyte> array_;
    GLubyte* array_;

    /** An intuitive assignment operator should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this operator cannot be used.
     */
    glfVertexBuffer& operator=(const glfVertexBuffer& b);
    
    /** An intuitive copy constructor should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this constructor cannot be used.
     */
    glfVertexBuffer(const glfVertexBuffer& b);    

  };

} // namespace BIAS

#endif // __glfVertexBuffer_hh__
