/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C)  2008, 2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __glfVertexFormat_hh__
#define __glfVertexFormat_hh__

#include "glfCommon.hh"
#include <vector>
#include <string>

namespace BIAS {

  /**
   * \brief A vertex format describes the attributes of a vertex. A list of
   * vertices with the same format are stored in a vertex buffer.
   * \ingroup g_openglframework   
   * \author jkollmann
   */
  class BIASOpenGLFramework_EXPORT glfVertexFormat {
  public:
    glfVertexFormat();
    ~glfVertexFormat();

    /**
     * Possible attributes of a vertex. Note that ATTRIB_TEXCOORD
     * and ATTRIB_GENERIC require an additional index, i.e. there
     * can be multiple of them.
     */
    enum Attribute
    {
      ATTRIB_POSITION,
      ATTRIB_NORMAL,
      ATTRIB_COLOR,
      ATTRIB_TEXCOORD,
      ATTRIB_GENERIC
    };

    /**
     * Adds a vertex attribute to the format.
     * \param attrib The attribute to be added.
     * \param index Additional index for ATTRIB_TEXCOORD and ATTRIB_GENERIC.
     * \param type Data type for this attribute. Can be GL_BYTE,
     *             GL_UNSIGNED_BYTE, GL_SHORT, GL_UNSIGNED_SHORT, GL_INT,
     *             GL_UNSIGNED_INT, GL_FLOAT, or GL_DOUBLE.
     * \param size Number of components for this attribute, i.e. 3 for a
     *             3-dimensional vector.
     * \param normalized True if values should be normalized. Only for ATTRIB_GENERIC.
     *
     * \note Not all combinations of attribute, type and size are supported by
     * OpenGL or may be unsupported by the current OpenGL implementation. In this
     * case an exception will be thrown.
     */
    void AddAttribute(Attribute attrib, int index, GLenum type, int size, bool normalized = false);

    /**
     * Convenience wrapper for other glfVertexFormat::AddAttribute method if no additional attribute
     * index is required. This can be used for all attributes except ATTRIB_TEXCOORD and ATTRIB_GENERIC.
     */
    void AddAttribute(Attribute attrib, GLenum type, int size) {
      AddAttribute(attrib, 0, type, size);
    }

    /**
     * Returns the size in bytes of a single vertex.
     */
    int GetVertexSize() const;

    /**
     * Returns <code>true</code> iff the given attribute is part of the format.
     */
    bool HasAttribute(Attribute attrib, int index = 0) const;

    /**
     * Retrieves information about a vertex attribute inside the vertex
     * structure. Returns <code>false</code> if the vertex attribute is not
     * part of this format.
     */
    bool GetAttributeInfo(Attribute attrib, int index, int& offset, GLenum& type, int& size) const;

    /**
     * Binds a vertex buffer or vertex array in OpenGL using this vertex format.
     * \attention For internal usage inside OpenGLFramework Base library only.
     */
    void Bind(const void* data) const;

    glfVertexFormat& operator=(const glfVertexFormat& v2);

  private:

    // Checks if an attribute is supported with the given parameters.
    bool IsAttributeSupported(Attribute attrib, int index, GLenum type, int size, std::string& reason);

    // description of a vertex attribute inside the vertex format.
    struct AttributeDesc {
      Attribute attrib;     // which attribute
      int       index;      // additional index for ATTRIB_TEXCOORD and ATTRIB_GENERIC
      int       offset;     // byte offset in vertex structure
      GLenum    type;       // data type
      int       size;       // number of components
      GLboolean normalized; // for ATTRIB_GENERIC
    };

    std::vector<AttributeDesc> attribDesc_;   // description of all used attributes.
    int vertexSize_;                          // total size of a vertex in bytes.
  };

} // namespace BIAS

#endif // __glfVertexFormat_hh__
