/*
 * ExampleMultipleDepthWarp.cpp
 *
 *  Created on: Jan 28, 2010
 *      Author: africk
 */

/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <OpenGLFramework/RenderingContext/glfPBuffer.hh>
#include <OpenGLFramework/SpecializedBatches/MultipleDepthWarp.hh>
#include <OpenGLFramework/Base/glfException.hh>

#include <Utils/Param.hh>
#include <Utils/IOUtils.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/Image.hh>
#include <Image/Camera.hh>
#include <Base/Debug/TimeMeasure.hh>
#include <OpenGLFramework/SpecializedBatches/DisparityInterpolator.hh>
#include <OpenGLFramework/SpecializedBatches/SeparabelBoxFilter.hh>

#include <Gui/biasgl.h>

using namespace std;
using namespace BIAS;

int main(int argc, char* argv[]) {
	Param params(true);

	string* firstDepthImageName = params.AddParamString("firstDepthImage",
			"image of first source camera", "", 'f');
	string* secondDepthImageName = params.AddParamString("secondDepthImage",
			"image of second source camera", "", 's');

	string* dispImageName = params.AddParamString("dispImage",
			"disparity image name", "result.mip", 'o');

	string* targetProjName = params.AddParamString("targetProj",
			"projection with params of target perspective cam", "", 't');
	string* firstSourceProjName = params.AddParamString("firstSourceProj",
			"projection with params of first source cam", "", 'a');
	string* secondSourceProjName = params.AddParamString("secondSourceProj",
			"projection with params of second source cam", "", 'b');

	double* zNear = params.AddParamDouble("zNear", "", 100);
	double* zFar = params.AddParamDouble("zFar", "", 8000);

	if (!IOUtils::ParseCommandLineEvalHelp(params, argc, argv))
		return 0;

	vector<ProjectionParametersPerspectiveDepth> pmdParams;

	vector<Projection> sourceProjs(2);
	Projection targetProj;

	if (sourceProjs[0].Load(*firstSourceProjName) != 0) {
		BIASERR("Could not load source tof projection: "<<*firstSourceProjName);
		return -1;
	}
	if (sourceProjs[1].Load(*secondSourceProjName) != 0) {
		BIASERR("Could not load source tof projection: "<<*secondSourceProjName);
		return -1;
	}

	if (targetProj.Load(*targetProjName) != 0) {
		BIASERR("Could not load target tof projection:"<<*targetProjName);
		return -1;
	}
	Image<float> firstDepthMap, secondDepthMap;
	if (ImageIO::Load(*firstDepthImageName, firstDepthMap)) {
		BIASERR("Could not load first depth map :"<< *firstDepthImageName);
		return -1;
	}

	//firstDepthMap.ScaleShift(1000, 0);

	if (ImageIO::Load(*secondDepthImageName, secondDepthMap)) {
		BIASERR("Could not load second depth map :"<< *secondDepthImageName);
		return -1;
	}

	//secondDepthMap.ScaleShift(1000, 0);

	MultipleDepthWarp mpmdWarp;
	DisparityInterpolator dispInterp;
	SeparabelBoxFilter boxFilter;

	ProjectionParametersBase* ppb = targetProj.GetParameters();
	ProjectionParametersPerspective* ppp =
			dynamic_cast<ProjectionParametersPerspective*> (ppb);
	if (ppp == NULL) {
		BIASERR("target parameters are not from type perspective");
		return -1;
	}

	TimeMeasure tm;
	Image<float> result;
	glfPBuffer pbuffer;
	try {

		glfRenderingContextConfig cfg;
		cfg.width = 5;
		cfg.height = 5;
		cfg.doubleBuffer = 0;
		cfg.redSize = 8;
		cfg.greenSize = 8;
		cfg.blueSize = 8;
		cfg.alphaSize = 0;
		cfg.depthSize = 24;
		cfg.stencilSize = 0;

		pbuffer.Init(cfg);

		mpmdWarp.SetDepthScaleBeforeWarping(1000.0f);

		ppp->Rescale((unsigned int) 960, (unsigned int) 540);
		mpmdWarp.Init(*ppp, sourceProjs);
		dispInterp.Init(960, 540);
		boxFilter.Init();
		boxFilter.SetFilterSize(10, 10);
		tm.Start();
	//	for (unsigned int i = 0; i < 100; i++) {
		mpmdWarp.UploadDepthMap(firstDepthMap, 0);
		mpmdWarp.UploadDepthMap(secondDepthMap, 1);
		mpmdWarp.SetDiscardAngle(91);
		mpmdWarp.SetDispForDistance(3000, 128);
		mpmdWarp.SetZNearAndZFar((float) *zNear, (float) *zFar);
		mpmdWarp.Warp();

	//	}
		//		dispInterp.SetDisparityMap(mapFusion.GetTexture());
		//		dispInterp.Draw();
		//		boxFilter.SetFilterSize(3,3);
		//		boxFilter.SetSourceAndDest(mpmdWarp.GetTexture(0), mpmdWarp.GetTexture(
		//				1));
		//		boxFilter.Draw();
		tm.Stop();

		tm.Print(cout);

		//mpmdWarp.GetWarpedDisparity(result);
		//boxFilter.GetFilteredTexture(result);

		result.Flip();
		if (ImageIO::Save(*dispImageName, result) == 0) {
			cout << "written disparity map " << *dispImageName << endl;
		} else {
			BIASERR("could not write " << *dispImageName);
			pbuffer.Destroy();
			return -1;
		}
	} catch (glfException& e) {
		std::cout << "Error: " << e.GetMessageString() << std::endl;
		pbuffer.Destroy();
		return -1;
	}

	pbuffer.Destroy();

	return 0;
}

