/*
 * MultipleDepthWarp.cpp
 *
 *  Created on: Jan 28, 2010
 *      Author: africk
 */

#include <OpenGLFramework/SpecializedBatches/MultipleDepthWarp.hh>

using namespace std;
using namespace BIAS;

MultipleDepthWarp::MultipleDepthWarp() {
	zNear_ = 100;
	zFar_ = 8000;
	SetDiscardAngle(91);
	initCalled_ = false;
	depthScaleBeforeWarping_ = 1.0;
	flop_ = false;
	createDepthMaps_ = true;
	current_index_ = 0;
	depthCorType_ = NONE;
	warpType_ = DISPARITY;
}

MultipleDepthWarp::~MultipleDepthWarp() {
	for (unsigned int i = 0; i < depthMaps_.size(); i++) {
		delete depthMaps_[i];
		delete resultDispTextures_[i];
	}
}

int MultipleDepthWarp::Init(const ProjectionParametersPerspective& targetCam,
		std::vector<Projection>& sourceCameras, bool createDepthMaps) {
	if (!initCalled_) {
		createDepthMaps_ = createDepthMaps;
		if (sourceCameras.empty()) {
			BIASERR("sourceCameras.size() == 0 ");
			return -1;
		}

		targetParams_ = targetCam;
		sourceParams_ = sourceCameras;

		unsigned int width, height;
		targetParams_.GetImageSize(width, height);
		viewport_.SetSize(width, height);
		viewport_.SetOrigin(0, 0);

		modelViewMatrix_.MakeViewMatrixNew(targetParams_);
		projectionMatrix_.MakeProjectionMatrixNew(targetParams_, zNear_, zFar_);

		depthMaps_.resize(sourceParams_.size());
		resultDispTextures_.resize(sourceParams_.size());

		for (unsigned int i = 0; i < depthMaps_.size(); i++) {
			if (createDepthMaps_) {
				depthMaps_[i] = new glfTexture2D();

				depthMaps_[i]->Create();
				depthMaps_[i]->SetMinFilter(GL_NEAREST);
				depthMaps_[i]->SetMagFilter(GL_NEAREST);
				depthMaps_[i]->SetWrapS(GL_CLAMP);
				depthMaps_[i]->SetWrapT(GL_CLAMP);
				batch_.SetTexture(depthMaps_[i], i);
			}

			resultDispTextures_[i] = new glfTexture2D;
			resultDispTextures_[i]->Create();
			resultDispTextures_[i]->SetMagFilter(GL_NEAREST);
			resultDispTextures_[i]->SetMinFilter(GL_NEAREST);
			resultDispTextures_[i]->SetWrapS(GL_CLAMP);
			resultDispTextures_[i]->SetWrapT(GL_CLAMP);
			resultDispTextures_[i]->Allocate(width, height, GL_LUMINANCE32F_ARB);
		}

		depthBufferMode_.SetDepthTest(true);
		batch_.SetDepthBufferMode(&depthBufferMode_);
		batch_.SetProjectionMatrix(&projectionMatrix_);
		batch_.SetModelViewMatrix(&modelViewMatrix_);
		batch_.SetViewport(&viewport_);

		resultZBufferTexture_.Create();
		resultZBufferTexture_.SetMagFilter(GL_NEAREST);
		resultZBufferTexture_.SetMinFilter(GL_NEAREST);
		resultZBufferTexture_.SetWrapS(GL_CLAMP);
		resultZBufferTexture_.SetWrapT(GL_CLAMP);
		resultZBufferTexture_.Allocate(width, height, GL_DEPTH_COMPONENT);

		renderTarget_.Create();
		renderTarget_.AttachTexture(*resultDispTextures_[0],
				GL_COLOR_ATTACHMENT0_EXT);
		renderTarget_.AttachTexture(resultZBufferTexture_,
				GL_DEPTH_ATTACHMENT_EXT);
		renderTarget_.CheckComplete();

		batch_.SetRenderTarget(&renderTarget_);
		batch_.SetShaderProgram(&shaderProgram_);
		SetupShaders_();
		SetupPrimitiveRendering_();
	}

	initCalled_ = true;
	return 0;
}

int MultipleDepthWarp::UploadDepthMap(Image<float>& depthMap,
		unsigned int index) {
	if (index >= depthMaps_.size()) {
		BIASERR("could not upload depth map, index " << index << " out of bound " << depthMaps_.size());
		return -1;
	}

	depthMaps_[index]->UploadImage(depthMap);

	return 0;
}

int MultipleDepthWarp::SetDepthMap(glfTexture2D* depthMap, unsigned int index) {
	if (createDepthMaps_) {
		BIASERR("to set depth map, Init must be called with createDepthMaps = false" << endl);
		return -2;
	}
	if (index >= depthMaps_.size()) {
		BIASERR("could not set depth map, index " << index << " out of bound " << depthMaps_.size());
		return -1;
	}

	depthMaps_[index] = depthMap;

	return 0;
}

int MultipleDepthWarp::Warp(unsigned int index) {
	if (initCalled_) {
		if (index < sourceParams_.size()) {
			current_index_ = index;
			renderTarget_.AttachTexture(*resultDispTextures_[index],
					GL_COLOR_ATTACHMENT0_EXT);
			renderTarget_.ClearDepthBuffer(1.0);
			renderTarget_.ClearColorBuffer(0, 0, 0, 0);
			unsigned int width, height;
			sourceParams_[index].GetParameters()->GetImageSize(width, height);
			unsigned int numIndicesPerStrip = width * 2;
			unsigned int first = index * numIndicesPerStrip * (height - 1);
			unsigned int targetWidth, targetHeight;
			targetParams_.GetImageSize(targetWidth, targetHeight);

			shaderProgram_.SetUniform("imageWidth", (float) targetWidth);
			shaderProgram_.SetUniform("depthScaleBefore",
					depthScaleBeforeWarping_);
			shaderProgram_.SetUniform("depthImage", (int) index);
			shaderProgram_.SetUniform("C", cameraCenters_[index]);
			shaderProgram_.SetUniform("thr", thr_);
			shaderProgram_.SetUniform("dist_from_cam", dist_from_cam_);
			shaderProgram_.SetUniform("disp_for_dist_from_cam", disp_);
			shaderProgram_.SetUniform("warpType", (int) warpType_);
			if (flop_) {
				shaderProgram_.SetUniform("flop", (int) 1);
			} else {
				shaderProgram_.SetUniform("flop", (int) 0);
			}

			for (unsigned int y = 0; y < height - 1; y++) {
				batch_.SetRange(first, numIndicesPerStrip);
				batch_.Draw();
				first += numIndicesPerStrip;
			}
		} else {
			BIASERR("index exceeds number of source parameters");
			return -1;
		}

	} else {
		BIASERR("call SetProjectios(...) first");
		return -1;
	}

	return 0;
}

int MultipleDepthWarp::Warp() {
	if (initCalled_) {
		renderTarget_.ClearDepthBuffer(1.0);
		renderTarget_.ClearColorBuffer(0, 0, 0, 0);
		unsigned int first = 0;
		for (unsigned int i = 0; i < sourceParams_.size(); i++) {

			unsigned int targetWidth, targetHeight;
			targetParams_.GetImageSize(targetWidth, targetHeight);

			shaderProgram_.SetUniform("imageWidth", (float) targetWidth);
			shaderProgram_.SetUniform("depthScaleBefore",
					depthScaleBeforeWarping_);
			shaderProgram_.SetUniform("depthImage", (int) i);
			shaderProgram_.SetUniform("C", cameraCenters_[i]);
			shaderProgram_.SetUniform("thr", thr_);
			shaderProgram_.SetUniform("dist_from_cam", dist_from_cam_);
			shaderProgram_.SetUniform("disp_for_dist_from_cam", disp_);
			shaderProgram_.SetUniform("warpType", (int) warpType_);
			if (flop_) {
				shaderProgram_.SetUniform("flop", (int) 1);
			} else {
				shaderProgram_.SetUniform("flop", (int) 0);
			}

			unsigned int width, height;
			sourceParams_[i].GetParameters()->GetImageSize(width, height);

			shaderProgram_.SetUniform("sourceImageWidth", (float) width);
			shaderProgram_.SetUniform("sourceImageHeight", (float) height);

			if (depthCorType_ == POLYNOM) {
				ProjectionParametersPerspectiveDepth
						* paramsDepth =
								dynamic_cast<ProjectionParametersPerspectiveDepth*> (sourceParams_[i].GetParameters());
				std::vector<double> params;
				paramsDepth->GetDepthCalibrationParameters(params);
				shaderProgram_.SetUniform("d0", (float) params[0]);
				shaderProgram_.SetUniform("d1", (float) params[1]);
				shaderProgram_.SetUniform("d2", (float) params[2]);
				shaderProgram_.SetUniform("d3", (float) params[3]);
				shaderProgram_.SetUniform("d4", (float) params[4]);
				shaderProgram_.SetUniform("d5", (float) params[5]);

			} else {
				if (depthCorType_ == LINEAR) {
					ProjectionParametersPerspectiveDepth
							* paramsDepth =
									dynamic_cast<ProjectionParametersPerspectiveDepth*> (sourceParams_[i].GetParameters());
					std::vector<double> params;
					paramsDepth->GetDepthCalibrationParameters(params);
					shaderProgram_.SetUniform("d0", (float) params[0]);
					shaderProgram_.SetUniform("d1", (float) params[1]);

				}
				shaderProgram_.SetUniform("depthCorType", (int) depthCorType_);
			}

			batch_.SetTexture(depthMaps_[i], i);

			unsigned int numIndicesPerStrip = width * 2;
			for (unsigned int y = 0; y < height - 1; y++) {
				batch_.SetRange(first, numIndicesPerStrip);
				batch_.Draw();
				first += numIndicesPerStrip;
			}
		}
	} else {
		BIASERR("call SetProjectios(...) first");
		return -1;
	}

	return 0;
}

int MultipleDepthWarp::GetWarpedDisparity(Image<float>& dispMap) {
	unsigned int width, height;
	targetParams_.GetImageSize(width, height);
	if (dispMap.GetWidth() != width || dispMap.GetHeight() != height
			|| dispMap.GetChannelCount() != 1) {
		dispMap.Release();
		dispMap.Init(width, height, 1);
	}
	resultDispTextures_[current_index_]->CopyToImage(dispMap);

	return 0;
}
int MultipleDepthWarp::GetWarpedZBuffer(Image<float>& zbuffer) {
	unsigned int width, height;
	targetParams_.GetImageSize(width, height);
	if (zbuffer.GetWidth() != width || zbuffer.GetHeight() != height
			|| zbuffer.GetChannelCount() != 1) {
		zbuffer.Release();
		zbuffer.Init(width, height, 1);
	}
	resultZBufferTexture_.CopyToImage(zbuffer);

	return 0;
}

void MultipleDepthWarp::SetDispForDistance(const float& dist_from_cam,
		const float& disp) {
	dist_from_cam_ = dist_from_cam;
	disp_ = disp;
}

void MultipleDepthWarp::SetZNearAndZFar(const float& zNear, const float& zFar) {
	zNear_ = zNear;
	zFar_ = zFar;

	if (initCalled_) {
		projectionMatrix_.MakeProjectionMatrixNew(targetParams_, zNear_, zFar_);
	}
}

void MultipleDepthWarp::SetDiscardAngle(float degree) {
	thr_ = cos(M_PI * degree / 180.0);
}

void MultipleDepthWarp::SetDepthScaleBeforeWarping(float scale) {
	depthScaleBeforeWarping_ = scale;
}

//#####################################
// PRIVATE
//#####################################
void MultipleDepthWarp::SetupPrimitiveRendering_() {
	glfVertexFormat vertexFormat;
	vertexFormat.AddAttribute(glfVertexFormat::ATTRIB_POSITION, GL_FLOAT, 4);
	vertexFormat.AddAttribute(glfVertexFormat::ATTRIB_TEXCOORD, 0, GL_FLOAT, 2);

	const int numComponenents = 6;//4 for vertex 2 for tex coord

	// count number of vertices
	unsigned int sizeOfVertexBuffer = 0;
	unsigned int sizeOfElementBuffer = 0;
	for (unsigned int i = 0; i < sourceParams_.size(); i++) {
		unsigned int width, height;
		sourceParams_[i].GetParameters()->GetImageSize(width, height);
		sizeOfElementBuffer += width * (height - 1) * 2;
		sizeOfVertexBuffer += width * height * numComponenents;
	}

	float* vertexData = new float[sizeOfVertexBuffer];
	GLuint* triangleIndices = new GLuint[sizeOfElementBuffer];

	unsigned int numTriangleIndex = 0;
	unsigned int numVertexIndex = 0;
	unsigned int bufferOffset = 0;
	for (unsigned int i = 0; i < sourceParams_.size(); i++) {
		unsigned int width, height;
		sourceParams_[i].GetParameters()->GetImageSize(width, height);
		cameraCenters_.push_back(sourceParams_[i].GetC());

		for (unsigned int y = 0; y < height - 1; y++) {
			for (unsigned int x = 0; x < width; x++) {
				/*         v1
				 *      i0--- i2
				 *       |   / |
				 *       |  /  |
				 *       i1    i3
				 */
				triangleIndices[numTriangleIndex++] = y * width + x
						+ bufferOffset;//up
				triangleIndices[numTriangleIndex++] = (y + 1) * width + x
						+ bufferOffset;//below
			}
		}

		bufferOffset += width * height;

		// create vertices for source camera i
		for (unsigned int y = 0; y < height; y++) {
			for (unsigned int x = 0; x < width; x++) {
				HomgPoint2D imgPoint(x, y);
				Vector3<double> ray = sourceParams_[i].UnProjectToRay(imgPoint);

				if (ray.NormL2() > 1e-12) {
					ray.Normalize();
				} else {
					BIASERR("error");
				}

				vertexData[numVertexIndex++] = static_cast<float> (ray[0]);
				vertexData[numVertexIndex++] = static_cast<float> (ray[1]);
				vertexData[numVertexIndex++] = static_cast<float> (ray[2]);
				vertexData[numVertexIndex++] = 1.0;

				vertexData[numVertexIndex++] = (static_cast<float> (x) + 0.5)
						/ static_cast<float> (width);
				vertexData[numVertexIndex++] = (static_cast<float> (y) + 0.5)
						/ static_cast<float> (height);
			}
		}
	}

	vertices_.Create(sizeOfVertexBuffer / numComponenents, vertexFormat,
			vertexData);
	elements_.Create(sizeOfElementBuffer, GL_UNSIGNED_INT, GL_TRIANGLE_STRIP,
			triangleIndices);

	batch_.SetElementBuffer(&elements_);
	batch_.SetVertexBuffer(&vertices_);
	delete[] triangleIndices;
	delete[] vertexData;

}

void MultipleDepthWarp::SetupShaders_() {
	string
			vertexShaderCode =
					""
						"uniform sampler2D depthImage;\n"
						"uniform vec3 C;\n"
						"uniform float depthScaleBefore;\n"
						"uniform float imageWidth;\n"
						"uniform float sourceImageHeight;"
						"uniform float sourceImageWidth;"
						"uniform float dist_from_cam;\n"
						"uniform float disp_for_dist_from_cam;\n"
						"uniform int flop;"
						"uniform int warpType;"
						"uniform float d0;"
						"uniform float d1;"
						"uniform float d2;"
						"uniform float d3;"
						"uniform float d4;"
						"uniform float d5;"
						"uniform int depthCorType;"
						""
						"varying float disp;\n"
						""
						"void main() {\n"
						""
						"vec2 coord = gl_MultiTexCoord0.st;"
						"if (flop==1)"
						"coord[0] = 1.0-coord[0];"
						""
						"float depth = texture2D(depthImage, coord).r * depthScaleBefore;\n"
						""
						"if (depthCorType == 1) {"
						"float tmp_depth = depth * depth;"
						"depth = d0 + (1.0 + d1)*depth + d2 * sourceImageWidth * coord[0] + d3 * sourceImageHeight * coord[1] + d4*tmp_depth + d5 *depth * tmp_depth; "
						"} else {"
						"if (depthCorType == 2) {"
						"depth = (d1 + 1)* depth + d0;"
						"}"
						"}"
						""
						"if (depth == 0) depth = 10000;\n"
						"vec4 depthAdjustedVertex = gl_Vertex * depth;\n"
						"depthAdjustedVertex.w = 1.0;\n"
						"depthAdjustedVertex[0]+=C[0];\n"
						"depthAdjustedVertex[1]+=C[1];\n"
						"depthAdjustedVertex[2]+=C[2];\n"
						""
						"if (warpType == 0) {"
						"vec4 tmp = gl_ModelViewProjectionMatrix * vec4(imageWidth,0.0,0.0,0.0);\n"
						"vec4 t = vec4(abs(((dist_from_cam*disp_for_dist_from_cam)/tmp[0])), 0.0,0.0,0.0);\n"
						"vec4 proj_t = gl_ModelViewProjectionMatrix * t;\n"
						""
						""
						"disp = proj_t[0]*imageWidth;\n"
						"} else {"
						"disp = length(depthAdjustedVertex.xyz);"
						"}"
						"gl_Position = depthAdjustedVertex;\n"
						""
						"}\n";

	string fragmentShaderCode = ""
		"varying float resDisp;\n"
		""
		"void main() {\n"
		"gl_FragColor.r = resDisp;\n"
		"}\n";

	string geometryShaderCode = ""
		"#extension GL_EXT_geometry_shader4 : enable\n"
		""
		"varying in float disp[3];\n"
		"varying out float resDisp;\n"
		"uniform vec3 C;\n"
		"uniform float thr;\n"
		"uniform int warpType;"
		""
		"void main() {\n"
		""
		""
		"vec4 v0 = gl_PositionIn[0];"
		"vec4 v1 = gl_PositionIn[1];"
		"vec4 v2 = gl_PositionIn[2];"
		""
		"vec3 mean_vec = (v0.xyz+v1.xyz+v2.xyz)/3.0;"
		""
		"vec3 cam_dir = C-mean_vec;"
		"vec3 dir1 = v0.xyz - v1.xyz;"
		"vec3 dir2 = v2.xyz - v1.xyz;"
		""
		"vec3 n = cross(dir1,dir2);"
		"vec3 norm_n = normalize(n);"
		"vec3 norm_cam = normalize(cam_dir);"
		""
		"float res = dot(norm_n,norm_cam);"
		""
		"if (res < thr) {"
		""
		"gl_Position = gl_ModelViewProjectionMatrix*v0;"
		"if (warpType == 0) {"
		"resDisp = disp[0]/gl_Position[3];\n"
		"} else {"
		"resDisp = disp[0];\n"
		"}"
		"EmitVertex();"
		""
		"gl_Position = gl_ModelViewProjectionMatrix*v1;"
		"if (warpType == 0) {"
		"resDisp = disp[1]/gl_Position[3];\n"
		"} else {"
		"resDisp = disp[1];\n"
		"}"
		"EmitVertex();"
		""
		"gl_Position = gl_ModelViewProjectionMatrix*v2;"
		"if (warpType == 0) {"
		"resDisp = disp[2]/gl_Position[3];\n"
		"} else {"
		"resDisp = disp[2];\n"
		"}"
		"EmitVertex();"
		""
		"}"
		""
		""
		"}\n";

	vertexShader_.Create(GL_VERTEX_SHADER, vertexShaderCode);
	fragmentShader_.Create(GL_FRAGMENT_SHADER, fragmentShaderCode);
	geometryShader_.Create(GL_GEOMETRY_SHADER_EXT, geometryShaderCode);
	shaderProgram_.Create();
	shaderProgram_.AttachShader(vertexShader_);
	shaderProgram_.AttachShader(fragmentShader_);
	shaderProgram_.AttachShader(geometryShader_);

	glProgramParameteriEXT(shaderProgram_.GetProgramId(),
			GL_GEOMETRY_VERTICES_OUT_EXT, 3);
	glProgramParameteriEXT(shaderProgram_.GetProgramId(),
			GL_GEOMETRY_INPUT_TYPE_EXT, GL_TRIANGLES);
	glProgramParameteriEXT(shaderProgram_.GetProgramId(),
			GL_GEOMETRY_OUTPUT_TYPE_EXT, GL_TRIANGLE_STRIP);
	shaderProgram_.Link();

}
