/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __MULTIPLE_DEPTH_WARP_HH__
#define __MULTIPLE_DEPTH_WARP_HH__

#include <Geometry/Projection.hh>
#include <Geometry/ProjectionParametersPerspectiveDepth.hh>
#include <Geometry/ProjectionParametersPerspective.hh>

#include <OpenGLFramework/Base/glfTexture2D.hh>
#include <OpenGLFramework/Base/glfBatch.hh>
#include <OpenGLFramework/Base/glfFramebufferObject.hh>

#include <bias_config.h>

namespace BIAS {

class BIASOpenGLFramework_EXPORT MultipleDepthWarp {
public:

	enum DEPTH_COR_TYPE {
		NONE = 0, POLYNOM = 1, LINEAR = 2
	};

	enum WARPING_TYPE {
		DISPARITY = 0, DEPTH = 1
	};

	MultipleDepthWarp();
	~MultipleDepthWarp();

	/**
	 * @brief
	 * targetCam must be a perspective camera
	 * sourceCameras must be a vector of cameras
	 */
	int
	Init(const ProjectionParametersPerspective& targetCam, std::vector<
			Projection>& sourceCameras, bool createDepthMaps = true);

	/**
	 * @brief
	 * uploads the given depthMap to the gpu as texture
	 * depthMap is a depth map from source camera i
	 */
	int UploadDepthMap(Image<float>& depthMap, unsigned int index);

	/**
	 * @brief
	 * sets a texture for depthMap
	 * can only be used if createDepthMaps = false by calling Init
	 */
	int SetDepthMap(glfTexture2D* depthMap, unsigned int index);

	/**
	 * @brief
	 * warps the depths maps for all sourceCameras to the target camera view as disparity map
	 */
	int Warp();

	/**
	 * @brief
	 * warps the depths maps for all sourceCameras to the target camera view as disparity map
	 */
	int Warp(unsigned int index);

	/**
	 * @brief
	 * gives the warped disparity map as bias image
	 */
	int GetWarpedDisparity(Image<float>& dispMap);

	/**
	 * @brief
	 * gives the warped zbuffer as bias image
	 */
	int GetWarpedZBuffer(Image<float>& zbuffer);

	/**
	 * @brief
	 * if disparity output is required, then points with the distance
	 * (dist_from_cam) will get disparity (disp)
	 */
	void SetDispForDistance(const float& dist_from_cam, const float& disp);

	/**
	 * @brief
	 * this function sets near and far clipping planes for the target camera
	 * default parameters defined in the constructor are zNear = 100, zFar = 8000
	 */
	void SetZNearAndZFar(const float& zNear, const float& zFar);

	void SetDiscardAngle(float degree);

	void SetDepthScaleBeforeWarping(float scale);

	void FlopTextureCoordinates(bool flop) {
		flop_ = flop;
	}

	glfTexture2D* GetWarpedDispTexture() {
		return resultDispTextures_[current_index_];
	}

	int GetTextureAsImage(Image<float>& dispMap, unsigned int index) {
		if (index < resultDispTextures_.size()) {
			resultDispTextures_[index]->CopyToImage(dispMap);
			return 0;
		} else {
			return -1;
		}
	}

	glfTexture2D* GetTexture(unsigned int index) {
		if (index < resultDispTextures_.size()) {
			return resultDispTextures_[index];
		} else {
			return NULL;
		}
	}

	glfFramebufferObject* GetFrameBufferObject() {
		return &renderTarget_;
	}

	void SetDepthCorType(DEPTH_COR_TYPE type) {
		depthCorType_ = type;
	}

	void SetWarpType(WARPING_TYPE type) {
		warpType_ = type;
	}

private:

	DEPTH_COR_TYPE depthCorType_;
	WARPING_TYPE warpType_;

	bool initCalled_, flop_, createDepthMaps_;

	float zNear_, zFar_, thr_, dist_from_cam_, disp_, depthScaleBeforeWarping_;

	unsigned int current_index_;

	std::vector<glfTexture2D*> depthMaps_;
	glfTexture2D resultDispTexture_;
	std::vector<glfTexture2D*> resultDispTextures_;
	glfTexture2D resultZBufferTexture_;

	ProjectionParametersPerspective targetParams_;
	std::vector<Projection> sourceParams_;
	std::vector<Vector3<double> > cameraCenters_;

	glfBatch batch_;

	glfMatrix modelViewMatrix_;
	glfMatrix projectionMatrix_;

	glfViewport viewport_;

	glfVertexBuffer vertices_;//the placeholders for the mesh vertices
	glfElementBuffer elements_;//indices forming the mesh

	glfDepthBufferMode depthBufferMode_;

	glfFramebufferObject renderTarget_;

	BIAS::glfShaderProgram shaderProgram_;
	BIAS::glfShader vertexShader_;
	BIAS::glfShader fragmentShader_;
	BIAS::glfShader geometryShader_;

	void SetupPrimitiveRendering_();
	void SetupShaders_();

};
}

#endif /* __MULTIPLE_DEPTH_WARP_HH__ */
