/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C)  2008, 2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <OpenGLFramework/Utils/FramebufferSetup.hh>
#include <OpenGLFramework/Base/glfFBOAttachmentReleaseException.hh>

using namespace std;
using namespace BIAS;

FramebufferSetup::FramebufferSetup() :
  fbo_(NULL), clearColorBuffer_(false), clearDepthBuffer_(false), clearDepth_(
      1.0f), depthAttachment_(NULL)
{
  for (unsigned int i = 0; i < 4; i++)
    clearColor_[i] = 0.0f;
}

void
FramebufferSetup::LinkFBO(BIAS::glfFramebufferObject* fbo)
{
  BIASASSERT(fbo != NULL);
  fbo_ = fbo;
}

void
FramebufferSetup::SetColorAttachment(BIAS::glfTexture* texture,
    int AttachmentPoint)
{
  if (AttachmentPoint >= static_cast<int> (colorAttachments_.size()))
    {
      colorAttachments_.resize(AttachmentPoint + 1, NULL);
    }
  colorAttachments_[AttachmentPoint] = texture;
}


void
FramebufferSetup::ClearColorAttachments()
{
  colorAttachments_.clear();
}



void
FramebufferSetup::SetColorAttachments(const vector<glfTexture*>& textures)
{
  colorAttachments_.clear();
  colorAttachments_ = textures;
}

void
FramebufferSetup::SetDepthAttachment(BIAS::glfTexture* texture)
{
  depthAttachment_ = texture;
}

glfTexture2D*
FramebufferSetup::GetColorAttachment2D(int AttachmentPoint)
{
  if (AttachmentPoint >= static_cast<int> (colorAttachments_.size()))
    {
      return NULL;
    }
  else
    {
      return dynamic_cast<glfTexture2D*> (colorAttachments_[AttachmentPoint]);
    }
}

void
FramebufferSetup::EnableColorClearance(float red, float green, float blue,
    float alpha)
{
  clearColor_[0] = red;
  clearColor_[1] = green;
  clearColor_[2] = blue;
  clearColor_[3] = alpha;
  clearColorBuffer_ = true;
}

void
FramebufferSetup::EnableDepthClearance(float depth)
{
  clearDepth_ = depth;
  clearDepthBuffer_ = true;
}

void
FramebufferSetup::DisableColorClearance()
{
  clearColorBuffer_ = false;
}

void
FramebufferSetup::DisableDepthClearance()
{
  clearDepthBuffer_ = false;
}

void
FramebufferSetup::Execute()
{
  BIASASSERT(fbo_ != NULL);
  const unsigned int numRT =
      static_cast<GLuint> (glfFramebufferObject::MaxRenderTargets());
  vector<GLenum> renderBuffers;
  renderBuffers.reserve(numRT);
  for (unsigned int i = 0; i < colorAttachments_.size(); i++)
    {
      if (colorAttachments_[i] != NULL)
        {
          fbo_->AttachTexture(*(colorAttachments_[i]), GL_COLOR_ATTACHMENT0_EXT
          + i);
          renderBuffers.push_back(GL_COLOR_ATTACHMENT0_EXT + i);
        }
      else
        {
          try
            {
              fbo_->ReleaseAttachment(GL_COLOR_ATTACHMENT0_EXT + i);
            }
          catch (glfFBOAttachmentReleaseException&)
            {
              //we ignore this error, this is just fine for us
            }
        }
    }
  fbo_->SetDrawBuffers(renderBuffers);
  //release all others
  for (unsigned int i = colorAttachments_.size(); i < numRT; i++)
    {
      try
        {
          fbo_->ReleaseAttachment(GL_COLOR_ATTACHMENT0_EXT + i);
        }
      catch (glfFBOAttachmentReleaseException&)
        {
          //we ignore this error, this is just fine for us
        }
    }

  if (depthAttachment_ != NULL)
    {
      fbo_->AttachTexture(*depthAttachment_, GL_DEPTH_ATTACHMENT_EXT);
    }
  else
    {
      try
        {
          fbo_->ReleaseAttachment(GL_DEPTH_ATTACHMENT_EXT);
        }
      catch (glfFBOAttachmentReleaseException&)
        {
          //we ignore this error, this is just fine for us
        }
    }

#ifdef BIAS_DEBUG
  fbo_->CheckComplete();
#endif

  if (clearColorBuffer_)
    {
      fbo_->ClearColorBuffer(clearColor_[0], clearColor_[1], clearColor_[2],
          clearColor_[3]);
    }
  if (clearDepthBuffer_)
    {
      fbo_->ClearDepthBuffer(clearDepth_);
    }

}
