/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __FramebufferSetupPool_hh__
#define __FramebufferSetupPool_hh__

#include <OpenGLFramework/Utils/FramebufferSetup.hh>
#include <OpenGLFramework/Utils/Texture2DPool.hh>
#include <OpenGLFramework/Base/glfFramebufferObject.hh>

#include <bias_config.h>
#include <map>

namespace BIAS
{

  /** Class containing a framebuffer and a map of setups.
   * Class can be used instead of an stand alone glfFramebufferObject, to manage
   * different frambuffer setups. This class so far only supports 2D textures.
   *
   * \author bartczak 03/2009
   **/

  class BIASOpenGLFramework_EXPORT FramebufferSetupPool
  {
  public:
    ~FramebufferSetupPool();
    FramebufferSetupPool()
    {
      texturePoolP_ = NULL;
    }
    ;

    /** Call this once the GLContext is valid.
     * \attention must be called before using any other member method.
     **/
    void
    Create(Texture2DPool& texturePoolP);

    FramebufferSetup*
    CreateSetup(const std::string& setupName);

    /** Create a setup and the required 2D textures.
     * \param setupName is the name used to identify the setup.
     * \param width horizontal resolution.
     * \param width vertical resolution.
     * \param colorAttachmentName name of the glfTexture2D attached to COLOR_ATTACHMENT_0, identifying it in the internal
     *        Texture2DPool, this name must be unique across all setups.
     * \param internalFormat is the internal format of the target glfTexture2D.
     * \param clearColor if true the target texture is cleared when the setup is activated.
     * \param depthAttachmentName if you want to render the depth output to a texture, give it a name for the internal Texture2DPool,
     *  if left empty to texture for the depth is attached, this name must be unique across all setups.
     * \param clearDepth if true the target 'depth textur'e is cleared when the setup is activated.
     * \param *Clear values the different channels are cleared, if clearing is enabled.
     **/
    FramebufferSetup*
    CreateSetup(const std::string& setupName, unsigned int width,
        unsigned int height, const std::string& colorAttachmentName,
        GLenum internalFormat, const bool clearColors = true,
        const std::string& depthAttachmentName = "", const bool clearDepth =
            true, const float rClear = 0.f, const float gClear = 0.f,
        const float bClear = 0.f, const float aClear = 0.f, const float zClear =
            1.f);

    FramebufferSetup*
    CreateSetup(const std::string& setupName, unsigned int width,
        unsigned int height,
        const std::vector<std::string>& colorAttachmentNames,
        GLenum internalFormat, const bool clearColors = true,
        const std::string& depthAttachmentName = "", const bool clearDepth =
            true, const float rClear = 0.f, const float gClear = 0.f,
        const float bClear = 0.f, const float aClear = 0.f, const float zClear =
            1.f);

    void
    SetColorAttachments(const std::string& setupName, const std::vector<
        std::string>& colorAttachmentNames);

    void
    ClearAll();

    /** Does not remove the attachment. **/
    void
    ClearSetup(const std::string& setupName);

    void
    Activate(const std::string& setupName);

    //########################
    //this method will be removed because a texture with the given name can exist in texture pool but may be not attached to a FrameBufferSetup
    //########################
//    /** Returns the pointer to a texture attached during a setup.
//     * \attention it is not possible to read from and write into the same texture.
//     **/
//    glfTexture2D*
//    GetAttachment(const std::string& attachmentName);
    //########################

    glfTexture2D*
    GetAttachment(const std::string& setupName,
        const std::string& attachmentName);

    FramebufferSetup*
    GetSetup(const std::string& setupName);

    Texture2DPool*
    GetTexturePool();

    glfFramebufferObject*
    GetFBO();

   void GetSetupAttachmentNames(const std::string& setupName,  std::vector<std::string>& attachmentNames);

  private:

    void
    SetColorAttachments_(FramebufferSetup& frameBufferSetup, const std::vector<
        std::string>& colorAttachmentNames);

    FramebufferSetupPool(const FramebufferSetupPool&)
    {
    }
    FramebufferSetupPool&
    operator=(const FramebufferSetupPool&)
    {
      return (*this);
    }

    glfFramebufferObject fbo_;
    // key: setup name - value: pair FramebufferSetup pointer - vector of attachment names
    std::map<std::string,
        std::pair<FramebufferSetup*, std::vector<std::string> > > fboSetupPool_;

    Texture2DPool* texturePoolP_;

  };

}

#endif
