/*
 * ShaderProgramPool.cpp
 *
 *  Created on: Mar 6, 2009
 *      Author: africk
 */

#include "ShaderProgramPool.hh"
#include <Base/Common/FileHandling.hh>
#include <Base/Math/Vector.hh>
#include <Utils/IOUtils.hh>
#include <algorithm>

using namespace std;
using namespace BIAS;

ShaderProgramPool::ShaderProgramPool()
{
  all_program_linked_ = false;

}

ShaderProgramPool::~ShaderProgramPool()
{
  ClearAll();
}

void
ShaderProgramPool::ClearAll()
{
  all_program_linked_ = false;

  program_map_t::iterator program_iter = program_map_.begin();

  for (; program_iter != program_map_.end(); program_iter++)
    {
      glfShaderProgram* programP = GetShaderProgram_(program_iter->second);

      map<string, shader_map_t>::iterator shader_iter = shaders_.begin();
      for (; shader_iter != shaders_.end(); shader_iter++)
        {
          shader_map_t& shader_map = shader_iter->second;
          shader_map_t::iterator iter = shader_map.begin();
          for (; iter != shader_map.end(); iter++)
            {
              delete GetShader_(iter->second);
            }
        }
      shaders_.clear();

      delete programP;
    }

  program_map_.clear();
  validitlyLinked_.clear();
}

//###############################################################

/**
 * adds an empty shader program identified through the shader_program_name to the pool
 * @author frick
 */
void
ShaderProgramPool::AddShaderProgram(const string& shader_program_name)
{
  glfShaderProgram* programP = new glfShaderProgram();

  //try to insert
  pair<program_map_t::iterator, bool> result = program_map_.insert(make_pair(
      shader_program_name, make_pair(programP, uniforms_map_t())));
  if (result.second)
    {
      programP->Create();
      validitlyLinked_[shader_program_name] = false;
    }
  else
    {
      delete programP;
        GLF_THROW_EXCEPTION("program name :" << shader_program_name
            << " already exists !" << endl);
    }
  all_program_linked_ = false;
}

/**
 * adds a fragment shader with shader_name and shader_source to the program with shader_program_name
 * the program with the given name must exist
 * @author frick
 */
void
ShaderProgramPool::AddFragmentShader(const string& shader_source,
    const string& shader_name, const string& shader_program_name)
{
  AddShader_(GL_FRAGMENT_SHADER,shader_source, shader_name, shader_program_name);
  all_program_linked_ = false;
}

/**
 * adds a vertex shader with shader_name and shader_source to the program with shader_program_name
 * the program with the given name must exist
 * @author frick
 */
void
ShaderProgramPool::AddVertexShader(const string& shader_source,
    const string& shader_name, const string& shader_program_name)
{
  AddShader_(GL_VERTEX_SHADER,shader_source, shader_name, shader_program_name);
  all_program_linked_ = false;
}

void
ShaderProgramPool::AddFragmentShaderFromFile(const string& path_to_shader,
    const string& shader_name, const string& shader_program_name)
{
  string shader_source;
  LoadShaderFromFile_(path_to_shader, shader_source);
  AddFragmentShader(shader_source, shader_name, shader_program_name);
}

void
ShaderProgramPool::AddVertexShaderFromFile(const string& path_to_shader,
    const string& shader_name, const string& shader_program_name)
{
  string shader_source;
  LoadShaderFromFile_(path_to_shader, shader_source);
  AddVertexShader(shader_source, shader_name, shader_program_name);
}

/**
 * removes the shader program from program pool; deletes all attached shaders
 */
void
ShaderProgramPool::RemoveShaderProgram(const string& shader_program_name)
{
  glfShaderProgram* programP = GetShaderProgram(shader_program_name);
  if (programP != NULL)
    {
      program_map_.erase(shader_program_name);
      validitlyLinked_.erase(shader_program_name);
      map<string, shader_map_t>::iterator shader_iter = shaders_.find(
          shader_program_name);
      if (shader_iter != shaders_.end())
        {
          shader_map_t& shader_map = shader_iter->second;
          shader_map_t::iterator iter = shader_map.begin();
          for (; iter != shader_map.end(); iter++)
            {
              delete GetShader_(iter->second);
            }

          shader_map.clear();
          shaders_.erase(shader_iter);
        }

      delete programP;
    }
}

/**
 * returns the program pointer or null if the program with shader_program_name does not exist
 * @author frick
 */
glfShaderProgram*
ShaderProgramPool::GetShaderProgram(const string& shader_program_name)
{
  program_data_t* program_dataP = GetShaderProgramData_(shader_program_name);
  if (program_dataP != NULL)
    {
      return program_dataP->first;
    }
  else
    {
      return NULL;
    }
}

void
ShaderProgramPool::BindShaderProgram(const string& shader_program_name)
{
  glfShaderProgram* programP = GetShaderProgram(shader_program_name);
  if (programP != NULL)
    {
      programP->Bind();
    }
}

bool
ShaderProgramPool::IsLinked()
{
  return all_program_linked_;
}

void
ShaderProgramPool::LinkAll()
{
  if (!all_program_linked_)
    {
      program_map_t::iterator iter = program_map_.begin();
      for (; iter != program_map_.end(); iter++)
        {
          program_data_t& program_data = iter->second;
          //link only when required
          if (!validitlyLinked_[iter->first])
            {
              (program_data.first)->Link();
              validitlyLinked_[iter->first] = true;
            }
        }

      all_program_linked_ = true;
    }
}

void
ShaderProgramPool::GetShaderNames(const string& shader_program_name, vector<
    string>& shader_names)
{
  shader_map_t* shaders_mapP = GetShadersMap_(shader_program_name);
  if (shaders_mapP != NULL)
    {
      shader_map_t::iterator iter = shaders_mapP->begin();
      for (; iter != shaders_mapP->end(); iter++)
        {
          shader_names.push_back(iter->first);
        }
    }
}

string
ShaderProgramPool::GetUniformType(const string& shader_program_name,
    const string& varName)
{
  string result = "";
  uniforms_map_t* uniforms_mapP = GetUniformMap_(shader_program_name);
  if (uniforms_mapP == NULL)
    {
        GLF_THROW_EXCEPTION("program name :" << shader_program_name
            << " does not exist !" << endl);
    }
  else
    {
      uniforms_map_t::iterator iter = uniforms_mapP->begin();
      for (; iter != uniforms_mapP->end(); iter++)
        {
          list<string>& variable_list = iter->second;
          list<string>::iterator var_iter = find(variable_list.begin(),
              variable_list.end(), varName);
          if (var_iter != variable_list.end())
            {
              return iter->first;
            }
        }
    }

  return result;
}

template<class StorageType>
  void
  ShaderProgramPool::SetUniformArray(const std::string& shader_program_name,
      const std::string& varName, StorageType value, int arg_number)
  {
    if (all_program_linked_)
      {
        program_data_t* program_dataP = GetShaderProgramData_(
            shader_program_name);
        if (program_dataP == NULL)
          {
              GLF_THROW_EXCEPTION("program name :" << shader_program_name
                  << " does not exist !" << endl);
          }
        else
          {
            uniforms_map_t* uniforms_mapP = GetUniformMap_(*program_dataP);
            uniforms_map_t::iterator iter;
            if (typeid(StorageType) == typeid(int*))
              {
                iter = uniforms_mapP->find("int_array");
              }
            else
              {
                  GLF_THROW_EXCEPTION("SetUniform() is not implemented for type (" << typeid(StorageType).name() << ")"<<endl);
              }

            if (iter != uniforms_mapP->end())
              {
                list<string>& uniforms = iter->second;
                list<string>::iterator uniform_iter = find(uniforms.begin(),
                    uniforms.end(), varName);
                if (uniform_iter != uniforms.end())
                  {
                    glfShaderProgram* programP = program_dataP->first;
                   // cout << varName << endl;
                  //  cout << arg_number << endl;
                  //  cout << programP->GetProgramId() << endl;
                  //  cout << programP->GetInfoLog() << endl;
                    programP->SetUniform(varName, arg_number, value);

                  }
                else
                  {
                      GLF_THROW_EXCEPTION("no uniform variable with name (" << varName << ") and type (" << typeid(StorageType).name()
                          << ") exists in shader program : " << shader_program_name << endl);
                  }
              }
            else
              {
                  GLF_THROW_EXCEPTION("no uniform variable with name (" << varName << ") and type (" << typeid(StorageType).name()
                      << ") exists in shader program : " << shader_program_name << endl);
              }
          }
      }
    else
      {
          GLF_THROW_EXCEPTION("all programs must be linked to set uniform variable, call LinkAll() first ! " << endl);
      }
  }

template<class StorageType>
  void
  ShaderProgramPool::SetUniform(const std::string& shader_program_name,
      const std::string& varName, StorageType value)
  {
    // cout<<"trying to set uniform "<<varName<<" with value "<<value<<" to program "<<shader_program_name<<endl;
    if (all_program_linked_)
      {
        program_data_t* program_dataP = GetShaderProgramData_(
            shader_program_name);
        if (program_dataP == NULL)
          {
              GLF_THROW_EXCEPTION("program name :" << shader_program_name
                  << " does not exist !" << endl);
          }
        else
          {
            uniforms_map_t* uniforms_mapP = GetUniformMap_(*program_dataP);
            uniforms_map_t::iterator iter;
            if (typeid(StorageType) == typeid(bool))
              {
                iter = uniforms_mapP->find("bool");
              }
            else
              {

                if (typeid(StorageType) == typeid(int))
                  {
                    iter = uniforms_mapP->find("int");
                  }
                else
                  {
                    if (typeid(StorageType) == typeid(float)
                        || typeid(StorageType) == typeid(double))
                      {
                        iter = uniforms_mapP->find("float");
                      }
                    else
                      {

                        if (typeid(StorageType) == typeid(const Vector<float>&)
                            || typeid(StorageType)
                                == typeid(const Vector<float>&))
                          {
                            iter = uniforms_mapP->find("vec");
                          }
                        else
                          {
                              GLF_THROW_EXCEPTION("SetUniform() is not implemented for type (" << typeid(StorageType).name() << ")"<<endl);
                          }
                      }

                  }
              }
            //----------------------
            if (iter != uniforms_mapP->end())
              {
                list<string>& uniforms = iter->second;
                list<string>::iterator uniform_iter = find(uniforms.begin(),
                    uniforms.end(), varName);
                if (uniform_iter != uniforms.end())
                  {
                    glfShaderProgram* programP = program_dataP->first;
                    programP->SetUniform(varName, value);
                  }
                else
                  {
                    iter = uniforms_mapP->find("sampler2D");
                    if (iter != uniforms_mapP->end())
                      {
                        list<string> & uniforms = iter->second;
                        list<string>::iterator uniform_iter = find(
                            uniforms.begin(), uniforms.end(), varName);
                        if (uniform_iter != uniforms.end())
                          {
                            glfShaderProgram* programP = program_dataP->first;
                            programP->SetUniform(varName, value);
                          }
                        else
                          {
                              GLF_THROW_EXCEPTION("no uniform variable with name (" << varName << ") and type (" << typeid(StorageType).name()
                                  << ") exists in shader program : " << shader_program_name << endl);
                          }
                      }
                    else
                      {
                          GLF_THROW_EXCEPTION("no uniform variable with name (" << varName << ") and type (" << typeid(StorageType).name()
                              << ") exists in shader program : " << shader_program_name << endl);
                      }
                  }
              }
            else
              {
                if (typeid(StorageType) == typeid(int))
                  {
                    iter = uniforms_mapP->find("sampler2D");
                    if (iter != uniforms_mapP->end())
                      {
                        list<string>& uniforms = iter->second;
                        list<string>::iterator uniform_iter = find(
                            uniforms.begin(), uniforms.end(), varName);
                        if (uniform_iter != uniforms.end())
                          {
                            glfShaderProgram* programP = program_dataP->first;
                            programP->SetUniform(varName, value);
                          }
                        else
                          {
                              GLF_THROW_EXCEPTION("no uniform variable with name (" << varName << ") and type (" << typeid(StorageType).name()
                                  << ") exists in shader program : " << shader_program_name << endl);
                          }
                      }
                    else
                      {
                          GLF_THROW_EXCEPTION("no uniform variable with  name (" << varName << ") and type (" << typeid(StorageType).name()
                              << ") exists in shader program : " << shader_program_name << endl);
                      }
                  }
                else
                  {
                      GLF_THROW_EXCEPTION("no uniform variable with type (" << typeid(StorageType).name()
                          << ") exists in shader program : " << shader_program_name << endl);
                  }
              }
          }
      }
    else
      {
          GLF_THROW_EXCEPTION("all programs must be linked to set uniform variable, call LinkAll() first ! " << endl);
      }
  }

//###############################################################

ShaderProgramPool::program_data_t*
ShaderProgramPool::GetShaderProgramData_(const string& shader_program_name)
{
  program_map_t::iterator iter = program_map_.find(shader_program_name);
  if (iter != program_map_.end())
    {
      return &(iter->second);
    }
  else
    {
      return NULL;
    }
}

void
ShaderProgramPool::AddShader_(GLenum type, const string& shader_source,
    const string& shader_name, const string& shader_program_name)
{
  program_data_t* program_dataP = GetShaderProgramData_(shader_program_name);
  if (program_dataP == NULL)
    {
        GLF_THROW_EXCEPTION("program with name:" << shader_program_name
            << " does not exist !" << endl);
    }
  else
    {
      glfShader* shaderP = GetShader_(shader_program_name, shader_name);
      if (shaderP == NULL)
        {
          glfShaderProgram* programP = program_dataP->first;
          shaderP = new glfShader();
          shaderP->Create(type, shader_source);
          programP->AttachShader(*shaderP);
          // shader_map_t tmp;
          shader_map_t& shader_map = (shaders_.insert(make_pair(
              shader_program_name, shader_map_t())).first)->second;
          shader_map[shader_name] = make_pair(shaderP, shader_source);
          AddUniformDataFromSource_(*program_dataP, shader_source);
        }
      else
        {
            GLF_THROW_EXCEPTION("shader with name:" << shader_name
                << " already exists !" << endl);
        }
    }
}

ShaderProgramPool::shader_map_t*
ShaderProgramPool::GetShadersMap_(const std::string& shader_program_name)
{
  map<string, shader_map_t>::iterator iter = shaders_.find(shader_program_name);
  if (iter != shaders_.end())
    {
      return &(iter->second);
    }
  else
    {
      return NULL;
    }
}

ShaderProgramPool::shader_data_t*
ShaderProgramPool::GetShaderData_(ShaderProgramPool::shader_map_t& shader_map,
    const std::string& shader_name)
{
  shader_map_t::iterator iter = shader_map.find(shader_name);
  if (iter != shader_map.end())
    {
      return &(iter->second);
    }
  else
    {
      return NULL;
    }
}

ShaderProgramPool::uniforms_map_t*
ShaderProgramPool::GetUniformMap_(program_data_t& program_data)
{
  return &((program_data).second);
}

ShaderProgramPool::uniforms_map_t*
ShaderProgramPool::GetUniformMap_(const std::string& shader_program_name)
{
  program_map_t::iterator iter = program_map_.find(shader_program_name);
  if (iter != program_map_.end())
    {
      return GetUniformMap_(iter->second);
    }
  else
    {
      return NULL;
    }
}

string*
ShaderProgramPool::GetShaderSource_(shader_data_t& shader_data)
{
  return &(shader_data.second);
}

glfShader*
ShaderProgramPool::GetShader_(shader_data_t& shader_data)
{
  return shader_data.first;
}

glfShader*
ShaderProgramPool::GetShader_(const string& shader_program_name,
    const string& shader_name)
{
  shader_map_t* shader_mapP = GetShadersMap_(shader_program_name);
  if (shader_mapP != NULL)
    {
      shader_data_t* shader_dataP = GetShaderData_(*(shader_mapP), shader_name);
      if (shader_dataP != NULL)
        {
          return GetShader_(*shader_dataP);
        }
      else
        {
          return NULL;
        }
    }
  else
    {
      return NULL;
    }
}

void
ShaderProgramPool::AddUniformDataFromSource_(program_data_t& program_data,
    string shader_source)
{
  //Remove all commentaries, which do not follow the expected grammer here.
  shader_source = RemoveCommentariesFromSource_(shader_source);

  string::size_type loc = shader_source.find("uniform ", 0);
  //  IOUtils::PrintWithLineNumbers(shader_source);
  while (loc != string::npos)
    {
      string::size_type pos = shader_source.find(';', loc);
      if (pos == string::npos)
        {
          IOUtils::PrintWithLineNumbers(shader_source);
            BIASERR("; missing " <<endl);
            BIASABORT;
        }

      string::size_type length = pos - loc;
      char* data = new char[length + 1];
      shader_source.copy(data, length, loc);
      data[length] = '\0';

      // data type is by default no array
      bool array = false;
      //check if data type is an array (search after '[')
      for (int i = length; i >= 0; i--)
        {
          if (data[i] == '[' || data[i] == ']')
            {
              array = true;
            }
        }

      vector<string> str_v(3);
      unsigned int index = 0;
      char* tokenP;
      tokenP = strtok(data, " = [");
      while (tokenP != NULL)
        {
          if (index >= str_v.size())
            {
              break;
            }
          else
            {
              str_v[index] = string(tokenP);
            }
          tokenP = strtok(NULL, " = [");
          index++;
        }

      if (index < str_v.size() - 1)
        {
          IOUtils::PrintWithLineNumbers(shader_source);
            BIASERR("error parsing uniforms !"<<endl);
            BIASABORT;
        }

      delete data;

      if (array)
        {
          AddUniformVariable_(program_data, str_v[1] + "_array", str_v[2]);
        }
      else
        {
          //     cout<<"adding uniform "<<str_v[2]<<" of type "<<str_v[1]<<endl;
          AddUniformVariable_(program_data, str_v[1], str_v[2]);
        }
      loc = shader_source.find("uniform", pos);
    }
}

void
ShaderProgramPool::AddUniformVariable_(program_data_t& program_data,
    const string& type, const string& name)
{
  uniforms_map_t& uniform_map = *(GetUniformMap_(program_data));

  string::size_type loc = type.find("vec");
  if (loc != string::npos)
    {
      uniform_map[string("vec")].push_back(name);
    }
  else
    {
      uniform_map[type].push_back(name);
    }

}

void
ShaderProgramPool::LoadShaderFromFile_(const string& shader_path,
    string& shader_source)
{
  ifstream ifs(shader_path.c_str());
  if (!ifs)
    {
        GLF_THROW_EXCEPTION("Failed to load shader source file: " << shader_path);
    }

  while (!ifs.eof())
    {
      string line;
      getline(ifs, line);
      shader_source += line + "\r\n";
    }
}

glfShaderProgram*
ShaderProgramPool::GetShaderProgram_(
    ShaderProgramPool::program_data_t& program_data)
{
  return program_data.first;
}

string
ShaderProgramPool::RemoveCommentariesFromSource_(const string& source)
{
  string copyString = source;
  string::size_type pos = 0;
  while (pos != string::npos)
    {
      pos = copyString.find("//");
      if (pos != string::npos)
        {
          string::size_type new_pos = copyString.find('\n',pos);
          copyString.erase(pos, new_pos - pos + 1);
        }
    }

  return copyString;
}

//#################################################
template void BIASOpenGLFramework_EXPORT
ShaderProgramPool::SetUniform<int>(const string& shader_program_name,
    const string& varName, int value);
template void BIASOpenGLFramework_EXPORT
ShaderProgramPool::SetUniform<bool>(const string& shader_program_name,
    const string& varName, bool value);
template void BIASOpenGLFramework_EXPORT
ShaderProgramPool::SetUniform<float>(const string& shader_program_name,
    const string& varName, float value);
template void BIASOpenGLFramework_EXPORT
ShaderProgramPool::SetUniform<double>(const string& shader_program_name,
    const string& varName, double value);
template void BIASOpenGLFramework_EXPORT
ShaderProgramPool::SetUniform<Vector<float> >(
    const string& shader_program_name, const string& varName,
    Vector<float> value);

//#################################################
template void BIASOpenGLFramework_EXPORT
ShaderProgramPool::SetUniformArray<int*>(const string& shader_program_name,
    const string& varName, int* value, int arg_number);
