/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIASWXGLviewer_hh__
#define __BIASWXGLviewer_hh__
#include <bias_config.h>

#include <wx/string.h>
#include <wx/button.h>
#include <wx/gdicmn.h>
#include <wx/font.h>
#include <wx/colour.h>
#include <wx/settings.h>
#include <wx/toolbar.h>
#include <wx/bitmap.h>
#include <wx/image.h>
#include <wx/icon.h>
#include <wx/menu.h>
#include <wx/scrolwin.h>
#include <wx/panel.h>
#include <wx/notebook.h>
#include <wx/splitter.h>

#include <wx/sizer.h>
#include <wx/frame.h>
#include <Gui/ImageCanvas.hh>
#include <Gui/ScaledImageCanvas.hh>
#include <Base/Image/Image.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <Utils/Param.hh>
#include <Gui/ParamGUI.hh>
#include <Filter/Bilateral.hh>
#include <Filter/Median.hh>
#include <GLviewer/Scenes/SceneThreeDOutWrapper.hh>
#include <GLviewer/Controller/DistanceMeasureControl.hh>
#include <GLviewer/ScreenShotListener.hh>
#include <GLviewer/Scenes/PointLightScene.hh>

#ifdef BIAS_HAVE_OPENSCENEGRAPH
#include <GLviewer/ThreeDOutOpenSceneGraph.hh>
#include <GLviewer/Scenes/SceneOpenSceneGraph.hh>
#include <GLviewer/Scenes/ScenePlainOpenSceneGraph.hh>
#endif
#ifdef BIAS_HAVE_GLUT
#include <GLviewer/Scenes/SceneGlutPrimitives.hh>
#endif
#include <GLviewer/ContextWX.hh>
#include <GLviewer/GLProjectionParametersPerspective.hh>
#include <Geometry/ProjectionParametersPerspective.hh>

#include <Base/Debug/TimeMeasure.hh>
///////////////////////////////////////////////////////////////////////////

enum BIASOpenGLViewerFrameIDs{
	
	// standard ids
	ID_LOADMODEL=wxID_HIGHEST+1223,
  ID_DELETEMODELS,  
  ID_SAVE_PARAMETER,
  ID_SHOW_PARAMETER,
  ID_SCREENSHOT,
  ID_FULLSCREEN,    
  ID_SOLID,
  ID_LINES,
  ID_POINTS,
  ID_TIMER,
  ID_CHECK_MEASURE
};


///////////////////////////////////////////////////////////////////////////////
/// Class BIASOpenGLViewerFrame
///////////////////////////////////////////////////////////////////////////////
class BIASOpenGLViewerFrame : public wxFrame, public BIAS::ParamGUI, public BIAS::ParamCallbackInterface
{
	DECLARE_EVENT_TABLE()
	
private:
	// Virtual event handlers, overide them in your derived class
	virtual void OnFullscreen( wxCommandEvent& event );
	virtual void OnViewSolid( wxCommandEvent& event );
	virtual void OnViewLines( wxCommandEvent& event );
	virtual void OnViewPoints( wxCommandEvent& event );
	virtual void OnAddModel(wxCommandEvent& event);
	virtual void OnQuit( wxCloseEvent& event );
	virtual void OnExit( wxCommandEvent& event );
	virtual void OnSaveParameters(wxCommandEvent& event);
	virtual void OnShowParameters(wxCommandEvent& event);
	virtual void OnTimer(wxTimerEvent& event);
	virtual void OnDeleteModels(wxCommandEvent& event);		
	virtual void OnAbout(wxCommandEvent& event);		
	virtual void OnMeasure(wxCommandEvent& event);	
	virtual void OnScreenshot(wxCommandEvent& event);	
	
protected:
	std::string* animationFile_;
	double* fixedFPS_;
	bool* useFlyMode_;
	bool* useRotMode_;

	// View Params
	double* zNear_;
	double* zFar_;
	double* fov_;
	bool* autoClipping_;
	bool* backfaceCulling_;
#ifdef BIAS_HAVE_GLUT
	bool* teapot_;
	bool* cube_;
	bool* tetrahedron_;
	bool* icosahedron_;
#endif
	// Camera Params  
	std::string* projectionFile_;
	bool* useSceneCenter_;
	bool* useJustIntrinsics_;
	int* autoReshapeBehaviour_;
	int firstUnknownParam_;
	
	// Misc Params
	int* winwidth_;
	int* winheight_;
	std::string* screenShotName_;
	bool* greyScreenShot_; 
	BIAS::Vector<double>* backGroundColor_;
	bool* help_;


	protected:
		wxToolBar* toolBar_;
	
		wxButton* loadModelButton_;
		wxButton* fullscreenButton_;
		wxButton* deleteModelsButton_;
		wxButton* screenshotButton_;
		wxCheckBox *measureCheckbox_;
		
		wxMenuBar* menubar_;
		wxMenu* fileMenu_;	
		wxMenu* viewMenu_;	
		wxMenu* helpMenu_;
		
		wxSplitterWindow* splitterWindow_;
		
		wxFrame* paramFrame_;
		wxBoxSizer* panelSizer_;
		
		BIAS::ContextWX*   contextWX_;
		
		wxTimer timer_;
		wxMenuItem* menuItemFullscreen_;
		wxMenuItem* menuItemShowParameter_;
		wxMenuItem* menuItemPoints_,*menuItemLines_,*menuItemSolid_;
				
		
		void    			ToggleFullscreen_();
		void          InitParameters_();

		void 		      LoadDefaultModel_();
		void          RemoveAllModels_();				
		
#ifdef BIAS_HAVE_OPENSCENEGRAPH		
		BIAS::ScenePlainOpenSceneGraph     scene_;
#endif
		
#ifdef BIAS_HAVE_GLUT
		BIAS::SceneGlutPrimitives      sceneGlut_;
#endif
		BIAS::SceneBGImage             bgImageScene_;
		BIAS::PointLightScene          lightScene_;
		
		BIAS::GLProjectionParametersBase *virtualCamera_;
	  BIAS::DistanceMeasureControl      Control_;
	  BIAS::ScreenShotListener screenShotControl_;
	  
 public:
	 BIASOpenGLViewerFrame( wxWindow* parent, wxWindowID id = wxID_ANY, const wxString& title = wxT("MixIn3D"),
                const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxSize( 1024,768) ,
                long style = wxDEFAULT_FRAME_STYLE|wxTAB_TRAVERSAL|wxSYSTEM_MENU );
  ~BIASOpenGLViewerFrame();
  int InitModels();
  int ResetSceneCenter(BIAS::Vector3<double> *C =NULL,BIAS::Quaternion<double>* Q =NULL); 
  void SetParameterFile(int argc, char** argv, std::string& paramFile );
  void ParameterChanged(const std::string &paramname, const void *data);
};

#endif //__MixIn3D__
