/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __ShowCamWxFrame_hh__
#define __ShowCamWxFrame_hh__

#include <Base/Common/BIASpragmaStart.hh>
#include <bias_config.h>

#include <wx/wx.h>
#include <wx/toolbar.h>

#ifdef BIAS_HAVE_DSHOW
#  include <VideoSource/VideoSource_DSHOW.hh>
#endif

#include <Gui/StringConv.hh>
#include <Gui/ImageCanvas.hh>
#include <Gui/ParamGUI.hh>
#include <Gui/VideoSource_Controller.hh>
#include <Image/Camera.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Debug/TimeMeasure.hh>
#include <Filter/Rescale.hh>
#include <Utils/Param.hh>

#include <VideoSource/VideoSource_Base.hh>
#include <VideoSource/VideoSource_Net.hh>
#include <VideoSource/VideoSource_Shm.hh>
#include "ShowCamWxThreads.hh"

#ifdef BIAS_HAVE_PanTilt
#include <PanTilt/DPPanTiltControl.hh>
#include <PanTilt/DPPanTiltControlDialog.hh>
#endif

#ifdef BIAS_HAVE_TIFF
#ifdef BIAS_HAVE_XML2
#include <VideoSource/VideoSource_ShmPMD.hh>
#endif
#endif

#ifdef BIAS_HAVE_V4L
#  include <VideoSource/VideoSource_V4L.hh>
#endif // BIAS_HAVE_V4L

#ifdef BIAS_HAVE_UEYE
#  include <VideoSource/VideoSource_uEye.hh>
#endif // BIAS_HAVE_UEYE

#ifdef BIAS_HAVE_DCAM
#  include "ShowCamWxFormat7Frame.hh"
#  include <VideoSource/VideoSource_DCAM.hh>
#  include <VideoSource/VideoSource_DCAM_XB3.hh>
#  include <Gui/VideoSource_Controller_XB3.hh>
#endif // BIAS_HAVE_DCAM

#ifdef BIAS_HAVE_PMD_PMDTec
#  include <Image/PMDImageProc.hh>
#  include <VideoSource/VideoSource_PMD.hh>
#  include <Gui/VideoSource_Controller_PMD.hh>
#endif // BIAS_HAVE_PMD_PMDTec

#ifdef BIAS_HAVE_PMD_PMDZess
#  include <VideoSource/VideoSource_usbPMD.hh>
#  include <Gui/VideoSource_Controller_usbPMD.hh>
#endif // BIAS_HAVE_PMD_PMDZess

#ifdef BIAS_HAVE_SWISSRANGER
#  include <VideoSource/VideoSource_SwissRanger.hh>
#  include <Gui/VideoSource_Controller_SwissRanger.hh>
#endif

/**
   @class  ShowCamWxFrame
   @brief  This class is the main frame of the ShowCamWx Application
   @author Christoffer Menk, Ingo Schiller, Benjamin Angerer
*/

#define NUMBER_CAMERAS 9

namespace BIAS {

  // event ids
  enum {
    ID_Quit = 42,
    ID_Timer,
    ID_Grab,
    ID_GrabContinuous,
    ID_GrabToStream,
    ID_Key,
    ID_Fit,  
    ID_KeepScroll,
    ID_Show1,
    ID_Show2,
    ID_Show3,
    ID_Show4,
    ID_Show5,
    ID_Show6,
    ID_Show7,
    ID_Show8,
    ID_Show9,
    ID_SaveParameters,
    ID_ShowParameters,
    ID_Controller,
    ID_Paramframe,
    ID_Disable,
    ID_Frame1,
    ID_Frame2,
    ID_FindIEEE1394,
    ID_FindUEYE,
    ID_FindSwissranger,
    ID_FindZess,
    ID_FindPMDTec,
    ID_FindWebcams
  };

  class ShowCamWxFrame : public wxFrame, public BIAS::ParamGUI
  {
  public:

    ShowCamWxFrame(const wxString& title, const wxPoint& pos,
                   const wxSize& size);
    virtual ~ShowCamWxFrame();

    /** Initializes all paramaters for param file
     */
    void InitParams(int &argc, char *argv[]);

    /** Initialize cameras with param file. Only call after InitParams().
     */
    void InitCameras();
#ifdef BIAS_HAVE_PanTilt
    /** Initialize Pan Tilt Unit. Only call after InitParams().
     */
    void InitPanTiltUnit();
#endif
    /** Tries to open all specified camera devices. Only call after
        cameras are initialized.
    */
    void OpenDevices();

    /** Creates the menu of the main window.
     */
    void CreateMenu();

    /** Creates the menu for the available cameras
     */
    void CreateCameraMenu();

    /**
        Starts the saving from all images to disk from all activated cameras
     */
    void StartSavingImagesToDisk();

    /**
        Stops the saving from all images to disk from all activated cameras
     */
    void StopSavingImagesToDisk();

    void ShowParamGUI();
#ifdef BIAS_HAVE_DCAM
    int ShowFormat7Selector(int camindex, dc1394video_mode_t mode);
#endif

  protected:
    void SaveImagePMDZessUnthreaded_(int cameraIndex);
    void SaveImagePMDSRUnthreaded_(int cameraIndex);
    void SaveImagePMDUnthreaded_(int cameraIndex);
    void SaveImageXB3Unthreaded_(int cameraIndex);
    void SaveImageUnthreaded_(int cameraIndex);
    void SaveAllImagesUnthreaded_(int cameraIndex);

    int CreateMasterThread_();
    int CreateGrabThreadForCamera_(int cameraIndex);
    int CreateImageSaveThreadForCamera_(int cameraIndex);
#ifdef BIAS_HAVE_PanTilt
    void GetProjectionFromPanTilt_(BIAS::Projection &projHeader);
#endif
    void GrabSingleImageUnthreadedFromCamera_(int cameraIndex);

    void SelectAndShowCameraImage_();


    /** Parses the color models from param file to enum EColorModel
     */
    void ParseColorModels();

    /** Initializes combined image for selected camera (usb pmd)
     */
    void InitCombinedImage();

    void CleanUpAndExit();

    int OpenDevice(int camindex);
    int CloseDevice(int camindex);
    void InitCamera(int camindex);

    void AddUEyeCameraToMenu(int camindex, bool fromFile);
    void AddWebcamToMenu(std::string name,int camindex, bool fromFile);

#ifdef BIAS_HAVE_DCAM
    void AddIEEECameraToMenu(int camindex,int IEEEIndex, bool fromFile);
    dc1394video_mode_t SelectedModeFromMenu(int camindex);
    std::string VideoModeToString(dc1394video_mode_t mode);
#endif
    // functions for event handling
    void OnWindowClose(wxCloseEvent& event);
    void OnQuit(wxCommandEvent& event);
    void OnAbout(wxCommandEvent &event);
    void OnGrab(wxCommandEvent& event);
    void OnGrabContinuous(wxCommandEvent& event);
    void OnGrabToStream(wxCommandEvent& event);
    void OnShowImage1(wxCommandEvent& event);
    void OnShowImage2(wxCommandEvent& event);
    void OnShowImage3(wxCommandEvent& event);
    void OnShowImage4(wxCommandEvent& event);
    void OnShowImage5(wxCommandEvent& event);
    void OnShowImage6(wxCommandEvent& event);
    void OnShowImage7(wxCommandEvent& event);
    void OnShowImage8(wxCommandEvent& event);
    void OnShowImage9(wxCommandEvent& event);
    void OnShowController(wxCommandEvent& event);
    void OnSaveParameters(wxCommandEvent& event);
    void OnShowParameters(wxCommandEvent& event);
    void OnFindIEEE1394(wxCommandEvent& event);
    void OnFindUEye(wxCommandEvent& event);
    void OnFindPMDZess(wxCommandEvent& event);
    void OnFindPMDTec(wxCommandEvent& event);
    void OnFindSwissRanger(wxCommandEvent& event);
    void OnFindWebcams(wxCommandEvent& event);
    
    void OnChangeVideoMode(wxCommandEvent& event);
    void OnChangeActivate(wxCommandEvent& event);
    void OnChangeTrigger(wxCommandEvent& event);
    void OnChangeFirewireB(wxCommandEvent& event);
    void OnTimer(wxTimerEvent& event);
    void OnKey(wxKeyEvent& event);
    
    void OnFit(wxCommandEvent& event);
    void OnKeepScrollPosition(wxCommandEvent& event);

    // variables for parameters
    int *ParamCameras_[NUMBER_CAMERAS];                 // camera model
    int *ParamWidth_[NUMBER_CAMERAS];                   // width of cameras
    int *ParamHeight_[NUMBER_CAMERAS];                  // height of camera
    int *ParamDown_[NUMBER_CAMERAS];                    // rescale factor
    int *ParamPort_[NUMBER_CAMERAS];                    // port of NET cameras
    std::string *ParamIP_[NUMBER_CAMERAS];              // ip of NET cameras
    std::string *ParamDevice_[NUMBER_CAMERAS];          // device name of camera
    std::string *ParamBitFile_[NUMBER_CAMERAS];         // bit file for pmd cameras
    int *ParamColorModel_[NUMBER_CAMERAS];              // color model of camera
    double *ParamFPS_[NUMBER_CAMERAS];                  // frames per second for camera

#ifdef BIAS_HAVE_PanTilt
    // include the Pan-Tilt-Unit?
    bool *bUsePanTiltUnit_;
#endif
    // vector for all cameras read from param file
    std::vector<BIAS::VideoSource*> Cameras_;
    std::vector<BIAS::ImageBase::EColorModel> ColorModels_;
    std::vector<BIAS::VideoSource_Controller*> VideoControllers_;
#ifdef BIAS_HAVE_DCAM
    std::vector<BIAS::VideoSource_Controller_XB3*> VideoControllersXB3_;
#endif
#ifdef BIAS_HAVE_PMD_PMDZess
    std::vector<BIAS::VideoSource_Controller_usbPMD*> VideoControllersUSBPMD_;
    void AddZessCameraToMenu(int camindex, bool fromFile, std::string paramFile);
#endif // BIAS_HAVE_PMD_PMDZess
#ifdef BIAS_HAVE_PMD_PMDTec
    std::vector<BIAS::VideoSource_Controller_PMD*> VideoControllersPMD_;
    void AddPMDTecCameraToMenu(int camindex, bool fromFile);
#endif // BIAS_HAVE_PMD_PMDTec
#ifdef BIAS_HAVE_SWISSRANGER
    std::vector<BIAS::VideoSource_Controller_SwissRanger*> VideoControllersSwissRanger_;
    void AddSwissRangerCameraToMenu(int camindex, bool fromFile);
#endif // BIAS_HAVE_SWISSRANGER

#ifdef BIAS_HAVE_PanTilt
    //Pan Tilt Control
    BIAS::DPPanTiltControl *panTiltControl_;
    BIAS::DPPanTiltControlDialog *panTiltControlDialog_;
#endif //BIAS_HAVE_PanTilt

    bool InitCameras_; 
    bool InitParams_;
    bool OpenedDevices_;
    unsigned int FrameCounter_;
    unsigned int GrabCounter_;
    unsigned int LastWidth_, LastHeight_;
    unsigned int Win32DCAMNumber_;
    int NumCamSel_;
    int Delay_;
    int NumOfCams_;
    int ActiveCameras_;
    unsigned int ShowXB3_;
    bool bSaveImages_;
    bool bSaveToStream_;
    bool bWriteBuffered_;
    bool EmptyDevice_;
    bool bSingleSaveDirCreated_;
    std::string currentSaveDir_;
    std::string currentSingleSaveDir_;
    unsigned int dDirIndex_;
    unsigned int dSingleDirIndex_;

    // camera images
    std::vector<BIAS::Camera<unsigned char>* > CamImg_;
    std::vector<BIAS::Camera<float>* > CamImgPMDDepth_;
    std::vector<BIAS::Camera<float>* > CamImgPMDInt_;
    std::vector<BIAS::Camera<float>* > CamImgPMDAmp_;

    // images for usb pmd cameras
    BIAS::Camera<unsigned char> CamImgCombined_;

    // extra images for bumblebee bx3
    std::vector<BIAS::Camera<unsigned char>* > CamImgMiddle_;
    std::vector<BIAS::Camera<unsigned char>* > CamImgRight_;
    std::vector<BIAS::Camera<unsigned char>* > XB3AllInOne_;

    // thread handling variables
    std::vector<wxCondition*> ServerCondition_;
    std::vector<wxCondition*> ClientsCondition_;
    std::vector<wxMutex*> ServerMutex_;
    std::vector<wxMutex*> ClientsMutex_;

    std::vector<bool> ThreadInited_;

    std::vector<BIAS::GrabSingleThread*> grabSingleThread_;    
    std::vector<BIAS::ImageSaveThread*>  saveThreads_;
    // the master thread syncronising the grab and save threads
    BIAS::GrabMasterThread* grabMasterThread_;

    std::vector<std::vector<Camera<unsigned char>* >* > ImageBuffer_;
    std::vector<std::vector<Camera<unsigned char>* >* > ImageBufferMiddle_;
    std::vector<std::vector<Camera<unsigned char>* >* > ImageBufferRight_;

    // for pmd
    std::vector<std::vector<Camera<float>* >* > ImageBufferPMDDepth_;
    std::vector<std::vector<Camera<float>* >* > ImageBufferPMDAmp_;
    std::vector<std::vector<Camera<float>* >* > ImageBufferPMDInt_;

    std::vector<unsigned int> WriteIndex_;
    std::vector<unsigned int> ReadIndex_;
    std::vector<wxMutex*> BufferAccess_;

    BIAS::Camera<unsigned char> ShowImage_;

    BIAS::ImageCanvas* IC_;

    std::vector<bool> CameraActive_;
    std::vector<bool> CameraInitialized_;
    std::vector<bool> SetFirewireB_;
    std::vector<bool> SetTrigger_;
    std::vector<bool> SetUserMode_;
    std::vector<float> SelectedFps_;
    std::vector<BIAS::Vector2<int> > Format7LeftTop_;
    std::vector<BIAS::Vector2<int> > Format7WidthHeight_;
    std::vector<int> Format7BpP_;
    std::vector<int> Format7ColorMode_;
    std::vector<ImageBase::EColorModel> ViewColorMode_;

    wxButton *pButtonGrab, *pButtonContinuous, *pButtonStream;
    wxButton *pButtonIEEE1394,*pButtonSR,*pButtonPMDTec,*pButtonWebCams;
    wxCheckBox *pCheckBoxFit,*pCheckBoxKeepScrollPosition;
    
    bool TimerActive_;
    bool CleanedUp_;

    wxTimer Timer_;
    TimeMeasure StopWatch_;
    wxFrame *paramFrame_;
    wxNotebook *paramNotebook_;
    wxMenu *menuShow;
    wxMenu* CameraMenu_;

    int FirstCameraID_;
    int LastCameraID_;

    int NumIEEECams_;
#ifdef BIAS_HAVE_DCAM
    std::vector<CameraInfo> IEEECameraInfo_;
    ShowCamWxFormat7Frame *format7Frame_;
#endif


    DECLARE_EVENT_TABLE()
  }; // end of class

} // end of namespace MIP

#include <Base/Common/BIASpragmaEnd.hh>

#endif // __ShowCamWxFrame_hh__
