/*
   This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel


   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.

   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __BVWXMainFrame_HH
#define __BVWXMainFrame_HH

#include <Base/Common/W32Compat.hh>

#include <wx/wx.h>
#include <wx/spinctrl.h>

#include "BVWXApp.hh"
//#include "BVWXDnDTarget.hh"

#include <Gui/ImageCanvas.hh>
#include <Gui/ScaledImageCanvas.hh>
#include <Base/Image/ImageConvert.hh>
#include <Gui/ConsoleRedirectIO.hh>
#include <Gui/StringConv.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Common/FileHandling.hh>
#include <Image/Camera.hh>
#include <Utils/Param.hh>
#include <Gui/FilterDialogBase.hh>


namespace BIAS {

/**
    @class BVWXMainFrame
    @ingroup g_tools
    @brief   The main frame of the BiasViewWX application
    @author  MIP
*/
class BVWXMainFrame : public wxFrame
{

public:
  //friend class BIAS::BVWXDnDTarget;

  BVWXMainFrame(const wxString& title,
                  const wxPoint& pos,
                  const wxSize& size);
  ~BVWXMainFrame();

  void SetFileList(const std::vector<std::string>& fileNames);
  void LoadImages(const wxArrayString &paths);

  /// fits the frame size if it is too small for curent image
  void FitSizetIfTooSmall();

  /// workaround adapting client canvases size
  void RefitClients();

  /// some compressed images display size is bigger then original size
  /// e.b. X3B images can be displayed three times a swide as loaded
  /// Uncompress after laod to keep same display and orig image size
  /// which eases coordinate transforms
  /// between mouse, display rgb and orig image. (JW)
  void UncompressImage(BIAS::ImageBase & image);

  /// print info such as dim, video format ro info stream.
  /// Used by GetInfo (JW)
  void AddInfo(const BIAS::ImageBase & image, std::stringstream & info);

  int ConvertFLtoUC(ImageBase& img);

  // some error messages and hints
  void PrintSavingError(const int errorNo, wxString fname);
  void PrintConversionError(const int errorNo, wxString fname);
  void PrintEmptyNameMsg();
  void PrintNoImageToFilter();

  // when an image is filtered then ask user, if she/he wants
  // to save the image
  // void SaveFilteredImgDlg();



protected:
  void LoadImageFromList(int index);

  //SizerWindow and its content, _MainSizer which handles
  //the sizing-events for the canvases
  wxWindow*           _SizerWindow;
  wxBoxSizer*         _MainSizer;

  //Canvases for UChar and Float images
  ImageCanvas*        _ImageCanvasUC;
  ScaledImageCanvas*  _ImageCanvasFloat;
  wxPoint             _HistoPosition;
  wxSize              _HistoSize;

  std::vector<std::string>  _FileNames;
  int             _CurrentImage;
  std::string     _MetaData;
  wxTimer         _Timer;
  bool           _BlockAnimation;
  wxSpinCtrl*     _DelaySpin;
  wxCheckBox*     _FitCheckBox;
  wxCheckBox*     _KeepScrollPositionCheckBox;
  wxCheckBox*     _ReverseCheckBox;
  wxMenuItem*     _DrawROICheckItem;
  wxButton*       _PreviousButton;
  wxButton*       _NewButton;
  wxButton*       _RunStopButton;
  wxButton*       _HistogrammButton;
  bool           _NeedReload;
  bool           _ShowHistogramm;
  unsigned int  _SelectedChannel;

  // for file open dialog
  wxString            _DefaultFile;
  wxString            _DefaultDir;

  // event handlers
  void OnLoad(wxCommandEvent& event);
  void OnRemoveCurrent(wxCommandEvent& event);
  void OnRemoveAllButCurrent(wxCommandEvent& event);
  void OnInfo(wxCommandEvent& event);
  void OnMetaData(wxCommandEvent& event);
  void OnExit(wxCommandEvent& event);
  void OnKeyInfo(wxCommandEvent& event);
  void OnAbout(wxCommandEvent& event);
  void OnImageValues(wxCommandEvent& event);

  /* ==== Saving file(s) operations ==== */  
  /* 
   * Save an image to disk with extension choice if no extension we use mip as 
   * default. On some cases storagetype conversion will be taken - from float
   * to uchar but the other way won't be supported (may be in the future).
   */
  void OnSaveAs(wxCommandEvent& event);
  
  /** @brief Save current displayed image with original dimension and .mip as default   */
  void OnSaveCurrent(wxCommandEvent& event);

  /** @brief Print current image   */
  void OnPrintCurrent(wxCommandEvent& event);

  /** @brief Print multiple images   */
  void OnPrint(wxCommandEvent& event);

  /// save all filenames to disk
  void OnSaveList(wxCommandEvent& event);

  void OnNext(wxCommandEvent& event);
  void OnPrev(wxCommandEvent& event);
  void OnRunStop(wxCommandEvent& event);

  void OnHistogramm(wxCommandEvent& event);
  void ShowHistogramm(ImageBase& image);

  void OnDelay(wxSpinEvent& event);
  void OnFit(wxCommandEvent& event);
  void OnKeepScrollPosition(wxCommandEvent& event);

  void OnChangeChannel_1(wxCommandEvent& event);
  void OnChangeChannel_2(wxCommandEvent& event);
  void OnChangeChannel_3(wxCommandEvent& event);
  void OnChangeChannel_4(wxCommandEvent& event);
  void OnChangeChannel_5(wxCommandEvent& event);
  void OnChangeChannel_6(wxCommandEvent& event);
  void OnChangeChannel_7(wxCommandEvent& event);
  void OnChangeChannel_8(wxCommandEvent& event);
  void OnChangeChannel_ALL(wxCommandEvent& event);
	
	void OnFilterCannyEdge(wxCommandEvent& event);
	void OnFilterGauss(wxCommandEvent& event);
	void OnFilterGradientSobel3x3(wxCommandEvent& event);
	void OnFilterMean(wxCommandEvent& WXUNUSED(event));
	void OnFilterMedian(wxCommandEvent& event);
	void OnFilterRescale(wxCommandEvent& event);
	void OnDrawROI(wxCommandEvent& event);

  void OnTimer(wxTimerEvent& event);
  void OnIdle(wxIdleEvent& event);
  void OnSize(wxSizeEvent& event);

  void OnFilterSave( wxCommandEvent &event );

  std::string infoRaw_;//< global info field filled by Uncompress

  enum biasviewxx_IDS
  {

    // standard ids
    ID_About = wxID_ABOUT,
    ID_Exit  = wxID_EXIT,

    // our custom ids
    ID_Load = wxID_HIGHEST+2001, /* first custom id. */

    ID_RemoveCurrent,
    ID_RemoveAllButCurrent,

    ID_Info,
    ID_MetaData,
    ID_KeyInfo,

    // toolbar item IDs
    ID_Prev,
    ID_Next,
    ID_RunStop,
    ID_Histogramm,
    ID_Delay,
    ID_Fit,
    ID_DrawROI,
    // timer ID
    ID_Timer,

    ID_ChangeChannel_1,
    ID_ChangeChannel_2,
    ID_ChangeChannel_3,
    ID_ChangeChannel_4,
    ID_ChangeChannel_5,
    ID_ChangeChannel_6,
    ID_ChangeChannel_7,
    ID_ChangeChannel_8,
    ID_ChangeChannel_ALL,
    
    ID_FilterCannyEdge,
    ID_FilterGauss,
    ID_FilterGradientSobel3x3,
    ID_FilterMean,
    ID_FilterMedian,
    ID_FilterRescale,    

    ID_SaveAs,
    ID_SaveCurrent,
    ID_PrintCurrent,
    ID_Print,
    ID_SaveList,
    ID_ImageValues,

    ID_KeepScrollPosition,
    
    ID_FilterWindow

  };

  DECLARE_EVENT_TABLE()
};

} // namespace BIAS

#endif // __BVWXMainFrame_HH
