/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef WIN32
#include <Base/Common/BIASpragmaStart.hh>
#endif
#include <bias_config.h>
#include <Base/Image/Image.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <Gui/ImageCanvasCheckerboardExtract.hh>
#include <Image/ConvertHDR.hh>
#include <iostream>
#include <fstream>
#include <wx/wx.h>
#include <wx/spinctrl.h>
#include <Utils/Param.hh>
#include <Filter/Rescale.hh>

#ifndef  WIN32
#include <unistd.h>
#else  //WIN32
#include "Base/Common/getopt_W32.h"
#endif //WIN32

using namespace std;
using namespace BIAS;

#define PARAM_FILE_NAME "clickcorners.param"

#define HDR_STORAGETYPE float
/**
    @class biasClickImageCornersWx
    @ingroup g_tools
    @brief The tool allows to click corners/point in image and saves them on demand to a file.
    @author ischiller 11/09
*/

namespace BIAS{

  enum {
    // standard ids
    ID_About= wxID_ABOUT,
    ID_Exit = wxID_EXIT,
    ID_Next = wxID_HIGHEST+2051,
    ID_Prev,
    ID_Save,
    ID_Invalid,
    ID_Help,
    ID_RescaleImage
  };

  class ClickImageCornersApp: public wxApp
  {
    virtual bool OnInit();
  };

  /**
     @class ClickCornersFrame
     is derived from wxFrame, ICECCallBackInterface
  */
  class ClickCornersFrame :
  	public wxFrame, public BIAS::ICECCallbackInterface, public BIAS::Param
  {
  public:

    ClickCornersFrame(const wxString& title, const wxPoint& pos, const wxSize& size);
    virtual ~ClickCornersFrame();

    void Usage();
    int Init(int argc, char* argv[]);
    void Write();
    void CheckerboardSelected(std::vector<BIAS::HomgPoint2D> & corners,
    		bool bOnlyFourPointsGiven, bool valid=true);
    void HandleImage();
    void OnRescaleImage(wxSpinEvent& event);
    
    void OnNext(wxCommandEvent& event);
    void OnPrev(wxCommandEvent& event);
    void OnSave(wxCommandEvent& event);
    void OnInvalid(wxCommandEvent& event);
    void OnHelp(wxCommandEvent& event);
    void ShowText(const string text, int slot);

  protected:
  	BIAS::Rescale<unsigned char, unsigned char> rescaler_;
    string MatlabFilename_;
    double fRescaleImageFactor_;
    int corners_;
    int colorDepth_;
    bool bWriteDebug_;
    std::vector<std::string> fileNames_;
    vector<vector<BIAS::HomgPoint2D> > allCoords_;
    vector<bool>  bImageDone_;
    BIAS::ImageCanvasCheckerboardExtract *ic_;
    unsigned int imageNumber_;
    wxFlexGridSizer *Sizer_;
 private:
    DECLARE_EVENT_TABLE()
  };
}//end namespace




///////////////////////////////////////////////////////
//declaration
DECLARE_APP(ClickImageCornersApp)

// implementation
IMPLEMENT_APP(ClickImageCornersApp)

bool ClickImageCornersApp::OnInit()
{
    ClickCornersFrame *frame_;
    frame_ = new ClickCornersFrame(wxT("ClickCorners"), wxPoint(50,50),
					  wxSize(700,550) );

    // read params from file
    int ret =0;
    ret = frame_->Init(argc, WxToAsciiArray(argv, argc));
    if(ret <0){
      frame_->Usage();
      exit(-1);
    }
    frame_->Show(TRUE);
    frame_->HandleImage();
    SetTopWindow(frame_);
    return TRUE;
    
}

///////////////////////////////////////////////////////
ClickCornersFrame::
ClickCornersFrame(const wxString& title,const wxPoint& pos,const wxSize& size):
  wxFrame((wxFrame *)NULL, -1, title, pos, size)
{
  // Sizer and ImageCanvas
  Sizer_ = new wxFlexGridSizer(1,2,0,0);
  Sizer_->AddGrowableCol(0);
  Sizer_->AddGrowableRow(0);

  ic_=new ImageCanvasCheckerboardExtract(this, NULL, 1,-1,wxDefaultPosition, wxSize(640,480));
  ic_->KeepScrollPosition(true);
  fRescaleImageFactor_ = 1.0;
  
  wxSize buttonSize(65,30);
 // create tool bar
  wxToolBar* toolBar = CreateToolBar();
  wxButton *pBNext= new wxButton(toolBar, ID_Next, wxT("Next"),wxDefaultPosition,buttonSize);
  wxButton *pBPrev= new wxButton(toolBar, ID_Prev, wxT("Previous"),wxDefaultPosition,buttonSize);
  wxButton *pBSave= new wxButton(toolBar, ID_Save, wxT("Save"),wxDefaultPosition,buttonSize);
  wxButton *pBInvalid= new wxButton(toolBar, ID_Invalid, wxT("Invalid"),wxDefaultPosition,buttonSize);
  wxButton *pBHelp= new wxButton(toolBar, ID_Help, wxT("Help"),wxDefaultPosition,buttonSize);
  wxSpinCtrl *pRescaleImageSpinCtrl_ = new wxSpinCtrl(toolBar,ID_RescaleImage,
                                            wxT("Rescale"), wxDefaultPosition,wxSize(60,30),
                                            wxSP_ARROW_KEYS,1,60,10,wxT("10"));
  pRescaleImageSpinCtrl_->SetToolTip(wxT("Rescale image for corners selection."));

  toolBar->AddControl(pBPrev);
  toolBar->AddControl(pBNext);
  toolBar->AddControl(pBSave);
  toolBar->AddControl(pBInvalid);
  toolBar->AddControl(pBHelp);
  toolBar->AddControl(pRescaleImageSpinCtrl_);
  toolBar->Realize();
  //finally add to sizer
  Sizer_->Add(ic_,0,wxALL|wxEXPAND, 5);
  SetSizer(Sizer_);
  Layout();
  SetMinSize(wxSize(700,550));
  Fit();
  CreateStatusBar(2);
}

ClickCornersFrame::
~ClickCornersFrame(){
  if(ic_ != NULL)
    delete ic_;

  if(Sizer_ != NULL)
    delete Sizer_;
}

void  ClickCornersFrame::
OnNext(wxCommandEvent& event){
  if(imageNumber_ < fileNames_.size()-1 && bImageDone_[imageNumber_]){
    imageNumber_++;
    HandleImage();
  }
  else if(imageNumber_ == fileNames_.size()-1){
    string message ="Done! Now save the corners.";
    wxMessageBox(AsciiToWx(message),
    wxT("Detector"), wxOK , this);
    ShowText(message,0);
  }
  else
    ShowText("Select Corners!",0);
}

void  ClickCornersFrame::
OnPrev(wxCommandEvent& event){
  if(imageNumber_ >=1 && bImageDone_[imageNumber_]){
    imageNumber_--;
    HandleImage();
  }
  else
    ShowText("Select Corners!",0);
}

void ClickCornersFrame::
OnInvalid(wxCommandEvent& event){
	std::vector<BIAS::HomgPoint2D> corners;
	BIAS::HomgPoint2D c(-1,-1,-1);
	corners.push_back(c);
	CheckerboardSelected(corners,false,true);
}

void  ClickCornersFrame::
OnSave(wxCommandEvent& event){
  Write();
}

void  ClickCornersFrame::
OnHelp(wxCommandEvent& event){
  string message;
  message = "This is the MIP corners clicking tool \nAfter selecting all corners in one image press \"Next\".\n\nFinally save the corners, pressing \"Save\".";

  wxMessageBox(AsciiToWx(message),
               wxT("Detector"), wxOK , this);
}

void ClickCornersFrame::
OnRescaleImage(wxSpinEvent& event)
{
  fRescaleImageFactor_ = event.GetPosition()/10.0;
  HandleImage();
}

void ClickCornersFrame::
Usage()
{
  cout <<"Usage: biasClickImageCornersWx -f Filename.m -x xnum -y ynum -d image1 image2 ..."<<endl
       <<endl;
  cout <<" The tool tries to find checkerboards in the input images" << endl
       <<" finds corners also in fisheye distorted images and " << endl
       <<" writes these to the Matlab file 'Filename.m'." << endl
       <<" xnum and ynum are the number of the checkerboards inner corners."
       << endl
       <<" -d writes debug images." << endl
       <<" Do not use .mip images since Matlab is unable to read them." << endl
       <<endl<<endl;
}


int ClickCornersFrame::
Init(int argc, char* argv[])
{
  MatlabFilename_ = "ExtractedCorners.m";
  corners_ = 1;
  bWriteDebug_ = false;
  imageNumber_=0;

  int *colorDepth = AddParamInt("colorDepth", "color depth in bit", 8);
  int *corners = AddParamInt("corners","Number of Points", 1,1,1000,'c');
  string *imageList = AddParamString("ImageList","List with images","");
  // update parameters from file with arguments from command line
  if (argc > 1){
    UpdateParameter(argc, argv, PARAM_FILE_NAME);
  }
  else
    return -1;
  Param::ParseListFile(*imageList,fileNames_);
  allCoords_.resize(fileNames_.size());
  bImageDone_.resize(fileNames_.size());
  for(unsigned i=0;i<fileNames_.size();i++)
  	bImageDone_[i] = false;
  corners_ = *corners;
  colorDepth_ = *colorDepth;
  //yCorners_ = *yc;
  ic_->SetPatternParameters( corners_, 1);
  ic_->ManuallySetSelectionMode(true);
  imageNumber_=0;
  return 0;
}

void ClickCornersFrame::
HandleImage(){
  BIAS::ImageBase image;
  BIAS::Image<unsigned char> imageUC,imageUC2;

  if(BIAS::ImageIO::Load(fileNames_[imageNumber_],image)==0){

    if(colorDepth_ != 8){
      Image<float> imageFl;
      ImageConvert::ConvertST(image, imageFl, ImageBase::ST_float);
      ConvertHDR<HDR_STORAGETYPE> hdrHandler;
      hdrHandler.ToUnsignedCharGamma(imageFl, imageUC, 1.0, pow(2.0, colorDepth_));
    } else if(image.GetStorageType() != ImageBase::ST_unsignedchar){
      ImageConvert::ConvertST(image, imageUC, ImageBase::ST_unsignedchar);
    } else {
      imageUC = image;
    }

    if(imageUC.GetColorModel() !=ImageBase::CM_Grey){
      BIAS::Image<unsigned char> dbImage2;
      BIAS::ImageConvert::ToRGB(imageUC, dbImage2);
      BIAS::ImageConvert::ToGrey(dbImage2,imageUC);
    }
    
    if(fRescaleImageFactor_ != 1.0){
    	if(fRescaleImageFactor_ >1.0){
    		rescaler_.Upsample (imageUC, imageUC2,
    				imageUC.GetWidth() * fRescaleImageFactor_,
    				imageUC.GetHeight() * fRescaleImageFactor_);
    	}
    	else if(fRescaleImageFactor_ <1.0){
    		rescaler_.Downsample (imageUC, imageUC2,
    				imageUC.GetWidth() * fRescaleImageFactor_,
    				imageUC.GetHeight() * fRescaleImageFactor_);
    	}
    	ic_->Show(imageUC2,"Select");
    }
    else
    	ic_->Show(imageUC,"Select");

    ic_->SetPatternParameters (corners_,1);
    ic_->ManuallySetSelectionMode(true);
    if( bImageDone_[imageNumber_])
    	ic_->DisplayCheckerboardCorners(allCoords_[imageNumber_],fRescaleImageFactor_);
    else
    	ic_->SelectCheckerboardCorners(*this,true,false);
  }
}

void  ClickCornersFrame::
CheckerboardSelected(std::vector<BIAS::HomgPoint2D> & corners, bool bOnlyFourPointsGiven, bool valid){
  if(bOnlyFourPointsGiven){
    ShowText("Only four points, this is invalid!",0);
    return;
  }
    
  // delete old corners
  allCoords_[imageNumber_].clear();
  for(unsigned int k=0;k< corners.size();k++){
  	corners[k][0] /= fRescaleImageFactor_;
    corners[k][1] /= fRescaleImageFactor_;
    allCoords_[imageNumber_].push_back(corners[k]);
  }
  ShowText("Points selected!",0);
  bImageDone_[imageNumber_] = true;
}


void ClickCornersFrame::
Write(){

  wxString fileName;
  wxFileDialog dialog(this,wxT("Save Corners"),wxT(""),wxT(""),
                      wxT("Matlab (*.m)|*.m"),
                      wxSAVE|wxOVERWRITE_PROMPT);
  int ret = dialog.ShowModal();
  if(ret == wxID_OK){
    fileName = dialog.GetPath();
    ofstream ofs(WxToAscii(fileName));
   // ofs << "numpoints = [" << corners_ <<"];" << endl;

   // ofs << "imagenames = {";
   // for (unsigned int i=0; i<fileNames_.size(); i++) {
   //   ofs << "'" << fileNames_[i] << "' ";
   // }
   // ofs << "};" << endl;

   // ofs << "xpos = [";
    for (unsigned int i=0; i<allCoords_.size(); i++) {
      for (unsigned int j=0; j<allCoords_[i].size(); j++) {
        ofs << (allCoords_[i][j])[0] << " "<<(allCoords_[i][j])[1]<< endl;
      }
    }
    ofs.close();
  }
}

void ClickCornersFrame::
ShowText(const string text, int slot){
  if(GetStatusBar() != NULL)
    GetStatusBar()->SetStatusText(AsciiToWx(text),slot);
}

BEGIN_EVENT_TABLE(BIAS::ClickCornersFrame, wxFrame)
  EVT_BUTTON    (ID_Next,  ClickCornersFrame::OnNext)
  EVT_BUTTON    (ID_Prev,  ClickCornersFrame::OnPrev)
  EVT_BUTTON    (ID_Save,  ClickCornersFrame::OnSave)
  EVT_BUTTON    (ID_Invalid, ClickCornersFrame::OnInvalid)
  EVT_BUTTON    (ID_Help,  ClickCornersFrame::OnHelp)
  EVT_SPINCTRL  (ID_RescaleImage,ClickCornersFrame::OnRescaleImage)
END_EVENT_TABLE()

#ifdef WIN32
#include <Base/Common/BIASpragmaEnd.hh>
#endif
