/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <stdio.h>
#include <string>
#include "Geometry/PMatrix.hh"
#include <Geometry/FMatrix.hh>
#include <Geometry/EpipolarLine.hh>
#include <Image/Camera.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/ImageUtils/ImageDraw.hh>
#include <FeatureDetector/CornerDetectorHarris.hh>
#include <Utils/IOUtils.hh>
#include <Utils/Param.hh>


#define BIASOUT(theArg) cout<<theArg<<endl<<flush

using namespace BIAS;
using namespace std;


#define STORAGETYPE unsigned char
#define CALCTYPE float

/**
   @class biasshowepi
   @ingroup g_tools
   @brief show epipolar lines for image pair (GTK Version)
   @author koeser
*/
//void usage() {
//
//  cout <<"This program opens the image files given by the command line"
//       <<" and extracts PMatrices from the headers, computes FMatrices " <<endl
//       <<"Then runs corner detector in image1 and draws the feature points found"
//       <<"into image 1"
//       << "The epipolar lines are computed for image 2 and drawn there"<<endl;
//
//
//}

#ifdef BIAS_HAVE_OPENCV
int main (int argc, char *argv[])
{


  Param params(true);

  int groupID = params.GetFreeGroupID();
  string* inputCam1 = params.AddParamString("FirstImage",
                           "filename of a BIAS readable image with projection in it",
                           "",
                           'i',groupID);
  string* inputCam2 = params.AddParamString("SecondImage",
                                            "filename of a BIAS readable image with projection in it",
                                            "",
                                            'j',groupID);


  string* outName1 = params.AddParamString("outName1",
                                          "name of the first result image",
                                          "epi1.mip",
                                          'o', groupID);

  string* outName2 = params.AddParamString("outName2",
                                          "name of the second result image",
                                          "epi2.mip",
                                          'p', groupID);



  params.SetGroupName(groupID, "I/O");

  groupID = params.GetFreeGroupID();


  int* numCorners = params.AddParamInt("numCorners",
                                       "max number of corners to be detected by corner detector",
                                        100, 1, INT_MAX, 'C', groupID);
  int* minDist = params.AddParamInt("minDist",
                                    "min distance between corners",
                                     10, 1, INT_MAX, 'D', groupID);

  double* minCornerness = params.AddParamDouble("minCornerness",
                                                "min cornerness determines the quality of the corners found",
                                                50.0, 1.0, DBL_MAX,'Q', groupID);

  params.SetGroupName(groupID, "Corner Params");

  groupID = params.GetFreeGroupID();

  int* thickness = params.AddParamInt("thickness",
                                      "line and point thickness",
                                      1, 1, 5, 't', groupID);
  Vector<double>* color =
  params.AddParamVecDbl("color",
                        "color for point and line drawing",
                        "255 255 255",
                        'c', groupID);

  params.SetGroupName(groupID,"drawing params");

  if(!IOUtils::ParseCommandLineEvalHelp(params, argc, argv) )
     return 0;




  Camera<STORAGETYPE> img1, img2;
  if(ImageIO::Load(*inputCam1, img1) != 0){
    cerr << "could not load image 1 " << *inputCam1 << endl;
    return -1;
  }

  if(ImageIO::Load(*inputCam2, img2) != 0){
    cerr << "could not load image 2 " << *inputCam2 << endl;
    return -1;
  }

  Camera<STORAGETYPE> imgRGB1, imgRGB2;
  if(ImageConvert::ToRGB(img1, imgRGB1) != 0){
    cerr << "could not convert image to rgb " << endl;
    return -1;
  }

  if(ImageConvert::ToRGB(img2, imgRGB2) != 0){
    cerr << "could not convert image to rgb " << endl;
    return -1;
  }


  // read projections from image meta data
  Projection *P1 = NULL;
  Projection *P2 = NULL;

  img1.ParseMetaData();
  img2.ParseMetaData();

  if (img1.IsProjValid()) {
    //    cout << "Found projection in meta data !" << endl;
    P1 = new Projection();
    *P1 = img1.GetProj();
    if (!P1->GetParameters()->PoseValid()) {
      BIASERR("received projection with invalid pose!");
      return -1;
    }
  } else {
    BIASERR("could not read projection from meta data");
  }
  if (img2.IsProjValid()) {
    //    cout << "Found projection in meta data !" << endl;
    P2 = new Projection();
    *P2 = img2.GetProj();
    if (!P2->GetParameters()->PoseValid()) {
      BIASERR("received projection with invalid pose!");
      return -1;
    }
  } else {
    BIASERR("could not read projection from meta data");
  }


  cout << "found projections " << *P1 << *P2 << endl;






  CornerDetectorHarris<STORAGETYPE, CALCTYPE> harris;
  harris.SetMaxNumFeatures(*numCorners);
  harris.SetMinCornerness(*minCornerness);
  harris.SetMinDistance(*minDist);

  vector<HomgPoint2D> features;
  vector<QUAL> quality;
  harris.Detect(img1, features, quality);


  unsigned int numfeatures = features.size();

  FMatrix F;
  ProjectionParametersPerspective *ppp1, *ppp2;
  ppp1 = dynamic_cast<ProjectionParametersPerspective* >(P1->GetParameters());
  ppp2 = dynamic_cast<ProjectionParametersPerspective* >(P2->GetParameters());
  if(ppp1 == NULL || ppp2 == NULL){
    cout << "only implemented for ProjectionParametersPerspective so far and cast did not work" << endl;
    return -1;
  }
  PMatrix Pmatrix1, Pmatrix2;
  Pmatrix1 = ppp1->GetP();
  Pmatrix2 = ppp2->GetP();
  F.ComputeFromPMatrices(Pmatrix1, Pmatrix2);

  for(unsigned int i = 0; i < numfeatures; i++){

    HomgLine2D line = F.GetEpipolarLineImage2(features[i]);
    double y1 = -line[0]/line[1] * 0.0 - line[2]/line[1];
    double y2 = -line[0]/line[1] * (double)img1.GetWidth() - line[2]/line[1];
    ImageDraw<STORAGETYPE>::Line(imgRGB2, 0, (int)rint(y1), (int)img1.GetWidth(), (int)rint(y2),ColourRGB<STORAGETYPE>((STORAGETYPE)(*color)[0], (STORAGETYPE)(*color)[1], (STORAGETYPE)(*color)[2]), *thickness);
    ImageDraw<STORAGETYPE>::Circle(imgRGB1, (int)rint(features[i][0]), (int)rint(features[i][1]), 3, ColourRGB<STORAGETYPE>((STORAGETYPE)(*color)[0], (STORAGETYPE)(*color)[1], (STORAGETYPE)(*color)[2]), *thickness);
  }

  ImageIO::Save(*outName1, imgRGB1);
  ImageIO::Save(*outName2, imgRGB2);




  delete P1;
  delete P2;
  return 0;

}

#endif










