/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <Base/Image/Image.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageIO.hh>
#include <Image/Camera.hh>
#include <Utils/IOUtils.hh>
#include <Utils/Param.hh>
#include <Base/Common/FileHandling.hh>

#include <Image/ConvertHDR.hh>

using namespace std;
using namespace BIAS;

/**
 @class biasrgb2luv
 @ingroup g_tools
 @brief convert a (hdr) bayer image to unsigned char rgb
 @author bartczak 12/2008
 */
int main(int argc, char* argv[]) {

  Param *p = new Param();
  p->AddParamString("image", "image in bayer pattern", "", 'i');
  p->AddParamString("imageList", "images in bayer pattern", "", 'l');
  p->AddParamString("bayer", "bayer pattern RGB", "", 'b');
  p->AddParamInt("colorDepth", "specify color depth e.g. in case of 8 bit image set to 255", 255);
  if (!IOUtils::ParseCommandLineEvalHelp(*p, argc, argv)) {
    return -1;
  }

  string imageName = *(p->GetParamString("image"));
  string imageListName = *(p->GetParamString("imageList"));
  string bayer = *(p->GetParamString("bayer"));
  int colorDepth = *(p->GetParamInt("colorDepth"));
  
  if (imageListName.compare("") != 0) {
    vector<string> imageNames;
    if (Param::ParseListFile(imageListName, imageNames) != 0) {
      BIASERR("error: loading list " << imageListName << endl);
      return -1;
    } else {
      cout << "loaded image list : " << imageListName << endl;
    }

    for (unsigned int i = 0; i < imageNames.size(); i++) {
      string imageName = imageNames[i];
      string newImageName = "rgb_" + imageName;

      ImageBase tmp;

      if (ImageIO::Load(imageName, tmp) != 0) {
        BIASERR("error loading input image named :"<<imageName);
        return -1;
      } else {
        if(colorDepth > 256){ // convert hdr image
          // convert to float first
          Image<float> hdrFloat;
          ImageConvert::ConvertST(tmp, hdrFloat, ImageBase::ST_float);
          ConvertHDR<float> hdrconverter;
          Image<unsigned char> tmpUC;
          hdrconverter.ToUnsignedCharGamma(hdrFloat, tmpUC, 1.0, colorDepth);
          tmp = tmpUC;
        }
        
        if (tmp.GetStorageType() == ImageBase::ST_unsignedchar) {

          BIAS::ImageBase::EColorModel pattern = BIAS::ImageBase::CM_invalid;
          if (strcmp(bayer.c_str(), "BGGR") == 0) {
            cout << "converting bayer BGGR to RGB" << endl;
            pattern = ImageBase::CM_Bayer_BGGR;
          } else if (strcmp(bayer.c_str(), "GBRG") == 0) {
            cout << "converting bayer GBRG to RGB" << endl;
            pattern = ImageBase::CM_Bayer_GBRG;
          } else if (strcmp(bayer.c_str(), "GRBG") == 0) {
            cout << "converting bayer GRBG to RGB" << endl;
            pattern = ImageBase::CM_Bayer_GRBG;
          } else if (strcmp(bayer.c_str(), "RGGB") == 0) {
            cout << "converting bayer RGGB to RGB" << endl;
            pattern = ImageBase::CM_Bayer_RGGB;
          } else {
            cout << bayer << "? i dont know this pattern... abort!" << endl;
            BIASABORT;
          }

          tmp.SetColorModel(pattern);

          Camera<unsigned char> bayerImage(tmp);
          Camera<unsigned char> rgbImage;
          bayerImage.ParseMetaData();
          ImageConvert::ToRGB(bayerImage, rgbImage);
          cout << "converted: " << imageName << " "
              << bayerImage.GetColorModel() << " ---> " << newImageName << " "
              << rgbImage.GetColorModel() << endl;
          rgbImage.SetMetaData(*bayerImage.GetMetaData());
          if (ImageIO::Save(newImageName, rgbImage) != 0) {
            BIASERR("error saving rgb image !");
            return -1;
          }
        } else {
          BIASERR("storage type is not unsigned char or color model is not CM_grey");
          return -1;
        }
      }
    }
  } else {
    ImageBase tmp;

    if (ImageIO::Load(imageName, tmp) != 0) {
      BIASERR("error loading input image named :"<<imageName);
      return -1;
    } else {
      cout << "loaded image " << imageName << endl;
      if(colorDepth > 256){ // convert hdr image
        // convert to float first
        Image<float> hdrFloat;
        ImageConvert::ConvertST(tmp, hdrFloat, ImageBase::ST_float);
        ConvertHDR<float> hdrconverter;
        Image<unsigned char> tempUC;
        hdrconverter.ToUnsignedCharGamma(hdrFloat, tempUC, 1.0, colorDepth);
        tmp = tempUC;
      }
      if (tmp.GetStorageType() == ImageBase::ST_unsignedchar){
//          && tmp.GetColorModel() == ImageBase::CM_Grey) {

        BIAS::ImageBase::EColorModel pattern = BIAS::ImageBase::CM_invalid;
        if (strcmp(bayer.c_str(), "BGGR") == 0) {
          cout << "converting bayer BGGR to RGB" << endl;
          pattern = ImageBase::CM_Bayer_BGGR;
        } else if (strcmp(bayer.c_str(), "GBRG") == 0) {
          cout << "converting bayer GBRG to RGB" << endl;
          pattern = ImageBase::CM_Bayer_GBRG;
        } else if (strcmp(bayer.c_str(), "GRBG") == 0) {
          cout << "converting bayer GRBG to RGB" << endl;
          pattern = ImageBase::CM_Bayer_GRBG;
        } else if (strcmp(bayer.c_str(), "RGGB") == 0) {
          cout << "converting bayer RGGB to RGB" << endl;
          pattern = ImageBase::CM_Bayer_RGGB;
        } else {
          cout << bayer << "? i dont know this pattern... abort!" << endl;
          BIASABORT;
        }

        tmp.SetColorModel(pattern);

        Camera<unsigned char> bayerImage(tmp);
        Camera<unsigned char> rgbImage;
        bayerImage.ParseMetaData();
        ImageConvert::ToRGB(bayerImage, rgbImage);
        cout << "converted " << bayerImage.GetColorModel() << " ---> "
            << rgbImage.GetColorModel() << endl;
        rgbImage.SetMetaData(*bayerImage.GetMetaData());
        if (ImageIO::Save("rgb_" + imageName, rgbImage) != 0) {
          BIASERR("error saving rgb image !");
          return -1;
        }
      } else {
        BIASERR("storage type is not unsigned char or color model is not CM_grey");
        return -1;
      }
    }
  }

  return 0;
}
