/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/** Tool crops a rectangular image region of specified image, 
    writing the result either into a file from specified name. 
    If no output name was specified the input file name is extended by 
    "_crop". If projection parameters are found in the meta data,
    the principle point is shifted accordingly and the image size is adapted.
    \author bartczak 10/2009 
**/

#include <iostream>
#include <Utils/IOUtils.hh>


using namespace std;
using namespace BIAS;

int main(int argc, char* argv[]) 
{
  Param params(true);
  Vector<int>* ulIn = params.AddParamVecInt("UL", "upper left corner", "");
  Vector<int>* lrIn = params.AddParamVecInt("LR", "lower left corner", "");

  int fup;
  if(!IOUtils::ParseCommandLineEvalHelp(params, argc, argv, fup) ) {
    cout<<argv[0]<<" --UL=\"x y\" --LR=\"x y\" <sourceImageName> [targetImageName]\n";
    cout<<" while the upper left corner (UL) is contained in the cropped region, "
      "the LR corner is excluded, as are the lower and the right rectangle border.\n";
    return 0;
  }
  
  if(argc<4) {
    cout<<argv[0]<<" --UL=\"x y\" --LR=\"x y\" <sourceImageName> [targetImageName]\n";
    cout<<" while the upper left corner (UL) is contained in the cropped region, "
      "the LR corner is excluded, as are the lower and the right rectangle border.\n";
    return -1;
  }

  //determine output name
  string inName(argv[fup]);
  string outName;
  if(argc == 5) {
    outName = argv[fup+1];
  } else {
    outName = IOUtils::ExtendOutName(inName, "crop");
  }
  //  cout<<"outName = "<<outName<<endl;
  int ul[2] = {(*ulIn)[0], (*ulIn)[1]};
  int lr[2] = {(*lrIn)[0], (*lrIn)[1]};

  if(ul[0] < 0 || ul[1] < 0) {
    BIASERR("negative coordinates are nonsense! : ("<<ul[0]<<", "<<ul[1]<<") -> ("<<lr[0]<<", "<<lr[1]<<")");
    return -1;
  }  
  if(ul[0] >= lr[0] || ul[1] >= lr[1]) {
    BIASERR("zero area rectangel defined : ("<<ul[0]<<", "<<ul[1]<<") -> ("<<lr[0]<<", "<<lr[1]<<")");
    return -1;
  }


  //load input image
  ImageBase inputImg;
  if(ImageIO::Load(inName, inputImg) !=0 ) {
    BIASERR("error loading src image named : "<<inName);
    return -1;
  }

  bool haveProjection = false;
  Projection proj;
#ifdef BIAS_HAVE_XML2
  //load projection
  haveProjection = IOUtils::GetProjection(inputImg, proj);  
#endif
  inputImg.SetROICorners(static_cast<unsigned int>(ul[0]), 
                         static_cast<unsigned int>(ul[1]),
                         static_cast<unsigned int>(lr[0]), 
                         static_cast<unsigned int>(lr[1]));
  inputImg.Cut2ROI();

  if(haveProjection) {
    unsigned int newWidth = static_cast<unsigned int>(lr[0] - ul[0]);
    unsigned int newHeight = static_cast<unsigned int>(lr[1] - ul[1]);
    proj.GetParameters()->SetImageSize(newWidth, newHeight);
    proj.GetParameters()->SetPrincipal(static_cast<double>(ul[0]), static_cast<double>(ul[1]));
    //    BIASWARN("not completely implemented!");
  }
  
  if(ImageIO::Save(outName, inputImg) !=0 ) {
    BIASERR("error loading src image named : "<<inName);
    return -1;
  }
  
  
  
  return 0;
}
