/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef WIN32
#include <Base/Common/BIASpragmaStart.hh>
#endif

#include <Base/Image/Image.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <FeatureDetector/CheckerboardDetector.hh>
#include <Gui/ImageCanvasCheckerboardExtract.hh>
#include <iostream>
#include <fstream>
#include <wx/wx.h>
#include <Utils/Param.hh>

#ifndef  WIN32
#include <unistd.h>
#else  //WIN32
#include "Base/Common/getopt_W32.h"
#endif //WIN32

using namespace std;
using namespace BIAS;

#define PARAM_FILE_NAME "checkerboard.param"

/** 
    @class biasextractcheckerboardWX
    @ingroup g_tools
    @brief The tool tries to find checkerboards in the input images and writes these to the Matlab file 'Filename.m'. Additionally it offers the posssibility to select them manually.
    @author ischiller 09/07
*/

namespace BIAS{

  enum { 
    // standard ids
    ID_About= wxID_ABOUT,
    ID_Exit = wxID_EXIT,
    ID_Next = wxID_HIGHEST+2051,
    ID_Prev,
    ID_Save,    
    ID_Help
  };

  class CheckerBoardExtractApp: public wxApp
  {
    virtual bool OnInit();
  };


  /**
     @class DepthCalibrationFrame
     is derived from wxFrame, ICECCallBackInterface
  */
  class CheckerBoardExtractFrame:public wxFrame, public BIAS::ICECCallbackInterface, public BIAS::Param
  {
  public:

    CheckerBoardExtractFrame(const wxString& title, const wxPoint& pos, const wxSize& size);
    virtual ~CheckerBoardExtractFrame();
    
    void Usage();
    int Init(int argc, char* argv[]);
    void WriteMatlab(); 
    void CheckerboardSelected(std::vector<BIAS::HomgPoint2D> & corners, bool bOnlyFourPointsGiven);
    void HandleImage();
    
    void OnNext(wxCommandEvent& event);
    void OnPrev(wxCommandEvent& event);
    void OnSave(wxCommandEvent& event);
    void OnHelp(wxCommandEvent& event);
    void ShowText(const string text, int slot);

  protected:
    bool bImageDone_;
    string MatlabFilename_;
    int xCorners_;
    int yCorners_;
    bool bWriteDebug_; 
    std::vector<std::string> fileNames_;
    vector<vector<BIAS::HomgPoint2D> > allCoords_;
    BIAS::ImageCanvasCheckerboardExtract *ic_; 
    unsigned int imageNumber_;
    wxFlexGridSizer *Sizer_;
 private:
    DECLARE_EVENT_TABLE()
  };
}//end namespace




///////////////////////////////////////////////////////
//declaration
DECLARE_APP(CheckerBoardExtractApp)

// implementation
IMPLEMENT_APP(CheckerBoardExtractApp ) 

bool CheckerBoardExtractApp::OnInit()
{
    CheckerBoardExtractFrame *frame_;
    frame_ = new CheckerBoardExtractFrame(wxT("CheckerBoardExtract"), wxPoint(50,50), 
					  wxSize(700,550) );
    
    // read params from file
    int ret =0;
    ret = frame_->Init(argc, WxToAsciiArray(argv, argc));
    if(ret <0){
      frame_->Usage();
      exit(-1);
    }
    frame_->Show(TRUE); 
    frame_->HandleImage();
    SetTopWindow(frame_);  
    return TRUE;
} 

///////////////////////////////////////////////////////
CheckerBoardExtractFrame::
CheckerBoardExtractFrame(const wxString& title,const wxPoint& pos,const wxSize& size):
  wxFrame((wxFrame *)NULL, -1, title, pos, size)
{
  // Sizer and ImageCanvas
  Sizer_ = new wxFlexGridSizer(1,2,0,0);  
  Sizer_->AddGrowableCol(0);
  Sizer_->AddGrowableRow(0);

  ic_=new ImageCanvasCheckerboardExtract(this, NULL, 1,-1,
                                         wxDefaultPosition, wxSize(640,480));

  wxSize buttonSize(65,30);
 // create tool bar
  wxToolBar* toolBar = CreateToolBar();
  wxButton *pBNext= new wxButton(toolBar, ID_Next, wxT("Next"),wxDefaultPosition,buttonSize);
  wxButton *pBPrev= new wxButton(toolBar, ID_Prev, wxT("Previous"),wxDefaultPosition,buttonSize);
  wxButton *pBSave= new wxButton(toolBar, ID_Save, wxT("Save"),wxDefaultPosition,buttonSize);
  wxButton *pBHelp= new wxButton(toolBar, ID_Help, wxT("Help"),wxDefaultPosition,buttonSize);

  toolBar->AddControl(pBPrev);
  toolBar->AddControl(pBNext);
  toolBar->AddControl(pBSave);
  toolBar->AddControl(pBHelp);

  toolBar->Realize();
  //finally add to sizer
  Sizer_->Add(ic_,0,wxALL|wxEXPAND, 5);
  SetSizer(Sizer_);
  Layout();
  SetMinSize(wxSize(700,550));
  Fit(); 
  CreateStatusBar(2); 
}

CheckerBoardExtractFrame::
~CheckerBoardExtractFrame(){
  if(ic_ != NULL)
    delete ic_;
  
  if(Sizer_ != NULL)
    delete Sizer_;
}

void  CheckerBoardExtractFrame::
OnNext(wxCommandEvent& event){ 
  if(imageNumber_ < fileNames_.size()-1 && bImageDone_){
    imageNumber_++;
    HandleImage();
  }
  else if(imageNumber_ == fileNames_.size()-1){
    string message ="Done! Now save the corners.";
    wxMessageBox(AsciiToWx(message),
    wxT("Detector"), wxOK , this);
    ShowText(message,0);
  }
  else
    ShowText("Select Corners!",0);
}

void  CheckerBoardExtractFrame::
OnPrev(wxCommandEvent& event){
  if(imageNumber_ >=1 && bImageDone_){
    imageNumber_--;
    HandleImage();
  } 
  else
    ShowText("Select Corners!",0);
}

void  CheckerBoardExtractFrame::
OnSave(wxCommandEvent& event){
  WriteMatlab();
}

void  CheckerBoardExtractFrame::
OnHelp(wxCommandEvent& event){
  string message;
  message = "This is the MIP checkerboard detector utility.\n It can be used in combination with Scaramuzza's OcamCalib.\n\nCorner Selection:\nIf the corners are not selected automatically\nselect them ROWISE, starting in the top left corner.\nAfter selecting all corners in one image press \"Next\".\n\nFinally save the corners, pressing \"Save\".";
  
  wxMessageBox(AsciiToWx(message),
               wxT("Detector"), wxOK , this);
}


void CheckerBoardExtractFrame::
Usage()
{
  cout <<"Usage: biasextractcheckerboardWX -f Filename.m -x xnum -y ynum -d image1 image2 ..."<<endl
       <<endl;
  cout <<" The tool tries to find checkerboards in the input images" << endl
       <<" finds corners also in fisheye distorted images and " << endl
       <<" writes these to the Matlab file 'Filename.m'." << endl
       <<" xnum and ynum are the number of the checkerboards inner corners." 
       << endl
       <<" -d writes debug images." << endl
       <<" Do not use .mip images since Matlab is unable to read them." << endl
       <<endl<<endl;
}


int CheckerBoardExtractFrame::
Init(int argc, char* argv[])
{
  MatlabFilename_ = "ExtractedCorners.m";
  xCorners_ = 7;
  yCorners_ = 4;
  bWriteDebug_ = false; 
  imageNumber_=0;

  int *xc = AddParamInt("xCorners",
                "Number of Corners in x direction", 7);
  int *yc = AddParamInt("yCorners",
                "Number of Corners in y direction", 4);
  string *imageList = AddParamString("ImageList","List with images",""); 
  // update parameters from file with arguments from command line
  if (argc > 1){
    UpdateParameter(argc, argv, PARAM_FILE_NAME);
  }
  else
    return -1;
  Param::ParseListFile(*imageList,fileNames_);
  allCoords_.resize(fileNames_.size());
  xCorners_ = *xc;
  yCorners_ = *yc;
  ic_->SetPatternParameters( xCorners_, yCorners_);
  imageNumber_=0;  
  return 0;
}

void CheckerBoardExtractFrame::
HandleImage(){
  BIAS::ImageBase image;
  BIAS::Image<unsigned char> imageUC; 
 
  if(BIAS::ImageIO::Load(fileNames_[imageNumber_],image)==0){
   
    if(image.GetStorageType() != ImageBase::ST_unsignedchar){
      ImageConvert::ConvertST(image, imageUC, ImageBase::ST_unsignedchar);    
    }
    else
      imageUC = image;

    if(imageUC.GetColorModel() !=ImageBase::CM_Grey){
      BIAS::Image<unsigned char> dbImage2;
      BIAS::ImageConvert::ToRGB(imageUC, dbImage2);
      BIAS::ImageConvert::ToGrey(dbImage2,imageUC);
    }

    ic_->Show(imageUC,"Select");   
    bImageDone_ = false;
    ic_->SetPatternParameters (xCorners_,yCorners_);
    ic_->SelectCheckerboardCorners(*this,true,false);    
  }
}

void  CheckerBoardExtractFrame::
CheckerboardSelected(std::vector<BIAS::HomgPoint2D> & corners, bool bOnlyFourPointsGiven){
  if(bOnlyFourPointsGiven){
    ShowText("Only four points, this is invalid!",0);
    return;
  }
  // delete old corners
  allCoords_[imageNumber_].clear();

  for(unsigned int k=0;k< corners.size();k++){     
    allCoords_[imageNumber_].push_back(corners[k]); 
  }
  ShowText("Checkerboard selected!",0);
  bImageDone_ = true;
}


void CheckerBoardExtractFrame::
WriteMatlab(){

  wxString fileName;
  wxFileDialog dialog(this,wxT("Save Corners"),wxT(""),wxT(""),
                      wxT("Matlab (*.m)|*.m"),
                      wxSAVE|wxOVERWRITE_PROMPT);
  int ret = dialog.ShowModal();
  if(ret == wxID_OK){
    fileName = dialog.GetPath();
           
    ofstream ofs(WxToAscii(fileName));

    ofs << "numpoints = [" << xCorners_ << " " << yCorners_ << "];" << endl;

    ofs << "imagenames = {";
    for (unsigned int i=0; i<fileNames_.size(); i++) {
      ofs << "'" << fileNames_[i] << "' ";
    }
    ofs << "};" << endl;

    ofs << "xpos = [";
    for (unsigned int i=0; i<allCoords_.size(); i++) {
      for (unsigned int j=0; j<allCoords_[i].size(); j++) {
        ofs << (allCoords_[i][j])[0] << " ";      
      }
      ofs << ";" << endl;
    }
    ofs << "];" << endl;

    ofs << "ypos = [";
    for (unsigned int i=0; i<allCoords_.size(); i++) {
      for (unsigned int j=0; j<allCoords_[i].size(); j++) {
        ofs << (allCoords_[i][j])[1] << " ";      
      }
      ofs << ";" << endl;
    }
    ofs << "];" << endl;

    ofs.close(); 
  } 

}

void CheckerBoardExtractFrame::
ShowText(const string text, int slot){
  if(GetStatusBar() != NULL)
    GetStatusBar()->SetStatusText(AsciiToWx(text),slot);
}

BEGIN_EVENT_TABLE(BIAS::CheckerBoardExtractFrame, wxFrame)
  EVT_BUTTON    (ID_Next,  CheckerBoardExtractFrame::OnNext)
  EVT_BUTTON    (ID_Prev,  CheckerBoardExtractFrame::OnPrev) 
  EVT_BUTTON    (ID_Save,  CheckerBoardExtractFrame::OnSave)
  EVT_BUTTON    (ID_Help,  CheckerBoardExtractFrame::OnHelp)
END_EVENT_TABLE()

#ifdef WIN32
#include <Base/Common/BIASpragmaEnd.hh>
#endif
