/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <Geometry/ProjectionParametersPerspectiveDepth.hh>
#include <Utils/Param.hh>
#include <Utils/IOUtils.hh>
#include <Base/Common/FileHandling.hh>

using namespace std;
using namespace BIAS;

/**
 @class biasshowsm
 @ingroup g_tools
 @brief Undistorts the depth map in the sense of ToF-camera depth distortion,
 without touching the lens distortion. Expects parameters in the header and
 is adding fitting params to output image's header.
 @author bartzcak
 */
int
main(int argc, char* argv[])
{

  Param params(true);
  string* inputCameraName = params.AddParamString("inputCam",
      "input camera/image with params in the header", "", 'i');

  string* inputCameraNameList = params.AddParamString("inputCamList",
      "input cameras/images with params in the header", "", 'l');

  if (!IOUtils::ParseCommandLineEvalHelp(params, argc, argv))
    {
      return -1;
    }

  Camera<float> inputCamera;
  Projection proj;

  if (!((*inputCameraName).compare("") == 0))
    {
      if (!IOUtils::LoadCamera(inputCameraName, inputCamera, proj))
        {
          return -1;
        }
      ProjectionParametersPerspectiveDepth* pppd = NULL;
      if (!IOUtils::CheckType<ProjectionParametersPerspectiveDepth,
          ProjectionParametersBase>(proj.GetParameters(), pppd))
        {
            BIASERR("expected parameters of type ProjectionParametersPerspectiveDepth!");
          return -1;
        }

      if (!pppd->HasDepthDistortion())
        {
          cout << "nothing to un-distort!\n";
          return 0;
        }

      if (pppd->UnDistortDepthMapIP(inputCamera) != 0)
        {
            BIASERR("something went wrong!");
          return -1;
        }

      pppd->SetDepthDistortionToZero();

      string outputCameraName;
      outputCameraName = FileHandling::Basename(*inputCameraName);
      if (!IOUtils::SaveCamera("dud-" + outputCameraName + ".mip", inputCamera,
          Projection(*pppd)))
        {
          return -1;
        }
    }

  vector<string> imageList;
  if (!((*inputCameraNameList).compare("") == 0))
    {
      if (Param::ParseListFile(*inputCameraNameList, imageList) != 0)
        {
            BIASERR("error parsing "<< (*inputCameraNameList) << endl);
          return -1;
        }
      else
        {
          cout << "loaded " << (*inputCameraNameList) << endl;

          for (unsigned int i = 0; i < imageList.size(); i++)
            {
              string inputCameraName = imageList[i];
              if (!IOUtils::LoadCamera(inputCameraName, inputCamera, proj))
                {
                  return -1;
                }
              ProjectionParametersPerspectiveDepth* pppd = NULL;
              if (!IOUtils::CheckType<ProjectionParametersPerspectiveDepth,
                  ProjectionParametersBase>(proj.GetParameters(), pppd))
                {
                    BIASERR("expected parameters of type ProjectionParametersPerspectiveDepth!");
                  return -1;
                }

              if (!pppd->HasDepthDistortion())
                {
                  cout << "nothing to un-distort!\n";
                  return 0;
                }

              if (pppd->UnDistortDepthMapIP(inputCamera) != 0)
                {
                    BIASERR("something went wrong!");
                  return -1;
                }

              pppd->SetDepthDistortionToZero();

              string outputCameraName;
              outputCameraName = FileHandling::Basename(inputCameraName);
              if (!IOUtils::SaveCamera("dud-" + outputCameraName + ".mip",
                  inputCamera, Projection(*pppd)))
                {
                  return -1;
                } else {
                  cout << "written: dud_" << outputCameraName << endl;
                }
            }
        }
    }

  return 0;
}

