/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include <iostream>
#include <vector>

#include <Geometry/PMatrix.hh>
#include <Geometry/Projection.hh>
#include <Geometry/ProjectionParametersPerspective.hh>
#include <Utils/Param.hh>

using namespace std;
using namespace BIAS;


/**
   @class pmatrix2projection
   @ingroup g_tools  
   @brief Convert PMatrix to BIAS::Projection
   @author bartczak
*/
Param param(true);

void usage(char* name) 
{
  param.Usage();
  cout<<"\nConverts PMatrix files into Projection readable files.\n";

  cout<<"\n usage:\n";
  cout<<name<<"params <PMatrix> ...\n";
}

string GetBaseName(const std::string& FileName)
{
  int begin = FileName.find_last_of('/');
  int end = FileName.find_last_of('.');
  return FileName.substr(begin+1, end-begin-1);
}


void PtoPPP(Projection  &proj,const string &file, 
	    unsigned int w, unsigned int h, bool overwriteIntrinsics)
{
  PMatrix P;
  ProjectionParametersPerspective* params;
  params =dynamic_cast<ProjectionParametersPerspective*>(proj.GetParameters());
  if(params==NULL) {
    cout<<"no perspective parameters?\n";
    exit(0) ;
  }
    Quaternion<double> quat;
  int qres;
  if(!P.Load(file)) {
    cout<<"could not read "<<file<<endl;
    return;
  }  
  qres = static_cast<RMatrixBase>(P.GetR()).GetQuaternion(quat);
  BIASASSERT(qres==0);
  params->SetQ(quat);
  params->SetC(P.GetC());
  if (overwriteIntrinsics) {
    params->SetK(P.GetK());
    params->SetImageSize(w, h);
  }
#ifdef BIAS_HAVE_XML2
  proj.XMLWrite(GetBaseName(file)+".xml");
  cout<<GetBaseName(file)+".xml"<<endl;
#endif //BIAS_HAVE_XML2
}




/** @author bartczak */
int main(int argc, char* argv[])
{
  int* width;
  int* height;
  string* listFile;
  string *CameraParamFile;
  width = param.AddParamInt("width", 
			    "width of corresponding image", 
			    0, 0, INT_MAX, 'w');
  height = param.AddParamInt("height", 
			    "height of corresponding image", 
			    0, 0, INT_MAX, 'h');
  listFile = param.AddParamString("listFile", 
				  "file containing a list of pamtrices");
  CameraParamFile = param.AddParamString("cameraParam", 
				  "XML-file with camera parameters");

  int firstPMatrix = param.ParseCommandLine(argc, argv);
  if( argc<2) {
    usage(argv[0]);
    return 0;
  }


  PMatrix P;

  std::vector< std::string > LinesInFile;
  if(*listFile!="") {
    Param::ParseListFile (*listFile, LinesInFile);
    cout<<"found"<<LinesInFile.size()<<"entries in specified listFile"<<endl;
  }

  bool UseKFromP = false;
  Projection *proj;

  if (*CameraParamFile == "") {
    cout<<"width = "<<*width<<"  height = "<<*height<<endl;
    ProjectionParametersPerspective pp;
    proj = new Projection(pp);
    UseKFromP = true;
  }
  else { 
    proj = new Projection;
#ifdef BIAS_HAVE_XML2
    proj->ReadFromCamParamFile(*CameraParamFile);
#endif //BIAS_HAVE_XML2
  }


  int i;
  for(i=0; i<(int)LinesInFile.size(); i++) {
    PtoPPP(*proj, LinesInFile[i], *width, *height, UseKFromP);
  }
  
  for(i=firstPMatrix; i<argc; i++) {
     PtoPPP(*proj, argv[i],  *width, *height, UseKFromP);
  }
    

  return 0;
}
