#include "ColourGradient.hh"

using namespace std;
using namespace BIAS;

template <class StorageType>
BIAS::ColourGradient<StorageType>::
ColourGradient(std::vector<StorageType> colourLeft,
               std::vector<StorageType> colourRight) {
  BIASASSERT(colourLeft.size() ==  colourRight.size())
  BIASASSERT(colourLeft.size() > 0)

  numChannels_ = colourLeft.size();

  GradientStep<StorageType> left;
  left.pos = 0.0;
  left.colour = colourLeft;
  gradientSteps_.push_front(left);

  GradientStep<StorageType> right;
  right.pos = 1.0;
  right.colour = colourRight;
  gradientSteps_.push_back(right);
}

template <class StorageType>
void BIAS::ColourGradient<StorageType>::
AddColour(float pos,
          const std::vector<StorageType> &colour) {
  BIASASSERT(pos >= 0.0 && pos <= 1.0)
  BIASASSERT(colour.size() == numChannels_)

  typename std::list<GradientStep<StorageType> >::iterator left
    = gradientSteps_.begin();
  typename std::list<GradientStep<StorageType> >::iterator right
    = ++(gradientSteps_.begin());

  // replace left?
  if (pos == left->pos) {
    left->colour = colour;
    return;
  }

  do {
    // replace right?
    if (pos == right->pos) {
      right->colour = colour;
      break;
    }
    // insert between left and right?
    else if (pos > left->pos && pos < right->pos) {
      GradientStep<StorageType> step;
      step.pos = pos;
      step.colour = colour;
      gradientSteps_.insert(right, step);
      break;
    }

    // prepare next iteration
    left = right;
    ++right;
  }
  while (right != gradientSteps_.end());
}


template <class StorageType>
void BIAS::ColourGradient<StorageType>::
GetColour(float pos,
          std::vector<StorageType> &colour) {
  BIASASSERT(pos >= 0.0 && pos <= 1.0)

  colour.clear();
  colour.resize(numChannels_);

  typename std::list<GradientStep<StorageType> >::iterator left
    = gradientSteps_.begin();
  typename std::list<GradientStep<StorageType> >::iterator right
    = ++(gradientSteps_.begin());

  do {
    if (pos >= left->pos && pos <= right->pos) {
      const float delta = right->pos - left->pos;
      const float weight = (pos - left->pos) / delta;

      for (unsigned int c = 0; c < numChannels_; c++) {
        colour[c] = (unsigned char)
        (((1.0 - weight) * left->colour[c]) +
         (weight * right->colour[c]));
      }

      return;
    }

    // prepare next iteration
    ++left;
    ++right;
  }
  while (right != gradientSteps_.end());
}

//
// solve explicit instantiation
//
#define INSTANCE_ColourGradient(type)\
template class BIASUtils_EXPORT ColourGradient<type>;

INSTANCE_ColourGradient(unsigned char)
INSTANCE_ColourGradient(float)
#ifdef BUILD_IMAGE_INT
  INSTANCE_ColourGradient(int)
#endif
#ifdef BUILD_IMAGE_CHAR
  INSTANCE_ColourGradient(char)
#endif
#ifdef BUILD_IMAGE_SHORT
  INSTANCE_ColourGradient(short)
#endif
#if defined(BUILD_IMAGE_USHORT)
  INSTANCE_ColourGradient(unsigned short)
#endif
#ifdef BUILD_IMAGE_DOUBLE
  INSTANCE_ColourGradient(double)
#endif
#ifdef BUILD_IMAGE_UINT
  INSTANCE_ColourGradient(unsigned int)
#endif