#ifndef __COLOUR_GRADIENT_HH__
#define __COLOUR_GRADIENT_HH__

// STL
#include <vector>
#include <list>

// BIAS
#include <Base/Debug/Error.hh>



namespace BIAS {

  /**
   * @brief Represents a colour gradient.
   *
   * The ColourGradient is a function [0,1] -> YOUR_COLOUR_SPACE.
   *
   * The gradient will constist of at least two colours: one colour on the
   * "left" side (at position 0) and one on the "right" side (at position 1).
   * Intermediate colours can be added with AddColour().
   *
   * To get a linearly interpolated colour at a certain position call
   * GetColour().
   *
   * @author rwulff 04/2010
   */
  template <class StorageType>
  class BIASUtils_EXPORT ColourGradient {

    public:

    /**
     * Constructor.
     *
     * @note Both colour vectors may not be empty and must have the same size.
     *
     * @param colourLeft
     *        the colour at position 0
     * @param colourRight
     *        the colour at position 1
     */
    ColourGradient(std::vector<StorageType> colourLeft,
                   std::vector<StorageType> colourRight);


    /**
     * @brief Adds a colour to this ColourGradient.
     *
     * If a colour already exists at the given pos, it is replaced by the new
     * value.
     *
     * @param pos
     *        the position where the new colour is to be set, must be in [0,1]
     * @param colour
     *        the new colour to set, must have the same size as the colours
     *        given to the constructor
     */
    void AddColour(float pos, const std::vector<StorageType> &colour);


    /**
     * @brief Returns the interpolated colour at a given position.
     *
     * The colour is interpolated linearly between the nearest colours that were
     * added before.
     *
     * @param pos
     *        the position for which the interpolated colour should be returned,
     *        must be in [0,1]
     * @param colour
     *        the returned interpolated colour
     */
    void GetColour(float pos, std::vector<StorageType> &colour);


    private:

    /**
     * struct to hold a single step of the colour gradient
     */
    template <class StructStorageType>
    struct GradientStep {
      float pos;
      std::vector<StructStorageType> colour;
    };


    // to ensure correct channel count
    unsigned int numChannels_;

    // sorted list of the colours
    std::list<GradientStep<StorageType> > gradientSteps_;


    /**
     * Hidden default constructor
     */
    ColourGradient() {}

  }; // end of class
} // end of namespace
#endif // __COLOUR_GRADIENT_HH__