#include "DataPlot.hh"
#include <Base/Debug/Exception.hh>
#include <Base/ImageUtils/ImageDraw.hh>
#include <limits>

using namespace BIAS;
using namespace std;
DataPlot::DataPlot(Image<unsigned char> &img)
{
  PlotArea_ = &img;
  PlotAreaWidth_ = PlotArea_->GetWidth();
  PlotAreaHeight_ = PlotArea_->GetHeight();
  BorderX_ = 2;
  BorderY_ = 2;
  AutoRangeX_ = true;
  AutoRangeY_ = true;
  ChannelCount_ = 3;
  if (img.GetChannelCount() != ChannelCount_) BEXCEPTION("");
  DrawIda_ = PlotArea_->GetImageDataArray();
  BGColour_ = ColourRGB<unsigned char>(0,0,0);
  PenColour_ = ColourRGB<unsigned char>(255,255,0);
  DrawStyle_ = Points;
  Thickness_ = 1;
}

DataPlot::~DataPlot()
{
}

void DataPlot::SetDrawingStyle(enum DrawingStyle style)
{
DrawStyle_ = style;
LastVX_=-1;
LastVY_=-1;

}


void DataPlot::SetXRange(float min,float max)
{
  AutoRangeX_ = false;
  MaxX_ = max;
  MinX_ = min;
}

void DataPlot::SetYRange(float min,float max)
{
  AutoRangeY_ = false;
  MaxY_ = max;
  MinY_ = min;
}

void DataPlot::SetAutoRange(bool x,bool y)
{
  AutoRangeX_ = x;
  AutoRangeY_ = y;
}

void DataPlot::DrawXAxis()
{
  unsigned int vx1,vy,vx2;
  int res = WindowToViewport_(MinX_,0,vx1,vy);
   res = WindowToViewport_(MaxX_,0,vx2,vy);
   DrawLine_(vx1,vy,vx2,vy);
  
}



void DataPlot::DrawYAxis()
{
  unsigned int vx,vy1,vy2;
  int res = WindowToViewport_(0,MinY_,vx,vy1);
   res = WindowToViewport_(0,MaxY_,vx,vy2);
   DrawLine_(vx,vy1,vx,vy2);
  
}

void DataPlot::Plot(const std::vector<float> &data)
{
  if (AutoRangeX_) {
    MinX_ = 0.0f;
    MaxX_ = float(data.size()-1);
  }
  if (AutoRangeY_) {
    MinY_ = numeric_limits<float>::max();
    MaxY_ = numeric_limits<float>::min();
    for (unsigned int i=0;i<data.size();i++){
      if (data[i]>MaxY_) MaxY_ = data[i];
      if (data[i]<MinY_) MinY_ = data[i];
    }
  }
  for (unsigned int i=0;i<data.size();i++)
    Plot(float(i),data[i]);
}

void DataPlot::Plot(const std::vector<float> &x,const std::vector<float> &y)
{
  if (AutoRangeX_) {
    MinX_ = numeric_limits<float>::max();
    MaxX_ = numeric_limits<float>::min();
    for (unsigned int i=0;i<x.size();i++){
      if (x[i]>MaxX_) MaxX_ = x[i];
      if (x[i]<MinX_) MinX_ = x[i];
    }
  }
  if (AutoRangeY_) {
    MinY_ = numeric_limits<float>::max();
    MaxY_ = numeric_limits<float>::min();
    for (unsigned int i=0;i<y.size();i++){
      if (y[i]>MaxY_) MaxY_ = y[i];
      if (y[i]<MinY_) MinY_ = y[i];
    }
  }
  for (unsigned int i=0;i<x.size();i++)
    Plot(x[i],y[i]);
}


int DataPlot::Plot(float x,float y)
{
  unsigned int vx=0,vy=0;
  int res = WindowToViewport_(x,y,vx,vy);
  if (res<0) return res;
  switch (DrawStyle_) {
    case Points:
      for (unsigned int c=0;c<ChannelCount_;c++){
         int d = Thickness_;
        for (int dx =-d;dx<d;dx++) 
          for (int dy =-d;dy<d;dy++){ 
            int xx = vx+dx;
            int yy = vy+dy;
            DrawIda_[yy][xx*ChannelCount_+c] = PenColour_[c];
          }
      }
      break;
    case Lines:
      if (LastVX_>=0&&LastVY_>=0)
        DrawLine_(LastVX_,LastVY_,vx,vy);
        LastVX_ = vx;
        LastVY_ = vy;
      break;
    case Cross:
      for (unsigned int c=0;c<ChannelCount_;c++) {
        for (int dx =-5;dx<5;dx++) {
          int xx =  int(vx)+dx;
          DrawIda_[vy][xx*ChannelCount_+c] = PenColour_[c];
        }
        for (int dy =-5;dy<5;dy++) {
          int yy = int(vy) + dy;
          DrawIda_[yy][vx*ChannelCount_+c] = PenColour_[c];
        }
      }
      break;
  }
  return 0;
}

void DataPlot::CutLinePlot()
{
  LastVY_ = -1;
  LastVX_ = -1;
}

void DataPlot::Clear()
{
  for (unsigned int y=0;y<PlotAreaHeight_;y++)
    for (unsigned int x=0;x<PlotAreaWidth_;x++)
      for (unsigned int c=0;c<ChannelCount_;c++)
          DrawIda_[y][x*ChannelCount_+c] = BGColour_[c];

}

int DataPlot::WindowToViewport_(float wx,float wy, unsigned int &vx,unsigned int &vy)
{
  float vpw = float(PlotAreaWidth_ - 2*BorderX_);
  float vph = float(PlotAreaHeight_ - 2*BorderY_);
  float RangeX = MaxX_ - MinX_;
  float RangeY = MaxY_ - MinY_;
  float tx = (wx - MinX_)/RangeX * vpw + float(BorderX_);
  float ty = PlotAreaHeight_ - ((wy - MinY_)/RangeY * vph + float(BorderY_));
  if (tx <0 || tx > PlotAreaWidth_) return -1;
  if (ty <0 || ty > PlotAreaHeight_) return -2;
  vx = (int)rint(tx);
  vy = (int)rint(ty);
  return 0;
}



void DataPlot::DrawLine_(unsigned int x,unsigned int y,unsigned int x2, unsigned int y2)
{
  ImageDraw<unsigned char>::Line(*PlotArea_,x,y,x2,y2,PenColour_,Thickness_);
}
