/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "GenGroundTruth.hh"

#include <Base/Debug/Debug.hh>
#include <Base/Math/Random.hh>
#include <Geometry/RMatrix.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Base/Math/Vector3.hh>
#include <Base/Geometry/HomgLine2D.hh>
#include <Base/Geometry/HomgPoint3D.hh>
#include <Base/ImageUtils/ImageDraw.hh>

using namespace BIAS;
using namespace std;

void GenGroundTruth::NormalizeMatches(vector<BIAS::HomgPoint2D>& match1,
                      vector<BIAS::HomgPoint2D>& match2,
                      KMatrix K)
{
  KMatrix Kinv=K.Invert();
  for (unsigned int i=0; i<match1.size(); i++){
    match1[i]=Kinv*match1[i];
    match2[i]=Kinv*match2[i];
  }
}

void GenGroundTruth::DeNormalizeMatches(vector<BIAS::HomgPoint2D>& match1,
                        vector<BIAS::HomgPoint2D>& match2,
                        KMatrix K)
{
  for (unsigned int i=0; i<match1.size(); i++){
    match1[i]=K*match1[i];
    match2[i]=K*match2[i];
  }
}

void GenGroundTruth::DrawMatches(BIAS::Image<unsigned char>& im,
                      vector<BIAS::HomgPoint2D>& match1,
                      vector<BIAS::HomgPoint2D>& match2,
                      unsigned char lcolor, unsigned char mcolor)
{
  unsigned int start[2], end[2];
  HomgLine2D line;
  unsigned int coo[4];

  for (unsigned int i=0; i<match1.size(); i++){
    start[0]=(unsigned int)rint(match1[i][0]);
    start[1]=(unsigned int)rint(match1[i][1]);
    end[0]=(unsigned int)rint(match2[i][0]);
    end[1]=(unsigned int)rint(match2[i][1]);
    line.Set(match1[i], match2[i]);
    line.GetIntersectionsWithImage(im.GetWidth(), im.GetHeight(), coo);
    if (ImageDraw<unsigned char>::Line(im, &(coo[0]), &(coo[2]), lcolor)!=0){
      BIASERR("error drawing epipolar line "<<i);
    }
  }
  for (unsigned int i=0; i<match1.size(); i++){
    start[0]=(unsigned int)rint(match1[i][0]);
    start[1]=(unsigned int)rint(match1[i][1]);
    end[0]=(unsigned int)rint(match2[i][0]);
    end[1]=(unsigned int)rint(match2[i][1]);
    if (ImageDraw<unsigned char>::Line(im, start, end, mcolor)!=0){
      BIASERR("error drawing match "<<i<<" ("<<start[0]<<", "<<start[1]
              <<")  ("<<end[0]<<", "<<end[1]<<")   "<<match1[i]<<"  <-->  "
              <<match2[i]);
    }
  }
}


void GenGroundTruth::CreateIdealMatches(vector<PMatrix>& pvec, unsigned num,
                        unsigned width, unsigned height,
                        vector<vector<HomgPoint2D> >& matches,
                        vector<HomgPoint3D>& points,
                        double minx, double maxx, double miny, double maxy,
                        double minz, double maxz)
{
  Random ran;
  unsigned i;
  HomgPoint3D p3d(0.0, 0.0, 0.0);
  HomgPoint2D p2d(0.0, 0.0);
  vector<HomgPoint2D> p2vec;
  unsigned numim=pvec.size();
  bool done;

  p2vec.resize(numim);
  points.clear();
  for (i=0; i<matches.size(); i++)
    matches[i].clear();
  matches.clear();

  for (i=0; i<num; i++){
    do {
      p3d[0]=ran.GetUniformDistributed(minx, maxx);
      p3d[1]=ran.GetUniformDistributed(miny, maxy);
      p3d[2]=ran.GetUniformDistributed(minz, maxz);
      done=true;
      p2vec.clear();
      for (unsigned j=0; j<numim; j++){
        p2d=pvec[i]*p3d;
        p2d.Homogenize();
        if (p2d[0]>=0 && p2d[0]<width-1 && p2d[1]>=0 && p2d[1]<height-1){
          p2vec.push_back(p2d);
        } else {
          done=false;
          break;
        }
      }
    } while (!done);
    points.push_back(p3d);
    matches.push_back(p2vec);
  }
}


void GenGroundTruth::CreateMatches(vector<PMatrix>& pvec, unsigned num,
                   unsigned width, unsigned height, double spatial_noise,
                   vector<vector<HomgPoint2D> >& i_matches,
                   vector<vector<HomgPoint2D> >& n_matches,
                   vector<HomgPoint3D>& points,
                   double minx, double maxx, double miny, double maxy,
                   double minz, double maxz)
{
  Random ran;
  unsigned i;
  HomgPoint3D p3d(0.0, 0.0, 0.0);
  HomgPoint2D p2d(0.0, 0.0), n_p2d(0.0, 0.0);
  vector<HomgPoint2D> p2vec;
  vector<HomgPoint2D> noisy_p2vec;
  unsigned numim=pvec.size();
  bool done;
  p2vec.resize(numim);
  points.clear();
  for (i=0; i<i_matches.size(); i++)
    i_matches[i].clear();
  for (i=0; i<n_matches.size(); i++)
    n_matches[i].clear();
  i_matches.clear();
  n_matches.clear();

  for (i=0; i<num; i++){
    do {
      p3d[0]=ran.GetUniformDistributed(minx, maxx);
      p3d[1]=ran.GetUniformDistributed(miny, maxy);
      p3d[2]=ran.GetUniformDistributed(minz, maxz);
      done=true;
      p2vec.clear();
      noisy_p2vec.clear();
      for (unsigned j=0; j<numim; j++){
        p2d=pvec[i]*p3d;
        p2d.Homogenize();
        n_p2d[0]=p2d[0]+ran.GetNormalDistributed(0.0, spatial_noise);
        n_p2d[1]=p2d[1]+ran.GetNormalDistributed(0.0, spatial_noise);
        if (p2d[0]>=0 && p2d[0]<width-1 && p2d[1]>=0 && p2d[1]<height-1 &&
            n_p2d[0]>=0 && n_p2d[0]<width-1
            && n_p2d[1]>=0 && n_p2d[1]<height-1){
          p2vec.push_back(p2d);
          noisy_p2vec.push_back(n_p2d);
        } else {
          done=false;
          break;
        }
      }
    } while (!done);
    points.push_back(p3d);
    i_matches.push_back(p2vec);
    n_matches.push_back(noisy_p2vec);
  }
}

void GenGroundTruth::CreateMatches(vector<PMatrix>& pvec, vector<PMatrix>& n_pvec,
                        unsigned num, unsigned width, unsigned height,
                        double spatial_noise, double r_noise, double c_noise,
                        vector<vector<HomgPoint2D> >& i_matches,
                        vector<vector<HomgPoint2D> >& n_matches,
                        vector<HomgPoint3D>& points,
                        double minx, double maxx, double miny, double maxy,
                        double minz, double maxz, bool print)
{
  Random ran;
  unsigned i;
  HomgPoint3D p3d(0.0, 0.0, 0.0);
  HomgPoint2D p2d(0.0, 0.0), n_p2d(0.0, 0.0);
  vector<HomgPoint2D> p2vec;
  vector<HomgPoint2D> noisy_p2vec;
  unsigned numim=pvec.size();
  bool done;
  RMatrix R;
  KMatrix K;
  Vector3<double> C, r;
  PMatrix P;

  p2vec.resize(numim);
  points.clear();
  for (i=0; i<i_matches.size(); i++)
    i_matches[i].clear();
  for (i=0; i<n_matches.size(); i++)
    n_matches[i].clear();
  i_matches.clear();
  n_matches.clear();


  // create noisy P matrices
  n_pvec.clear();
  for (i=0; i<numim; i++){
    if (pvec[i].GetK(K) || pvec[i].GetR(R) || pvec[i].GetC(C)){
      BIASERR("error decomposing pvec["<<i<<"] : "<<pvec[i]<<"\nK: "<<K
              <<"\nR: "<<R<<"\nC: "<<C);
      return;
    }
    if (R.GetRotationAnglesXYZ(r[0], r[1], r[2])){
      BIASERR("error extracting rotation angles from R: "<<R);
      return;
    }
    cerr << setw(4) <<i<<"   : true C : "<<C
         <<"\n         true R : "<<r<<endl;
    if (i>0){ // first camera is usually known exact
      r[0]+=ran.GetNormalDistributed(0.0, r_noise);
      r[1]+=ran.GetNormalDistributed(0.0, r_noise);
      r[2]+=ran.GetNormalDistributed(0.0, r_noise);
      R.SetXYZ(r[0], r[1], r[2]);
      C[0]+=ran.GetNormalDistributed(0.0, c_noise);
      C[1]+=ran.GetNormalDistributed(0.0, c_noise);
      C[2]+=ran.GetNormalDistributed(0.0, c_noise);
    }
    cerr <<  "         noisy C : "<<C
         <<"\n         noisy R : "<<r<<endl;

    P.Compose(K, R, C);
    n_pvec.push_back(P);
  }

  // create points
  cerr <<"creating 3D points and matches"<<endl;
  for (i=0; i<num; i++){
    do {
      p3d[0]=ran.GetUniformDistributed(minx, maxx);
      p3d[1]=ran.GetUniformDistributed(miny, maxy);
      p3d[2]=ran.GetUniformDistributed(minz, maxz);
      done=true;
      p2vec.clear();
      noisy_p2vec.clear();
      for (unsigned j=0; j<numim; j++){
        p2d=pvec[j]*p3d;
        p2d.Homogenize();
        n_p2d=n_pvec[j]*p3d;
        n_p2d.Homogenize();
        n_p2d[0]+=ran.GetNormalDistributed(0.0, spatial_noise);
        n_p2d[1]+=ran.GetNormalDistributed(0.0, spatial_noise);
        if (p2d[0]>=0 && p2d[0]<width-1 && p2d[1]>=0 && p2d[1]<height-1 &&
            n_p2d[0]>=0 && n_p2d[0]<width-1 &&
            n_p2d[1]>=0 && n_p2d[1]<height-1){
          p2vec.push_back(p2d);
          noisy_p2vec.push_back(n_p2d);
        } else {
          done=false;
          break;
        }
      }
      if (done) cout << "*"; else cout <<".";
      cout.flush();
    } while (!done);
    points.push_back(p3d);
    i_matches.push_back(p2vec);
    n_matches.push_back(noisy_p2vec);
  }
  cout << endl;

  if (print){
    cout <<"3D points and matches:"<<endl;
    for (unsigned i=0; i<num; i++){
      cout <<setw(3) << i <<": true  "<<points[i]<<"    \t: ";
      for (unsigned k=0; k<numim; k++)
        cout <<k<<i_matches[i][k]<<"  ";
      cout <<endl;
      cout <<"     noisy "<<points[i]<<"    \t: ";
      for (unsigned k=0; k<numim; k++)
        cout <<k<<n_matches[i][k]<<"  ";
      cout <<endl;
    }
    cout << endl;
  }

}


void GenGroundTruth::CreateMatchesOutlier(vector<BIAS::PMatrix>& pvec,
                               vector<BIAS::PMatrix>& n_pvec, unsigned num,
                               unsigned width, unsigned height,
                               double spatial_noise, double r_noise,
                               double c_noise, double outlier_fraction,
                               vector<vector<BIAS::HomgPoint2D> >& i_matches,
                               vector<vector<BIAS::HomgPoint2D> >& n_matches,
                               vector<BIAS::HomgPoint3D>& points,
                               double minx/*=-50.0*/, double maxx/*=50.0*/,
                               double miny/*=-50.0*/, double maxy/*=50.0*/,
                               double minz/*=-50.0*/, double maxz/*=50.0*/,
                               bool print)
{
  Random ran;
  unsigned inlier, outlier;
  outlier=(unsigned)rint((double)num*outlier_fraction);
  inlier=num-outlier;
  unsigned numim=pvec.size();
  double offsetx, offsety;

  CreateMatches(pvec, n_pvec, num, width, height, spatial_noise, r_noise,
                c_noise, i_matches, n_matches, points, minx, maxx, miny, maxy,
                minz, maxz, false);

  for (unsigned i=inlier; i<num; i++){
    cerr << i << "is outlier" << endl;
    offsetx=ran.GetUniformDistributed(3+2*spatial_noise, 10);
    offsety=ran.GetUniformDistributed(3+2*spatial_noise, 10);
    for (unsigned j=1; j<numim;j++){
      n_matches[i][j][0]+=j*offsetx;
      n_matches[i][j][1]+=j*offsety;
      i_matches[i][j][0]+=j*offsetx;
      i_matches[i][j][1]+=j*offsety;
    }
  }

  if (print){
    inlier=num-outlier;

    cout <<"3D points and matches:"<<endl;
    for (unsigned i=0; i<num; i++){

      if (i<inlier) cout << "inl  "; else cout << "outl ";
      cout << setw(3) << i <<": true  ";//<<points[i]<<"    \t: ";
      for (unsigned k=0; k<numim; k++){
        cout <<k<<i_matches[i][k]<<"-"
             <<pvec[k].GetK().Invert()*i_matches[i][k]<<"\t";
      }
      cout <<endl;

      cout <<"          noisy ";//<<points[i]<<"    \t: ";
      for (unsigned k=0; k<numim; k++)
        cout <<k<<n_matches[i][k]<<"-"
             <<n_pvec[k].GetK().Invert()*n_matches[i][k]<<"\t";
      cout <<endl;

    }
    cout << endl;
  }

}

void GenGroundTruth::GenerateRandomPlane(vector<BIAS::HomgPoint3D> &Points,
                         int n,
                         double aX, double aY, double aZ,
                         double alphaX, double alphaY, double alphaZ,
                         double betaX, double betaY, double betaZ,
                         double alphamax,
                         double betamax)
{
  GenerateRandomCube(Points, n, aX, aY, aZ,
                     alphaX, alphaY, alphaZ,
                     betaX, betaY, betaZ,
                     0.0, 0.0, 0.0,
                     alphamax, betamax, 0);
}

void GenGroundTruth::GenerateRandomCube(vector<BIAS::HomgPoint3D> &Points,
                      int n,
                      double aX, double aY, double aZ,
                      double alphaX, double alphaY, double alphaZ,
                      double betaX, double betaY, double betaZ,
                      double gammaX, double gammaY, double gammaZ,
                      double alphamax,
                      double betamax,
                      double gammamax)
{
  Random ran;
  double f1,f2,f3;
  Vector3<double> newpoint;
  HomgPoint3D newHomgPoint;
  Vector3<double> a(aX,aY,aZ);
  Vector3<double> alpha(alphaX,alphaY,alphaZ);
  Vector3<double> beta(betaX,betaY,betaZ);
  Vector3<double> gamma(gammaX,gammaY,gammaZ);
  alpha.Normalize();
  beta.Normalize();
  gamma.Normalize();
  for(int i=0;i<n;i++)
    {
      f1=ran.GetUniformDistributed(0.0,alphamax);
      f2=ran.GetUniformDistributed(0.0,betamax);
      f3=ran.GetUniformDistributed(0.0,gammamax);
      newpoint = a + f1*alpha + f2*beta + f3*gamma;
      newHomgPoint=newpoint;
      Points.push_back(newHomgPoint);
    }
}

int GenGroundTruth::ReadCorrespondences(const char *Filename,
                                vector<HomgPoint2D> &Points1,
                                vector<HomgPoint2D> &Points2)
{
  ifstream filein;
  double x,y,w;
  Points1.clear();
  Points2.clear();
  filein.open(Filename, ios::in);
  if ( !filein.is_open() )
    {
      cerr << "Error opening file '" << Filename << "'" << endl;
      return -1;
    }
  while  ( !filein.eof() )
    {
      filein.read( (char*) &x,sizeof(HOMGPOINT2D_TYPE) );
      filein.read( (char*) &y,sizeof(HOMGPOINT2D_TYPE) );
      filein.read( (char*) &w,sizeof(HOMGPOINT2D_TYPE) );
      Points1.push_back(HomgPoint2D(x,y,w));
      filein.read( (char*) &x,sizeof(HOMGPOINT2D_TYPE) );
      filein.read( (char*) &y,sizeof(HOMGPOINT2D_TYPE) );
      filein.read( (char*) &w,sizeof(HOMGPOINT2D_TYPE) );
      Points2.push_back(HomgPoint2D(x,y,w));
    }
  // the last correspondence is always read in twice so we cut one off
  Points1.pop_back();
  Points2.pop_back();
  filein.close();
  assert( Points1.size() == Points2.size() );
  return Points1.size();
}


int GenGroundTruth::WriteCorrespondences(const char *Filename,
                                vector<HomgPoint2D> &Points1,
                                vector<HomgPoint2D> &Points2)
{
  assert( Points1.size() == Points2.size() );
  ofstream fileout;
  fileout.open(Filename, ios::out);
  if ( !fileout.is_open() )
    {
      cerr << "Error opening file '" << Filename << "'" << endl;
      return -1;
    }
  for(unsigned int i=0;i<Points1.size();i++)
    {
      fileout.write( (char*) &Points1[i][0],sizeof(HOMGPOINT2D_TYPE) );
      fileout.write( (char*) &Points1[i][1],sizeof(HOMGPOINT2D_TYPE) );
      fileout.write( (char*) &Points1[i][2],sizeof(HOMGPOINT2D_TYPE) );
      fileout.write( (char*) &Points2[i][0],sizeof(HOMGPOINT2D_TYPE) );
      fileout.write( (char*) &Points2[i][1],sizeof(HOMGPOINT2D_TYPE) );
      fileout.write( (char*) &Points2[i][2],sizeof(HOMGPOINT2D_TYPE) );
    }
  fileout.close();
  return Points1.size();
}


int GenGroundTruth::ReadWorldPoints( const char *Filename,
                            vector<HomgPoint3D> & Points )
{
  ifstream filein;
  double x,y,z,w;
  Points.clear();
  filein.open(Filename, ios::in);
  if ( !filein.is_open() )
    {
      cerr << "Error opening file '" << Filename << "'" << endl;
      return -1;
    }
  while  ( !filein.eof() )
    {
      filein.read( (char*) &x,sizeof(HOMGPOINT3D_TYPE) );
      filein.read( (char*) &y,sizeof(HOMGPOINT3D_TYPE) );
      filein.read( (char*) &z,sizeof(HOMGPOINT3D_TYPE) );
      filein.read( (char*) &w,sizeof(HOMGPOINT3D_TYPE) );
      Points.push_back(HomgPoint3D(x,y,z,w));
    }
  // the last coordinate is always read in twice so we cut one off
  Points.pop_back();
  filein.close();
  return Points.size();
}


int GenGroundTruth::WriteWorldPoints( const char *Filename,
                              vector<HomgPoint3D> & Points )
{
  ofstream fileout;
  fileout.open(Filename, ios::out);
  if ( !fileout.is_open() )
    {
      cerr << "Error opening file '" << Filename << "'" << endl;
      return -1;
    }
  for(unsigned int i=0;i<Points.size();i++)
    {
      fileout.write( (char*) &Points[i][0],sizeof(HOMGPOINT3D_TYPE) );
      fileout.write( (char*) &Points[i][1],sizeof(HOMGPOINT3D_TYPE) );
      fileout.write( (char*) &Points[i][2],sizeof(HOMGPOINT3D_TYPE) );
      fileout.write( (char*) &Points[i][3],sizeof(HOMGPOINT3D_TYPE) );
    }
  fileout.close();
  return Points.size();
}


