/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sys/io.h>
#include <unistd.h>
#include <fstream>

#include "MDcam.hh"


//extern int GetCamIDs(int busid ,int &ncams, int *&camids);

using namespace BIAS;
using namespace std;

MDcam::MDcam()
{
  Camera_ = NULL;
  NumCams_ = 0;
  ParPort_ = 0x378;
  ExternalTrigger_ = false;
  ActiveTrigger_ = false;
  BytesPerPixel_ = 1;
}



MDcam::~MDcam()
{
  if (NumCams_ != 0 )
    ReleaseCameras();

}


void MDcam::SetDebugLevelDcam(int d) {
  DebuglevelDcam_ = d;
  for (int i=0; i<NumCams_; i++)
    Camera_[i]->SetDebugLevel(d) ;
}



int MDcam::ScanBus(std::ofstream &ss)
{

  BIASDOUT(D_MDCAM_INIT,"Scanning bus ");


  VideoSource_DCAM *cam;
  int ncams = 0;
  // int *camids = NULL;

  dc1394_t *d = NULL;
  dc1394camera_t* cams;
  dc1394camera_list_t* camsList;
  dc1394error_t err;

  err = dc1394_camera_enumerate( d, &camsList);
  ncams = camsList->num;

  if(err != DC1394_SUCCESS){
    BIASERR("Scanning bus failed");
  }

  if(ncams > 0){
    cam = new VideoSource_DCAM();
    cam->SetDebugLevel(DebuglevelDcam_);

    for (int i=0; i<ncams; i++) {
      cams = dc1394_camera_new(d, camsList->ids[i].guid);
      cam->SetSize(0,0);
      cam->OpenDevice(cams->guid & 0x00000FFF);
      if (ss) {
        ss << "0" <<"  "<<(cams->guid & 0x00000FFF)<<"  "<<cam->GetWidth()<<"  "
        <<cam->GetHeight()<<" 0"<<" 0" <<endl;
      }
      cam->CloseDevice();
    }
    delete cam;
  }else{
    BIASERR("No cameras found on bus ");
    return -1;
  }
  return ncams;
}




int MDcam::InitCameras(const string filename)
{
  ifstream ifs;
  int bus,camid, width,height,xoffset,yoffset;
  int res;
  BIASDOUT(D_MDCAM_INIT,"InitCameras() from "<<filename);
  if (NumCams_ != 0) {
    BIASERR("MDcam::InitCameras(): Already initialized");
    return -1;
  }
  ifs.open(filename.c_str());
  if (!ifs) {
    perror(filename.c_str());
    return -1;
  }
  while (ifs) {
    char tmpline[128];
    ifs.getline(tmpline,128);
    if (tmpline[0]=='0' || tmpline[0]=='1'|| tmpline[0]=='2')
      NumCams_ ++;
  }

  ifs.close();
  ifs.clear();
  if (!ifs) {
    cerr<<"after close"<<endl;
    perror(filename.c_str());
    return -1;
  }
  ifs.open(filename.c_str());
  if (!ifs) {
    perror(filename.c_str());
    return -1;
  }

  Camera_ = new VideoSource_DCAM * [NumCams_];
  for (int i=0; i< NumCams_; i++){
    ifs >> bus >> camid >> width >> height >>xoffset >>yoffset;
    cout <<"Expecting cam "<<i<<" at bus "<<bus<<" with ID "<<camid<<endl;
    Camera_[i] = new VideoSource_DCAM(bus);
    Camera_[i]->SetDebugLevel(DebuglevelDcam_) ;
    //   Camera_[i]->ForceFormat7(true);
    //Camera_[i]->SetSize(width,height);
    //Camera_[i]->SetBytesPerPixel(BytesPerPixel_);
    //Camera_[i]->SetLeftTop(xoffset,yoffset);
    Camera_[i]->SetFPS(Fps_);
    cout << "Frames Per Second" << Fps_ << endl;

    res = Camera_[i]->OpenDevice(camid);
    if (res < 0) {
      BIASERR("Can not open camera "<<camid<<" on bus "<<bus);
      return -1;
    }
 }

  ifs.close();


  return NumCams_;

}


int MDcam::ReleaseCameras()
{
  if (NumCams_ == 0) {
    BIASERR("MDcam::Release(): not initialized");
    return -1;
  }

  for (int i=0; i<NumCams_; i++) {
    Camera_[i]->CloseDevice();
    delete Camera_[i];
  }
  delete[] Camera_;
  NumCams_ = 0;
  return 0;
}


void MDcam::ActivateScorpionTimestamp()
{
  for (int i=0; i<NumCams_; i++)
    Camera_[i]->ActivateScorpionTimestamp();
}


int MDcam::PreGrab()
{
  int res;
  for (int i=0; i<NumCams_; i++) {
    res =  Camera_[i]->PreGrab();
    if (res != 0) return res;
  }

  return 0;
}




int MDcam::PostGrab()
{
  int res;
  for (int i=0; i<NumCams_; i++) {
    res =  Camera_[i]->PostGrab();
    if (res != 0) return res;
  }

  return 0;
}


int MDcam::GrabSingle(Camera<unsigned char> *Image)
{
  struct timeval now;
  if (ExternalTrigger_ && ActiveTrigger_ ) {
    outb(0, ParPort_);
    usleep(0);
    outb(255,ParPort_);
  }
  if (ExternalTrigger_)
    gettimeofday(&now,NULL);

  for (int i=0; i<NumCams_; i++) {
    Camera_[i]->GrabSingle(Image[i]);
  }

  // if external trigger use accurate timestamps
  if (ExternalTrigger_) {
    stringstream time_ss;
    time_ss.str("");
    time_ss<<now.tv_sec<<":"<<now.tv_usec;
    // reset trigger
    if (  ActiveTrigger_ )
      outb(255,ParPort_);
    for (int i=0; i<NumCams_; i++)
      //  Image[i].SetTime(now.tv_sec,now.tv_usec);
      Image[i].SetASCIIString(time_ss.str());
  }
  return 0;
}

#ifdef BUILD_IMAGE_USHORT
int MDcam::GrabSingle(BIAS::Camera <unsigned short> *Image)
{
  struct timeval now;
  if (ExternalTrigger_ && ActiveTrigger_ ) {
    outb(0, ParPort_);
    usleep(0);
    outb(255,ParPort_);
  }
  if (ExternalTrigger_)
    gettimeofday(&now,NULL);

  for (int i=0; i<NumCams_; i++) {
    Camera_[i]->GrabSingle(Image[i]);
  }

  // if external trigger use accurate timestamps
  if (ExternalTrigger_) {
    stringstream time_ss;
    time_ss.str("");
    time_ss<<now.tv_sec<<":"<<now.tv_usec;
    // reset trigger
    if (  ActiveTrigger_ )
      outb(255,ParPort_);
    for (int i=0; i<NumCams_; i++)
      //  Image[i].SetTime(now.tv_sec,now.tv_usec);
      Image[i].SetASCIIString(time_ss.str());
  }
  return 0;
}
#endif


int MDcam::SetFeature(dc1394feature_t feature, unsigned int value)
{
  int res;
  for (int i=0; i<NumCams_; i++) {
    res = Camera_[i]->SetFeature(feature, value);
    if (res<0) return res;
  }
  return 0;
}

int MDcam::SetGain(float g)
{
  int res;
  for (int i=0; i<NumCams_; i++) {
    res = Camera_[i]->SetGain(g);
    if (res<0) return res;
  }
  return 0;
}

int MDcam::SetShutter(float exptime)
{
  int res;
  for (int i=0; i<NumCams_; i++) {
    res = Camera_[i]->SetShutter(exptime);
    if (res<0) return res;
  }
  return 0;
}


//    int GetFeature( int feature);
int MDcam::AutoMode( dc1394feature_t feature, dc1394feature_mode_t mode)
{
  int res;
  for (int i=0; i<NumCams_; i++) {
    res = Camera_[i]->AutoMode(feature,mode);
    if (res<0) return res;
  }
  return 0;
}

int MDcam::OnePushAuto(dc1394feature_t feature)
{
  int res;
  for (int i=0; i<NumCams_; i++) {
    res = Camera_[i]->OnePushAuto(feature);
    if (res<0) return res;
  }
  return 0;
}

int MDcam::ExternalTrigger(bool on, bool active)
{
  int res;
  if (on && active) {
    if (ioperm(ParPort_, 3, 1)) {
      BIASERR("Can not acces ParPort at "<<ParPort_<<": Permission denied."<<endl
	     <<" External Trigger disabled");
      ExternalTrigger_ = false;
      return -1;
    }
    // set all bits high
    else outb(255, ParPort_);
  }

  for (int i=0; i<NumCams_; i++) {
    res = 0;
    Camera_[i]->SetExternalTrigger(on);
    if (res<0) return res;
  }
  ExternalTrigger_ = on;
  ActiveTrigger_ = active;
  return 0;
}




int MDcam::EqualizeFeature(int master, dc1394feature_t feature)
{
  int value = Camera_[master]->GetFeature(feature);
  for (int i=0; i<NumCams_; i++) {
    if (i != master)
      Camera_[i]->SetFeature(feature, value);
  }

  return 0;
}


int MDcam::SetCamera(CameraParameterSet camparam)
{
  int res=0;
  for (int i=0; i<NumCams_; i++) {
    res = Camera_[i]->SetCamera(camparam);
    if (res <0) return res;
  }

  return res;
}
