/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __BIASVIDEOSOURCE_HH__
#define __BIASVIDEOSOURCE_HH__

#include <bias_config.h>


#include <vector>
#include <Base/Image/Image.hh>
  //#include <Image/Camera.hh>
#include <Image/VideoStream.hh>

			 //#include <VideoSource/VideoSourceCapabilities.hh>

#define D_V4L  (1<<0)
#define D_DCAM (1<<1)
#define D_VSO_DISK (1<<2)
#define D_DCAM_F7 (1<<3)
#define D_DV (1<<4)
#define D_VS_WHITEB (1<<5)
#define D_NET (1<<6)
#define D_DV2 (1<<8)
#define D_DTD (1<<9)

namespace BIAS{

  // forward declaration
  template <class T> class BIASVideoSource_EXPORT Camera;

  class BIASVideoSource_EXPORT VideoSourceCapabilities;

  /**
      \class VideoSource
      \ingroup g_videosource
      \author evers
      \brief  Defines a common interface to different devices.

      Specialized classes for each deviceclass should be derived and most methods
      have to be reimplemented.

      Usage:

      VideoSource *foo;

      Image bar;


      /// assume to have a class for DCAM cams
      foo = new VideoSource_DCAM;

      /// select a colormode with SetColorMode() if required
      //  the default colormode depends on the used Camera(kind of VideoSource)

      // set a image size for grabbing foo->SetSize(width,height)
      foo->SetSize(768,576);

      /// Want to take the second channel (what ever this means)
      foo->SetChannel(2);

      /// take the first appropriate device
      foo->OpenDevice();

      foo->InitImage(bar);

      // do last preparations
      foo->PreGrab();

      // get one frame
      foo->GrabSingle(bar);

      // stop enetually running things (isochronous transfers for example)
      foo->PostGrab();

      // clean up
      foo->CloseDevice();

  */

  class BIASVideoSource_EXPORT VideoSource :public BIAS::Debug{

  public:
    VideoSource();  
    virtual ~VideoSource();

    /**
       this simplest method to open a device selects the first available device
       E.g.: V4L: /dev/video0
       DCAM/DV : Card 0, first camera
    */
    virtual int OpenDevice();

    /// a device name like "/dev/video1" can be specified here
    virtual int OpenDevice(const char *device);

    /// a device number  can be specified here (DCAM) or port number (Net)
    virtual int OpenDevice(int device);

    /** used in child class VideoSource_Disk */
    virtual int OpenDevice(std::string Prefix, std::string Postfix,
                           int DigitCount, int PictureCount,int StartCount=0);

    /** used in child class VideoSource_Disk */
    virtual int OpenDevice(const std::vector<std::string> &FileNames);

    virtual int CloseDevice();

    virtual int GetAllDevices(std::vector<std::string> &devices);

    virtual int GrabSingle(BIAS::Camera <unsigned char> &image);
    virtual int GrabSingle(BIAS::Camera <float> &image);
#ifdef BUILD_IMAGE_CHAR
    virtual int GrabSingle(BIAS::Camera <char> &image);
#endif
#ifdef BUILD_IMAGE_SHORT
    virtual int GrabSingle(BIAS::Camera <short int> &image);
#endif
#ifdef BUILD_IMAGE_INT
    virtual int GrabSingle(BIAS::Camera <int> &image);
#endif
#ifdef BUILD_IMAGE_USHORT
    virtual int GrabSingle(BIAS::Camera <unsigned short int> &image);
#endif
#ifdef BUILD_IMAGE_UINT
    virtual int GrabSingle(BIAS::Camera <unsigned int> &image);
#endif
#ifdef BUILD_IMAGE_DOUBLE
    virtual int GrabSingle(BIAS::Camera <double> &image);
#endif
    /// Do last preparations before grabbing
    /** E.g.: start isochronous transfer
     */
    virtual int PreGrab();

    /// Stop anythig started in PreGrab()
    virtual int PostGrab();

    /** These functions are not inline to let other class overload them<br>
        But they are implemented for this class
        Must be used before OpenDevice() !
    */
    virtual void SetSize(int w, int h, int bytesperpixel=1);

    /** Select colormodel to use
        Not all sources support all colormodels.
        Must be used before OpenDevice() !
+    */
    virtual int SetColorModel(BIAS::ImageBase::EColorModel  mode);

    /** Use channel to select one of many cameras on the same
        Must be used before OpenDevice() !
    */
    void SetChannel(int channel) { DeviceChannel_ = channel;};

    /// Use channel to select one of many cameras on the same
    /**      Must be used before OpenDevice() ! */
    void SetFPS(float fps) { FramesPerSecond_ = fps;};

    /** @brief specify if there's an external trigger
    * @param trigger true if external trigger should be acticated, false if deactivated   
    * */
    virtual void SetExternalTrigger(bool trigger = true) { bExternalTrigger_ = trigger;};
    
    /**      Must be used before OpenDevice() ! */
    void SetBytesPerPixel(unsigned char bpp) { BytesPerPixel_ = (float) bpp;};

    inline int GetWidth() { return Width_;};

    inline int GetHeight() { return Height_;};

    inline int GetColorChannels() {return ColorChannels_;};

    inline BIAS::ImageBase::EColorModel  GetColorModel() {return ColorMode_;};

    inline float GetBytesPerPixel() const { return BytesPerPixel_; }

    /** Call this only after opening the device successfully
     */
    virtual int InitImage(BIAS::ImageBase &Image);

    /**
     * Use this method to learn something about the capabilities of the source
     *(only usefull for v4l-sources e.g. /dev/video0). The output is send to
     * stdout.
     */
 
    virtual int GetCapabilities(VideoSourceCapabilities &caps);
    virtual int GetCapabilities(const char *device, VideoSourceCapabilities &caps);

    /// Use this to enable/disable any automatic modes addressed via enums
    virtual int AutoMode(int, bool);
    inline int GetID(){return ID_;};

    inline bool IsActive()
    { return Active_; };

    /// tell if SetShutter, SetGain... would work
    inline bool IsControllable(){ return IsControllable_;};

    /// this function is used to control things like shutter, zoom...
    /** normaly, feature is an enum
     */
    virtual int  SetFeature( int feature, unsigned int value);

    /// this function is used to query things like shutter, zoom...
    /** normaly, feature is an enum
     */
    virtual unsigned int GetFeature( int feature);

    /* more convinient function to control camera parameter */

    virtual bool HasControlShutter(){return false;};
    virtual float GetShutter();
    /// Set shutter (exposure time) to exptime in seconds
    virtual int SetShutter(float exptime);
    virtual void SetAutoShutter(bool b){};
    virtual bool GetAutoShutter(){return false;};
    virtual void GetShutterRange(double &min, double &max){
      min = minShutter_, max = maxShutter_;};
    
    virtual bool HasControlGain(){return false;};
    virtual float GetGain();
    /// Set gain to g with g[dB]
    virtual int SetGain(float g);
    virtual void SetAutoGain(bool b){};
    virtual bool GetAutoGain(){return false;};
    virtual void GetGainRange(double &min, double &max){
      min = minGain_, max = maxGain_;};
    
    virtual bool HasControlBrightness(){return false;};
    /* Value: 0<=g<=100 */
    virtual float GetBrightness(){return 0.0;};
    /* Value: 0<=g<=100 */
    virtual int SetBrightness(float g){return 0;};
    virtual void SetAutoBrightness(bool b){};
    virtual bool GetAutoBrightness(){return false;};
    virtual void GetBrightnessRange(double &min,double &max){
         min = minBrightness_, max = maxBrightness_;};

    virtual bool HasControlContrast(){return false;};
    /* Value: 0<=g<=100 */
    virtual float GetContrast(){return 0.0;};
    /* Value: 0<=g<=100 */
    virtual int SetContrast(float g){return 0;};
    virtual void SetAutoContrast(bool b){};
    virtual bool GetAutoContrast(){return false;};
    virtual void GetContrastRange(double &min,double &max){
      min = minContrast_, max = maxContrast_;};

    virtual bool HasControlWhiteBalance(){return false;};

    /// Reset digital zoom, pan and tilt to defaults to ensure
    /// compliance with calibration.
    virtual void ResetPanTiltZoom(){};

    virtual int OnePushAuto( int feature);
    /// range is [-1.0 , 1.0)
    virtual int SetWhiteBalance(float rvalue,float bvalue);
    virtual int GetWhiteBalance(float &rvalue,float &bvalue);

    //sets the whitebalance to optimal value in software
    virtual int SoftwareWhiteBalance();
    /**
       this enables the very fast direct-to-disk mode.
       Use this _after_ OpenDevice() because the image size must be known.
       @author evers
    */
    int SetDirectToDisk(const std::string &filename);

    virtual void EnsureLatestFrame(const bool b) {};

    std::string GetIdentifier() {return Identifier_;};

    /** If you want as many images as possbile, and dont care about equal ones
        call this with "false". Default is that GrabSingle blocks until
        a new frame arrives from the camera. Only used in V4L.
        @author grest, Oct. 2005 */
    inline void SetWaitForNew(bool waitForNew) { WaitForNew_=waitForNew; }


  protected:

    int WriteDirect_(const unsigned char *data, const BIAS::UUID &uuid,
                     const long int &sec, const long int &usec);

    int Width_;
    int Height_;
    int ColorChannels_;
    float FramesPerSecond_;

    /// BytesPerPixel cannot be int, becaus of e.g. YUV420p: 1.5
    float BytesPerPixel_;

    const char *DefaultDevice_;

    int DeviceChannel_;
    BIAS::ImageBase::EColorModel ColorMode_;

    /// Some state variables

    /// Complete_initialized_ is set when OpenDevice(),
    // UseChannel() ... are done
    bool CompleteInitialized_;

    /// active_ is set in PreGrab() und unset in PostGrab()
    bool Active_;

    /// grabbing_ is set and unset in Grab*()
    bool Grabbing_;

    /// external trigger flag
    bool bExternalTrigger_;
    
    int ID_;

    bool WaitForNew_;
    bool DirectToDisk_;
    int DtDFd_; // DirectToDisk File descriptor
    // 512 byte aligned buffer, contains VSImageHeader and image data
    char *DtDBuffer_;
    // must be multiple of 512
    unsigned int DtDBufferSize_;
    unsigned int ImageSize_; // useful for allocating and copying memory
    // this point to the beginning of DtDBuffer_
    BIAS::VideoStream::VSImageHeader *DtDImageHeader_;
    char *DtDStartOfImage_; // points into DtDBuffer after VSImageHeader
    // must be initialized be the derived classes
    bool IsControllable_;

    double minContrast_, maxContrast_;
    double minGain_, maxGain_;
    double minShutter_, maxShutter_;
    double minBrightness_, maxBrightness_;
    
    // verbose descriptors
    std::string Identifier_;

  };

}// namespace MIP

#endif







