/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __BIASVIDEOSOURCE_USBPMD_HH__
#define __BIASVIDEOSOURCE_USBPMD_HH__

#ifndef WIN32
#  error The PMD-cameras can only be run under windows.
#endif //not WIN32

#include <VideoSource/VideoSource_Base.hh>
#include <Image/Camera.hh>
#include <bias_config.h>
#include <vector>
#include <Image/PMDImageProc.hh>

namespace BIAS {

  enum PMDtype {
    PMD_1K = 0, // 64*16
    PMD_3K = 1, // 66*48
    MultiCam_3K = 2, // 64*48
    MultiCam_19K = 3, // 160*120
    NoPMD = 99 // kein PMD-Chip
  };

  /** @class VideoSource_usbPMD
      @ingroup g_videosource
      @brief reads standardPMD-tiffs(2D) with depth images and modulation coefficients
             TODO: include support for normal images (2DImg_), not supported by prototype
      @author apetersen 5/2007 */
  class BIASVideoSource_EXPORT VideoSource_usbPMD : public VideoSource {

  public:
    VideoSource_usbPMD();

    ~VideoSource_usbPMD();

    //call before OpenDevice: set the !needed! bitfile with full path
    inline void SetBitFile(std::string file) {
      bitFile_ = file;
    }
    //call before OpenDevice: needed if desired type != MultiCam_3K
    inline int SetType(int type) {
      if ((type < 0)||(type > 3)) {
        BIASERR("Invalid type given!!!");
        return -1;
      }
      type_ = type;
      return 0;
    }

    //try to open first unused PMD-camera
    virtual int OpenDevice();
    //try to open a specific camera-device
    virtual int OpenDevice(int camNr);
    //close an active camera
    virtual int CloseDevice();

    //no needed for now
    virtual int PreGrab();
    virtual int PostGrab();

 
  	/** 
      Returns current framerate
	  */
    double GetFPS();

    virtual int InitImage(ImageBase &img);
    virtual int Init2DImage(ImageBase &img);

    /** Returns the 2D image, and reads depth and modulation coefficients */
    virtual int GrabSingle(Camera<unsigned char> &image);
    
    /** Capture a single phase image only */
    int CaptureSinglePhase(Camera<unsigned char> &image);

    /// Set shutter (exposure time) to exptime in seconds [==] for PMD-chip
    virtual int SetShutter3D(float exptime);
    /// Set shutter (exposure time) to exptime in seconds [==] for 2D-chip
    int SetShutter2D(float exptime);
    /// Set gain to g with g[dB] 
    int SetGain2D(float g);
  
    virtual float GetShutter3D();
    float GetShutter2D();
    float GetGain2D();

    /* Set the Modulation frequency - for now only 15 or 20 supported */
    int SetModulationFrequency(int dModulationFrequency);
    /* Get the Modulation frequency*/
    int GetModulationFrequency();

    void SetTime(long sec, long usec) {
      DepthImg_.SetTime(sec, usec);
      ModCoeffImg_.SetTime(sec, usec);
      img2D_.SetTime(sec, usec);
    }

    /** Returns the depth image. Must be called after GrabSingle() */
    inline int GrabSingleDepth(Camera<float> &image)
    { 
      if (!DepthImg_.IsEmpty()) {
        image = DepthImg_; 
        return 0;
      } 
      return -1;
    }

    /** Returns the amplitude image. Must be called after GrabSingle() */
    inline int GrabSingleModCoeff(Camera<float> &image)
    {       
      if (!ModCoeffImg_.IsEmpty()) {
        image = ModCoeffImg_; 
        return 0;
      } 
      return -1;
    }

    /** Returns the High Res 2D image. Must be called after GrabSingle() */
    inline int GrabSingle2D(Camera<unsigned char> &image)
    {       
      if (!img2D_.IsEmpty()) {
        image = img2D_; 
        return 0;
      } 
      return -1;
    }



    void SetGenerateNewUIDs(bool s) {  
      BIASERR("Always generating new UUIDS for PMD images");
      BIASABORT;
    }

  protected:
    void DepthImageFloatFromDouble_(double* data,Image<float> &dst);
    void ImageUcharFromDouble_(double* data, Image<unsigned char> &dst);
    void ImageUcharFromShort_(signed short* data, Image<unsigned char> &dst);
    void ImageFloatFromDouble_(double* data, Image<float> &dst);
    void Convert2DImage_(unsigned char* data, Image<unsigned char> &dst);

    int type_, use2D_, doCalcs_, phaseAlgorithm_;
    unsigned int modulationsFrequenz_, Width2D_, Height2D_;
    bool bIsOpen_;
    std::string devicePort_, bitFile_;

    //MIP images, holding grabbed data
    BIAS::Camera<unsigned char> img2D_, img2DPMD_, imgPMD2D_;
    BIAS::Camera<float> DepthImg_, ModCoeffImg_;

    double *dist_; //depth image from PMD-lib
    double *mod_; //Coefficient image from PMD-lib
    signed short *grey_; //PMD-grey image from PMD-lib
    signed short *total_; //raw img data from PMD-lib
    unsigned char *grey2D_; //2D-grey image, from PMD-lib 
  };

}

#endif //__VIDEOSOURCE_USBPMD_HH__
